#!/usr/bin/perl

# ====================================================================
# Copyright (c) 2010 The Regents of the University of Michigan and 
# The European Organization for Nuclear Research (CERN). Licensed under the
# Educational Community License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
# http://www.osedu.org/licenses/ECL-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
# ====================================================================

#####################################################################
# usage_reports.pl
#
# NB: for string_grep, you must enter in starting at root level of the repository, e.g. umich/courses/2006-D
#
# Author: Jeremy Herr (herrj@umich.edu)
# Copyright 2006 by the Regents of the University of Michigan
#
# TODO: note referrer.  This means I'll have to get the referrer from the first hit to the website,
# so I'll have to pay attention to visits, not just sessions.  I should probably deal with visits in any case
# (keep track of how many lectures they look at per visit, etc)
#
# USAGE:
#   usage_reports.pl --begin=YYYY-MM-DD --end=YYYY-MM-DD --string=pattern --report=??? --debug=N --verb=N
#
#####################################################################

use strict;
use Time::Local;

$::FS             = '/';
$::VERBOSITYLEVEL = 5;
$::DEBUGLEVEL     = 0;
$::LOGSTYLE       = 'umich'; # the other possibility is 'umich'

&main(@ARGV);

#####################################################################
# main
#
#
#####################################################################
sub main
{
    my @args = @_;
    my $parameters = {};
    my ($time_start, $time_stop);

    &debug(5, "entering main");

    my $time_start = time();

    &pr(2, "setting default parameters\n");
    $parameters = &set_default_parameters();
    &pr(2, "parsing command line and setting parameters\n");
    $parameters = &get_command_line($parameters, @args);

    &pr(2, "finding lecture objects and series\n");
    my $hash_lecture_series = &find_lecture_objects($parameters);

    &pr(1, "desired begin date: " . localtime($parameters->{'date_begin'}) . "\n");
    &pr(1, "desired end date:   " . localtime($parameters->{'date_end'}) . "\n");
    &pr(1, "filter by url string: " . $parameters->{'string_grep'} . "\n");

    &populate_logfile_list($parameters);

    my $hash_robots = &find_robots($parameters->{'list_all_log_files'});
#    foreach my $robot (@$list_robots) {
#	&pr(1, "robot = $robot\n");
#    }

    my $list_needed_log_files = &find_needed_log_files($parameters->{'date_begin'},
						$parameters->{'date_end'}, 
						$parameters->{'list_all_log_files'});
    foreach my $file_log (@$list_needed_log_files) {
	&pr(1, "log file " . $file_log->{'path'} . "");
	&pr(1, " begins [" . localtime($file_log->{'date_begin'}) . "]");
	&pr(1, " and ends [" . localtime($file_log->{'date_end'}) . "]\n");
    }

    my ($list_sessions, $hash_hosts) = &populate_data($parameters,
						      $list_needed_log_files,
						      $hash_lecture_series,
						      $hash_robots);
    
    &display_summary($list_sessions, $hash_lecture_series, $parameters);


    &pr(1, "creating graphs of time vs. sessions\n");
#    &make_graph_time_vs_sessions($parameters, $list_sessions, 300);
#    &make_graph_time_vs_sessions($parameters, $list_sessions, 600);

    &make_graph_time_vs_sessions($parameters, $list_sessions, 3600);
    &make_graph_time_vs_sessions($parameters, $list_sessions, 24*3600);
    &make_graph_time_vs_sessions($parameters, $list_sessions, 7*24*3600);

#    foreach my $session (@$list_sessions) {
#	if ($session->{'time_end'} - $session->{'time_begin'} > 5) {
#	    &display_session_brief($session);
#	}
#    }

    &make_report_ip_list($parameters, $list_sessions, 'detail');

    &make_report_host_list($parameters, $hash_hosts);

    &make_report_most_popular($parameters, $list_sessions, '');

    if ($parameters->{'report'} eq '2007-A-weekly') {
	&make_report_for_education($parameters, $list_sessions, '2007-A-weekly');
    }
    elsif ($parameters->{'report'} eq '2006-A-physics140') {
	&make_report_for_education($parameters, $list_sessions, '2006-A-physics140');
    }
    elsif ($parameters->{'report'} eq '2007-D-weekly') {
	&make_report_for_education($parameters, $list_sessions, '2007-D-weekly');
    }

    my $time_stop = time();

    &pr(1, "started: " . scalar(localtime($time_start)) . "\n");
    &pr(1, "stopped: " . scalar(localtime($time_stop)) . "\n");

    &pr(1, "total running time: " . int($time_stop - $time_start) . " seconds\n");
    &debug(5, "exiting main");
}

#####################################################################
# set_default_parameters
#
# Description:
# populate $parameters hash with default values
#
# Arguments: none
#
# Returns:
# $parameters: a hash
#
#####################################################################
sub set_default_parameters
{
    my $parameters = {};

    &debug(5, "entering set_default_parameters");

    $parameters->{'date_begin'}            = '';
    $parameters->{'date_end'}              = '';
    $parameters->{'string_grep'}           = '';
    $parameters->{'list_duration_visit'}   = [0, 1*60, 5*60, 10*60, 30*60, 1*3600, 2*3600];
    $parameters->{'list_duration_session'} = [0, 1*60, 5*60, 10*60, 30*60, 1*3600, 2*3600];
    $parameters->{'timeout_visit'}         = 1*3600;
    $parameters->{'timeout_session'}       = 1*3600;
    $parameters->{'path_lo'}               = '/data1/lo-repository';
    $parameters->{'path_wl'}               = '/data4/wl';
    $parameters->{'file_log_list'}         = '/var/www/html/usage-stats/lecb-logfiles.txt';
    $parameters->{'list_all_log_files'}    = [];
    $parameters->{'dir_output'}            = '/var/www/html/usage-stats/output';

    &debug(5, "exiting set_default_parameters");

    return $parameters;
}

#####################################################################
# get_command_line
#
# Description:
#
# Arguments: 
# $parameters: pointer to a hash
# @args: array of command line args
#
# Returns:
# $parameters: pointer to a hash populated with values from command line
#####################################################################
sub get_command_line
{
    my ($parameters, @args) = @_;

    while (scalar(@args) > 0) {
	my $arg = shift @args;
	if ( $arg =~ m/^\-\-begin\=(\d\d\d\d\-\d\d\-\d\d)/ ) {
	    $parameters->{'date_begin'} = &convert_YMD_time($1);
	}
	elsif ( $arg =~ m/^\-\-end\=(\d\d\d\d\-\d\d\-\d\d)/ ) {
	    $parameters->{'date_end'} = &convert_YMD_time($1);
	}
	elsif ( $arg =~ m/^\-\-string\=(.*)/ ) {
	    $parameters->{'string_grep'} = $1;
	    &debug(4, "got arg: " . $parameters->{'string_grep'} );
	}
	elsif ( $arg =~ m/^\-\-report\=(.*)/ ) {
	    $parameters->{'report'} = $1;
	    &debug(4, "got arg: " . $parameters->{'report'} );
	}
	elsif ( $arg =~ m/^\-\-logs\=(.*)/ ) {
	    $parameters->{'file_log_list'} = $1;
	    &debug(4, "got arg: " . $parameters->{'file_log_list'} );
	}
	elsif ( $arg =~ m/^\-\-debug\=(\d+)/ ) {
	    $::DEBUGLEVEL = $1;
	}
	elsif ( $arg =~ m/^\-\-verb\=(\d+)/ ) {
	    $::VERBOSITYLEVEL = $1;
	}
	else {
	    &pr(0, "invalid command line argument: $arg\n");
	    exit(1);
	}

    }

    return $parameters;
}

#####################################################################
# find_needed_log_files
#
# Description:
# Given starting and ending dates, look through given list of log
# files to find which ones cover the desired time period.
#
# Returns:
# 
#####################################################################
sub find_needed_log_files
{
    my ($date_begin, $date_end, $list_all_log_files) = @_;
    my $list_needed_log_files = [];

    &debug(5, "entering find_needed_log_files");

    foreach my $path_log (@$list_all_log_files) {
	my $date_log_begin;
	my $date_log_end;

#	print "first line: " . &get_first_line($path_log) . "\n";
#	print "last line:  " . &get_last_line($path_log) . "\n";

# First line of Michigan Apache logs looks like:
#
# [11/Apr/2005:14:28:42 -0400]

	my %hash_date_log_begin;
	my %hash_date_log_end;

	if ($::LOGSTYLE eq 'umich') {
	    &get_first_line($path_log)
		=~ m!\[(\d+)\/(\w+)\/(\d+)\:(\d+)\:(\d+)\:(\d+)\s*[\-\d]*\]!;
	    %hash_date_log_begin = ( 'day'    => $1,
				     'month'  => $2,
				     'year'   => $3,
				     'hour'   => $4,
				     'minute' => $5,
				     'second' => $6 );
	    
	    &get_last_line($path_log)
		=~ m!\[(\d+)\/(\w+)\/(\d+)\:(\d+)\:(\d+)\:(\d+)\s*[\-\d]*\]!;
	    %hash_date_log_end = ( 'day'    => $1,
				   'month'  => $2,
				   'year'   => $3,
				   'hour'   => $4,
				   'minute' => $5,
				   'second' => $6 );
	}

# First few lines of CERN IIS logs look like:
#
# #Software: Microsoft Internet Information Services 6.0
# #Version: 1.0
# #Date: 2009-07-05 00:00:00
# #Fields: date time s-sitename s-ip cs-method cs-uri-stem cs-uri-query s-port cs-username c-ip cs(User-Agent) cs(Referer) sc-status sc-substatus sc-win32-status sc-bytes
# 2009-07-04 23:59:59 W3SVC20 137.138.240.82 GET /Webping.htm - 80 - 137.138.140.62 MOM_URL_Monitoring - 200 0 0 502

	if ($::LOGSTYLE eq 'cern') {
	    &get_first_line($path_log)
		=~ m!^\s*(\d+)\-(\d+)\-(\d+)\s+(\d+)\:(\d+)\:(\d+)!;
	    %hash_date_log_begin = ( 'year'   => $1,
				     'month'  => $2,
				     'day'    => $3,
				     'hour'   => $4,
				     'minute' => $5,
				     'second' => $6 );
	    
	    &get_last_line($path_log)
		=~ m!\[(\d+)\/(\w+)\/(\d+)\:(\d+)\:(\d+)\:(\d+)\s*[\-\d]*\]!;
	    %hash_date_log_end = ( 'year'   => $1,
				   'month'  => $2,
				   'day'    => $3,
				   'hour'   => $4,
				   'minute' => $5,
				   'second' => $6 );
	}

#	print "second " . $hash_date_log_begin{'second'} . "\n";
#	print "minute " . $hash_date_log_begin{'minute'} . "\n";
#	print "hour   " . $hash_date_log_begin{'hour'} . "\n";
#	print "day    " . $hash_date_log_begin{'day'} . "\n";
#	print "month  " . &convert_month_value( $hash_date_log_begin{'month'} ). "\n";
#	print "year   " . $hash_date_log_begin{'year'} . "\n";

	$date_log_begin = timelocal( $hash_date_log_begin{'second'},
				     $hash_date_log_begin{'minute'},
				     $hash_date_log_begin{'hour'},
				     $hash_date_log_begin{'day'},
				     &convert_month_value( $hash_date_log_begin{'month'} ),
				     $hash_date_log_begin{'year'} );

	$date_log_end = timelocal( $hash_date_log_end{'second'},
				   $hash_date_log_end{'minute'},
				   $hash_date_log_end{'hour'},
				   $hash_date_log_end{'day'},
				   &convert_month_value( $hash_date_log_end{'month'} ),
				   $hash_date_log_end{'year'} );
	    
	&debug(3, "reading path_log = [$path_log]");
	&debug(3, "date_log_begin = [" . localtime($date_log_begin) . "]");
	&debug(3, "date_log_end   = [" . localtime($date_log_end) . "]");

	my $hash_log_file_summary;
	%$hash_log_file_summary = ( 'path'  => $path_log,
				    'date_begin' => $date_log_begin,
				    'date_end'   => $date_log_end );
	push(@$list_needed_log_files, $hash_log_file_summary);
    }

    &debug(5, "exiting find_needed_log_files");

    return $list_needed_log_files;
}

#####################################################################
# find_lecture_objects
#
# I need to change this so that each lecture is instead a hash
# containing all the slide timing info and basic metadata
#####################################################################
sub find_lecture_objects
{
    my ($parameters) = @_;
    my $hash_lecture_series = {};
    my $list_lectures = [];
    my $list_lectures_sorted = [];

    my $command = "find " . $parameters->{'path_wl'} . $::FS . $parameters->{'string_grep'} . "* -follow -name \"lecture.xml\" | sort";
    &pr(1, "comand: $command\n");

    my $result = qx ($command);
    &debug(7, "result: $result\n");

    my $parameters_path_wl = $parameters->{'path_wl'};
    my $string_tailing = '/lecture.xml';
    $result =~ s!$parameters_path_wl(.*?)$string_tailing!$1!sg;

    @$list_lectures = split(m!\n!, $result);

    @$list_lectures_sorted = sort @$list_lectures;

#    foreach my $lecture (@$list_lectures_sorted) {
#	$lecture =~ s!$parameters_path_wl(.*?)$string_tailing!$1!s;
#    }

    # now organize the lecture objects into series
#    my $last_dir_parent = '';
#    my $dir_parent_ctr = 0;
    foreach my $lo (@$list_lectures_sorted) {
# /umich/courses/2006-D/histart112/20060905-umwlcd0007-02
	my ($dir_parent) = $lo =~ m!^(.*?)\/[^\/]+$!;
	
	my $hash_lo = &get_lo_metadata($lo, $parameters);

	if ($hash_lecture_series->{$dir_parent}) {
	    push( @{$hash_lecture_series->{$dir_parent}}, $hash_lo);
	}
	else {
	    $hash_lecture_series->{$dir_parent} = [];
	    push( @{$hash_lecture_series->{$dir_parent}}, $hash_lo);
	}
    }
    
    foreach my $series (sort keys %$hash_lecture_series) {
	&pr(1, "Series: $series\n");
	foreach my $hash_lo (@{$hash_lecture_series->{$series}}) {
	    &pr(1, "  Lecture: " . $hash_lo->{'id'} 
		. "  \"" . $hash_lo->{'title'} 
		. "\" (" . $hash_lo->{'creator'} . "), " 
		. $hash_lo->{'numslides'} . " slides\n");
	}
    }
    
    return $hash_lecture_series;
}

#####################################################################
# populate_logfile_list
#
# Read in file listing locations of logfiles and populate the 
# list of log files.
#####################################################################
sub populate_logfile_list
{
    my ($parameters) = @_;

    my $file_log_list = $parameters->{'file_log_list'};

    open (LISTFILE, "<$file_log_list")
	or die "can not open " . $file_log_list . ": $!\n";
    while (my $line = <LISTFILE>) {
	$line =~ m/(\S+)/;
	push(@{$parameters->{'list_all_log_files'}}, $1);
    }
    close (LISTFILE);

    if (! scalar($parameters->{'list_all_log_files'})) {
	&pr(0, "No log files listed in file " . $parameters->{'file_log_list'} . "\n");
	exit(1);
    }

    return $parameters;
}

#####################################################################
# get_lo_metadata
#
#
#####################################################################
sub get_lo_metadata
{
    my ($lo, $parameters) = @_;
    my $hash_lo = {};

    &debug(5, "entering get_lo_metadata");

    $hash_lo->{'id'} = $lo;

    my $lecture_xml = $parameters->{'path_lo'} . $::FS . $lo . $::FS . 'lecture.xml';
    if (! -r $lecture_xml) {
	&pr(0, sprintf("no such file: %s", $lecture_xml));
	&pr(1, "skipping the published lecture %s because no lecture object found.", $lo);
	return ($hash_lo);
    }

    open (LECTURE_XML, "<$lecture_xml") 
	or die "can not open $lecture_xml for reading: $!\n";
    my $slurp_lecture_xml = join('', <LECTURE_XML>);
    close (LECTURE_XML);

    my $metadata_xml = $parameters->{'path_lo'} . $::FS . $lo . $::FS . 'metadata.xml';
    open (METADATA_XML, "<$metadata_xml")
    or die "can not open $metadata_xml for reading: $!\n";
    my $slurp_metadata_xml = join('', <METADATA_XML>);
    close (METADATA_XML);

    ($hash_lo->{'title'}) = $slurp_lecture_xml =~ m!\<lecture\s+title\s*\=\s*\"([^\"]*)\"!isg;
    ($hash_lo->{'creator'}) = $slurp_metadata_xml =~ m!\<creator\>([^\<]*)\<\/creator\>!isg;
    ($hash_lo->{'string_duration'}) = $slurp_metadata_xml =~ m!\<duration\>([^\<]*)\<\/duration\>!isg;
    ($hash_lo->{'string_date'}) = $slurp_metadata_xml =~ m!\<date\>([^\<]*)\<\/date\>!isg;
    ($hash_lo->{'string_time'}) = $slurp_metadata_xml =~ m!\<time\>([^\<]*)\<\/time\>!isg;

    my ($lo_hour, $lo_minute, $lo_second) = $hash_lo->{'string_time'} =~ m!(\d\d)\:(\d\d)\:(\d\d)!;
    my ($lo_year, $lo_month, $lo_day)     = $hash_lo->{'string_date'} =~ m!(\d\d\d\d)\-(\d\d)\-(\d\d)!;

#    &pr(1, "getting time for lo = $lo\n");

# need to actually fix this:
#    $hash_lo->{'time'} = timelocal($lo_second, $lo_minute, $lo_hour, $lo_day, $lo_month - 1, $lo_year);

    my ($wl_dev,$wl_ino,$wl_mode,$wl_nlink,$wl_uid,$wl_gid,$wl_rdev,$wl_size,
                      $wl_atime,$wl_mtime,$wl_ctime,$wl_blksize,$wl_blocks)
		       = stat($parameters->{'path_wl'} . $::FS . $lo . $::FS . 'lecture.xml');
    $hash_lo->{'time_published'} = $wl_mtime;

    &debug(6, "lecture occurred  => " . scalar(localtime( $hash_lo->{'time'} )));
    &debug(6, "lecture published => " . scalar(localtime( $hash_lo->{'time_published'} )));



# Get time lecture was published !!! use stat and look it up under wl

#<SLIDE SRC="slides/20060905114800.jpg" BEGIN="00:00:00" TYPE="image/jpg" REGION="slide" TITLE="00:00:00" />
#<SLIDE SRC="slides/20060905115646.jpg" BEGIN="00:08:46" TYPE="image/jpg" REGION="slide" TITLE="00:08:46" />
#<SLIDE SRC="slides/20060905115650.jpg" BEGIN="00:08:50" TYPE="image/jpg" REGION="slide" TITLE="00:08:50" />

    $hash_lo->{'numslides'} = 0;
    $hash_lo->{'list_slides'} = [];
    my $save_pos = 0;
    while ( $slurp_lecture_xml =~ m!\<slide\s+src\=\"slides\/(.*?)\"\s+begin\=\"([\:\d]+)\".*?\>!isg ) {
#	&debug(2, "before match, pos = " . pos($slurp_lecture_xml) );
#	&debug(2, "after match, pos = " . pos($slurp_lecture_xml) );

	$hash_lo->{'numslides'}++;

#	if ( ! pos($slurp_lecture_xml) < $save_pos) {
#	    last;
#	}
	$save_pos = pos;
	my $hash_slide = {};
	$hash_slide->{'num'} = $hash_lo->{'numslides'};
	$hash_slide->{'src'} = $1;
	$hash_slide->{'string_timestamp'} = $2;
	my ($slide_hour, $slide_minute, $slide_second) = $hash_slide->{'string_timestamp'} =~ m!(\d\d)\:(\d\d)\:(\d\d)!;
	$hash_slide->{'timestamp'} = 3600*$slide_hour + 60*$slide_minute + $slide_second;

	foreach my $key (keys %$hash_slide) {
	    &debug(3, "slide: $key => " . $hash_slide->{$key});
	}

	push( @{$hash_lo->{'list_slides'}}, $hash_slide );
    }


    &debug(3, "title = " . $hash_lo->{'title'});   
    &debug(3, "creator = " . $hash_lo->{'creator'});   


    &debug(5, "exiting get_lo_metadata");

    return $hash_lo;
}

#####################################################################
# find_robots
#
#
#
#####################################################################
sub find_robots
{
    my ($list_all_log_files) = @_;
    my $hash_robots = {};

    &pr(1, "finding robots\n");

    my $numlines = 0;
# really simple: just loop through log files, and do a grep for robots.txt on it, and put those ip's in the list
    foreach my $file (@$list_all_log_files) {
	my $command = "grep \"robots.txt\" $file";
#	my $result = qx ($command);

	open (GREP, "$command |")
	    or die "can not fork process [$command]: $!\n";
	while (my $result_line = <GREP>) {
#	    &pr(1, $result_line);

	    my $robot;

	    if ($::LOGSTYLE eq 'umich') {
		($robot) = $result_line =~ m!^([\d\.]+)!;
	    }
	    elsif ($::LOGSTYLE eq 'cern') {
		$result_line =~ m!^([\d\-]+\s+[\d\:]+)\s+(\S+)\s+([\d\.]+)\s+(\w+)\s+(\S+)\s+\S+\s+\S+\s+\S+\s+([\d\.]+)\s+(\S+)\s+(\S+)!;
		$robot = $6;
	    }

	    $hash_robots->{$robot} = 1;
#	    if (! &member_of($robot, $list_robots) ) {
#		push(@$list_robots, $robot);
#	    }
	}

    }

    &pr(1, "counted " . scalar(keys %$hash_robots) . " robots\n");
    
    return $hash_robots;
}

#####################################################################
# get_first_line
#
# Description:
# Retrieve the first line of a text file
#####################################################################
sub get_first_line
{
    my ($path) = @_;
    my $line;

    &debug(5, "entering get_first_line");

    # Get first non-comment line
    open (FILE, "<$path")
	or die "can not open $path for reading: $!\n";
    do {
	$line = <FILE>;
    } until $line !~ m/^\s*\#/;
    close (FILE);

    &debug(5, "exiting get_first_line");

    return $line;
}

#####################################################################
# get_last_line
#
# Description:
# Retrieve the last line of a text file
#####################################################################
sub get_last_line
{
    my ($path) = @_;
    my $line;

    &debug(5, "entering get_last_line");

    open (FILE, "<$path")
	or die "can not open $path for reading: $!\n";
    seek(FILE, -1000, 2); # go to near the end of the file
    while (my $line_read = <FILE>) {
	if ($line_read =~ m!\[\d.*?\]!) {
	    $line = $line_read;
	}
    }
    close (FILE);

    &debug(5, "exiting get_last_line");

    return $line;
}

#####################################################################
# populate_data
#
#
#####################################################################
sub populate_data
{
    my ($parameters, $list_needed_log_files, $hash_lecture_series, $hash_robots) = @_;
    my $hash_log = {};
    my $hash_summary_stats = {};
    my $hash_hosts = {};

    $hash_summary_stats->{'hits'} = 0;
    $hash_summary_stats->{'hits_grep'} = 0;
    $hash_summary_stats->{'hits_date'} = 0;
    $hash_summary_stats->{'hits_robots'} = 0;

    my $total_line_ctr = 0;
    my $ctr = 0;
    my $list_sessions = [];

# sort needed log files in chronological order
# loop through log files
# read each line and do fast grep (substr) to see if string_grep is present before doing anything.
# read each line and get all fields
# check date to see if this line is within range and needs to be processed
# populate hash
# should I just stick every line into the hash directly, or do something smart to reduce size of hash??

    foreach my $hash_log (@$list_needed_log_files) {
	open (PATH_LOG, '<' . $hash_log->{'path'})
	    or die "can not open " . $hash_log->{'path'} . " for reading: $!\n";
	while (my $line = <PATH_LOG>) {
	    $hash_summary_stats->{'hits'}++;

	    # First only read in lines that contain the grep string
	    if ( index($line, $parameters->{'string_grep'}) != -1 ) {
		
		$hash_summary_stats->{'hits_grep'}++;

		# read in all fields
		if ($::LOGSTYLE eq 'umich' && $line =~ m!^([\d\.]+)\s+(\S+)\s+(\S+)\s+\[([^\]]+)\]\s+\"(.*?)\"\s+(\S+)\s+(\S+)!) {

		    # ignore corrupt line
		    
		    if ( index($line, '27/Au65.54.188.67') != -1 ) {
			next;
		    }
		    
		    
		    my %dummy_hit; # to avoid memory leaks (not sure if it's necessary or works)
		    my $hit = \%dummy_hit;
		    
		    $hit->{'ip'}          = $1;
		    $hit->{'unknown'}     = $2;
		    $hit->{'login'}       = $3;
		    $hit->{'string_date'} = $4;
		    $hit->{'request'}     = $5;
		    $hit->{'unknown2'}    = $6;
		    $hit->{'unknown3'}    = $7;

		    $hit->{'time'} = &convert_date_log_time($hit->{'string_date'});
		    $hit->{'lecture'} = &does_request_contain_lecture($hit->{'request'}, $hash_lecture_series);
		    
		    # Even though we've done a quick and dirty grep to eliminate log entries with grep_string in them,
		    # it is still possible that the search string
		    # could be found in a totally irrelevant hit, in a different part of the log entry (e.g. the conference
		    # number 47356 could happen to match tne number of bytes transferred ... this actually happened to me)
		    # So do not add anything unless does_request_contain_lecture has actually returned a value
		    if ($hit->{'lecture'} ne '') {
			if ( $hit->{'time'} >= $parameters->{'date_begin'} && $hit->{'time'} <= $parameters->{'date_end'} ) {
			    $hash_summary_stats->{'hits_date'}++;
			    
#			&pr(1, "$line");
			    
			    if ( $hit->{'request'} =~ m!^(\S*)\s*(\S+)\s*(\S*)$! ) {
				$hit->{'request_method'} = $1;
				$hit->{'request_path'}   = $2;
				$hit->{'request_agent'}  = $3;
			    }
			    
			    if ( $hash_robots->{ $hit->{'ip'} } ){
				$hash_summary_stats->{'hits_robots'}++;
			    }
			    else {
				$hash_hosts->{ $hit->{'ip'} } = 1;

				my $ptr_session = &does_hit_belong_to_existing_session($hit, $list_sessions, $parameters);
				
				if ($ptr_session) {
				    &update_session($ptr_session, $hit);
#			    &display_session($ptr_session);
				}
				else {
				    my $new_session = &create_new_session($hit, scalar(@$list_sessions));
				    push(@$list_sessions, $new_session);
#			    &display_session($new_session);
				}
			    }
			} # end if block testing for date range
			elsif ( $hit->{'time'} > $parameters->{'date_end'} ) {
			    last;
			}
		    } # end if ($hit->{'lecture'} ne '') {
#		    if ($parameters->{'date_begin'} < $hit->{'time'} && $hash_summary_stats->{'hits'} % 10000 == 0 ) {
#			&pr(1, sprintf("total: %08d, grep: %08d, in range: %08d, sessions: %04d  %s\n", 
#				       $hash_summary_stats->{'hits'}, 
#				       $hash_summary_stats->{'hits_grep'}, 
#				       $hash_summary_stats->{'hits_date'},
#				       scalar(@$list_sessions),
#				       substr($line, 0, 40)));
#		    } # end if for printing display every XXX hits
		    

		} # end if block testing for proper log format
		elsif ($::LOGSTYLE eq 'cern' && $line =~ m!^([\d\-]+\s+[\d\:]+)\s+(\S+)\s+([\d\.]+)\s+(\w+)\s+(\S+)\s+\S+\s+\S+\s+\S+\s+([\d\.]+)\s+(\S+)\s+(\S+)!) {


#		    2009-10-06 13:58:16 W3SVC20 137.138.240.80 GET /MediaArchive/Video/Public2/WebLectures/2009/47256/5.0/thumbs/slides/20091005121554.jpg - 80 - 158.109.1.9 Mozilla/5.0+(Windows;+U;+Windows+NT+5.1;+en-US;+rv:1.9.0.14)+Gecko/2009082707+Firefox/3.0.14+(.NET+CLR+3.5.30729) - 200 0 0 5606

# 2009-08-06 00:00:13 W3SVC20 137.138.240.80 GET /MediaArchive/Photo/Public/2007/0704006/0704006_01/0704006_01-A5-at-72-dpi.jpg - 80 - 70.79.179.31 Mozilla/5.0+(Windows;+U;+Windows+NT+6.0;+en-GB;+rv:1.9.0.13)+Gecko/2009073022+Firefox/3.0.13 http://www.stumbleupon.com/refer.php?url=http%3A%2F%2Fmediaarchive.cern.ch%2FMediaArchive%2FPhoto%2FPublic%2F2007%2F0704006%2F0704006_01%2F0704006_01-A5-at-72-dpi.jpg 200 0 0 40827
		    my %dummy_hit; # to avoid memory leaks (not sure if it's necessary or works)
		    my $hit = \%dummy_hit;
		    
		    $hit->{'string_date'}     = $1;
#		    $hit->{'unknown'}         = $2;
#		    $hit->{'unknown'}         = $3;
		    $hit->{'request_method'}  = $4;
		    $hit->{'request_path'}    = $5;
		    $hit->{'ip'}              = $6;
		    $hit->{'request_agent'}   = $7;
		    $hit->{'referer'}         = $8;

		    $hit->{'time'}    = &convert_date_log_time($hit->{'string_date'});
		    $hit->{'lecture'} = &does_request_contain_lecture($hit->{'request_path'}, $hash_lecture_series);
		    

#print "DATE_BEGIN: " . $parameters->{'date_begin'} . " " . localtime($parameters->{'date_begin'}) . "\n";
#print "DATE_END:   " . $parameters->{'date_end'}   . " " . localtime($parameters->{'date_end'})   . "\n";
#print "TIME:       " . $hit->{'time'}              . " " . localtime($hit->{'time'}            )  . " " . $line . "\n";
		    
		    # Even though we've done a quick and dirty grep to eliminate log entries with grep_string in them,
		    # it is still possible that the search string
		    # could be found in a totally irrelevant hit, in a different part of the log entry (e.g. the conference
		    # number 47356 could happen to match tne number of bytes transferred ... this actually happened to me)
		    # So do not add anything unless does_request_contain_lecture has actually returned a value
		    if ($hit->{'lecture'} ne '') {
			if ($hit->{'time'} >= $parameters->{'date_begin'} && $hit->{'time'} <= $parameters->{'date_end'} ) {
			    
			    $hash_summary_stats->{'hits_date'}++;
			    
#			&pr(1, "$line");
			    
			    if ( $hash_robots->{ $hit->{'ip'} } ){
				$hash_summary_stats->{'hits_robots'}++;
			    }
			    else {
				$hash_hosts->{ $hit->{'ip'} } = 1;
				
				my $ptr_session = &does_hit_belong_to_existing_session($hit, $list_sessions, $parameters);
				
				if ($ptr_session) {
				    &update_session($ptr_session, $hit);
#			    &display_session($ptr_session);
				}
				else {
				    my $new_session = &create_new_session($hit, scalar(@$list_sessions));
				    push(@$list_sessions, $new_session);
#			    &display_session($new_session);
				}
			    }
			} # end if block testing for date range
			elsif ( $hit->{'time'} > $parameters->{'date_end'} ) {
			    last;
			} 
		    } # end if ($hit->{'lecture'} ne '')

		} # end if block testing for proper log format
		else {
		    &pr(1, "Sorry, this script doesn't know how to read logfiles in this format yet.\n");
		    &pr(1, $line);
		    exit(0);
		} # end if..else testing for proper log format

	    } # end if testing for string_grep

	    if ($hash_summary_stats->{'hits'} % 10000 == 0 ) {
		&pr(1, sprintf("total: %08d, grep: %08d, in range: %08d, robots: %08d, sessions: %04d  %s\n", 
			       $hash_summary_stats->{'hits'},
			       $hash_summary_stats->{'hits_grep'}, 
			       $hash_summary_stats->{'hits_date'},
			       $hash_summary_stats->{'hits_robots'},
			       scalar(@$list_sessions),
			       substr($line, 0, 40)));
	    } # end if for printing display every XXX hits

	} # end while loop for reading each line
	close (PATH_LOG);
    } # end for loop looping through each log file

#    foreach my $session (@$list_sessions) {
#	&display_session($session);
#    }


    &pr(1, "Total lines read:   " . $hash_summary_stats->{'hits'} . "\n");
    &pr(1, "Hits of interest:   " . $hash_summary_stats->{'hits_grep'} . "\n");
    &pr(1, "Hits in date range: " . $hash_summary_stats->{'hits_date'} . "\n");
    &pr(1, "Hits of interest by robots:     " . $hash_summary_stats->{'hits_robots'} . "\n");
    &pr(1, "Total sessions:     " . scalar(@$list_sessions) . "\n");


    return ($list_sessions, $hash_hosts);
}



#####################################################################
# display_summary
#
#
#####################################################################
sub display_summary
{
    my ($list_sessions, $hash_lecture_series, $parameters) = @_;
    my $tally_sessions = {};
    my $list_lecture_series = [];
    
	
    foreach my $duration_session (@{$parameters->{'list_duration_session'}}) {
	$tally_sessions->{$duration_session} = 0;
    }

    foreach my $session (@$list_sessions) {
	foreach my $duration_session (@{$parameters->{'list_duration_session'}}) {
	    if ($session->{'time_end'} - $session->{'time_begin'} >= $duration_session) {
		$tally_sessions->{$duration_session}++;
	    }
	}
    }

    foreach my $duration_session ( @{$parameters->{'list_duration_session'}} ) {
	&pr(1, "sessions longer than $duration_session: " . $tally_sessions->{$duration_session} . "\n");
    }

    @$list_lecture_series = keys %$hash_lecture_series;
# do something here...



}

#####################################################################
# make_graph_time_vs_session
#
# Arguments:
# $list_sessions
# $bin_width: bin width in seconds
#
# $hash_time_vs_sessions->{$bin}->{$duration_value} = $counts
#####################################################################
sub make_graph_time_vs_sessions
{
    my ($parameters, $list_sessions, $bin_width) = @_;
    my %dummy_hash;
    my $hash_time_vs_sessions = \%dummy_hash; # to avoid memory leaks

    # Initialize hash with 0 values
    my $hash_time_vs_sessions = {};
    for(my $ctr_time = int( $parameters->{'date_begin'} / $bin_width); 
	$ctr_time <= int( $parameters->{'date_end'} / $bin_width); 
	$ctr_time++) {
	foreach my $duration_value (@{$parameters->{'list_duration_session'}}) {
	    $hash_time_vs_sessions->{$ctr_time*$bin_width}->{$duration_value} = 0;
	}
    }

    # Loop through each session
    foreach my $session (@$list_sessions) {

#	&pr(1, sprintf("Session #%04d. duration: %04d seconds, bin width: %08d\n", 
#		       $session->{'number'}, 
#		       $session->{'time_end'} - $session->{'time_begin'},
#		       $bin_width));
#	&pr(1, "duration of current session is " . $session->{'time_end'} . ' - ' . $session->{'time_begin'} . "\n");

	# make 2-d array here, looping through array of given duration values
	# I should only count each session once.
#	for(my $ctr_time = int($session->{'time_begin'} / $bin_width);
#	    $ctr_time <= int($session->{'time_end'} / $bin_width);
#	    $ctr_time++) {

	    foreach my $duration_value (@{$parameters->{'list_duration_session'}}) {
		if ($session->{'time_end'} - $session->{'time_begin'} >= $duration_value) {
#		    $hash_time_vs_sessions->{$ctr_time*$bin_width}->{$duration_value}++;
		    $hash_time_vs_sessions->{int($session->{'time_begin'} / $bin_width) * $bin_width}->{$duration_value}++;

#		    &pr(1, sprintf("For time %s, duration_value: %08d\n",
#				   scalar(localtime(int($session->{'time_begin'} / $bin_width) * $bin_width)),
#				   $duration_value) );
		}
	    }

#	}
    }

    # write information to file
    my $nice_string_grep = $parameters->{'string_grep'};
    $nice_string_grep =~ s!\/!_!g;

    my $dir_path_output = $parameters->{'dir_output'} . 
	$::FS . $nice_string_grep;
    if (! -d $dir_path_output) {
	mkdir($dir_path_output);
    }
    my $path_output = $dir_path_output . $::FS  .
	sprintf("%s_%s_%06d_%s.csv", $nice_string_grep, &nice_timestamp(), "time_sessions", $bin_width);

    open (FILE_OUT, ">$path_output")
	or die "can not open $path_output for writing: $!\n";

    # print column names
    print FILE_OUT "date,   time";
    foreach my $duration_value (@{$parameters->{'list_duration_session'}}) {
	print FILE_OUT ", " . $duration_value;
    }
    print FILE_OUT "\n";

    # print table of values
    foreach my $bin (sort keys %$hash_time_vs_sessions) {

	my ($tsec,$tmin,$thour,$tmday,$tmon,$tyear,$twday,$tyday,$tisdst) = localtime($bin);
	print FILE_OUT sprintf("%04d-%02d-%02d, %02d:%02d:%02d", $tyear+1900, $tmon+1, $tmday, $thour, $tmin, $tsec);
#	print sprintf("%04d-%02d-%02d %02d:%02d:%02d", $tyear+1900, $tmon+1, $tmday, $thour, $tmin, $tsec);
	foreach my $duration_value (@{$parameters->{'list_duration_session'}}) {
	    print FILE_OUT ", " . $hash_time_vs_sessions->{$bin}->{$duration_value};
#	    print " " . $hash_time_vs_sessions->{$bin}->{$duration_value};
	}
	print FILE_OUT "\n";
#	print "\n";
    }
    close (FILE_OUT);

}

#####################################################################
# make_report_ip_list
#
# this needs serious work
# list of ip's containing hashes... $hash_ip_info
#   number of sessions
#   number of lectures viewed
#   list of lectures
#     lecture id
#     total duration of all sessions for this lecture
#     
# $hash_ip = { 
#              $ip => {  
#                       'list_durations' => { 
#                                             $lecture1 => $duration1,
#                                             $lecture2 => $duration2,
#                                              ...
#                                           },
#                       'num_sessions' = $ctr_s,
#                       'num_lectures' = $ctr_l,
#                       'duration'     = $duration1 + $duration2 + ...
#                     },
#              $ip2 => ...,
#              ,,,
#            };
#
#
#####################################################################
sub make_report_ip_list
{
    my ($parameters, $list_sessions, $view) = @_;
    my $hash_ip = {};

    # initialize
    foreach my $session (@$list_sessions) {
	$hash_ip->{$session->{'ip'}} = { 'list_durations' => {},
					 'num_sessions'   => 0,
					 'num_lectures'   => 0,
					 'duration'       => 0 };
    }

    foreach my $session (@$list_sessions) {
	my $ip = $session->{'ip'};
	my $duration = int($session->{'time_end'} - $session->{'time_begin'});

	$hash_ip->{$ip}->{'num_sessions'}++;

#	$hash_ip->{$ip}->{'list_durations'} = {};

	if ($hash_ip->{$ip}->{'list_durations'}->{$session->{'lecture'}}) {
	    $hash_ip->{$ip}->{'list_durations'}->{$session->{'lecture'}} 
	    += $duration;
	    $hash_ip->{$ip}->{'duration'} += $duration;
	}
	else {
	    $hash_ip->{$ip}->{'list_durations'}->{$session->{'lecture'}}
	    = $duration;
	    $hash_ip->{$ip}->{'duration'} += $duration;
	    $hash_ip->{$ip}->{'num_lectures'}++;
	}
    }

####################################################
# Uncomment following section to display details on every IP 
#
#    foreach my $ip (keys %{$hash_ip}) {
#	&pr(1, sprintf("%-15s viewed %-3d lectures in %-3d sessions for over %-5d seconds", 
#		       $ip, 
#		       $hash_ip->{$ip}->{'num_lectures'}, 
#		       $hash_ip->{$ip}->{'num_sessions'},
#		       $hash_ip->{$ip}->{'duration'}));
#	if ($view eq 'detail') {

#	    foreach my $lecture (keys %{$hash_ip->{$ip}->{'list_durations'}} ) {
#		my ($lecture_short) = $lecture =~ m!\/([^\/]+)$!;
#		&pr(1, sprintf("\n %-30s %-5d", $lecture_short, $hash_ip->{$ip}->{'list_durations'}->{$lecture}));
#	    }
#	}
	
#	&pr(1, "\n");
#    }

# which lectures did they look at
# how long did they look at each one
# how long did they look at stuff total

}

#####################################################################
#
#
#
#####################################################################
sub make_report_host_list
{
    my ($parameters, $hash_hosts) = @_;
    
    # write information to file
    my $nice_string_grep = $parameters->{'string_grep'};
    $nice_string_grep =~ s!\/!_!g;

    my $dir_path_output = $parameters->{'dir_output'} . 
	$::FS . $nice_string_grep;
    if (! -d $dir_path_output) {
	mkdir($dir_path_output);
    }
    my $path_output = $dir_path_output . $::FS  .
	sprintf("%s_%s_%s.csv", $nice_string_grep, &nice_timestamp(), "ip_list");

    open (FILE_OUT, ">$path_output")
	or die "can not open $path_output for writing: $!\n";

    foreach my $ip (sort keys %{$hash_hosts}) {
	my $result = `host $ip`;
	$result =~ m!\s+((\S+)\.([^\.]+)\.([^\.]+))\.$!;
	print FILE_OUT "$ip, $1, $3, $4\n";
	#133.209.247.190.in-addr.arpa domain name pointer 133-209-247-190.fibertel.com.ar.
    }

    close (FILE_OUT);
    
}

#####################################################################
# make_report_most_popular
#
#
#####################################################################
sub make_report_most_popular
{
    my ($parameters, $list_sessions, $view) = @_;
    my $hash_lecture = {};
   
    # initialize
    foreach my $session (@$list_sessions) {
	$hash_lecture->{$session->{'lecture'}} = { 'creator' => '',
						   'num_sessions'     => 0,
						   'num_lectures'     => 0,
						   'num_ips'          => 0,
						   'viewing_duration' => 0,
	                                           'list_ips' => []        };
    }

    foreach my $session (@$list_sessions) {
	my $lecture = $session->{'lecture'};

	if ($hash_lecture->{$lecture}->{'creator'} eq '') {
	    my $hash_lo = &get_lo_metadata($lecture, $parameters);
	    $hash_lecture->{$lecture}->{'creator'} = $hash_lo->{'creator'};
	}

	my $duration = int($session->{'time_end'} - $session->{'time_begin'});

	$hash_lecture->{$lecture}->{'num_sessions'}++;
	$hash_lecture->{$lecture}->{'viewing_duration'} += $duration;

	# if the current session IP is not already in the list of IPs for this lecture,
	# then add it
	if (! &member_of($session->{'ip'}, $hash_lecture->{$lecture}->{'list_ips'})) {
	    push(@{$hash_lecture->{$lecture}->{'list_ips'}}, $session->{'ip'});
	    $hash_lecture->{$lecture}->{'num_ips'}++;
	}
    }

#	if ($hash_lecture->{$lecture}->{'list_ips'}->{$session->{'ip'}}) {
#	    $hash_ip->{$ip}->{'list_durations'}->{$session->{'lecture'}} 
#	    += $duration;
#	    $hash_ip->{$ip}->{'duration'} += $duration;
#	}


#	if ($hash_ip->{$ip}->{'list_durations'}->{$session->{'lecture'}}) {
#	    $hash_ip->{$ip}->{'list_durations'}->{$session->{'lecture'}} 
#	    += $duration;
#	    $hash_ip->{$ip}->{'duration'} += $duration;
#	}
#	else {
#	    $hash_ip->{$ip}->{'list_durations'}->{$session->{'lecture'}}
#	    = $duration;
#	    $hash_ip->{$ip}->{'duration'} += $duration;
#	    $hash_ip->{$ip}->{'num_lectures'}++;
#	}
#    }


    # write information to file
    my $nice_string_grep = $parameters->{'string_grep'};
    $nice_string_grep =~ s!\/!_!g;

    my $dir_path_output = $parameters->{'dir_output'} . 
	$::FS . $nice_string_grep;
    if (! -d $dir_path_output) {
	mkdir($dir_path_output);
    }
    my $path_output = $dir_path_output . $::FS  .
	sprintf("%s_%s_%s.csv", $nice_string_grep, &nice_timestamp(), "lo_list");

    open (FILE_OUT, ">$path_output")
	or die "can not open $path_output for writing: $!\n";

    # print column headings
    print FILE_OUT "lecture ID, speaker, num IPs, num sessions, duration at least (sec)\n";

    foreach my $lecture (sort keys %{$hash_lecture}) {
	print FILE_OUT sprintf("%-20s, %-20s, %-3d, %-3d, %-5d\n", 
			       $lecture,
			       $hash_lecture->{$lecture}->{'creator'}, 
			       $hash_lecture->{$lecture}->{'num_ips'}, 
			       $hash_lecture->{$lecture}->{'num_sessions'},
			       $hash_lecture->{$lecture}->{'viewing_duration'});
	if ($view eq 'detail') {

	    foreach my $ip (keys %{$hash_lecture->{$lecture}->{'list_ips'}} ) {
#		my ($lecture_short) = $lecture =~ m!\/([^\/]+)$!;
#		&pr(1, sprintf("\n %-30s %-5d", $lecture_short, $hash_ip->{$ip}->{'list_durations'}->{$lecture}));
		print FILE_OUT sprintf("\n %-30s %-5d", $lecture, $hash_lecture->{$lecture}->{'list_ips'}->{$ip});
	    }
	}
	
    }

    close (FILE_OUT);

# which lectures did they look at
# how long did they look at each one
# how long did they look at stuff total

}

#####################################################################
# make_report_for_education
#####################################################################
sub make_report_for_education
{
    my ($parameters, $list_sessions, $logname) = @_;

    my ($csec,$cmin,$chour,$cmday,$cmon,$cyear,$cwday,$cyday,$cisdst) = localtime(time);
    my $currdate = sprintf("%04d%02d%02d", 1900+$cyear, 1+$cmon, $cmday);

    my $file = "/home/umacp/usage-stats/reports/${logname}_${currdate}.txt";

    my $ctr = {};

    open (FILE, ">$file")
	or "can no open $file for writing: $!\n";

    print "$file opened, looping through sessions...\n";

    print FILE "ip,uniqname,date,time,course,lecture id,requested file,session id\n";
    foreach my $session (@$list_sessions) {
	my ($course, $lecture) = $session->{'lecture'} =~ m!200\d\-\w\/([^\/]+)\/([^\/]+)!;
	$ctr->{$lecture}->{$session->{'login'}}++;
	print "course = [$course], lecture = [$lecture]\n";
	foreach my $session_hit (@{$session->{'time_vs_request'}}) {
	    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) =
		localtime($session_hit->{'time'});

	    my $datestamp = sprintf("%04d-%02d-%02d", 1900+$year, 1+$mon, $mday);
	    my $timestamp = sprintf("%02d:%02d:%02d", $hour, $min, $sec);

	    my $session_id = sprintf("%s_%s_%03d", $session->{'login'}, 
				     $lecture,
				     $ctr->{$lecture}->{$session->{'login'}});

	    print FILE $session->{'ip'}
	        . ',' . $session->{'login'}
		. ',' . $datestamp
		. ',' . $timestamp
	        . ',' . $course
		. ',' . $lecture
		. ',' . $session_hit->{'file'}
	        . ',' . $session_id
	        . "\n";
	}
    }

    close (FILE);

}


#####################################################################
# make_report_lecture_list
#
#     
# 
#####################################################################
sub make_report_lecture_list
{

}

#####################################################################
# does_hit_belong_to_existing_session
#
# parameters:
# date_begin
# date_end
# string_grep
# durations_visit
# durations_session
# timeout_visit
# timeout_session
# path_lo
# path_wl
# 
# list of keys in hit hash:
# ip
# login
# unknown
# string_date
# request
# unknown2
# unknown3
# time
# request
# request_method
# request_path
# request_agent
# file_requested
#
# structure of session hash:
# number
# login
# ip
# lecture
# time_begin
# time_end
# duration?
# time_vs_request
# time_vs_slidenum
#
# structure of hash_lecture_series:
# 
# 
#
#
#####################################################################
sub does_hit_belong_to_existing_session
{
    my ($hit, $list_sessions, $parameters) = @_;
    my $ptr_session;

# somewhere check for presence of LO_ID/fXXX.htm to verify the beginning of a session

    &debug(5, "entering does_hit_belong_to_existing_session");

    my $flag_session_found = 0;

    my $lecture = $hit->{'lecture'};

#print "testing to see if session exists...\n";
#print "lecture = $lecture\n";
#print "request_path = " . $hit->{'request_path'} . "\n";

    if ( $lecture ne '' || index($hit->{'request_path'}, 'feedback') > -1 )  {
	($hit->{'file_requested'}) = $hit->{'request_path'} =~ m!$lecture\/(.*)$!;

	my $numsessions = scalar(@$list_sessions);
	
	# loop through all sessions (I can probably make this more efficient)
	for(my $ctr = 0; $ctr < $numsessions; $ctr++) {
	    # if the ip of the hit matches the current session,
	    # and either the lecture of the hit matches the lecture of the current session,
	    # or the hit is the feedback form, then we are close
	    if ($list_sessions->[$ctr]->{'ip'} eq $hit->{'ip'}
		&& ( $list_sessions->[$ctr]->{'lecture'} eq $lecture
		     || index($hit->{'request_path'}, 'feedback') > -1 ) ) {

		&debug(3, "ip definitely matches and probably lecture does too.");
		&debug(3, "next we'll test the time.");
		&debug(3, "hit time: " . $hit->{'time'});
		&debug(3, "curr session time_begin: " . $list_sessions->[$ctr]->{'time_begin'});
		&debug(3, "curr session time_end: " . $list_sessions->[$ctr]->{'time_end'});
		# Now compare the time of the hit to the current session.
		# We're looping forward through all sessions, time is increasing.
		# So if the time of the hit happens before the start of the current session,
		# then we've already passed it, and it's time to stop looping (hit doesn't belong in any session).
		# If the time of the hit happens within the timeout duration of the current session's
		# last hit, then the hit belongs to the current session.
		# If the time of the hit happens after the timeout duration of the current session's last hit,
		# then we need to keep looping through the sessions (do nothing).
		&debug(3, "TEST: " . $hit->{'time'} . ' < ' . $list_sessions->[$ctr]->{'time_begin'});
		if ( $hit->{'time'} > $list_sessions->[$ctr]->{'time_begin'} - $parameters->{'timeout_session'}
		     && $hit->{'time'} < $list_sessions->[$ctr]->{'time_end'} + $parameters->{'timeout_session'} ) {
		    $ptr_session = $list_sessions->[$ctr];
		    $flag_session_found = 1;
		    last;
		}
		else {
		    &debug(3, "TEST: " . $hit->{'time'} . ' < ' . $list_sessions->[$ctr]->{'time_end'} . ' + ' . $parameters->{'timeout_session'} );
		    &debug(3, "continue looping (ctr = $ctr)");
		    # do nothing
		}
	    } # end if testing if ip, time and lecture match

#	&pr(1, "looping through $ctr of $numsessions sessions\n");
	}
	
    }
    else {
	# I don't know how this could happen
#	&pr(1, "the request does not correspond to an existing lecture\n");
#	foreach my $key (sort keys %$hit) {
#	    &pr(1, "hit: $key -> " . $hit->{$key} . "\n");
#	}
    }

    &debug(5, "exiting does_hit_belong_to_existing_session");

    if ($flag_session_found) {
	return $ptr_session;
    }
    else {
	return 0;
    }

}

#####################################################################
# Now loop through session backwards
#####################################################################
sub does_hit_belong_to_existing_session2
{
    my ($hit, $list_sessions, $parameters) = @_;
    my $ptr_session;

# somewhere check for presence of LO_ID/fXXX.htm to verify the beginning of a session

    &debug(5, "entering does_hit_belong_to_existing_session");

    my $flag_session_found = 0;

    my $lecture = $hit->{'lecture'};

    if ( $lecture ne '' || index($hit->{'request_path'}, 'feedback') > -1 )  {
	($hit->{'file_requested'}) = $hit->{'request_path'} =~ m!$lecture\/(.*)$!;

	my $numsessions = scalar(@$list_sessions);
	
	# loop through all sessions (I can probably make this more efficient)
	for(my $ctr = $numsessions-1; $ctr >= 0; $ctr--) {
	    # if the ip of the hit matches the current session,
	    # and either the lecture of the hit matches the lecture of the current session,
	    # or the hit is the feedback form, then we are close
	    if ($list_sessions->[$ctr]->{'ip'} eq $hit->{'ip'}
		&& ( $list_sessions->[$ctr]->{'lecture'} eq $lecture
		     || index($hit->{'request_path'}, 'feedback') > -1 ) ) {

		&debug(3, "ip definitely matches and probably lecture does too.");
		&debug(3, "next we'll test the time.");
		&debug(3, "hit time: " . $hit->{'time'});
		&debug(3, "curr session time_begin: " . $list_sessions->[$ctr]->{'time_begin'});
		&debug(3, "curr session time_end: " . $list_sessions->[$ctr]->{'time_end'});
		# Now compare the time of the hit to the current session.
		# We're looping forward through all sessions, time is increasing.
		# So if the time of the hit happens before the start of the current session,
		# then we've already passed it, and it's time to stop looping (hit doesn't belong in any session).
		# If the time of the hit happens within the timeout duration of the current session's
		# last hit, then the hit belongs to the current session.
		# If the time of the hit happens after the timeout duration of the current session's last hit,
		# then we need to keep looping through the sessions (do nothing).
		&debug(3, "TEST: " . $hit->{'time'} . ' < ' . $list_sessions->[$ctr]->{'time_begin'});
		if ( $hit->{'time'} < $list_sessions->[$ctr]->{'time_begin'} ) {
		    &debug(3, "TEST: " . $hit->{'time'} . ' < ' . $list_sessions->[$ctr]->{'time_end'} . ' + ' . $parameters->{'timeout_session'} );
		    &debug(3, "continue looping (ctr = $ctr)");
		    # do nothing
		}
		elsif ( $hit->{'time'} < $list_sessions->[$ctr]->{'time_end'} + $parameters->{'timeout_session'} ) {
		    $ptr_session = $list_sessions->[$ctr];
		    $flag_session_found = 1;
		    last;
		}
		else {
		    last;
		}
	    } # end if testing if ip, time and lecture match

#	&pr(1, "looping through $ctr of $numsessions sessions\n");
	}
	
    }
    else {
	# I don't know how this could happen
#	&pr(1, "the request does not correspond to an existing lecture\n");
#	foreach my $key (sort keys %$hit) {
#	    &pr(1, "hit: $key -> " . $hit->{$key} . "\n");
#	}
    }

    &debug(5, "exiting does_hit_belong_to_existing_session");

    if ($flag_session_found) {
	return $ptr_session;
    }
    else {
	return 0;
    }

}

# (C)
# Session #263.  195.113.228.155 (-) viewing [/atlas/meetings/2005/rome/20050610-umwlap002-17-wilson]  at Tue Nov 14 05:38:24 2006 (89 seconds)
# Session #265.  66.249.72.199 (-) viewing [/atlas/meetings/2005/rome/20050610-umwlap002-17-wilson]  at Tue Nov 14 16:35:55 2006 (4609 seconds)
# Session #267.  74.6.71.92 (-) viewing [/atlas/meetings/2005/rome/20050610-umwlap002-17-wilson]  at Wed Nov 15 14:22:08 2006 (17 seconds)


#####################################################################
# does_request_contain_lecture
#
# This is supposed to search through the list of known lectures and
# see if any of them are contained in the given request path,
# and return the lecture object ID.
# For CERN this is currently kluged to just check that ANY lecture
# is in the request path (because I don't have easy access to the media archive on this machine at the moment)
# and return CERN's "lecture ID" which is conferenceID/contributionID
#####################################################################
sub does_request_contain_lecture
{
    my ($request_path, $hash_lecture_series) = @_;
    my $matched_lecture = '';

    if ($::LOGSTYLE eq 'umich') {
	foreach my $series (keys %$hash_lecture_series) {
	    foreach my $hash_lecture ( @{$hash_lecture_series->{$series}} ) {
		my $lecture = $hash_lecture->{'id'};
		if ($request_path =~ m!$lecture!) {
		    return $lecture;
		}
	    }
	}
    }
    # There are multiple possibilities here because I have done three different style of LO names
    elsif ($::LOGSTYLE eq 'cern') {
	# 1) Michigan style: <YEAR>/<conferenceID>/<MichiganStyleID>
	if ($request_path =~ m!Public2\/WebLectures\/\d+\/(\d+\/[\d\w\-]+)\/!) {
	    $matched_lecture = $1;
	}
	# 2) briefly used (for Oct 2009 ATLAS Week) style <YEAR>/<conferenceID>/<contributionID>.<subcontributionID>
	elsif ($request_path =~ m!Public2\/WebLectures\/\d+\/(\d+\/[\d\.]+)\/!) {
	    $matched_lecture = $1;
	}
	# 3) agreed on CERN/Indico style: <YEAR>/<conferenceID>c<contributionID> or <YEAR>/<conferenceID>s<sessionID> or <YEAR>/<conferenceID>c<contributionID>sc<subcontributionID>
	elsif ($request_path =~ m!Public2\/WebLectures\/\d+\/([sc\d]+)\/!) {
	    $matched_lecture = $1;
	}
    }

#    if ($matched_lecture eq '') {
#	print "empty lecture request path: $request_path\n";
#    }

    return $matched_lecture;
}

#####################################################################
# create_new_session
#
#
#####################################################################
sub create_new_session
{
    my ($hit, $session_number) = @_;
    my $hash_session = {};

    &debug(5, "entering create_new_session");

# structure of session hash:
# number
# login
# ip
# lecture
# time_begin
# time_end
# duration?
# time_vs_request: array of hashes { 'time' => internal_time, 'file' => file_requested }
# time_vs_slide: array of hashes { 'time' => internal_time, 'num' => slide number }
#
# list of keys in hit hash:
# ip
# login
# unknown
# string_date
# request
# unknown2
# unknown3
# time
# request
# request_method
# request_path
# request_agent
# file_requested

    &debug(5, "lecture: " . $hit->{'lecture'});

    $hash_session->{'number'}          = $session_number;
    $hash_session->{'ip'}              = $hit->{'ip'};
    $hash_session->{'login'}           = $hit->{'login'};
    $hash_session->{'lecture'}         = $hit->{'lecture'};
    $hash_session->{'time_begin'}      = $hit->{'time'};
    $hash_session->{'time_end'}        = $hit->{'time'};
    $hash_session->{'time_vs_request'} = [];
    $hash_session->{'time_vs_slide'}   = [];

    my $hash_time_file = {};
    $hash_time_file->{'time'} = $hit->{'time'};
    $hash_time_file->{'file'} = $hit->{'file_requested'};

    push(@{$hash_session->{'time_vs_request'}}, $hash_time_file);

    if ($hit->{'file_requested'} =~ m!sld(\d\d\d)\.htm!) {
	my $slidenum = int($1);
	my $hash_time_slide = {};
	$hash_time_slide->{'time'} = $hit->{'time'};
	$hash_time_slide->{'slide'} = $slidenum;
	push(@{$hash_session->{'time_vs_slide'}}, $hash_time_slide);
    }

    &debug(5, "exiting create_new_session");

    return $hash_session;
}

#####################################################################
# update_session
#
#
#####################################################################
sub update_session
{
    my ($hash_session, $hit) = @_;

# structure of session hash:
# ip
# lecture
# time_begin
# time_end
# duration?
# time_vs_request: array of hashes { 'time' => internal_time, 'file' => file_requested }
# time_vs_slidenum: array of hashes { 'time' => internal_time, 'slide' => slide number }
#
# list of keys in hit hash:
# ip
# login
# unknown
# string_date
# request
# unknown2
# unknown3
# time
# request
# request_method
# request_path
# request_agent
# file_requested

    if ($hit->{'login'} ne '-') {
	$hash_session->{'login'} = $hit->{'login'};
    }

    # The web log entries are NOT ALWAYS IN CHRONOLOGICAL ORDER!!!!!
    if ($hit->{'time'} > $hash_session->{'time_end'} ) {
	$hash_session->{'time_end'}   = $hit->{'time'};
    }
    elsif ($hit->{'time'} < $hash_session->{'time_begin'} ) {
	$hash_session->{'time_begin'}   = $hit->{'time'};
    }


    my %hash_time_file = ( 'time' => $hit->{'time'},
			   'file' => $hit->{'file_requested'} );

    push(@{$hash_session->{'time_vs_request'}}, \%hash_time_file);

    if ($hit->{'file_requested'} =~ m!sld(\d\d\d)\.htm!) {
	my $slidenum = $1;
	my $hash_time_slide = {};
	$hash_time_slide->{'time'} = $hit->{'time'};
	$hash_time_slide->{'slide'} = $slidenum;
	push(@{$hash_session->{'time_vs_slide'}}, $hash_time_slide);
    }

}

#####################################################################
# display_session
#
#
#####################################################################
sub display_session
{
    my ($session) = @_;

# structure of session hash:
# number
# ip
# lecture
# time_begin
# time_end
# duration?
# time_vs_request: array of hashes { 'time' => internal_time, 'file' => file_requested }
# time_vs_slide: array of hashes { 'time' => internal_time, 'num' => slide number }

# display: times on separate line, duration, uniqname
    &pr(1, "Session #" . $session->{'number'} . ".  " . $session->{'ip'} . " (" . $session->{'login'} . ") viewing [" . $session->{'lecture'} . "]\n");
	&pr(1, "  from " . localtime($session->{'time_begin'}) . " until " . localtime($session->{'time_end'}) 
	    . " (" . ($session->{'time_end'} - $session->{'time_begin'}) . " seconds)\n");
    
    &pr(1, "List of files requested: \n");
    my $rctr = 1;
    foreach my $request (@{$session->{'time_vs_request'}}) {
	&pr(1, "   " . $rctr++ . ': ' . localtime($request->{'time'}) . "  " 
	    . $request->{'file'} . "\n");
    }
    &pr(1, "List of slides requested: \n");
    my $sctr = 1;
    foreach my $request (@{$session->{'time_vs_slide'}}) {
	&pr(1, "   " . $sctr++ . ': ' . localtime($request->{'time'}) . "  " 
	    . $request->{'slide'} . "\n");
    }
    &pr(1, "----------------------------------\n");

}

#####################################################################
# display_session
#
#
#####################################################################
sub display_session_brief
{
    my ($session) = @_;

# structure of session hash:
# number
# ip
# lecture
# time_begin
# time_end
# duration?
# time_vs_request: array of hashes { 'time' => internal_time, 'file' => file_requested }
# time_vs_slide: array of hashes { 'time' => internal_time, 'num' => slide number }

# display: times on separate line, duration, uniqname
    &pr(1, "Session #" . $session->{'number'} . ".  " . $session->{'ip'} . " (" . $session->{'login'} . ") viewing [" . $session->{'lecture'} . "]");
	&pr(1, "  at " . localtime($session->{'time_begin'}) 
	    . " (" . ($session->{'time_end'} - $session->{'time_begin'}) . " seconds)\n");
    
#    &pr(1, "List of files requested: \n");
#    my $rctr = 1;
#    foreach my $request (@{$session->{'time_vs_request'}}) {
#	&pr(1, "   " . $rctr++ . ': ' . localtime($request->{'time'}) . "  " 
#	    . $request->{'file'} . "\n");
#    }
#    &pr(1, "List of slides requested: \n");
#    my $sctr = 1;
#    foreach my $request (@{$session->{'time_vs_slide'}}) {
#	&pr(1, "   " . $sctr++ . ': ' . localtime($request->{'time'}) . "  " 
#	    . $request->{'slide'} . "\n");
#    }
#    &pr(1, "----------------------------------\n");

}

#####################################################################
# nice_timestamp
#
#
#####################################################################
sub nice_timestamp
{

    my ($tsec,$tmin,$thour,$tmday,$tmon,$tyear,$twday,$tyday,$tisdst) = localtime();
    
    return(sprintf("%04d%02d%02d_%02d%02d%02d", $tyear, $tmon, $tmday, $thour, $tmin, $tsec));
}

#####################################################################
# convert_YMD_time
#
#
#####################################################################
sub convert_YMD_time
{
    my ($string_YMD) = @_;

    &debug(5, "entering convert_YMD_time with argument $string_YMD");

    my ($year, $month, $day) = $string_YMD =~ m/^(\d\d\d\d)\-(\d\d)\-(\d\d)$/;
    my $time_YMD = timelocal(0, 0, 0, $day, $month-1, $year);



# total: 03830000, grep: 00419685, in range: 00060936, robots: 00049157, sessions: 3179  24.248.222.80 - - [09/Aug/2006:13:2Day '' out of range 1..31 at /home/herrj/bin/usage_reports.pl line 1233
# 9:46
# total: 03840000, grep: 00420577, in range: 00061828, robots: 00049958, sessions: 3244  66.30.15.18 - - [10/Aug/2006:03:01:17 -0
# total: 03850000, grep: 00421601, in range: 00062852, robots: 00050792, sessions: 3328  72.30.132.209 - - [10/Aug/2006:18:32:23
# total: 03860000, grep: 00422437, in range: 00063688, robots: 00051474, sessions: 3404  66.249.65.138 - - [11/Aug/2006:14:17:38
# total: 03870000, grep: 00423378, in range: 00064629, robots: 00052195, sessions: 3523  65.54.188.68 - - [12/Aug/2006:17:16:20 -
# total: 03880000, grep: 00424241, in range: 00065492, robots: 00052918, sessions: 3622  64.4.8.112 - - [14/Aug/2006:00:08:51 -04
# total: 03890000, grep: 00425050, in range: 00066301, robots: 00053326, sessions: 3762  202.113.7.177 - - [15/Aug/2006:02:06:34
# total: 03900000, grep: 00425589, in range: 00066840, robots: 00053496, sessions: 3961  137.138.164.234 - - [16/Aug/2006:07:56:1
# total: 03910000, grep: 00426356, in range: 00067607, robots: 00053924, sessions: 4115  131.154.12.56 - - [17/Aug/2006:09:43:33
# total: 03920000, grep: 00427279, in range: 00068530, robots: 00054579, sessions: 4212  65.54.188.68 - - [18/Aug/2006:14:30:33 -
# total: 03930000, grep: 00428642, in range: 00069893, robots: 00055835, sessions: 4266  66.249.65.201 - - [19/Aug/2006:11:17:12
# total: 03940000, grep: 00429118, in range: 00070369, robots: 00056086, sessions: 4396  206.116.6.111 - - [20/Aug/2006:21:08:28
# total: 03950000, grep: 00429720, in range: 00070971, robots: 00056480, sessions: 4502  65.54.188.68 - - [22/Aug/2006:02:39:31 -
# total: 03960000, grep: 00430294, in range: 00071545, robots: 00056895, sessions: 4577  66.249.66.242 - - [23/Aug/2006:03:19:57
# total: 03970000, grep: 00431341, in range: 00072592, robots: 00057793, sessions: 4637  164.58.68.3 - - [24/Aug/2006:09:27:17 -0
# total: 03980000, grep: 00432159, in range: 00073410, robots: 00058289, sessions: 4684  66.249.65.202 - - [25/Aug/2006:17:28:37
# total: 03990000, grep: 00433918, in range: 00075169, robots: 00059970, sessions: 4705  64.4.8.112 - - [26/Aug/2006:22:59:21 -04
# bad date? ... [27/Au65.54.188.67 - - [27/Aug/2006:12:30:50 -0400]






#    = timelocal($hash_lo_schedule->{'SS'},
#		$hash_lo_schedule->{'MM'},
#		$hash_lo_schedule->{'HH'}, 
#		$hash_lo_schedule->{'day'}, 
#		$hash_lo_schedule->{'month'} - 1, 
#		$hash_lo_schedule->{'year'});

    &debug(5, "exiting convert_YMD_time");

    return $time_YMD;
}

#####################################################################
# convert_date_log_time
#
# Description:
# convert stuff that looks like this (umich):
# 11/Nov/2006:14:45:35 -0500
# into an internal time stamp
# 
# or like this (cern):
# 2009-07-06 13:52:56
#
#
# Arguments:
# $string_date
#
# Returns:
# $time: internal time stamp
#####################################################################
sub convert_date_log_time
{
    my ($string_date) = @_;

    &debug(5, "entering convert_date_log_time with argument $string_date");

    my ($day, $month, $year, $hour, $minute, $second, $zone);
    my $time_internal;

    if ($::LOGSTYLE eq 'umich') {
	($day, $month, $year, $hour, $minute, $second, $zone) 
	    = $string_date 
	    =~ m/^(\d+)\/(\w+)\/(\d+)\:(\d+)\:(\d+)\:(\d+)\s*([\-\d]*)$/;
	$time_internal = timelocal($second, $minute, $hour, $day, &convert_month_value($month), $year);
    }
    elsif ($::LOGSTYLE eq 'cern') {
	($year, $month, $day, $hour, $minute, $second) 
	    = $string_date 
	    =~ m/^(\d+)\-(\d+)\-(\d+)\s+(\d+)\:(\d+)\:(\d+)$/;
	$time_internal = timelocal($second, $minute, $hour, $day, $month-1, $year);
    }

    if ($day < 1 || $day > 31) {
	&pr(1, "bad date? ... XXX" . $string_date . "XXX\n");
    }

    

    &debug(5, "exiting convert_date_log_time");

    return $time_internal;
}

#####################################################################
# convert_month_value
#
# Description:
# Given a string name for a month, return a numerical value
# (0-indexed)
#
# **********
# I should rewrite this to use a hash - it will get accessed constantly
# so speed is critical - or actually I should provide a special function just for log dates, and keep this one for a more general purpose one
# **********
#####################################################################
sub convert_month_value
{
    my ($month) = @_;

    &debug(5, "entering convert_month_value");

    my @names_Mon = ('Jan', 'Feb', 'Mar',
		     'Apr', 'May', 'Jun',
		     'Jul', 'Aug', 'Sep',
		     'Oct', 'Nov', 'Dec');
    my @names_mon = ('jan', 'feb', 'mar',
		     'apr', 'may', 'jun',
		     'jul', 'aug', 'sep',
		     'oct', 'nov', 'dec');
    my @names_Month = ('January', 'February', 'March',
		       'April',   'May',      'June',
		       'July',    'August',   'September',
		       'October', 'November', 'December');

    for(my $i=0; $i < 12; $i++) {
	if( $month eq @names_Mon[$i] ) {
	    return $i;
	}
    }
    for(my $i=0; $i < 12; $i++) {
	if( $month eq @names_mon[$i] ) {
	    return $i;
	}
    }
    for(my $i=0; $i < 12; $i++) {
	if( $month eq @names_Month[$i] ) {
	    return $i;
	}
    }

    &debug(5, "exiting convert_month_value");

    return 0;
}

#####################################################################
# pr
#
#
#####################################################################
sub pr
{
    my ($level, $message) = @_;

    if ($level == 0) {
	print STDERR "ERROR: $message";
    }
    elsif ($level <= $::VERBOSITYLEVEL) {
	print $message;
    }

}

#####################################################################
# debug
#
#
#####################################################################
sub debug
{
    my ($level, $message) = @_;

    if ($level <= $::DEBUGLEVEL) {
	print "DEBUG $level: $message\n";
    }

}


##########################################################
# member_of
#
#
##########################################################
sub member_of
{
    my ($elt, $lptr) = @_;
    my $ismember = 0;

    foreach my $list_elt (@$lptr) {
	if ($list_elt eq $elt) { $ismember = 1; last;}
    }

    return $ismember;
}

##########################################################
# member_of_sorted
#
#
##########################################################
sub member_of_sorted
{
    my ($elt, $lptr) = @_;
    my $ismember = 0;

    foreach my $list_elt (@$lptr) {
	if ($list_elt eq $elt) { $ismember = 1; last;}
    }

    return $ismember;
}
