package Micala::Common;

# ====================================================================
# Copyright (c) 2010 The Regents of the University of Michigan and 
# The European Organization for Nuclear Research (CERN). Licensed under the
# Educational Community License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
# http://www.osedu.org/licenses/ECL-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
# ====================================================================

use 5.010001;
use strict;
use warnings;

use DBI();

use vars qw($VERSION $DEBUG_LEVEL $VERBOSITY_LEVEL);

require Exporter;

our @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

# This allows declaration	use Micala::Common ':all';
# If you do not need this, moving things directly into @EXPORT or @EXPORT_OK
# will save memory.
our %EXPORT_TAGS = ( 'all' => [ ('pr', 
				 'debug', 
				 'member_of', 
				 'syscom', 
				 'getMatchingIndicoID',
				 'getIdMachine',
				 'getIdTask',
				 'getIdLecture',
				 'reportStatus',)  ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our @EXPORT = qw();

our $VERSION = '0.1';

# Set default verbosity and debug levels.
our $DEBUG_LEVEL     = 0;
our $VERBOSITY_LEVEL = 1;

# Declare package-global variables to be set in configuration file
our ($DBHost, $DBPort, $DBName, $DBUser, $DBPW, $DBReaderUser, $DBReaderPW);

# Preloaded methods go here.

##########################################################
# member_of
#
# Description: determines if an element is found in an array
#
# Arguments:
# $elt: the element (string) in question
# $arr: the array in question, can be an array or the string "*"
#
# Returns:
# 1 if $elt is a member of @$arr
# 0 if it is not
##########################################################
sub member_of
{
    my ($elt, $arr) = @_;

    if ($arr->[0] eq "*") {
        return 1;
    }
    else {
	foreach my $elt_curr (@$arr) {
	    if ($elt_curr eq $elt) {
		return 1;
	    }
	}
    }

    return 0;
}

###############################################
# syscom
#
#
###############################################
sub syscom
{
    my ($level, $command) = @_;

    &pr($level, "\nCOMMAND: $command\n");
    return system($command);
}

##########################################################
# pr
#
# this is my own subroutine to replace the print statement.
# All output should be sent to this sub.  The advantage of
# this is that you can specify verbosity levels, and also
# it makes it easy to turn it into a CGI or ASP script.
##########################################################
sub pr()
{
    my ($level, $message) = @_;

    if ($level == 0) {
        print STDERR "ERROR: $message";
    }
    elsif ($level <= $VERBOSITY_LEVEL) {
        print $message;
    }

}

##########################################################
# debug
#
#
##########################################################
sub debug()
{
    my ($level, $message) = @_;

    if ($level <= $DEBUG_LEVEL) {
	print "DBG$level $message\n";
    }
}

##########################################################
# getMatchingIndicoID
#
# Given an LOID, search in the database for a matching IndicoID
##########################################################
sub getMatchingIndicoID()
{
    my ($LOID) = @_;

    # Connect to the database.
    my $dbh = DBI->connect("DBI:mysql:database=$DBName;host=$DBHost",
			   "$DBReaderUser", 
			   "$DBReaderPW",
			   {'RaiseError' => 1});
    
    # retrieve data
    my $sth = $dbh->prepare("SELECT IndicoID FROM Lectures WHERE LOID = '$LOID'");
    $sth->execute();
    my $ref = $sth->fetchrow_hashref();
    my $IndicoID = $ref->{'IndicoID'};
    $sth->finish();
    
    # Disconnect from the database.
    $dbh->disconnect();

    return ($IndicoID);
}

##########################################################
# getIdMachine
#
# Given a machine name, return the internal DB ID for that machine.
##########################################################
sub getIdMachine()
{
    my ($machine_name) = @_;

    # Connect to the database.
    my $dbh = DBI->connect("DBI:mysql:database=$DBName;host=$DBHost",
			   "$DBReaderUser", 
			   "$DBReaderPW",
			   {'RaiseError' => 1});
    
    # retrieve data
    my $sth = $dbh->prepare("SELECT idMachine FROM Machines WHERE Hostname = '$machine_name'");
    $sth->execute();
    my $ref = $sth->fetchrow_hashref();
    my $idMachine = $ref->{'idMachine'};
    $sth->finish();
    
    # Disconnect from the database.
    $dbh->disconnect();

    return ($idMachine);
}

##########################################################
# getIdTask
#
# Given a task name, return the internal DB ID for that task.
##########################################################
sub getIdTask()
{
    my ($task_name) = @_;

    # Connect to the database.
    my $dbh = DBI->connect("DBI:mysql:database=$DBName;host=$DBHost",
			   "$DBReaderUser", 
			   "$DBReaderPW",
			   {'RaiseError' => 1});
    
    # retrieve data
    my $sth = $dbh->prepare("SELECT idTask,Name FROM Tasks WHERE Name = '$task_name'");
    $sth->execute();
    my $ref = $sth->fetchrow_hashref();
    my $idTask = $ref->{'idTask'};
    $sth->finish();
    
    # Disconnect from the database.
    $dbh->disconnect();

    return ($idTask);
}

##########################################################
# getIdLecture
#
# Given an LOID or IndicoID, return the internal DB ID for that lecture.
##########################################################
sub getIdLecture()
{
    my ($lecture_name) = @_;

    # Connect to the database.
    my $dbh = DBI->connect("DBI:mysql:database=$DBName;host=$DBHost",
			   "$DBReaderUser", 
			   "$DBReaderPW",
			   {'RaiseError' => 1});
    
    # retrieve data
    my $sth = $dbh->prepare("SELECT idLecture FROM Lectures WHERE LOID = '$lecture_name' OR IndicoID = '$lecture_name'");
    $sth->execute();
    my $ref = $sth->fetchrow_hashref();
    my $idLecture = $ref->{'idLecture'};
    $sth->finish();
    
    # Disconnect from the database.
    $dbh->disconnect();

    return ($idLecture);
}

##########################################################
# reportStatus
#
# Report current task and status to database.
##########################################################
sub reportStatus()
{
    my ($status, $message, $idMachine, $idTask, $idLecture) = @_;

    # Connect to the database.
    my $dbh = DBI->connect("DBI:mysql:database=$DBName;host=$DBHost",
			   "$DBUser", 
			   "$DBPW",
			   {'RaiseError' => 1});
    
    # retrieve data
    my $sth = $dbh->prepare(sprintf("INSERT INTO Status
            (idMachine, idTask, idLecture, Status, Message, DateReported)
            VALUES('%s', '%s', '%s', '%s', '%s', NOW());",
				    $idMachine, $idTask, $idLecture, $status, $message));

    $sth->execute();
    $sth->finish();
    
    # Disconnect from the database.
    $dbh->disconnect();

    return ();
}

1; # Perl wants to see a "true" value

__END__
# Below is stub documentation for your module. You'd better edit it!

=head1 NAME

Micala::Common - Commonly used functions for the micala perl scripts, e.g. print/debugging functions, etc.

=head1 SYNOPSIS

  use Micala::Common;
  blah blah blah

=head1 DESCRIPTION

This module contains some useful functions, including the following:

=head2 FUNCTIONS

pr

debug

member_of

syscom

=head2 VARIABLES

VERBOSITY_LEVEL

DEBUG_LEVEL

=head2 EXPORT

None by default.


=head1 SEE ALSO

Mention other useful documentation such as the documentation of
related modules or operating system documentation (such as man pages
in UNIX), or any relevant external documentation such as RFCs or
standards.

Code available at http://www.micala.org

=head1 AUTHOR

Jeremy Herr <herrj@umich.edu>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2010 by the Regents of the University of Michigan and CERN

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.10.0 or,
at your option, any later version of Perl 5 you may have available.


=cut
