# ====================================================================
# Copyright (c) 2010 The Regents of the University of Michigan and 
# The European Organization for Nuclear Research (CERN). Licensed under the
# Educational Community License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
# http://www.osedu.org/licenses/ECL-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
# ====================================================================

##########################################################
# micala.py
#
# Authors: Jeremy Herr (herrj@umich.edu) and Nicola Tarocco (ntarocci@cern.ch)
#
# This file contains frequently used classmethods, separated
# into two classes:
#  - MicalaCommunication (for interacting with the micala MySQL database)
#  - Common (for print and debug output)
##########################################################

import MySQLdb
import sys
import datetime
import time
import calendar
import os
import conf
import re
import subprocess
from types import IntType

class MicalaCommunication(object):

    @classmethod
    def getIp(cls):
        '''This should return the IP of the current machine.'''
        return("")

    @classmethod
    def getIdMachine(cls, machine_name):
        '''Look up ID of this machine in database'''

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBReaderUser,
                                         passwd = conf.DBReaderPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
            Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(2, """SELECT idMachine,Hostname FROM Machines WHERE Hostname = %s\n""" % machine_name)

        cursor = connection.cursor()
        # believe it or not, the comma following machine_name is supposed to be there for MySQLdb's sake
        cursor.execute("""SELECT idMachine,Hostname FROM Machines WHERE Hostname = %s""",
            (machine_name,))
        connection.commit()

        result_set = cursor.fetchone()


        if result_set is not None and len(result_set) > 0:
            idMachine = result_set[0]
        else:
            idMachine = ''

        cursor.close()
        connection.close()

        return(idMachine)

    @classmethod
    def getIdTask(cls, task_name):
        '''Look up ID of this task in database'''

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBReaderUser,
                                         passwd = conf.DBReaderPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
            Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(2, """SELECT idTask,Name FROM Tasks WHERE Name = %s\n""" % task_name)

        cursor = connection.cursor()
        cursor.execute("""SELECT idTask,Name FROM Tasks WHERE Name = %s""",
            (task_name,))
        connection.commit()

        result_set = cursor.fetchone()

        if result_set is not None and len(result_set) > 0:
            idTask = result_set[0]
        else:
            idTask = result_set[0]

        cursor.close()
        connection.close()

        return(idTask)

    @classmethod
    def getIdLecture(cls, lecture_name):
        '''Look up ID of the given lecture in micala database'''

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBReaderUser,
                                         passwd = conf.DBReaderPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
            Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(2, """SELECT idLecture,LOID,IndicoID FROM Lectures WHERE LOID = %s OR IndicoID = %s\n""" % (lecture_name, lecture_name))

        # Depending on style of lecture ID, search under Michigan style column or CERN style column
        cursor = connection.cursor()
        cursor.execute("""SELECT idLecture,LOID,IndicoID FROM Lectures WHERE LOID = %s OR IndicoID = %s""",
            (lecture_name, lecture_name))
        connection.commit()

        result_set = cursor.fetchone()

        if result_set is not None and len(result_set) > 0:
            Common.debug(1, "result_set: %s" % str(result_set))
            idLecture = result_set[0]
        else:
            idLecture = ''

        cursor.close()
        connection.close()

        return(idLecture)

    @classmethod
    def isRequested(cls, idLecture, idTask):
        """Check to see if the given task has been requested for the given lecture."""

        flagRequested = False

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBReaderUser,
                                         passwd = conf.DBReaderPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
            Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(2, """SELECT S.Status FROM LectureLatestStatus LS, Status S
                            WHERE LS.idStatus = S.idStatus
                            AND LS.idLecture = %s
                            AND LS.idTask = %s;\n""" % (idLecture, idTask))

        cursor = connection.cursor()
        cursor.execute("""SELECT S.Status FROM LectureLatestStatus LS, Status S
                            WHERE LS.idStatus = S.idStatus
                            AND LS.idLecture = %s
                            AND LS.idTask = %s;""",
            (idLecture, idTask))
        connection.commit()

        result_set = cursor.fetchone()

        if result_set is not None and len(result_set) > 0:
            Common.debug(1, "isRequested(idLecture = %d, idTask = %d) = %s" % (idLecture, idTask, result_set[0]))
            if result_set[0] == "REQUESTED":
                flagRequested = True
        else:
            flagRequested = False

        cursor.close()
        connection.close()

        return flagRequested

    @classmethod
    def lectureExists(cls, idLecture):
        """Given a lecture ID, check if it exists into the database"""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBReaderUser,
                                         passwd = conf.DBReaderPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
            Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(2, """SELECT idLecture FROM Lectures WHERE idLecture = %s;\n""" % idLecture)

        cursor = connection.cursor()
        cursor.execute("""SELECT idLecture FROM Lectures WHERE idLecture = %s;""", idLecture)
        connection.commit()

        result_set = cursor.fetchone()

        if result_set is None:
            return False
        else:
            return True

    @classmethod
    def getCropTimes(cls, idLecture):
        """Given a lecture ID, look up CropBegin and CropEnd, return the values in seconds, as well as formatted into strings HH:MM:SS."""

        cropBegin = 0
        cropEnd   = 0

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBReaderUser,
                                         passwd = conf.DBReaderPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
            Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(2, """SELECT CropBegin, CropEnd FROM Lectures WHERE idLecture = %s;\n""" % idLecture)

        cursor = connection.cursor()
        cursor.execute("""SELECT CropBegin, CropEnd FROM Lectures WHERE idLecture = %s;""", idLecture)
        connection.commit()

        result_set = cursor.fetchall()

        if result_set is not None and len(result_set) > 0:
            row = result_set[0]

            Common.debug(2, "For lecture ID %d, got CropBegin: %s sec, CropEnd: %s sec" % \
                             (idLecture, row[0], row[1]))
            if row[0] is not None and row[0] != "":
                cropBegin = int(row[0])
            if row[1] is not None and row[1] != "":
                cropEnd = int(row[1])
        else:
            cropBegin = 0
            cropEnd   = 0

        cursor.close()
        connection.close()

        return (cropBegin, cropEnd, Common.formatCropTime(cropBegin), Common.formatCropTime(cropEnd))

    @classmethod
    def createNewMicalaLecture(cls, lecture_name, contentType, pathToOriginalFiles = None):
        '''insert a record into the micala database for a new lecture'''

        # only create a new record if one doesn't already exist
        idLectureExisting = cls.getIdLecture(lecture_name)
        if idLectureExisting != "":
            return idLectureExisting

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBUser,
                                         passwd = conf.DBPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
            Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        cursor = connection.cursor()

        if contentType == conf.contentTypeWebLecture:
            if pathToOriginalFiles is not None and pathToOriginalFiles != '':
                Common.pr(1, """INSERT INTO Lectures (LOID, contentType, OriginPath, DateCreated) VALUES(%s, %s, %s, NOW());\n""" % (lecture_name, contentType, pathToOriginalFiles))
                cursor.execute("""INSERT INTO Lectures (LOID, contentType, OriginPath, DateCreated) VALUES(%s, %s, %s, NOW());""", (lecture_name, contentType, pathToOriginalFiles))
            else:
                Common.pr(1, """INSERT INTO Lectures (LOID, contentType, DateCreated) VALUES(%s, %s, NOW());\n""" % (lecture_name, contentType))
                cursor.execute("""INSERT INTO Lectures (LOID, contentType, DateCreated) VALUES(%s, %s, NOW());""", (lecture_name, contentType))
        elif contentType == conf.contentTypePlainVideo:
            Common.pr(1, """INSERT INTO Lectures (IndicoID, contentType, DateCreated) VALUES(%s, %s, NOW());\n""" % (lecture_name, contentType))
            cursor.execute("""INSERT INTO Lectures (IndicoID, contentType, DateCreated) VALUES(%s, %s, NOW());""", (lecture_name, contentType))

        connection.commit()

        connection.close()

        return cls.getIdLecture(lecture_name)

    @classmethod
    def getLecturesDoneTaskANotTaskB(cls, idTaskA, idTaskB, contentType, originPath = ""):
        """Get a list of dictionaries, each dictionary is a lecture record for whom task A is completed,
        and task B has not yet been started (since task A finished). An additional clause can be added."""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBReaderUser,
                                         passwd = conf.DBReaderPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
            Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        cursor = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        if originPath == "NULL":
            originPathClause = " AND OriginPath IS NULL"
        elif originPath == "NOT NULL":
            originPathClause = " AND OriginPath IS NOT NULL"
        else:
            originPathClause = ""

        Common.pr(2, """SELECT LSA.idLecture AS idLecture, L.LOID, L.OriginPath
FROM LectureLatestStatus LSA, `Status` SA, LectureLatestStatus LSB, Lectures L
WHERE LSA.idLecture = LSB.idLecture
AND LSA.idLecture = L.idLecture
AND LSA.idStatus  = SA.idStatus
AND LSA.idTask    = %s
AND LSB.idTask    = %s
AND SA.Status     = 'COMPLETE'
AND LSB.idStatus IS NULL
AND L.contentType = %s %s
UNION
SELECT LSA.idLecture AS idLecture, L.LOID, L.OriginPath
FROM LectureLatestStatus LSA, `Status` SA, LectureLatestStatus LSB, `Status` SB, Lectures L
WHERE LSA.idLecture = LSB.idLecture
AND LSA.idLecture = L.idLecture
AND LSA.idStatus  = SA.idStatus
AND LSB.idStatus  = SB.idStatus
AND LSA.idTask    = %s
AND LSB.idTask    = %s
AND SA.Status     = 'COMPLETE'
AND LSB.idStatus  = SB.idStatus
AND ( SB.Status   = 'ERROR' OR SB.Status = 'REQUESTED' )
AND L.contentType = %s %s
ORDER BY idLecture;\n""" % (idTaskA, idTaskB, contentType, originPathClause, idTaskA, idTaskB, contentType, originPathClause))

        query = \
"""
SELECT LSA.idLecture AS idLecture, L.LOID, L.OriginPath
FROM LectureLatestStatus LSA, `Status` SA, LectureLatestStatus LSB, Lectures L
WHERE LSA.idLecture = LSB.idLecture
AND LSA.idLecture = L.idLecture
AND LSA.idStatus  = SA.idStatus
AND LSA.idTask    = %s
AND LSB.idTask    = %s
AND SA.Status     = 'COMPLETE'
AND LSB.idStatus IS NULL
AND L.contentType = %s""" + originPathClause + """
UNION
SELECT LSA.idLecture AS idLecture, L.LOID, L.OriginPath
FROM LectureLatestStatus LSA, `Status` SA, LectureLatestStatus LSB, `Status` SB, Lectures L
WHERE LSA.idLecture = LSB.idLecture
AND LSA.idLecture = L.idLecture
AND LSA.idStatus  = SA.idStatus
AND LSB.idStatus  = SB.idStatus
AND LSA.idTask    = %s
AND LSB.idTask    = %s
AND SA.Status     = 'COMPLETE'
AND LSB.idStatus  = SB.idStatus
AND ( SB.Status   = 'ERROR' OR SB.Status = 'REQUESTED' )
AND L.contentType = %s""" + originPathClause + """
ORDER BY idLecture;
"""

        cursor.execute(query, (idTaskA, idTaskB, contentType, idTaskA, idTaskB, contentType))

        connection.commit()
        rows = cursor.fetchall()
        cursor.close()
        connection.close()

        return rows

    @classmethod
    def getLectureCurrentlyProcessedByMachineTask(cls, idMachine, idTask, contentType):
        """Given a machine ID and task ID, return the record of the lecture currently being processed by this machine and this task."""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBReaderUser,
                                         passwd = conf.DBReaderPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
            Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(2, """SELECT L.LOID, L.idLecture
         FROM LectureLatestStatus LS, `Status` S, Lectures L
         WHERE LS.idStatus   = S.idStatus
           AND LS.idLecture  = L.idLecture
           AND S.idMachine   = %s
           AND S.idTask      = %s
           AND S.Status     != 'COMPLETE'
           AND L.contentType = %s
         ORDER BY L.idLecture;\n""" % (idMachine, idTask, contentType))

        cursor = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)

        query = """SELECT L.LOID, L.idLecture
         FROM LectureLatestStatus LS, `Status` S, Lectures L
         WHERE LS.idStatus   = S.idStatus
           AND LS.idLecture  = L.idLecture
           AND S.idMachine   = %s
           AND S.idTask      = %s
           AND S.Status     != 'COMPLETE'
           AND L.contentType = %s
         ORDER BY L.idLecture;"""

        cursor.execute(query, (idMachine, idTask, contentType))

        connection.commit()
        rows = cursor.fetchall()
        cursor.close()
        connection.close()

        rowlist = list(rows)

        if len(rowlist) > 1:
            return rowlist[-1]
        elif len(rowlist) == 1:
            return rowlist[0]
        elif len(rowlist) == 0:
            return None

    @classmethod
    def getTaskList(cls):
        """Returns a list of all tasks, returning idTask, Name and Ordering, in the Tasks table, ordered by the Ordering column."""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBReaderUser,
                                         passwd = conf.DBReaderPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
           Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(2, """SELECT idTask, Name, Ordering FROM Tasks ORDER BY Ordering\n""")

        cursor = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        cursor.execute("""SELECT idTask, Name, Ordering FROM Tasks ORDER BY Ordering""")
        connection.commit()
        rows = cursor.fetchall()
        cursor.close()
        connection.close()

        listTasks = []
        for row in rows:
            task = dict()
            task["idTask"] = row["idTask"]
            task["Name"] = row["Name"]
            task["Ordering"] = row["Ordering"]
            listTasks.append(task)

        return (listTasks)

    @classmethod
    def updateTaskOrder(cls, idTask, orderingValue):
        """Update the ordering position of a specific task"""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBUser,
                                         passwd = conf.DBPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
           Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(1, """UPDATE Tasks SET Ordering=%s WHERE idTask=%s\n""" % (orderingValue, idTask))

        cursor = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        cursor.execute("""UPDATE Tasks SET Ordering=%s WHERE idTask=%s""" , (orderingValue, idTask))
        connection.commit()
        cursor.close()
        connection.close()

    @classmethod
    def insertNewTaskAfter(cls, newTaskName, taskAfter):
        """Insert a new task after selected task, update ordering"""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBUser,
                                         passwd = conf.DBPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
           Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        cursor = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        # Get all task
        taskList = cls.getTaskList()
        # Get taskAfter ordering position, and set ordering for the new task after this
        taskAfterOrdering = 1
        taskAfter = long(taskAfter)
        for task in taskList:
            if task["idTask"] == taskAfter:
                taskAfterOrdering = task["Ordering"] + 1
        # Create a new free position for the new task, moving 1 position after others
        for task in taskList:
            if task["Ordering"] >= taskAfterOrdering:
                cls.updateTaskOrder(task["idTask"], task["Ordering"]+1)
        # Insert the new task in taskAfterOrdering position, it's now free
        Common.pr(1, """INSERT INTO Tasks (Name, Ordering, DateCreated) VALUES( %s, %s, NOW());\n""" % (newTaskName, taskAfterOrdering))
        cursor.execute("""INSERT INTO Tasks (Name, Ordering, DateCreated) VALUES( %s, %s, NOW());""" , (newTaskName, taskAfterOrdering))
        connection.commit()
        cursor.close()
        connection.close()

    @classmethod
    def editTaskName(cls, idTask, changedTaskName):
        """Edit the name of the task selected"""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBUser,
                                         passwd = conf.DBPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
           Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(1, """UPDATE Tasks SET Name=%s, DateModified=NOW() WHERE idTask=%s;\n""" % (changedTaskName, idTask))

        cursor = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        cursor.execute("""UPDATE Tasks SET Name=%s, DateModified=NOW() WHERE idTask=%s;""" , (changedTaskName, idTask))
        connection.commit()
        cursor.close()
        connection.close()

    @classmethod
    def getMachineList(cls):
        """Returns a list of all machines with all fields"""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBReaderUser,
                                         passwd = conf.DBReaderPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
           Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(2, """SELECT idMachine, Hostname, IP, CPU, Memory, Disk, OS FROM Machines ORDER BY Hostname\n""")

        cursor = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        cursor.execute("""SELECT idMachine, Hostname, IP, CPU, Memory, Disk, OS FROM Machines ORDER BY Hostname""")
        connection.commit()
        rows = cursor.fetchall()
        cursor.close()
        connection.close()

        listMachines = []
        for row in rows:
            machine = dict()
            machine["idMachine"] = row["idMachine"]
            machine["Hostname"] = row["Hostname"]
            machine["IP"] = row["IP"]
            machine["CPU"] = row["CPU"]
            machine["Memory"] = row["Memory"]
            machine["Disk"] = row["Disk"]
            machine["OS"] = row["OS"]
            listMachines.append(machine)

        return (listMachines)


    @classmethod
    def getMachine(cls, idMachine):
        """Returns a specified machine"""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBReaderUser,
                                         passwd = conf.DBReaderPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
           Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(2, """SELECT idMachine, Hostname, IP, CPU, Memory, Disk, OS, idVenue FROM Machines WHERE idMachine=%s\n""" % idMachine)

        cursor = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        cursor.execute("""SELECT idMachine, Hostname, IP, CPU, Memory, Disk, OS, idVenue FROM Machines WHERE idMachine=%s""" , idMachine)
        connection.commit()
        rows = cursor.fetchall()
        cursor.close()
        connection.close()

        for row in rows:
            machine = dict()
            machine["idMachine"] = row["idMachine"]
            machine["Hostname"] = row["Hostname"]
            machine["IP"] = row["IP"]
            machine["CPU"] = row["CPU"]
            machine["Memory"] = row["Memory"]
            machine["Disk"] = row["Disk"]
            machine["OS"] = row["OS"]
            machine["idVenue"] = row["idVenue"]

        return (machine)


    @classmethod
    def insertNewMachine(cls, hostname, ip, cpu, memory, disk, os, idVenue):
        """Insert a new machine"""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBUser,
                                         passwd = conf.DBPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
           Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        cursor = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        Common.pr(1, """INSERT INTO Machines (Hostname, IP, CPU, Memory, Disk, OS, idVenue, DateCreated) VALUES( %s, %s, %s, %s, %s, %s, %s, NOW());\n""" % (hostname, ip, cpu, memory, disk, os, idVenue))
        cursor.execute("""INSERT INTO Machines (Hostname, IP, CPU, Memory, Disk, OS, idVenue, DateCreated) VALUES( %s, %s, %s, %s, %s, %s, %s, NOW());""" , (hostname, ip, cpu, memory, disk, os, idVenue))
        connection.commit()
        cursor.close()
        connection.close()

    @classmethod
    def editMachine(cls, idMachine, hostname, ip, cpu, memory, disk, os, idVenue):
        """Edit machine details"""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBUser,
                                         passwd = conf.DBPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
           Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(1, """UPDATE Machines SET Hostname=%s, IP=%s, CPU=%s, Memory=%s, Disk=%s, OS=%s, idVenue=%s, DateModified=NOW() WHERE idMachine=%s;\n""" % (hostname, ip, cpu, memory, disk, os, idVenue, idMachine))

        cursor = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        cursor.execute("""UPDATE Machines SET Hostname=%s, IP=%s, CPU=%s, Memory=%s, Disk=%s, OS=%s, idVenue=%s, DateModified=NOW() WHERE idMachine=%s;""" , (hostname, ip, cpu, memory, disk, os, idVenue, idMachine))
        connection.commit()
        cursor.close()
        connection.close()


    @classmethod
    def getVenueList(cls):
        """Returns a list of all venues with  all fields"""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBReaderUser,
                                         passwd = conf.DBReaderPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
           Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(2, """SELECT idVenue, RoomName, IndicoRoomID FROM Venues ORDER BY RoomName\n""")

        cursor = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        cursor.execute("""SELECT idVenue, RoomName, IndicoRoomID FROM Venues ORDER BY RoomName""")
        connection.commit()
        rows = cursor.fetchall()
        cursor.close()
        connection.close()

        listVenues = []
        for row in rows:
            venue = dict()
            venue["idVenue"] = row["idVenue"]
            venue["RoomName"] = row["RoomName"]
            venue["IndicoRoomID"] = row["IndicoRoomID"]
            listVenues.append(venue)

        return (listVenues)


    @classmethod
    def getVenue(cls, idVenue):
        """Returns a specified venue"""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBReaderUser,
                                         passwd = conf.DBReaderPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
           Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(2, """SELECT idVenue, RoomName, IndicoRoomID FROM Venues WHERE idVenue=%s\n""" % idVenue)

        cursor = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        cursor.execute("""SELECT idVenue, RoomName, IndicoRoomID FROM Venues WHERE idVenue=%s""" , idVenue)
        connection.commit()
        rows = cursor.fetchall()
        cursor.close()
        connection.close()

        for row in rows:
            venue = dict()
            venue["idVenue"] = row["idVenue"]
            venue["RoomName"] = row["RoomName"]
            venue["IndicoRoomID"] = row["IndicoRoomID"]

        return (venue)


    @classmethod
    def insertNewVenue(cls, roomName, IndicoRoomID):
        """Insert a new venue"""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBUser,
                                         passwd = conf.DBPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
           Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        cursor = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        Common.pr(1, """INSERT INTO Venues (RoomName, IndicoRoomID, DateCreated) VALUES( %s, %s, NOW());\n""" % (roomName, IndicoRoomID))
        cursor.execute("""INSERT INTO Venues (RoomName, IndicoRoomID, DateCreated) VALUES( %s, %s, NOW());""" , (roomName, IndicoRoomID))
        connection.commit()
        cursor.close()
        connection.close()

    @classmethod
    def editVenue(cls, idVenue, roomName, IndicoRoomID):
        """Edit venue details"""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBUser,
                                         passwd = conf.DBPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
           Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(1, """UPDATE Venues SET RoomName=%s, IndicoRoomID=%s, DateModified=NOW() WHERE idVenue=%s;\n""" % (roomName, IndicoRoomID, idVenue))

        cursor = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        cursor.execute("""UPDATE Venues SET RoomName=%s, IndicoRoomID=%s, DateModified=NOW() WHERE idVenue=%s;""" , (roomName, IndicoRoomID, idVenue))
        connection.commit()
        cursor.close()
        connection.close()


    @classmethod
    def getLectureAndStatusLists(cls,
                                 strIdLecture,
                                 dateStart   = '', # not currently used
                                 dateEnd     = '', # not currently used
                                 ageLecture  = "1month",
                                 contentType = conf.contentTypeWebLecture,
                                 orphan      = -1,
                                 sortBy      = 'DateCreated',
                                 sortOrder   = 'DESC'):
        """idLecture can be a single internal DB id, or * to find all.
    Choices for ageLecture are minute, hour, second, day, month, year.
    orphan: 1 means return orphans only, 0 means return non-orphans only, -1 means don't care about orphan status."""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBReaderUser,
                                         passwd = conf.DBReaderPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
           Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        # Construct single lecture clause if given an individual lecture id
        strIdLecture = str(strIdLecture)
        matchIdLecture = re.match("(\d+)", strIdLecture)
        if strIdLecture == "*":
            clauseSingleLecture = ""
        elif matchIdLecture:
            idLecture = int(matchIdLecture.group(1))
            clauseSingleLecture = " AND L.idLecture = %d" % idLecture
        else:
            Common.pr(0, "getLectureAndStatusLists() expects either a valid internal DB idLecture, or \"*\" to match all.")
            sys.exit(1)

        # Construct content type clause
        if clauseSingleLecture == "" and contentType is not None and contentType != "":
            clauseContentType = " AND L.ContentType = '%s' " % contentType
        else:
            clauseContentType = ""

        # Construct orphan clause
        if clauseSingleLecture == "":
            if orphan == 1:
                clauseOrphan = " AND L.LOID IS NOT NULL AND (L.IndicoID IS NULL OR L.IndicoID = '')"
            elif orphan == 0:
                clauseOrphan = " AND L.LOID IS NOT NULL AND (L.IndicoID IS NOT NULL AND L.IndicoID != '')"
            elif orphan == -1:
                clauseOrphan = ""
            else:
                clauseOrphan = ""
        else:
            clauseOrphan = ""

        # Construct SQL clause to find lectures younger than a given age.
        # Replace given words like "seconds" with SQL "SECOND" to avoid passing uncontrolled statements to an SQL query.
        if clauseSingleLecture == "":
            if ageLecture is not None and ageLecture != "":
                match = re.search("(\d+)(\w+)", ageLecture)
                if match:
                    unitValue = match.group(1)
                    matchFound = False
                    if match.group(2) == "second":
                        unitType = "SECOND"
                        matchFound = True
                    elif match.group(2) == "minute":
                        unitType = "MINUTE"
                        matchFound = True
                    elif match.group(2) == "hour":
                        unitType = "HOUR"
                        matchFound = True
                    elif match.group(2) == "day":
                        unitType = "DAY"
                        matchFound = True
                    elif match.group(2) == "week":
                        unitType = "WEEK"
                        matchFound = True
                    elif match.group(2) == "month":
                        unitType = "MONTH"
                        matchFound = True
                    elif match.group(2) == "year":
                        unitType = "YEAR"
                        matchFound = True

                    if matchFound:
                        clauseAge = " AND L.DateCreated > NOW() - INTERVAL %d %s" % (int(unitValue), unitType)
                    else:
                        clauseAge = ""
                else:
                    clauseAge = ""
            else:
                clauseAge = ""
        else:
            clauseAge = ""

        # Create order by clause
        if sortOrder == 'DESC':
            sortOrder = 'DESC'
        else:
            sortOrder = 'ASC'

        if sortBy == 'LOID' or sortBy == 'IndicoID' or sortBy == 'CDSRecord' or sortBy == 'Duration':
            orderBySelected = sortBy + ' ' + sortOrder
        else:
            orderBySelected = 'DateCreated' + ' ' + sortOrder

        Common.pr(2, """SELECT L.idLecture AS idLecture, L.LOID AS LOID, L.IndicoID AS IndicoID, L.ContentType AS ContentType, L.DateCreated AS DateCreated, L.CDSRecord AS CDSRecord, L.Title AS Title, L.Creator AS Creator, L.Duration AS Duration, L.OriginPath AS OriginPath
    FROM Lectures L
    WHERE 1 = 1
    """ + clauseContentType   + """
    """ + clauseOrphan        + """
    """ + clauseAge           + """
    """ + clauseSingleLecture + """
    ORDER BY
    """ + orderBySelected + """;\n""")

        # Query to get list of lecture objects matching given filters
        cursor = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        cursor.execute(
            """SELECT L.idLecture AS idLecture, L.LOID AS LOID, L.IndicoID AS IndicoID, L.ContentType AS ContentType, L.DateCreated AS DateCreated, L.CDSRecord AS CDSRecord, L.Title AS Title, L.Creator AS Creator, L.Duration AS Duration, L.OriginPath AS OriginPath
    FROM Lectures L
    WHERE 1 = 1
    """ + clauseContentType   + """
    """ + clauseOrphan        + """
    """ + clauseAge           + """
    """ + clauseSingleLecture + """
    ORDER BY
    """ + orderBySelected + """;""")

        connection.commit()
        listLectures = cursor.fetchall()
        cursor.close()
    #    connection.close()


        Common.pr(2, """SELECT L.idLecture AS idLecture, L.LOID AS LOID, L.IndicoID AS IndicoID, L.DateCreated AS DateCreated, T.Name AS Name, S.idStatus AS idStatus, S.Status AS Status, S.Message AS Message, M.Hostname AS Hostname, S.DateReported AS DateReported, V.RoomName AS RoomName
    FROM Lectures L, Tasks T, Status S, Machines M, Venues V
    WHERE L.idLecture = S.idLecture
--    WHERE L.idLecture = LS.idLecture
--    AND LS.idTask = T.idTask
--    AND LS.idStatus = S.idStatus
    AND S.idTask = T.idTask
    AND S.idMachine = M.idMachine
    AND M.idVenue   = V.idVenue
    """ + clauseContentType + """
    """ + clauseOrphan      + """
    """ + clauseAge         + """
    """ + clauseSingleLecture + """
    ORDER BY S.DateReported;\n""")

        # Now get all status updates for the list of lectures we're interested in
        cursor2 = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        cursor2.execute(
            """SELECT L.idLecture AS idLecture, L.LOID AS LOID, L.IndicoID AS IndicoID, L.DateCreated AS DateCreated, T.Name AS Name, S.idStatus AS idStatus, S.Status AS Status, S.Message AS Message, M.Hostname AS Hostname, S.DateReported AS DateReported, V.RoomName AS RoomName
    FROM Lectures L, Tasks T, Status S, Machines M, Venues V
    WHERE L.idLecture = S.idLecture
--    WHERE L.idLecture = LS.idLecture
--    AND LS.idTask = T.idTask
--    AND LS.idStatus = S.idStatus
    AND S.idTask = T.idTask
    AND S.idMachine = M.idMachine
    AND M.idVenue   = V.idVenue
    """ + clauseContentType + """
    """ + clauseOrphan      + """
    """ + clauseAge         + """
    """ + clauseSingleLecture + """
    ORDER BY S.DateReported;""")

        connection.commit()
        listStatuses = cursor2.fetchall()
        cursor2.close()
        connection.close()

        # For single lectures, return dictionaries;
        # for multiple lectures, return tuples of dictionaries (one dict for each lecture)
        if clauseSingleLecture != "":
            return (listLectures[0], listStatuses)
        else:
            return (listLectures, listStatuses)

    @classmethod
    def getLatestStatus(cls, idLecture, taskName):
        """Given Lecture ID and Task Name, it returns the latest status ID and Name"""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBReaderUser,
                                         passwd = conf.DBReaderPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
           Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        cursor = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        idTask = cls.getIdTask(taskName)
        cursor.execute("""SELECT s.idStatus, s.Status FROM lecturelateststatus lls
                        INNER JOIN status s
                        ON lls.idStatus = s.idStatus
                        WHERE lls.idLecture=%s AND lls.idTask=%s""" , (idLecture, idTask))
        connection.commit()
        row = cursor.fetchone()
        cursor.close()
        connection.close()

        status = None
        if row is not None and len(row) > 0:
            status = dict()
            status["idStatus"] = row["idStatus"]
            status["Status"] = row["Status"]

        return (status)

    @classmethod
    def deleteLectureRecord(cls, idLecture):
        """Be careful with this!! It first deletes all status updates with the given idLecture, then the lecture record itself."""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBAdminUser,
                                         passwd = conf.DBAdminPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
           Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        cursor = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        cursor.execute(
            """DELETE FROM Status WHERE idLecture = %s""", (idLecture,))

        connection.commit()
        result = cursor.fetchall()
        cursor.close()

        cursor2 = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        cursor2.execute(
            """DELETE FROM Lectures WHERE idLecture = %s""", (idLecture,))

        connection.commit()
        result2 = cursor2.fetchall()
        cursor2.close()

        connection.close()

        return (result, result2)

    @classmethod
    def getMatches(cls, confID):
        '''For the current conference, get list from the database of IndicoID's already matched to Lecture Objects.'''

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBReaderUser,
                                         passwd = conf.DBReaderPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
            raise RecordingManagerException("MySQL database error %d: %s" % (e.args[0], e.args[1]))

        cursor = connection.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        cursor.execute('''SELECT IndicoID, LOID, ContentType FROM Lectures WHERE IndicoID LIKE "%s%%"''' % confID)
        connection.commit()
        rows = cursor.fetchall()
        cursor.close()
        connection.close()

        # Build dictionary of matches
        match_array = {}
        for row in rows:
            # We are only interested in reporting on lecture object matches here,
            # not whether plain video talks are in the micala database.
            # Also, it shouldn't ever happen, but ignore records in the micala DB with ContentType WEBLECTURE
            # that don't have a LOID.
            # LOID = NULL in the MySQL database translates in Python to being None.
            if row["ContentType"] == 'WEBLECTURE' and row["LOID"] is not None:
                match_array[row["IndicoID"]] = row["LOID"]

        return (match_array)

    @classmethod
    def reportStatus(cls, status, message, idMachine, idTask, idLecture):
        """Given IDs for lecture, task and machine, report status to the database.
        This simply adds a single record to the Status table."""

        # Set this flag to false if it's an error message and it's already been posted.
        flagSubmit = True

        if idLecture == '':
            idLecture = None

        # If this is an ERROR message, then first check to see if it has already been reported.
        # We don't want the daemon to report error messages over and over, filling the database with crap.
        if status == "ERROR":
            try:
                connectionCheck = MySQLdb.connect(host   = conf.DBHost,
                                             port   = int(conf.DBPort),
                                             user   = conf.DBReaderUser,
                                             passwd = conf.DBReaderPW,
                                             db     = conf.DBName)
            except MySQLdb.Error, e:
                Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

            Common.pr(2, """SELECT Message FROM LectureLatestStatus LS, Status S
                WHERE LS.idStatus = S.idStatus
                 AND LS.idLecture = %s
                 AND LS.idTask    = %s;\n""" %  (idLecture, idTask))

            cursorCheck = connectionCheck.cursor(cursorclass=MySQLdb.cursors.DictCursor)
            cursorCheck.execute("""SELECT Message FROM LectureLatestStatus LS, Status S
                WHERE LS.idStatus = S.idStatus
                 AND LS.idLecture = %s
                 AND LS.idTask    = %s;""", (idLecture, idTask))

            rows = cursorCheck.fetchall()
            for row in rows:
                if len(row) > 0:
                    if row["Message"] == message:
                        flagSubmit = False

            cursorCheck.close()
            connectionCheck.commit()
            connectionCheck.close()

        # If flagSubmit is still True then go ahead and submit this status report.
        if flagSubmit:
            try:
                connection = MySQLdb.connect(host   = conf.DBHost,
                                             port   = int(conf.DBPort),
                                             user   = conf.DBUser,
                                             passwd = conf.DBPW,
                                             db     = conf.DBName)
            except MySQLdb.Error, e:
                Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

            Common.pr(1, """INSERT INTO Status
            (idMachine, idTask, idLecture, Status, Message, DateReported)
            VALUES(%s, %s, %s, %s, %s, NOW());\n""" % (idMachine, idTask, idLecture, status, message))

            cursor = connection.cursor()
            cursor.execute("""INSERT INTO Status
            (idMachine, idTask, idLecture, Status, Message, DateReported)
            VALUES(%s, %s, %s, %s, %s, NOW());""", \
                               (idMachine, idTask, idLecture, status, message))
            cursor.close()

            connection.commit()
            connection.close()
        else:
            Common.pr(2, "Suppressing repeated ERROR message.")

    @classmethod
    def updateLectureCropTimes(cls, idLecture, cropBegin=None, cropEnd=None):
        '''Edit crop times for the lecture'''

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBUser,
                                         passwd = conf.DBPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
            raise RecordingManagerException("MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(1, """UPDATE Lectures SET CropBegin=%s, CropEnd=%s, DateModified=NOW() WHERE idLecture=%s;\n""" % (cropBegin, cropEnd, idLecture))

        cursor = connection.cursor()
        cursor.execute("""UPDATE Lectures SET CropBegin=%s, CropEnd=%s, DateModified=NOW() WHERE idLecture=%s;""", (cropBegin, cropEnd, idLecture))
        cursor.close()

        connection.commit()
        connection.close()

    @classmethod
    def updateDuration(cls, idLecture, duration):
        """Given lecture ID and duration in seconds (integer), update the appropriate the Lectures table record."""

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBUser,
                                         passwd = conf.DBPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
            Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(1, """UPDATE Lectures SET Duration = %s WHERE idLecture = %s;\n""" % (duration, idLecture))

        cursor = connection.cursor()
        cursor.execute("""UPDATE Lectures SET Duration = %s WHERE idLecture = %s;""",
                       (duration, idLecture))
        cursor.close()

        connection.commit()
        connection.close()


    @classmethod
    def getMatchingLOID(cls, lecture_name):
        """Given a lecture ID of either the LOID form or IndicoID form, query the micala DB to find the matching record."""

        flagMatchFound = False

        try:
            connection = MySQLdb.connect(host   = conf.DBHost,
                                         port   = int(conf.DBPort),
                                         user   = conf.DBReaderUser,
                                         passwd = conf.DBReaderPW,
                                         db     = conf.DBName)
        except MySQLdb.Error, e:
            Common.pr(0, "MySQL database error %d: %s" % (e.args[0], e.args[1]))

        Common.pr(2, """SELECT idLecture,LOID,IndicoID FROM Lectures WHERE LOID = %s OR IndicoID = %s\n""" % (lecture_name, lecture_name))

        # Search for any matching records
        cursor = connection.cursor()
        cursor.execute("""SELECT idLecture,LOID,IndicoID FROM Lectures WHERE LOID = %s OR IndicoID = %s""",
            (lecture_name, lecture_name))
        connection.commit()

        result_set = cursor.fetchone()

        cursor.close()
        connection.close()

        if result_set is not None and len(result_set) > 0:
            Common.debug(1, "result_set: %s" % str(result_set))

            # If both IDs found, then we succeeded in finding a match
            if result_set[1] is not None and result_set[1] != '' \
                    and result_set[2] is not None and result_set[2] != '':
                flagMatchFound = True
                idRecord = result_set[0]
                LOID     = result_set[1]
                IndicoID = result_set[2]

        if not flagMatchFound:
            idRecord     = ''
            LOID         = ''
            IndicoID     = ''

        return(flagMatchFound, idRecord, LOID, IndicoID)



class Common():

    @classmethod
    def syscom(cls, command_string, print_output = True):
        """Use subprocess module to fork given command, print output as it appears, and return output."""

        print "running: %s" % command_string

        # Spawn a subprocess to run the script
        proc = subprocess.Popen(command_string,
                                shell  = True,
                                stdin  = subprocess.PIPE,
                                stdout = subprocess.PIPE,
                                stderr = subprocess.PIPE,)

        # Remember the starting time
        start_time = calendar.timegm(time.gmtime())

        output = ''

        keep_going = True
        while keep_going:
            # If you read too many chars at once, the read statement never executes.
            text_block = proc.stdout.read(1)

            if not text_block:
                keep_going = False
                break

            output = output + text_block

            if print_output:
                Common.pr(1, text_block)

        return output

    @classmethod
    def pr(cls, this_verbosity_level, message, log_this=True):
        """Home made version of the print statement.

        this is my own subroutine to replace the print statement.
        All output should be sent to this sub.  The advantage of
        this is that you can change verbosity levels, and also
        it makes it easier to turn it into a CGI or ASP script.

        log_this specifies whether or not to print to log file.
        We might not want to log messages from the daemon appearing over and over"""

        outstr = ''

        if this_verbosity_level == 0:
            outstr += 'ERROR! '

        if this_verbosity_level <= conf.VERBOSITY_LEVEL:
            outstr += message
            if conf.flag_logging and log_this:
                cls.log_entry(outstr)

            # Only print to console if consoleDisplay is True.
            # Web applications should set conf.consoleDisplay = False so statements don't show up in the web pages.
            if conf.consoleDisplay:
                # To avoid the line break or extra space after the line, use sys.stdout.write instead of print
                sys.stdout.write(outstr)

    @classmethod
    def debug(cls, this_debug_level, message):
        """Home made debug output function"""

        outstr = "DBG" + str(this_debug_level) + " " + message

        if int(this_debug_level) <= int(conf.DEBUG_LEVEL) and conf.consoleDisplay:
            print outstr

        if conf.flag_logging and int(this_debug_level) <= int(conf.DEBUG_LEVEL_LOGGING):
            cls.log_entry(outstr + "\n")

    @classmethod
    def log_entry(cls, message):
        """Print a one-line message to the log file. At the beginning of the line, include
        the time stamp and name of this script. However, if there is no newline char at the end
        of the string, supress this time stamp stuff."""

        time_struct = time.localtime()
        logfilename = os.path.join(conf.micalaHome, "log", "%s%04d%02d%s" % \
            (conf.logfile_base,
             time_struct.tm_year,
             time_struct.tm_mon,
             conf.logfile_ext))
        if len(message) > 0:
            logfile = open(logfilename, 'a')
            if message[-1] == '\n':
                logfile.write(time.strftime("%Y-%m-%d %H:%M:%S", time_struct) + ' ' + os.path.basename(sys.argv[0]) + ' ' + message)
            else:
                logfile.write(message)
            logfile.close()

    @classmethod
    def getVideoDuration(cls, pathCameraFile):
        """Use ffmpeg to get the duration of the given video file, returning the number of seconds rounded to the nearest integer. If duration can not be determined, return -1."""

        # use ffmpeg to get video duration
        command = "%s -i \"%s\"" \
            % (conf.ffmpeg, pathCameraFile)

        Common.pr(1, "Getting duration of video file.\n")
        Common.pr(1, "running command: " + command + "\n")

        # Spawn a subprocess to run the ffmpeg command and capture output
        proc = subprocess.Popen(command,
                                shell  = False,
                                stdin  = subprocess.PIPE,
                                stdout = subprocess.PIPE,
                                stderr = subprocess.PIPE,)

        output = ""
        keep_going = True
        while (keep_going):
            text_block = proc.stderr.read(1)
            if not text_block: keep_going = False

            output = output + text_block

        match = re.search("Duration\:\s*(\d+)\:(\d+)\:(\d+)\.(\d+)\,", output)
        if match:
            lectureDuration = int(match.group(1)) * 3600 + int(match.group(2)) * 60 + int(match.group(3))
            if int(match.group(4)) >= 50:
                lectureDuration = lectureDuration + 1
        else:
            lectureDuration = -1

        return lectureDuration

    @classmethod
    def formatCropTime(cls, cropTime):
        """Given number of seconds, translate it into a string of the form HH:MM:SS for use by ffmpeg."""

        if cropTime is not None and cropTime != "" and cropTime != 0:
            totalSeconds = int(cropTime)

            hours   = int(totalSeconds / 3600)
            seconds = totalSeconds % 60
            minutes = int((totalSeconds - (hours * 3600)) / 60)

            return "%02d:%02d:%02d" % (hours, minutes, seconds)
        else:
            return ""

    @classmethod
    def formatDuration(cls, duration):
        """Given the number of seconds, format duration nicely in a human readable string."""

        seconds = duration % 60
        hours   = int(duration / 3600)
        minutes = int( (duration - 3600 * hours) / 60 )

        if duration < 60:
            niceDuration = "%d seconds" % seconds
        #    elif duration < 600:
        #        niceDuration = "%d min %d s" % (minutes, seconds)
        elif duration < 3600:
            niceDuration = "%d minutes" % minutes
        else:
            niceDuration = "%dh%02dm" % (hours, minutes)

        return niceDuration

    @classmethod
    def formatTaskAge(cls, mysqlTimestamp):
        age = ""

        tDelta = datetime.datetime.now() - mysqlTimestamp

        if tDelta.days > 0:
            age = "%d days" % tDelta.days
        elif tDelta.seconds > 3600:
            age = "%d hours" % int(tDelta.seconds / 3600)
        elif tDelta.seconds > 60:
            age = "%d minutes" % int(tDelta.seconds / 60)
        else:
            age = "%d seconds" % tDelta.seconds

        return (age)

    @classmethod
    def getLectureLatestStatus(cls, lectureID, listTasks, listStatus):
        """Given the entire list of status updates, search for the given LOID or IndicoID and determine the most recent status for each task."""

        flagFinished = False

        dictTaskStatus = {}
        for task in listTasks:
            dictTerms = {}
            dictTerms["status"]    = ""
            dictTerms["message"]   = ""
            dictTerms["machine"]   = ""
            dictTerms["room"]      = ""
            dictTerms["started"]   = "" # time stamp of the most recent time this task was started
            dictTerms["completed"] = "" # time stamp of the most recent time this task was completed
            dictTerms["duration"]  = "" # duration of completed task in seconds
            dictTaskStatus[task["Name"]] = dictTerms

        # The list of status updates is already sorted chronologically, so just keep overwriting status
        # and we'll get the last one.
        for status in listStatus:
            if status["LOID"] is not None and status["LOID"] == lectureID \
                    or status["IndicoID"] is not None and status["IndicoID"] == lectureID:
                dictTaskStatus[status["Name"]]["status"]  = status["Status"]
                dictTaskStatus[status["Name"]]["message"] = status["Message"]
                dictTaskStatus[status["Name"]]["machine"] = status["Hostname"]
                dictTaskStatus[status["Name"]]["room"]    = status["RoomName"]

                # Instead of explicitly checking for when status["Status"] == "START", we
                # check for the first time any status was reported for the current task.
                # This is in case a program incorrectly reports RUNNING instead of START for example,
                # or only reports "COMPLETE" (The recording script used to do this).
                if status["Status"] == "START" or dictTaskStatus[status["Name"]]["started"] == "":
                    dictTaskStatus[status["Name"]]["started"] = status["DateReported"]
                if status["Status"] == "COMPLETE":
                    dictTaskStatus[status["Name"]]["completed"] = status["DateReported"]
                if dictTaskStatus[status["Name"]]["started"] != "" and dictTaskStatus[status["Name"]]["completed"] != "":
                    tDelta = \
                        dictTaskStatus[status["Name"]]["completed"] - \
                        dictTaskStatus[status["Name"]]["started"]
                    dictTaskStatus[status["Name"]]["duration"] = tDelta.seconds

        # Now check to see if this lecture is "finished" or not.
        # TODO: essential tasks should be updated as REQUESTED as appropriate for each lecture. That's how we'll decide if a task is finished. For now we'll just make sure the flash publishing task is started and complete.
        if dictTaskStatus["publishing flash WL"]["status"] is not None and dictTaskStatus["publishing flash WL"]["status"] == "COMPLETE":
            flagFinished = True

        # Check to make sure there are no started-not-completed tasks
        for (name, status) in dictTaskStatus.items():
    #          print "iterating: %s, %s<br />" % (name, status)
            if status["status"] is not None and status["status"] != "" and status["status"] != "COMPLETE":
                flagFinished = False

        return (flagFinished, dictTaskStatus)

    @classmethod
    def getConfId(self, IndicoID):
        """Given an Indico ID for a particular talk, return the general conference ID."""

        if IndicoID is not None and IndicoID != '':
            pattern = re.compile('^(\w*\d+)[sc]*')
            match = pattern.search(IndicoID)

            if match:
                return match.group(1)
            else:
                return "ERROR"
        else:
            return "none"
