-- ====================================================================
-- Copyright (c) 2010 The Regents of the University of Michigan and 
-- The European Organization for Nuclear Research (CERN). Licensed under the
-- Educational Community License, Version 2.0 (the "License"); you may
-- not use this file except in compliance with the License. You may
-- obtain a copy of the License at
--
-- http://www.osedu.org/licenses/ECL-2.0
--
-- Unless required by applicable law or agreed to in writing,
-- software distributed under the License is distributed on an "AS IS"
-- BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing
-- permissions and limitations under the License.
-- ====================================================================

-- This SQL file should be run by the MySQL root user

-- Must drop triggers before database can be dropped.
DROP TRIGGER IF EXISTS TriggerLecturesInsert;
DROP TRIGGER IF EXISTS TriggerStatusInsert;
DROP PROCEDURE IF EXISTS CreateLectureLatestStatus;

-- Delete existing database (currently done by mysqladmin)
DROP DATABASE IF EXISTS micala;

-- Create new database (currently done by mysqladmin)
CREATE DATABASE IF NOT EXISTS micala;

-- -------------------------------------------------------------
-- Users
--
-- This section assumes that all the user accounts already exist
-- (when you drop the database the users are still there)
-- and drops them all first. If they don't exist, you'll
-- need to comment out the DROP statements so you do not
-- get errors.
-- -------------------------------------------------------------
SOURCE users.sql

-- -------------------------------------------------------------
-- Tables
--
-- This section creates the database tables.
-- -------------------------------------------------------------

use micala;

-- Use the id's in this table for UserCreated, UserModified fields
CREATE TABLE Users (
       idUser       INT NOT NULL AUTO_INCREMENT,
       Username     VARCHAR(255),
       DateCreated  DATETIME,
       PRIMARY KEY (idUser)
);

-- I don't think this table is ever used, and it probably shouldn't be necessary...
INSERT INTO Users (Username, DateCreated)
       VALUES('reader', NOW()),
       ('cernwlcd0001', NOW()),
       ('cernwlcd0002', NOW()),
       ('cernwlcd0003', NOW()),
       ('umwlcd0010',   NOW()),
       ('processing01', NOW()),
       ('metadata01',   NOW());

-- list of all machines that are involved with lecture archiving
CREATE TABLE Machines (
       idMachine    INT NOT NULL AUTO_INCREMENT,
       Hostname     VARCHAR(255),
       idVenue      INT,
       IP           VARCHAR(100),
       CPU          VARCHAR(255),
       Memory       INT, -- amount of total memory in MB
       Disk         INT,
       OS           VARCHAR(255),
       DateCreated  DATETIME,
       DateModified DATETIME,
       UserCreated  INT,
       UserModified INT,
       PRIMARY KEY (idMachine)
);

-- List of possible tasks and an ordering to know which ones are generally done before others
CREATE TABLE Tasks (
       idTask       INT NOT NULL AUTO_INCREMENT,
       Name         VARCHAR(255),
       Ordering     INT,
       DateCreated  DATETIME,
       DateModified DATETIME,
       UserCreated  INT,
       UserModified INT,
       PRIMARY KEY (idTask)
);

-- Status of each Lecture Object
CREATE TABLE Lectures (
       idLecture    INT NOT NULL AUTO_INCREMENT,
       LOID         VARCHAR(255),
       IndicoID     VARCHAR(100),
       CDSRecord    INT,
       Title        VARCHAR(1000),
       Creator      VARCHAR(1000),
       contentType  ENUM('WEBLECTURE', 'PLAINVIDEO'),
       Duration     MEDIUMINT, -- duration in seconds (should never need big numbers)
       CropBegin    SMALLINT UNSIGNED, -- seconds into video to crop at beginning
       CropEnd      SMALLINT UNSIGNED, -- seconds into video to crop at end
       OriginPath   VARCHAR(1000), -- full path to network drive holding original file from which this Lecture Object was derived.
       DateCreated  DATETIME,
       DateModified DATETIME,
       UserCreated  INT,
       UserModified INT,
       PRIMARY KEY (idLecture)
);

-- List of recording venues
CREATE TABLE Venues (
       idVenue      INT NOT NULL AUTO_INCREMENT,
       RoomName     VARCHAR(255),
       IndicoRoomID INT,
       DateCreated  DATETIME,
       DateModified DATETIME,
       PRIMARY KEY (idVenue)
);

-- Every time a task is done by a machine to a resource, log it here
CREATE TABLE Status (
       idStatus   INT NOT NULL AUTO_INCREMENT,
       idMachine  INT,
       idVenue    INT, -- need idVenue here as well as in Machines table because machines might change venues over time
       idTask     INT,
       idLecture  INT,
       Status     ENUM('START', 'RUNNING', 'COMPLETE', 'ERROR', 'REQUESTED'),
       Message    VARCHAR(1000),
       IP         VARCHAR(100),
       MemoryFree INT, -- amount of memory free in MB
       CPULoad    INT, -- percent of CPU currently used
       DateReported  DATETIME, -- 'DateReported' used to be 'Time'
       PRIMARY KEY (idStatus)
);

-- Useful views that perform joins so we can see what is going on at a glance
CREATE VIEW ViewStatusComprehensive AS
    SELECT S.idStatus, S.idMachine, S.idTask, S.idLecture, S.Status, S.Message, S.DateReported, S.idVenue,
        T.idTask AS TidTask, T.Name,
        M.idMachine AS MidMachine, M.Hostname,
        L.idLecture AS LidLecture, L.LOID, L.IndicoID, L.CDSRecord
        FROM Status S, Tasks T, Machines M, Lectures L
        WHERE S.idMachine = M.idMachine && S.idTask = T.idTask && S.idLecture = L.idLecture;

-- I created another view for debugging purposes that only shows
-- certain columns from the first view because nested select statements
-- inside a view definition are not allowed.
CREATE VIEW ViewStatus AS SELECT LOID, Hostname, Name, Status, Message, DateReported FROM ViewStatusComprehensive;

-- Here is a procedure to generate the table 
DELIMITER |

DROP PROCEDURE IF EXISTS CreateLectureLatestStatus|

-- CreateLectureLatestStatus
--
-- This procedure takes no arguments; it creates the table LectureLatestStatus
-- from scratch (and deletes it first if it already exists).
-- It should not normally be necessary to call this procedure, since the Status
-- table trigger builds it as we go. But it will come in useful e.g. when
-- recovering from backup.
--
CREATE PROCEDURE CreateLectureLatestStatus ()
COMMENT 'Drop and Create the LectureLatestStatus table from scratch'
BEGIN

  -- counters to keep track of progress through Lectures and Tasks tables
  DECLARE ctrLecture  INT DEFAULT 0;
  DECLARE ctrTask     INT DEFAULT 0;

  -- variables to hold the total number of lectures and tasks
  DECLARE numLectures INT DEFAULT 0;
  DECLARE numTasks    INT DEFAULT 0;

  -- variables to keep track of the current lecture, task and status IDs
  DECLARE currIdLecture INT;
  DECLARE currIdTask    INT;
  DECLARE lastIdStatus  INT;

  -- cursors for iterating over the Lectures and Tasks tables
  DECLARE cursorLectures CURSOR FOR SELECT idLecture FROM Lectures;
  DECLARE cursorTasks    CURSOR FOR SELECT idTask    FROM Tasks;

  -- initialize LectureLatestStatus table
  DROP TABLE IF EXISTS LectureLatestStatus;
  CREATE TABLE LectureLatestStatus (idLecture INT, idTask INT, idStatus INT);

  -- open cursor for Lectures table
  OPEN cursorLectures;

  -- initialize lecture counter and count how many lectures there are total
  SET ctrLecture = 0;
  SELECT COUNT(*) FROM Lectures INTO numLectures;

  -- iterate through each lecture in the table Lectures
  LoopLectures: WHILE ctrLecture < numLectures DO
    FETCH cursorLectures INTO currIdLecture;

    -- itialize task counter and count how many tasks there are in total
    SET ctrTask = 0;
    SELECT COUNT(*) FROM Tasks INTO numTasks;

    -- Open cursor for Tasks table
    OPEN cursorTasks;

    -- iterate through each task in the table Tasks
    LoopTasks: WHILE ctrTask < numTasks DO
      FETCH cursorTasks INTO currIdTask;

      -- Get the Status record with the last status for a given lecture and task
      -- and insert that idStatus into the LectureLatestStatus table.
      SET lastIdStatus = NULL;
      SELECT idStatus FROM Status 
	  WHERE idStatus = ( 
	      SELECT MAX(idStatus) FROM Status 
	      WHERE idLecture = currIdLecture AND idTask = currIdTask
	    ) INTO lastIdStatus;
        INSERT INTO LectureLatestStatus (idLecture, idTask, idStatus) VALUES(currIdLecture, currIdTask, lastIdStatus);
      SET ctrTask = ctrTask + 1;
    END WHILE LoopTasks;
    SET ctrLecture = ctrLecture + 1;

    -- close cursor for tasks table (we need to close and open it anew to iterate over the tasks table each time)
    CLOSE cursorTasks;
  END WHILE LoopLectures;

  -- close cursor for Lectures table
  CLOSE cursorLectures;

END
|

DELIMITER ;

-- INSERT Trigger for Lectures table
-- Every time a new lecture record is added, we create a new record in
-- LectureLatestStatus for each task that could be done to it.
DROP TRIGGER IF EXISTS TriggerLecturesInsert;
DELIMITER |

CREATE TRIGGER TriggerLecturesInsert AFTER INSERT ON Lectures FOR EACH ROW
BEGIN
  -- declare variables, cursor
  DECLARE ctrTask    INT DEFAULT 0;
  DECLARE currIdTask INT;
  DECLARE numTasks   INT;
  DECLARE cursorTasks CURSOR FOR SELECT idTask FROM Tasks;

  OPEN cursorTasks;

  SELECT COUNT(*) FROM Tasks INTO numTasks;
  LoopTasks: WHILE ctrTask < numTasks DO
    FETCH cursorTasks INTO currIdTask;
    INSERT INTO LectureLatestStatus (idLecture, idTask, idStatus) VALUES(NEW.idLecture, currIdTask, NULL);
    SET ctrTask = ctrTask + 1;  
  END WHILE LoopTasks;

  CLOSE cursorTasks;

END
|
DELIMITER ;

-- INSERT Trigger for Status table
-- Each time a status update record is added, we update the
-- appropriate record in LectureLatestStatus (we know that these
-- records already exist because they were created by
-- TriggerLecturesInsert)
DROP TRIGGER IF EXISTS TriggerStatusInsert;
CREATE TRIGGER TriggerStatusInsert AFTER INSERT ON Status FOR EACH ROW
  UPDATE LectureLatestStatus SET idStatus = NEW.idStatus WHERE idLecture = NEW.idLecture AND idTask = NEW.idTask;
