# ====================================================================
# Copyright (c) 2010 The Regents of the University of Michigan and
# The European Organization for Nuclear Research (CERN). Licensed under the
# Educational Community License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
# http://www.osedu.org/licenses/ECL-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
# ====================================================================

######################################################
# create_preview.py
#
# Author: Jeremy Herr (herrj@umich.edu)
# 
# Description: this script is just a container to run the
# createPreview function, and will eventually just be part of
# get_lo.py, when get_lo.pl gets translated to python.  It generates
# two animated gifs, one from the camera feed and one from the slides
# feed, and installs them both in the Lecture Object itself as well as
# in the web directory.
# 
######################################################

import os
import sys
import time
import shutil
import glob
import re
import conf
from micala import *
import Image

def main(argv):

    currTask = "building preview"
    currMachine = conf.curr_machine

    (success, message, LOID) = parse_command_line(argv)

    if not success:
        Common.pr(0, message)
        sys.exit(2)

    idTask       = MicalaCommunication.getIdTask(currTask)
    idMachine    = MicalaCommunication.getIdMachine(currMachine)
    idLecture    = MicalaCommunication.getIdLecture(LOID)

    MicalaCommunication.reportStatus("START", "", idMachine, idTask, idLecture)

    createPreview(LOID, 
                  os.path.join(conf.dir_lo, LOID), 
                  os.path.join(conf.dir_lo, LOID, "media", "camera.f4v"), 
                  os.path.join(conf.dir_lo, LOID, "slides-capture-analyzed"))

    MicalaCommunication.reportStatus("COMPLETE", "", idMachine, idTask, idLecture)

def parse_command_line(arg_list):
    """Parse command line arguments."""

    success             = True
    message             = ''
    lecture_path        = ''
    lecture_path_camera = ''
    lecture_path_slides = ''
    time_begin          = ''
    time_end            = ''
    flag_batch_mode     = False

    # All we want here is a single LOID
    if len(arg_list) == 1:
        match = re.search("\d+\-[\d\w]+\-\d+", arg_list[0])

        if match:
            LOID = arg_list[0]

        else:
            message = "The command line argument you supplied does not appear to be an LOID."
            success = 0
            LOID = ""
    else:
        message = "You must supply exactly one command line argument, an LOID"
        success = 0

    return (success, message, LOID)

def createPreview(LOID, pathLOID, fileCamera, dirSlides):
    """Given the absolute path to a video file (camera feed) and absolute path to a directory (containing slides), generate two preview files in the form of animated gifs, using ffmpeg."""

    success = True

    # unique dir name to work in
    dirWorking = "preview-%s-%s" % (LOID, time.mktime(time.localtime()))

    # Create subdirectories for the various stuff we will do
    for dir in ["camera1", "camera2", "camera3", LOID, "slides1", "slides2", "slides3"]:
        os.makedirs(os.path.join(conf.dir_temp, dirWorking, dir))

    # ffmpeg command to create image sequence from video file.
    commandCreateImageSequence = "%s -t 00:%02d:00 -i %s -r 1 -s %dx%d %s 2>&1" % \
        (conf.ffmpeg,
         conf.previewFirstMinutes,
         fileCamera,
         conf.PREVIEW_THUMB_WIDTH,
         conf.PREVIEW_THUMB_HEIGHT,
         os.path.join(conf.dir_temp, dirWorking, "camera1", "preview%08d.gif"))
    print commandCreateImageSequence
    Common.syscom(commandCreateImageSequence)

    # Count how many files were created
    numImages = len(glob.glob(os.path.join(conf.dir_temp, dirWorking, "camera1", "*.gif")))

    # Pick out every n-th camera image file
    for ctr in range(1, numImages, conf.previewSecondsPerFrame):
        shutil.copy(os.path.join(conf.dir_temp, dirWorking, "camera1", "preview%08d.gif" % ctr),
                os.path.join(conf.dir_temp, dirWorking, "camera2", "preview%08d.gif" % ctr))

    # rename the camera images to be in sequential order
    ctr = 1
    for fileImage in glob.glob(os.path.join(conf.dir_temp, dirWorking, "camera2", "*.gif")):
        print "fileImage: %s" % fileImage
        shutil.copy(fileImage,
                    os.path.join(conf.dir_temp, dirWorking, "camera3", "preview%08d.gif" % ctr))
        ctr += 1

    # ffmpeg command to create camera.gif animated gif
    commandCreateAnimateGifCamera = "%s -r %d -i %s -pix_fmt rgb24 -loop_output 0 -r %d %s 2>&1" % \
        (conf.ffmpeg,
         conf.previewFPSCamera,
         os.path.join(conf.dir_temp, dirWorking, "camera3", "preview%08d.gif"),
         conf.previewFPSCamera,
         os.path.join(conf.dir_temp, dirWorking, LOID, conf.previewCamera))
    Common.syscom(commandCreateAnimateGifCamera)

    #########################################
    #     Now for the slides preview        #
    #########################################

    Common.pr(1, "Creating slides preview\n")

    # pick first n slides in slides dir
    ctr = 1
    if os.path.exists(dirSlides):
        print "%s exists" % dirSlides

    for pathSlide in glob.glob(os.path.join(dirSlides, "*.jpg")):
        if ctr < conf.previewNumSlides:
            print "ctr = %d" % ctr
            print "pathSlide: %s" % pathSlide
            shutil.copy(pathSlide, 
                        os.path.join(conf.dir_temp, dirWorking, "slides1"))
        else:
            break
        ctr += 1

    # resize slides and convert to gif at the same time
    for pathSlide in glob.glob(os.path.join(conf.dir_temp, dirWorking, "slides1", "*.jpg")):
        imObj = Image.open(pathSlide)
        imNew = imObj.resize((conf.PREVIEW_THUMB_WIDTH, conf.PREVIEW_THUMB_HEIGHT), Image.ANTIALIAS)
        match = re.search("(.*?)\.jpg", os.path.basename(pathSlide), re.IGNORECASE)
        pathSlideNew = match.group(1) + ".gif"
        print "pathSlideNew: %s" % pathSlideNew
        imNew.save(os.path.join(conf.dir_temp, dirWorking, "slides2", pathSlideNew), "GIF")
    
    # rename slides to be in sequential order
    ctr = 1
    for slideImage in glob.glob(os.path.join(conf.dir_temp, dirWorking, "slides2", "*.gif")):
        shutil.copy(slideImage,
                    os.path.join(conf.dir_temp, dirWorking, "slides3", "preview%08d.gif" % ctr))
        ctr += 1
    
    # ffmpeg command to create slides.gif animated gif
    commandCreateAnimateGifSlides = "%s -r %d -i %s -pix_fmt rgb24 -loop_output 0 -r %d %s 2>&1" % \
        (conf.ffmpeg,
         conf.previewFPSSlides,
         os.path.join(conf.dir_temp, dirWorking, "slides3", "preview%08d.gif"),
         conf.previewFPSSlides,
         os.path.join(conf.dir_temp, dirWorking, LOID, conf.previewSlides))
    print commandCreateAnimateGifSlides
    Common.syscom(commandCreateAnimateGifSlides)

    # If there's already an existing preview in the web directory, delete it
    if os.path.exists(os.path.join(conf.dirWebPreviews, LOID)):
        shutil.rmtree(os.path.join(conf.dirWebPreviews, LOID))
    # Copy whole preview directory to web directory, creating destination directory if necessary
    shutil.copytree(os.path.join(conf.dir_temp, dirWorking, LOID),
                    os.path.join(conf.dirWebPreviews, LOID))
    # Copy individual preview files to Lecture Object preview directory
    if not os.path.exists(os.path.join(pathLOID, "preview")):
        os.makedirs(os.path.join(pathLOID, "preview"))
    shutil.copy(os.path.join(conf.dir_temp, dirWorking, LOID, conf.previewCamera),
                os.path.join(pathLOID, "preview"))
    shutil.copy(os.path.join(conf.dir_temp, dirWorking, LOID, conf.previewSlides),
                os.path.join(pathLOID, "preview"))
    
    # Delete all the temporary working stuff
    shutil.rmtree(os.path.join(conf.dir_temp, dirWorking))

    return success

if __name__ == "__main__":
    main(sys.argv[1:])
