#!/usr/bin perl
# ====================================================================
# Copyright (c) 2010 The Regents of the University of Michigan and
# The European Organization for Nuclear Research (CERN). Licensed under the
# Educational Community License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
# http://www.osedu.org/licenses/ECL-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
# ====================================================================
###############################################################
# create_xml_timing.pl - get slide files and create lecture.xml from that
#
# Copyright 2004 by the Regents of the University of Michigan
# Author: Jeremy Herr (herrj@umich.edu)
# January 2004
#
###############################################################
# description goes here
#
# usage example:
# c:\acp\WLCD\scripts\create_xml_timing.pl 20060220-umwlcd0004-01-jenni --start=20060220140337
# where 20060220-umwlcd0004-01-jenni is a lecture object in the $::LO_DIR directory
# and 20060220140337 is the starting time of the lecture
#
# NB: search this file for the string "PLATFORM SPECIFIC CODE"
# before running in linux or other OS's
###############################################################

###############################################################
# PLATFORM SPECIFIC CODE
#
# definitions: file separator, copy command, default lecture object directory
###############################################################
$::FILESEP         = '\\';
$::OS_COMMAND_COPY = 'copy';
$::OS_COMMAND_MOVE = 'move';
$::LO_DIR          = 'C:' . $::FILESEP . "LectureObjects";
###############################################################
# load modules
###############################################################
use Time::Local;
use Micala::Common ':all';

###############################################################
# set defaults
###############################################################
$Micala::Common::VERBOSITY_LEVEL  = 1;
$Micala::Common::DEBUG_LEVEL      = 0;

my $timing_xml_file = 'timing-xml.txt';
my $time_stamp_file = 'start.txt';
my $given_formatted_start_time_stamp; # what the user provides on the command line
my $flag_no_chalk = 0;

my $noslides_image = 'C:' . $::FILESEP . 'WLAP-software' . $::FILESEP . 'conf' . $::FILESEP . 'NoSlides.jpg';

&main(@ARGV);

sub main()
{
    my @args = @_;

    my ($success, $lo) = &get_command_line(@args);

    if (! $success) {
	&pr(0, "command line", "problems with command line");
	exit(0);
    }

    my $path_lo = $::LO_DIR . ${::FILESEP} . $lo;

    my $timing_xml_string = &create_xml_timestamps($path_lo);

    &create_update_timing_xml($path_lo, $timing_xml_string);
}

###############################################################
# create_xml_timestamps
#
#
###############################################################
sub create_xml_timestamps()
{
    my ($lo_fullpath) = @_;
    my $timing_xml_string = '';
    my $slide_files = [];
    my $chalkboard_files = [];
    my $combined_images = []; # array of hashes containing slides, chalkboard images, in hash form
    my $sorted_combined_images = []; # the same thing, but sorted
    my $raw_time_stamp;
    my ($diff_sec, 
	$diff_min, 
	$diff_hour, 
	$diff_mday, 
	$diff_mon, 
	$diff_year, 
	$diff_wday, 
	$diff_yday, 
	$diff_isdst);
    my ($sec, 
	$min, 
	$hour, 
	$mday, 
	$mon, 
	$mon_given,
	$year, 
	$year_given, 
	$wday, 
	$yday, 
	$isdst);
    my $raw_start_time_stamp;
    my $found_time_stamp;

    opendir (LODIR, "$lo_fullpath${::FILESEP}slides")
	or die "can not open directory $lo_fullpath${::FILESEP}slides for reading: $!\n";
    @$slide_files = sort grep { m/^\d+\.jpg|png|gif|bmp$/i } readdir(LODIR);
    closedir (LODIR);

    # if there is a chalkboard directory and there are files in it,
    # then read them into an array
    if (-d "$lo_fullpath${::FILESEP}chalkboard" && ! $flag_no_chalk) {
	opendir (LODIR, "$lo_fullpath${::FILESEP}chalkboard")
	    or die "can not open directory $lo_fullpath${::FILESEP}chalkboard for reading: $!\n";
	@$chalkboard_files = sort grep { m/^chalkboard\-\d+\.jpg$/i } readdir(LODIR);
	# get rid of the "chalkboard-" string in each file name
	for(my $i=0; $i<scalar(@$chalkboard_files); $i++) {
	    my $chalkboard_file_name = $chalkboard_files->[$i];
	    ( $chalkboard_files->[$i] ) = $chalkboard_file_name =~ m/^chalkboard\-(\d+\.jpg)$/i;
	}
	closedir (LODIR);
    }

    # Check for the existence of at least one slide or chalkboard file.
    # If there is none, then insert a "dummy" slide that says "No slides available" on it.
    if ( scalar(@$slide_files) == 0 && scalar(@$chalkboard_files) == 0) {
	&pr(1, "WARNING: there are no slides in the slides directory\n");
	&pr(1, "    and no chalkboard images either, so a dummy slide is being inserted\n");
	&pr(1, "    (otherwise later publishing scripts will fail).\n");
	$slide_files->[0] = &insert_dummy_slide($lo_fullpath);

    }

#    opendir (LODIR, "$lo_fullpath${::FILESEP}chalkboard")
#	or die "can not open directory $lo_fullpath${::FILESEP}chalkboard for reading: $!\n";
#    @$slide_files = sort grep { m/^\d+\.jpg$/i } readdir(LODIR);
#    closedir (LODIR);

    ###########################################################
    # Get starting time stamp
    ###########################################################
    if (! &is_valid_time($given_formatted_start_time_stamp)) {

	$found_time_stamp = 0;
	# if a timestamp file exists, get the time stamp from that
	if (-r "$lo_fullpath${::FILESEP}scratch${::FILESEP}$time_stamp_file") {
	    open (TIMESTAMPFILE, "$lo_fullpath${::FILESEP}scratch${::FILESEP}$time_stamp_file")
		or die "can not open file $lo_fullpath${::FILESEP}scratch${::FILESEP}$time_stamp_file: $!\n";
	    ($raw_start_time_stamp) = join("", <TIMESTAMPFILE>) =~ m/(\d+)/;
	    close (TIMESTAMPFILE);
	    
	    if ($raw_start_time_stamp =~ m/^\d+$/) { $found_time_stamp = 1; }
	}
	#
	# otherwise, get the starting time stamp from the first slide filename
	#
	else {
	    &pr(1, "no time stamp file ( $time_stamp_file ) found.\n");
	    &pr(1, "Trying to get starting timestamp from first slide.\n");
	    if (&check_slide_loname_agreement($lo_fullpath, $slide_files->[0]) ) {
		if ($slide_files->[0] =~ m!^(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)\.!) {
		    ($year_given, $mon_given, $mday, $hour, $min, $sec)
			= ($1, $2, $3, $4, $5, $6);
		}
		my $mon = $mon_given - 1;
		my $year = $year_given - 1900;

		$raw_start_time_stamp = timelocal($sec, $min, $hour, $mday, $mon, $year);

		$found_time_stamp = 1;
	
	    }
	    else {
		&pr(1, "WARNING: time stamp of first slide does not agree with LO name.\n");
		&pr(1, "Renaming first slide found so that it has the same timestamp as the Lecture Object.\n");
		&pr(1, "    NB: This may or may not be the correct thing to do, and could indicate a problem!\n");
		&pr(1, "    You should check over this lecture to make sure all the slides are there,\n");
		&pr(1, "    although most of the time it is just caused by a small lag in the capture software.\n");

		my $new_first_slide_name = &rename_first_slide($lo_fullpath, $slide_files->[0]);
		$slide_files->[0] = $new_first_slide_name;

		if ($new_first_slide_name =~ m!^(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)\.!) {
		    ($year_given, $mon_given, $mday, $hour, $min, $sec)
			= ($1, $2, $3, $4, $5, $6);
		}
		my $mon = $mon_given - 1;
		my $year = $year_given - 1900;

		$raw_start_time_stamp = timelocal($sec, $min, $hour, $mday, $mon, $year);

		$found_time_stamp = 1;

	    }
	}
    }
    else {
	$found_time_stamp = 1;

	($year, $mon_given, $mday, $hour, $min, $sec)
	    = $given_formatted_start_time_stamp
	    =~ m/^(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)/;
	
	my $mon = $mon_given - 1;
	my $year = $year_given - 1900;

	&pr(1, "start time: " 
	      . sprintf("%04d-%02d-%02d %02d:%02d:%02d", 
			$year_given, $mon_given, $mday, $hour, $min, $sec) . "\n");

	$raw_start_time_stamp = timelocal($sec, $min, $hour, $mday, $mon, $year);
	
    }

    if (! $found_time_stamp) {
	&pr(0, "starting time", "no starting time stamp could be found.");
    }

    # now that we have a starting time stamp, 
    # populate $combined_images, an array of hashes 
    # containing all slides and chalkboard images
    my $index_combined = 0;
    foreach my $slide_file_name (@$slide_files) {
	$combined_images->[$index_combined++]
	    = { "file name"       => $slide_file_name,
		"image type"      => "slide",
		"raw time stamp"  => "",
		"diff time stamp" => ""};
    }
    if (scalar(@$chalkboard_files) > 0) {
	foreach my $chalkboard_file_name (@$chalkboard_files) {
	    $combined_images->[$index_combined++] 
		= { "file name"       => $chalkboard_file_name,
		    "image type"      => "chalkboard",
		    "raw time stamp"  => "",
		    "diff time stamp" => ""};
	}
    }
    # sort combined_images array by filename (which is a timestamp)
    @$sorted_combined_images 
	= sort {$a->{"file name"} cmp $b->{"file name"}} @$combined_images;

    # display sorted array
    foreach my $image_record (@$sorted_combined_images) {
	&pr(2, $image_record->{"file name"}."  ".$image_record->{"image type"}."\n");
    }

    foreach my $image_record (@$sorted_combined_images) {
	# if it's in the form of seconds since the epoch,
	# read it in that way
	if ($image_record->{"file name"} =~ m/^\d\d\d\d\d\d\d\d\d\d?\./ ) {
	    ($image_record->{"raw time stamp"})
		= $image_record->{"file name"}
		=~ m!(\d\d\d\d\d\d\d\d\d\d?)\.!;
	    ($diff_sec, 
	     $diff_min, 
	     $diff_hour, 
	     $diff_mday, 
	     $diff_mon, 
	     $diff_year, 
	     $diff_wday, 
	     $diff_yday, 
	     $diff_isdst) 
		= gmtime($image_record->{"raw time stamp"} - $raw_start_time_stamp);

	    &debug(1, "diff seconds: " . $image_record->{"raw time stamp"} - $raw_start_time_stamp);
	}
	# if it's in the form YYYYMMDDHHMMSS, do that
	elsif ($image_record->{"file name"} =~ m/^\d\d\d\d\d\d\d\d\d\d\d\d\d\d\./ ) {
	    my ($formatted_time_stamp) 
		= $image_record->{"file name"} 
		=~ m!^(\d\d\d\d\d\d\d\d\d\d\d\d\d\d)\.!;

	    ($year, $mon_given, $mday, $hour, $min, $sec)
		= $formatted_time_stamp
		=~ m/^(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)/;

	    my $mon = $mon_given - 1;

	    $image_record->{"raw time stamp"} 
	    = timelocal($sec, $min, $hour, $mday, $mon, $year);

	    ($diff_sec, 
	     $diff_min, 
	     $diff_hour, 
	     $diff_mday, 
	     $diff_mon, 
	     $diff_year, 
	     $diff_wday, 
	     $diff_yday, 
	     $diff_isdst) 
		= gmtime($image_record->{"raw time stamp"} - $raw_start_time_stamp);

	    &debug(1, $image_record->{"raw_time_stamp"} . " - " . $raw_start_time_stamp);
	    &debug(1, "diff seconds: " . eval($image_record->{"raw time stamp"} - $raw_start_time_stamp) );
	}
	

	my $nice_time_stamp = sprintf("%02d:%02d:%02d", $diff_hour, $diff_min, $diff_sec);
	&pr(1, $image_record->{"file name"} . " - $nice_time_stamp\n");

	my $image_file_name = '';
	my $title_text = '';
	if ($image_record->{"image type"} eq "slide") {
	    $image_file_name = "slides/" . $image_record->{"file name"};
	    $title_text = 'slide';
	}
	elsif ($image_record->{"image type"} eq "chalkboard") {
	    $image_file_name = "slides/" . $image_record->{"file name"};
	    $title_text = 'chalkboard';
	}

	$timing_xml_string .= "<SLIDE SRC=\"${image_file_name}\" BEGIN=\"$nice_time_stamp\" TYPE=\"image/jpg\" REGION=\"slide\" TITLE=\"$nice_time_stamp $title_text\" />\n";
    }

    my $outfile = $lo_fullpath . $::FILESEP . $timing_xml_file;
    open (OUTFILE, ">$outfile")
	or die "can not open $outfile for writing: $!\n";
    print OUTFILE $timing_xml_string;
    close (OUTFILE);

    &pr(1, "wrote results to $outfile\n");

    return $timing_xml_string;
}

###############################################################
# create_update_timing_xml
#
#
###############################################################
sub create_update_timing_xml()
{
    my ($lo_fullpath, $timing_xml_string) = @_;

    if (-e "$lo_fullpath${::FILESEP}lecture.xml") {
	# PLATFORM SPECIFIC CODE:
	&syscom(1, "$::OS_COMMAND_COPY $lo_fullpath${::FILESEP}lecture.xml $lo_fullpath${::FILESEP}lecture-backup.xml");

	open(LECTUREXML, "<$lo_fullpath${::FILESEP}lecture.xml")
	    or die "can not open $lo_fullpath${::FILESEP}lecture.xml for reading: $!\n";
	my $lecture_xml_contents = join("", <LECTUREXML>);
	close (LECTUREXML);

	my $new_slide_section = "<SEQ TITLE=\"Sequence of slides\" REGION=\"slide\" >\n"
	    . "$timing_xml_string\n"
	    . "</SEQ>\n";

	# Do this for when there are <SEQ> ...  </SEQ> tags
	$lecture_xml_contents 
	    =~ s!\<\s*SEQ\s+TITLE\s*\=\s*\"Sequence\s+of\s+slides\"\s+REGION\=\"slide\"\s*\>.*?\<\/SEQ\>!$new_slide_section!igs;

	# Do this for when there is a <SEQ ... /> tag
	$lecture_xml_contents 
	    =~ s!\<\s*SEQ\s+TITLE\s*\=\s*\"Sequence\s+of\s+slides\"\s+REGION\=\"slide\"\s*\/\s*\>!$new_slide_section!igs;

	open(LECTUREXMLOUT, ">$lo_fullpath${::FILESEP}lecture.xml")
	    or die "can not open $lo_fullpath${::FILESEP}lecture.xml for writing: $!\n";
	print LECTUREXMLOUT $lecture_xml_contents;
	close (LECTUREXMLOUT);

	&pr(1, "updated lecture.xml with new timing information\n");
    }
    else {

    }
}

##########################################################
# check_slide_loname_agreement
#
# Just check to make sure the YYYYMM DD and HHMMSS in the 
# loname and first slide filename agree, otherwise return failure
#
##########################################################
sub check_slide_loname_agreement
{
    my ($lopath, $firstslide) = @_;
    my 	($lo_year, $lo_month, $lo_day, $lo_hour, $lo_min, $lo_sec,
	 $start_year, $start_month, $start_day, $start_hour, $start_min, $start_sec);

    if ($lopath =~ m!(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)\_([\w\d]+)!) {
	($lo_year, $lo_month, $lo_day, $lo_hour, $lo_min, $lo_sec)
	    = ($1, $2, $3, $4, $5, $6);
    }
    elsif ($lopath =~ m!(\d\d\d\d)(\d\d)(\d\d)\-([\w\d]+)\-(\d\d)(\d\d)(\d\d)!) {
	($lo_year, $lo_month, $lo_day, $lo_hour, $lo_min, $lo_sec)
	    = ($1, $2, $3, $5, $6, $7);
    }

    if ($firstslide =~ m!^(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)\.!) {
	($start_year, $start_month, $start_day, $start_hour, $start_min, $start_sec)
	    = ($1, $2, $3, $4, $5, $6);
    }

    if ($lo_year     == $start_year
	&& $lo_month == $start_month
	&& $lo_day   == $start_day
	&& $lo_hour  == $start_hour
	&& $lo_min   == $start_min
	&& $lo_sec   == $start_sec   ) {
	return (1);
    }
    else {
	return (0);
    }

}

##########################################################
# insert_dummy_slide
#
# This should only be called if no slides are found in the
# lecture object being processed.
#
# It simply copies a special image with the message "No slides are available for this talk"
# into the lecture object slides/ directory, so that later scripts won't fail.
##########################################################
sub insert_dummy_slide
{
    my ($lo_fullpath) = @_;

    my 	($lo_year, $lo_month, $lo_day, $lo_hour, $lo_min, $lo_sec);

    if ($lo_fullpath =~ m!(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)\_([\w\d]+)!) {
	($lo_year, $lo_month, $lo_day, $lo_hour, $lo_min, $lo_sec)
	    = ($1, $2, $3, $4, $5, $6);
    }
    elsif ($lo_fullpath =~ m!(\d\d\d\d)(\d\d)(\d\d)\-([\w\d]+)\-(\d\d)(\d\d)(\d\d)!) {
	($lo_year, $lo_month, $lo_day, $lo_hour, $lo_min, $lo_sec)
	    = ($1, $2, $3, $5, $6, $7);
    }

    my $new_first_slide_file = sprintf("%04d%02d%02d%02d%02d%02d\.jpg", 
				    $lo_year, $lo_month, $lo_day, $lo_hour, $lo_min, $lo_sec);


    &pr(0, "inserting the file $noslides_image into $lo_fullpath${::FILESEP}slides${::FILESEP} \n");

    &syscom(1, "$::OS_COMMAND_COPY $noslides_image $lo_fullpath${::FILESEP}slides${::FILESEP}$new_first_slide_file");

    return ($new_first_slide_file);
}

##########################################################
# rename_first_slide
#
# This should only be called if the timestamp in the lecture object name
# does not match the timestamp in the title of the first slide.
#
# It simply makes a copy of the first slide found and renames it
# to match the lecture object name so later scripts don't fail.
##########################################################
sub rename_first_slide
{
    my ($lo_fullpath, $first_slide_file) = @_;

    # first get the starting timestamp contained in the LO name
    my 	($lo_year, $lo_month, $lo_day, $lo_hour, $lo_min, $lo_sec);

    if ($lo_fullpath =~ m!(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)\_([\w\d]+)!) {
	($lo_year, $lo_month, $lo_day, $lo_hour, $lo_min, $lo_sec)
	    = ($1, $2, $3, $4, $5, $6);
    }
    elsif ($lo_fullpath =~ m!(\d\d\d\d)(\d\d)(\d\d)\-([\w\d]+)\-(\d\d)(\d\d)(\d\d)!) {
	($lo_year, $lo_month, $lo_day, $lo_hour, $lo_min, $lo_sec)
	    = ($1, $2, $3, $5, $6, $7);
    }

    my $new_first_slide_file = sprintf("%04d%02d%02d%02d%02d%02d\.jpg", 
				    $lo_year, $lo_month, $lo_day, $lo_hour, $lo_min, $lo_sec);

    &pr(0, "EXECUTING: $::OS_COMMAND_MOVE $lo_fullpath${::FILESEP}slides${::FILESEP}$first_slide_file"
	   . " $lo_fullpath${::FILESEP}slides${::FILESEP}$new_first_slide_file");

    &syscom(1, "$::OS_COMMAND_MOVE $lo_fullpath${::FILESEP}slides${::FILESEP}$first_slide_file"
	   . " $lo_fullpath${::FILESEP}slides${::FILESEP}$new_first_slide_file");

    return ($new_first_slide_file);
}

##########################################################
# get_command_line
#
#
##########################################################
sub get_command_line()
{
    my @arg_list = @_;
    my $success = 1;
    my $message = '';
    my $lecture_object = '';

    # parse command line,
    # retrieve flag settings
    while (scalar(@arg_list) > 0) {
	my $arg = shift @arg_list;
	if ($arg =~ m/^\-/s) {
	    if ($arg =~ m/^\-\-verbose\=/s) {
		($VERBOSITY_LEVEL) = $arg =~ m/^\-\-verbose\=(\d+)/s;
	    }
	    elsif ($arg =~ m/^\-\-verbose$/s) {
		$VERBOSITY_LEVEL = 2;
	    }
	    elsif ($arg =~ m/^\-v$/s) {
		$VERBOSITY_LEVEL = 2;
	    }
	    elsif ($arg =~ m/^\-vv$/s) {
		$VERBOSITY_LEVEL = 3;
	    }
	    elsif ($arg =~ m/^\-vvv$/s) {
		$VERBOSITY_LEVEL = 4;
	    }
	    elsif ($arg =~ m/^\-\-debug\=/s) {
		($DEBUG_LEVEL) = $arg =~ m/^\-\-debug\=(\d+)/s;
	    }
	    elsif ($arg =~ m/^\-\-debug$/s) {
		$DEBUG_LEVEL = 1;
	    }
	    elsif ($arg =~ m/^\-d$/s) {
		$DEBUG_LEVEL = 1;
	    }
	    elsif ($arg =~ m/^\-\-start\=\S+$/s) {
		($given_formatted_start_time_stamp) = $arg =~ m/^\-\-start\=(\S+)$/s;
		&debug(1, "given_start=" . $given_formatted_start_time_stamp );
	    }
	    elsif ($arg =~ m/^\-\-lodir\=\S+$/s) {
		($::LO_DIR) = $arg =~ m/^\-\-lodir\=(\S+)$/s;
	    }
	    elsif ($arg eq '--nochalk') {
		$flag_no_chalk = 1;
	    }
	}
	else {
	    $lecture_object = $arg;
	}
    }
    
    # check in/out directories for existence,
    # and flags for validity
    if (!($VERBOSITY_LEVEL >= 0)) {
	$success = 0;
	&pr(0, "command line", "verbosity level must be between 0 and 5");
    }

    if ($lecture_object eq '') {
	$success = 0;
	&pr(0, "command line", "you must specify a lecture object");
    }

    return ($success, $lecture_object);
}

##########################################################
# is_valid_time
#
#
##########################################################
sub is_valid_time()
{
    my ($time_stamp) = @_;
    my $validity = 0;

    if ($time_stamp =~ m/^\d+$/) {
	$validity = 1;
    }
    else {
	$validity = 0;
    }

    return $validity;
}
