# ====================================================================
# Copyright (c) 2010 The Regents of the University of Michigan and
# The European Organization for Nuclear Research (CERN). Licensed under the
# Educational Community License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
# http://www.osedu.org/licenses/ECL-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
# ====================================================================
#########################################################
# get_lo.pl
#
# Description: downloads Lecture Objects from the upload server, 
# and performs most of the processing needed.
#
# Author: Jeremy Herr (herrj@umich.edu)
# copyright 2007-2009 by the Regents of the University of Michigan
#
# Usage:
# get_lo.pl
# running the script with no command line args compares the list
# of LO's in the file on the server with the local list, 
# and it downloads the first LO not yet processed, and processes it.
#
# get_lo.pl --lo=LONAME
# this downloads the given LO and processes it.
#
# get_lo.pl -r
# This removes the lockfile (in case the processing was aborted abnormally)
# so that the process can run again.
#########################################################

use strict;                # provides better debugging info
use Time::Local;           # for getting internal time stamp from time/date string
use Image::ExifTool::RIFF; # for accessing image and video metadata
use Image::Magick;         # for creating thumbnails
use Micala::Common ':all';

# Load default variables from global configuration file
do 'c:\\micala\\micala\\etc\\processing.conf';

$Micala::Common::VERBOSITY_LEVEL  = 5;
$Micala::Common::DEBUG_LEVEL      = 3;

my $seconds_leeway    = '3600'; # this concerns a subroutine that is no longer used.

# TODO: get rid of this shit. Most of it is handled by the database
# now. Not sure about style and done_url though.
my $fields_formatting = ['upload', 
			 'viewer_realplayer', 
			 'viewer_ipod', 
			 'viewer_flash',
			 'use_cwis_metadata',
			 'auto_analyze_slides',
			 'style',
			 'done_url'];

&main(@ARGV);

#######################################################
# main
#
#
#######################################################
sub main
{
    my @list_args = @_;

    my $list_lo_server = [];
    my $list_lo_local  = [];
    my $raw_list_los_to_be_processed = [];
    my $list_los_to_be_processed = [];
    
    my ($script_mode, $lo_override) = &parse_command_line(@list_args);
    
    #if (&already_running()) {
    if (-r $::lockfile_getlo) {
	&pr(0, "get_lo.pl is already running.\n");
	&pr(0, "If you are sure that this message is in error,\n");
	&pr(0, "you can run this script using the -r argument\n");
	&pr(0, "or delete the file $::lockfile_getlo to re-enable it.\n");
	exit(0);
    }

    &create_lockfile($::lockfile_getlo);
    
    &pr(2, "Getting list of lecture objects from server...\n");

    # if we're in force mode, then just process the given LO and exit
    if ($script_mode eq 'force') {
	my $lo_server = {};
	push (@$list_los_to_be_processed, $lo_override);
    }
    else {
	
	# first get LO list from server
	&debug(3, "Getting directory listing from server...");
	my $result = qx("$::ssh $::upload_arg_user_host ls uploads");
	if ($result == -1) {
	    &pr(0, "Failed to obtain list of Lecture Objects from server: $!\n");
	    exit(1);
	}
	else {
	    @$raw_list_los_to_be_processed = grep { m/^\d\d\d\d\d\d\d\d/ } split("\n", $result);
	}

	# Apply whatever filter you like
	foreach my $lo (@$raw_list_los_to_be_processed) {
	    if (&filter($lo)) {
		push(@$list_los_to_be_processed, $lo);
	    }
	}
	
    }

    ##############################################
    # Finally, do the actual processing
    #
    # Whether there is only one LO to be processed or many,
    # they will be stored in an array.  
    foreach my $lo (@$list_los_to_be_processed) {
	my $idMachine = &getIdMachine($::curr_machine);
	my $idTask    = &getIdTask("retrieving raw LO");
	my $idLecture = &getIdLecture($lo);
	
	&reportStatus('START', '', $idMachine, $idTask, $idLecture);

	&pr(1, "Retrieving lecture object $lo from server\n");

	# download LO into local temp dir
	&retrieve_lo($lo);

	&reportStatus('RUNNING', 'getting lecture object template', $idMachine, $idTask, $idLecture);
	
	# copy files from template directory
	# into lecture object (which is still in $::dir_temp)
	# the /E argument in windows xcopy copies all subdirectories and
	# files including empty ones (like -r in linux).  Apparently /I is
	# necessary to create a directory.
	my $command = "xcopy /I /E /EXCLUDE:" . $::excludes_file . " " . $::dir_templates_lo . $::FS . "*.* "
	    . $::dir_temp . $::FS . $lo . $::FS;
	&syscom(1, $command);
	
	&save_mscribe_loname($lo);

	my $formatting = &retrieve_formatting_metadata("$::dir_temp$::FS$lo");

	&reportStatus('RUNNING', 'analyzing slides', $idMachine, $idTask, $idLecture);

	if ($formatting->{'auto_analyze_slides'}) {
	    &screen_analyze($lo, 'auto');
	    &create_xml_timing($lo);
	}
	else {
	    &screen_analyze($lo, 'check');
	}

	&reportStatus('RUNNING', "moving to $::dir_lo", $idMachine, $idTask, $idLecture);

	# Give it an official Lecture Object name 
	# and move it to the Lecture Object directory
#	my $official_lo_name = &retrieve_metadata($lo);
	my $command = "move $::dir_temp$::FS$lo $::dir_lo$::FS";
	&syscom(1, $command);

	# create a subdirectory in the LO with preview metadata and images
	&create_preview($lo);

	# run MakeFLV.au3 to create flv file
#	&do_flv_processing($::dir_lo . $::FS . $lo);

	# add to the list of processed LO's
	&add_to_processed_list($lo);

	&reportStatus('COMPLETE', '', $idMachine, $idTask, $idLecture);
	
	# just process one at a time for now.
	last;
	&pr(1, "\n\nChecking for new lecture objects on server...\n");
    } # end foreach $lo in $list_los_to_be_processed

    &destroy_lockfile($::lockfile_getlo);

}

###############################################
# already_running
#
#
###############################################
sub already_running
{
    qx('tasklist ');
}

###############################################
# create_lockfile
#
#
###############################################
sub create_lockfile
{
    my ($lockfile) = @_;

    open (LOCKFILE, ">$lockfile")
	or die "can not open $lockfile for writing: $!\n";
    print LOCKFILE "This is a lock file created by get_lo.pl.\n";
    print LOCKFILE "It indicates that the script is currently running.\n";
    close (LOCKFILE);
}

###############################################
# destroy_lockfile
#
#
###############################################
sub destroy_lockfile
{
    my ($lockfile) = @_;

    &debug(3, "entering destroy_lockfile");

    &pr(2, "lockfile is $lockfile\n");

    unlink $lockfile;

    &debug(3, "leaving destroy_lockfile");
}

###############################################
# parse_command_line
#
#
###############################################
sub parse_command_line
{
    my @args = @_;
    my $script_mode = 'normal';
    my $lo_override;

    while (scalar(@args) > 0) {
	my $arg = shift @args;
	if ($arg =~ m!\-\-lo\=(.*)!) {
	    $lo_override = $1;
	    $script_mode = 'force';
	}
	elsif ($arg eq '-r') {
	    &debug(2, "trying to enable by deleting lockfile " . $::lockfile_getlo);
	    &destroy_lockfile($::lockfile_getlo);
	    &pr(1, "This script has now been re-enabled.\n");
	    exit(0);
	}
	# quiet mode
	elsif ($arg eq '-q') {
	    $Micala::Common::VERBOSITY_LEVEL  = 1;
	    $Micala::Common::DEBUG_LEVEL      = 0;
	}
	else {
	    &pr(0, "command line argument not recognized: $arg\n");
	    &pr(0, "exiting.\n");
	    exit(1);
	}

    }

    return ($script_mode, $lo_override);
}

###############################################
# retrieve_lo
#
# This subroutine retrieves the given lecture object from the server, by doing this:
# 1) tar up LO remotely on server
# 2) download and remove from server
# 3) untar LO locally and remove local tarball
###############################################
sub retrieve_lo
{
    my ($lo) = @_;
    my $command;
    my $result;

    # Move lo on server to tmp dir so nobody else downloads it by mistake
    $command = "\"$::ssh\" $::upload_arg_user_host \"mv $::server_dir_uploads/$lo $::server_dir_temp/\"";
    if (&syscom(1, $command) ) {
	# exit gracefully.
	&pr(0, "Unable to move directory $::server_dir_uploads/$lo on server\n");
	&destroy_lockfile($::lockfile_getlo);
	exit(0);
    }

    # Before downloading, tar up the the lo on the server
    $command = "\"$::ssh\" $::upload_arg_user_host \"cd $::server_dir_temp ;"
	. " tar cf ${lo}.tar $lo/*\"";
    &syscom(1, $command);

    # Probably should delete the lo on the server at this point
    #
    #
    #
    #

    # download tarball from server tmp dir to $::dir_temp
    my $dest = $::dir_temp;
    $dest =~ s/.\:(.*)$/$1/;
    $command = "\"$::scp\" -r $::upload_arg_user_host:$::server_dir_temp/${lo}.tar $dest";
    &syscom(1, $command);

    # remove tarball from server tmp dir
    $command = "\"$::ssh\" $::upload_arg_user_host \"rm -f $::server_dir_temp/${lo}.tar\"";
    &syscom(1, $command);

    # remove Lecture Object from server tmp dir
    $command = "\"$::ssh\" $::upload_arg_user_host \"rm -rf $::server_dir_temp/$lo\"";
    &syscom(1, $command);

    # untar the downloaded lo, which is now sitting in the $::dir_temp directory
    chdir ($::dir_temp);
    $command = "$::tar xf $lo.tar";
    &syscom(1, $command);

    # remove the local tarball
    &pr(1, "removing $lo.tar from $::dir_temp\n");
    &syscom(1, "del $::dir_temp$::FS$lo.tar");

}

###############################################
# tar_lo
#
#
###############################################
sub tar_lo
{
    my ($lo) = @_;
    my $command;

    $command = "\"$::ssh\" $::upload_arg_user_host \"cd $::server_dir_uploads ;"
	. " tar cf /tmp/${lo}.tar $lo/*\"";
    &syscom(1, $command);

}


###############################################
# download_lo
#
#
###############################################
sub download_lo
{
    my ($lo) = @_;
    my $command;
    
    # download tarball from server /tmp dir to $::dir_temp
    # This is annoying. When you download stuff using scp, you need to strip out the C:
    # at the beginning or it interprets the : as the separator between host and path
    my $dest = $::dir_temp;
    $dest =~ s/.\:(.*)$/$1/;
    $command = "\"$::scp\" -r $::upload_arg_user_host:/tmp/${lo}.tar $dest";
    &syscom(1, $command);

    # remove tarball from /tmp dir on server
    $command = "\"$::ssh\" $::upload_arg_user_host \"rm -f /tmp/${lo}.tar\"";
    &syscom(1, $command);

}

###############################################
# untar_lo
#
# untar the downloaded lo, which is now sitting in the $::dir_temp directory
###############################################
sub untar_lo
{
    my ($lo) = @_;

    chdir ($::dir_temp);

    my $command = "$::tar xf $lo.tar";
    &syscom(1, $command);
    
}
###############################################
# move_lo
#
#
###############################################
sub move_lo
{
    my ($lo) = @_;

    my $command = "move \/Y $::dir_temp$::FS$lo $::dir_lo_downloads";
    &syscom(1, $command);

    &pr(1, "removing $lo.tar from $::dir_temp\n");
    &syscom(1, "del $::dir_temp$::FS$lo.tar");

    #my $source_dir = "$dir_unpack\\$lo";
    #move($source_dir, $::dir_lo_downloads);
    
}
###############################################
# screen_analyze
#
# just call screen_analyze.pl
# and move the log file to the logs directory
###############################################
sub screen_analyze
{
    my ($lo, $mode) = @_;

    # If there's already a slides directory with stuff in it, 
    # then don't bother with screen_analyze
    my $path_to_slides = $::dir_temp . $::FS . $lo . $::FS . 'slides';
    if (-d $path_to_slides) {
	opendir (SLIDESDIR, $path_to_slides)
	    or die "can not open directory $path_to_slides even though it exists: $!\n";
	my @list_images = grep { m/png/i || m/jpg/i || m/bmp/i || m/gif/i } readdir SLIDESDIR;
	closedir (SLIDESDIR);

	if (scalar(@list_images) > 0) {
	    return;
	}
    }

    my $path_to_slides_capture = $::dir_temp . $::FS . $lo . $::FS . 'slides-capture';

    my $command = "$::screen_analyze_pl --mode=$mode $path_to_slides_capture";
    
    &syscom(1, $command);

    my $command2 = "move ${path_to_slides_capture}-analyzed" . $::FS 
	. "screen-analyze-log.txt" . ' ' . $::dir_temp . $::FS . $lo . $::FS . "logs" . $::FS;

    &syscom(1, $command2);

}

###############################################
# create_xml_timing
#
#
###############################################
sub create_xml_timing
{
    my ($lo) = @_;

    # create_xml_timing.pl --start=YYYYMMDDHHMMSS <--nochalk> LONAME
}

###############################################
# flag_for_processing
#
#
###############################################
sub flag_for_processing
{
    my ($lo) = @_;

    my $command = "copy $::file_flag_processing " . $::dir_temp . $::FS . $lo . $::FS;
    &syscom(1, $command);
}

###############################################
# save_mscribe_loname
#
#
###############################################
sub save_mscribe_loname
{
    my ($lo) = @_;

    my $file_upload_name = $::dir_temp . $::FS . $lo . $::FS . 'logs' . $::FS . 'upload_name';

    open (FILE, ">$file_upload_name")
	or die "can not open file $file_upload_name for writing: $!\n";
    print FILE $lo;
    close (FILE);
}

###############################################
# do_rm_processing
#
# Run master2rm, which creates multiple RM files
###############################################
sub do_rm_processing
{
    my ($lo) = @_;

    my $command = $::master2rm_pl;
    &syscom(1, $command);
}

##########################################################
# do_flv_processing
#
# Use ffmpeg to generate the flv file
#
##########################################################
sub do_flv_processing
{
    my ($path_lo) = @_;

    my $command = "$::ffmpeg -i \"$path_lo$::FS" . "media$::FS" . "master.avi\""
	. " -s 240x160"
	. " -b 150000"
	. " -vcodec flv"
	. " -ar 44100"
	. " \"$path_lo$::FS" . "media$::FS" . "master.flv\"";

	print "FLV Command:|$command|\n";


    &syscom(1, $command);
}

###############################################
# retrieve_metadata
#
# parse metadata.xml to retreve lo id,
# and update it with the duration 
###############################################
sub retrieve_metadata
{
    my ($lo) = @_;
    my $metadata_to_update = {};
    my $metadata_retrieved = {};

#    $metadata_to_update->{'duration'} = &get_duration($::dir_lo_downloads . $::FS . $lo . $::FS . "media" . $::FS . "master.avi");
#    $metadata_to_update->{'time'} = &get_start_time_from_slides_capture($lo);


    # Retrieve metadata from metadata.xml
    my $path_metadata_xml = $::dir_temp . $::FS . $lo . $::FS . 'metadata.xml';
    open(METADATA_XML, "<$path_metadata_xml")
	or die "can not open $path_metadata_xml for reading: $!\n";
    my $slurp_metadata_xml = join("", <METADATA_XML>);
    close (METADATA_XML);

    # can put this in a loop if there are more terms to be retrieved
    my $tag = 'id';
    $slurp_metadata_xml =~ m!<$tag>([^<]*)<\/$tag>!i;
    $metadata_retrieved->{$tag} = $1;

    # put any new values in metadata_to_update into the XML files
    &update_xml_files($lo, $metadata_to_update);
    

    return ($metadata_retrieved->{'id'});
}

###############################################
# guess_which_class
# NOT CURRENTLY BEING USED!!
# based on what day/time the class was recorded, guess
# which class it is and put this in logs/course.txt 
###############################################
sub guess_which_class
{
    my ($lo) = @_;
    my $file_course = $::dir_temp . $::FS . $lo . $::FS . 'logs' . $::FS . 'course.txt';
    my $course = '';
    my $lo_num = '';
    my $list_lo_schedule = [];
    my $flag_found_schedule = 0;
    my $hash_lo_schedule_info = {};
    my $metadata_to_update = {};

    &debug(5, "entering guess_which_class");

#    $metadata_to_update->{'duration'} = &get_duration($::dir_lo_downloads . $::FS . $lo . $::FS . "media" . $::FS . "master.avi");
#    $metadata_to_update->{'time'} = &get_start_time_from_slides_capture($lo);

    my ($yyyymmdd, $wlcd_box) = $lo =~ m/^(\d\d\d\d\d\d\d\d)\d\d\d\d\d\d\_(\w+)$/;
    $metadata_to_update->{'id'} = $yyyymmdd . '-' . $wlcd_box . '-' 
	. sprintf("%02d", $hash_lo_schedule_info->{'lecnum'});
    
    # First populate a hash with contents of recording schedule file.
    open(FILE_SCHEDULE, "<$::file_recording_schedule")
	or die "can not open $::file_recording_schedule for reading: $!\n";
    while (my $line = <FILE_SCHEDULE>) {
	my $hash_lo_schedule = {};
	$line =~ m/^(\d\d\d\d)\-(\d\d)\-(\d\d)\s+(\d\d)\:(\d\d)\:(\d\d)\s+(\w+)\s+(\w+)\s+\"([^\"]*)\"\s+\"([^\"]*)\"\s+\"([^\"]*)\"/;
	$hash_lo_schedule->{'year'}     = $1;
	$hash_lo_schedule->{'month'}    = $2;
	$hash_lo_schedule->{'day'}      = $3;
	$hash_lo_schedule->{'HH'}       = $4;
	$hash_lo_schedule->{'MM'}       = $5;
	$hash_lo_schedule->{'SS'}       = $6;
	$hash_lo_schedule->{'wlcd_box'} = $7;
	$hash_lo_schedule->{'course'}   = $8;
	$hash_lo_schedule->{'title'}    = $9;
	$hash_lo_schedule->{'creator'}  = $10;
	$hash_lo_schedule->{'location'} = $11;
	
	if ($hash_lo_schedule->{'creator'} ne '') {
	    $hash_lo_schedule->{'creator'} =~ m/^([^\,]*)\,\s*(.*)$/;
	    $hash_lo_schedule->{'creator_last'} = $1;
	    $hash_lo_schedule->{'creator_first'} = $2;
	}

	$hash_lo_schedule->{'internal time'}
	= timelocal($hash_lo_schedule->{'SS'},
		    $hash_lo_schedule->{'MM'},
		    $hash_lo_schedule->{'HH'}, 
		    $hash_lo_schedule->{'day'}, 
		    $hash_lo_schedule->{'month'} - 1, 
		    $hash_lo_schedule->{'year'});

	push(@$list_lo_schedule, $hash_lo_schedule);
    }
    close (FILE_SCHEDULE);
    

    if (! -r $::dir_lo_downloads . $::FS . $lo . $::FS . 'logs' . $::FS . 'course.txt') {
	$lo =~ m/^(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)/;
	my ($lo_year, $lo_month, $lo_day, $lo_hour, $lo_min, $lo_sec) 
	    = ($1, $2, $3, $4, $5, $6);

	my $lo_time_internal = timelocal($lo_sec, $lo_min, $lo_hour, $lo_day, $lo_month-1, $lo_year);

	my ($lo_out_sec, $lo_out_min, $lo_out_hour, 
	    $lo_out_day, $lo_out_month, $lo_out_year,
	    $lo_out_wday, $lo_out_yday, $lo_out_isdst) 
	    = localtime($lo_time_internal);

	($flag_found_schedule, $hash_lo_schedule_info) 
	    = &get_schedule_info($list_lo_schedule, $lo_time_internal, $wlcd_box);
	
	if ($flag_found_schedule) {
	    &pr(1, "Found schedule information: \n");
	    &pr(1, sprintf("%04d-%02d-%02d %02d:%02d:%02d %10s \"%s %s\" \"%s\"\n", 
			  $hash_lo_schedule_info->{'year'},
			  $hash_lo_schedule_info->{'month'},
			  $hash_lo_schedule_info->{'day'},
			  $hash_lo_schedule_info->{'HH'},
			  $hash_lo_schedule_info->{'MM'},
			  $hash_lo_schedule_info->{'SS'},
			  $hash_lo_schedule_info->{'course'},
			  $hash_lo_schedule_info->{'creator_first'},
			  $hash_lo_schedule_info->{'creator_last'},
			  $hash_lo_schedule_info->{'title'}
			  ) );

	    $metadata_to_update->{'date'} = sprintf("%04d-%02d-%02d", 
						    $hash_lo_schedule_info->{'year'},
						    $hash_lo_schedule_info->{'month'},
						    $hash_lo_schedule_info->{'day'}  );

	    &pr(1, "The date is: " . $metadata_to_update->{'date'} . "\n");

	    $metadata_to_update->{'title'} = $hash_lo_schedule_info->{'title'};
	    $metadata_to_update->{'location_room'} = $hash_lo_schedule_info->{'location'};
	    $metadata_to_update->{'creator_first'} = $hash_lo_schedule_info->{'creator_first'};
	    $metadata_to_update->{'creator_last'} = $hash_lo_schedule_info->{'creator_last'};
	}
	else {
	    &pr(1, "No schedule information found.\n");
	}

	foreach my $tag (sort keys %$metadata_to_update) {
	    &pr(1, sprintf("%15s -> %s\n", $tag, $metadata_to_update->{$tag}));
	}

#
#	open (COURSE, ">$file_course")
#	    or die "can not open $file_course for writing: $!\n";
#	print COURSE $course;
#	close (COURSE);
#
#	# given the course, copy files from appropriate template directory
#	# into lecture object
#	# the /E argument in windows xcopy copies all subdirectories and
#	# files including empty ones (like -r in linux).  Apparently /I is
#	# necessary to create a directory.  Retarded.
#	my $command = "xcopy /I /E " . $::dir_lo_downloads . $::FS . "template-" . $course . $::FS . "*.* " 
#	    . $::dir_lo_downloads . $::FS . $lo . $::FS;
#	print "\nCOMMAND: $command\n\n";
#	system($command);
#
#        # Since we know which course it is, we can give it an official
#        # Lecture Object name and move it to the official Lecture
#        # Object directory
#	my ($yyyymmdd, $wlcd_box) = $lo =~ m/^(\d\d\d\d\d\d\d\d)\d\d\d\d\d\d\_(\w+)$/;
#
#	my $official_lo_name = $yyyymmdd . "-" . $wlcd_box . "-" . $lo_num;
#	my $command2 = "move " . $::dir_lo_downloads . $::FS . $lo . " " . $::dir_lo . $::FS . $official_lo_name;
#	print "\nCOMMAND: [$command2]\n\n";
#	system($command2);


	# copy files from template directory
	# into lecture object
	# the /E argument in windows xcopy copies all subdirectories and
	# files including empty ones (like -r in linux).  Apparently /I is
	# necessary to create a directory.  Retarded.
	my $command = "xcopy /I /E " . $::dir_templates_lo . $::FS . "*.* " 
	    . $::dir_lo_downloads . $::FS . $lo . $::FS;
	&syscom(1, $command);

	&update_xml_files($lo, $metadata_to_update);

	# insert flag file so master2rm will do its thing
	&flag_for_processing($lo);
	
        # Since we know which course it is, we can give it an official
        # Lecture Object name and move it to the official Lecture
        # Object directory
	my $official_lo_name = $metadata_to_update->{'id'};
	my $command2 = "move " . $::dir_lo_downloads . $::FS . $lo . " " . $::dir_lo . $::FS . $official_lo_name;
	&syscom(1, $command2);


	# Re-create the original directory so that we know it's
	# already been downloaded
	my $command3 = "mkdir " . $::dir_lo_downloads . $::FS . $lo;
	&syscom(1, $command3);

	# run master2rm_all.pl
	&do_rm_processing($lo);


    }

    &debug(5, "exiting guess_which_class...");

}

###################################################################
# get_schedule_info
#
#
###################################################################
sub get_schedule_info
{
    my ($list_schedule_info, $internal_time_stamp) = @_;
    my $hash_lo = {};
    my $flag_found = 0;
    my $hash_lecnum_counter = {};

    foreach my $lo (@$list_schedule_info) {
	$hash_lecnum_counter->{$lo->{'year'} . $lo->{'month'} . $lo->{'day'} . '-' . $lo->{'wlcd_box'} }++;
	if ($internal_time_stamp - $lo->{'internal time'} > 0 &&
	    $internal_time_stamp - $lo->{'internal time'} < $seconds_leeway ) {
	    $flag_found = 1;
	    $hash_lo->{'year'}     = $lo->{'year'};
	    $hash_lo->{'month'}    = $lo->{'month'};
	    $hash_lo->{'day'}      = $lo->{'day'};
	    $hash_lo->{'HH'}       = $lo->{'HH'};
	    $hash_lo->{'MM'}       = $lo->{'MM'};
	    $hash_lo->{'SS'}       = $lo->{'SS'};
	    $hash_lo->{'course'}   = $lo->{'course'};
	    $hash_lo->{'title'}    = $lo->{'title'};
	    $hash_lo->{'creator'}  = $lo->{'creator'};
	    $hash_lo->{'creator_last'}  = $lo->{'creator_last'};
	    $hash_lo->{'creator_first'} = $lo->{'creator_first'};
	    $hash_lo->{'wlcd_box'}      = $lo->{'wlcd_box'};
	    $hash_lo->{'location'}      = $lo->{'location'};
	    $hash_lo->{'internal time'} = $lo->{'internal time'};
	    $hash_lo->{'lecnum'}   = $hash_lecnum_counter->{$lo->{'year'} . $lo->{'month'} . $lo->{'day'}};

	    last;
	}
    }

    return ($flag_found, $hash_lo);
}

###################################################################
# update_xml_files
#
#
###################################################################
sub update_xml_files
{
    my ($lo, $metadata_to_update) = @_;

    my $file_lecture_xml  = $::dir_temp . $::FS . $lo . $::FS . 'lecture.xml';
    my $file_metadata_xml = $::dir_temp . $::FS . $lo . $::FS . 'metadata.xml';

#    $metadata_to_update->{'author'} 
#    = $metadata_to_update->{'creator_first'} 
#    . " " . $metadata_to_update->{'creator_last'};

#    $metadata_to_update->{'creator'} 
#    = $metadata_to_update->{'creator_last'} 
#    . ", " . $metadata_to_update->{'creator_first'};


    open (LECTURE_XML, "<$file_lecture_xml")
	or die "can not open $file_lecture_xml for reading: $!\n";
    my $slurp_lecture_xml = join("", <LECTURE_XML>);
    close (LECTURE_XML);

    foreach my $tag (sort keys %$metadata_to_update) {
	my $value = $metadata_to_update->{$tag};
	$slurp_lecture_xml =~ s!(\<$tag\>)[^\<]*(\<\/$tag\>)!$1$value$2!ig;
    }

#    # deal with the lecture title separately
#    my $lec_title = $metadata_to_update->{'title'};
#    $slurp_lecture_xml =~ s!LECTURE\s+TITLE\=\"[^\"]*\"!LECTURE TITLE\=\"$lec_title\"!ig;

    open (LECTURE_XML, ">$file_lecture_xml")
	or die "can not open $file_lecture_xml for writing: $!\n";
    print LECTURE_XML $slurp_lecture_xml;
    close (LECTURE_XML);

    open (METADATA_XML, "<$file_metadata_xml")
	or die "can not open $file_metadata_xml for reading: $!\n";
    my $slurp_metadata_xml = join("", <METADATA_XML>);
    close (METADATA_XML);

    foreach my $tag (sort keys %$metadata_to_update) {
	my $value = $metadata_to_update->{$tag};
	$slurp_metadata_xml =~ s!(\<$tag\>)[^\<]*(\<\/$tag\>)!$1$value$2!ig;
    }

    open (METADATA_XML, ">$file_metadata_xml")
	or die "can not open $file_metadata_xml for writing: $!\n";
    print METADATA_XML $slurp_metadata_xml;
    close (METADATA_XML);

}

###################################################################
# get_duration
#
# use Image::ExifTool::RIFF module to get duration information of
# master.avi from AVI header.  Return a nice readable time string.
#
# P.S. this doesn't work for files longer than 44:29 recorded with XXXXX on the cart.
# But it does work on all master.avi files recorded on the laptops.
#
###################################################################
sub get_duration
{
    my ($avi_file) = @_;

    my $exifTool = new Image::ExifTool;
    $exifTool->Options(Unknown => 1);
    my $info = $exifTool->ImageInfo($avi_file);
    
    print "Got duration from EXIF tool: [" . $info->{'Duration'} . "]\n\n";
    print "Got frame count from EXIF tool: [" . $info->{'FrameCount'} . "]\n\n";
    print "Got frame rate from EXIF tool: [" . $info->{'FrameRate'} . "]\n\n";

    my $frame_count = $info->{'FrameCount'};
    my $frame_rate  = $info->{'FrameRate'};
    my $duration = sprintf("%d", $frame_count / $frame_rate);

#    my ($duration) = $duration_string =~ m/(\d+)\..*/;

    print "duration = [$duration]\n";

    my $hours = sprintf("%d", $duration / 3600);
    my $minutes = sprintf("%d", ($duration - (3600 * $hours) ) / 60);
    my $seconds = $duration % 60;

    print "HH: [$hours]\n";
    print "MM: [$minutes]\n";
    print "SS: [$seconds]\n";

    my $duration_output = sprintf("%02d:%02d:%02d", 
				  $hours, $minutes, $seconds);

    print "duration_output = \"$duration_output\"\n";

    return $duration_output;

}

###################################################################
# get_start_time_from_slide_capture
#
# look in the slides-capture directory for the first slide captured
# and lift the time stamp from its title.
# (I originally did it this way because Jim's time stamp in the LO name
# was not dependable.  But I think he may have fixed that).
###################################################################
sub get_start_time_from_slides_capture
{
    my ($lo) = @_;
    my $success = 0;
    my $start_time;

    my $path_slides_capture = $::dir_temp . $::FS . $lo . $::FS . 'slides-capture';

    if (-d $path_slides_capture) {

    }


    if (! $success && -d $path_slides_capture) {
	opendir (DIR_SLIDES_CAPTURE, $path_slides_capture)
	    or die "can not open directory $path_slides_capture for reading: $!\n";
	my @file_list = sort grep { m/^vga\-/ } readdir (DIR_SLIDES_CAPTURE);
	closedir (DIR_SLIDES_CAPTURE);
	
	if (scalar(@file_list)) {
	    $success = 1;
	    $file_list[0] =~ m/^vga\-(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)/;
	    $start_time = sprintf("%02d\:%02d\:%02d", $4, $5, $6);
	}
    }
    
    if ($success == 0) {
	&pr(0, "\n");
    }

    
    return ($start_time);
}

##########################################################
# filter
#
# apply a filter
##########################################################
sub filter
{
    my ($lo_name) = @_;


    # for now, ignore everything from the bioinformatics cart
    if ($lo_name =~ m/umwlcd0009/) {
	return 0;
    }
    else {
	return 1;
    }

}

##########################################################
# add_to_processed_list
#
#
##########################################################
sub add_to_processed_list
{
    my ($lo_name) = @_;
    
    open(LIST_DOWNLOADED, ">>$::file_lo_list_downloaded")
	or die "can not open $::file_lo_list_downloaded for appending: $!\n";
    print LIST_DOWNLOADED "$lo_name\n";
    close (LIST_DOWNLOADED);

    &debug(5, "$lo_name added to list of downloaded/processed lo's.");
}

##########################################################
# retrieve_formatting_metadata
#
# read in the formatting.xml file
##########################################################
sub retrieve_formatting_metadata
{
    my ($path_lo) = @_;
    my $hash_formatting = {};

    &debug(5, "retrieve_formatting_metadata: path_lo = $path_lo");

    if (-r "$path_lo/formatting.xml") {
	open (FORMATTING, "<$path_lo/formatting.xml")
	    or die "can not open $path_lo/formatting.xml for reading: $!\n";
	my $slurp_formatting_xml = join('', <FORMATTING>);
	close (FORMATTING);

	&debug(6, "slurp = [[[$slurp_formatting_xml]]]");

	foreach my $field (@$fields_formatting) {
	    if ($slurp_formatting_xml =~ m!<$field>(.*?)<\/$field>!) {
		my $raw_value = $1;
		
		&debug(6, "FOUND: $field -> $raw_value");

		if ($raw_value eq 'true') {
		    $hash_formatting->{$field} = 1;
		}
		elsif ($raw_value eq 'false') {
		    $hash_formatting->{$field} = 0;
		}
		else {
		    $hash_formatting->{$field} = $raw_value;
		}
	    }
	    else {
		$hash_formatting->{$field} = '';
	    }
	}

    }
    else {
	foreach my $field (@$fields_formatting) {
	    $hash_formatting->{$field} = '';
	}
    }
    
# <formatting>
#  <upload>false</upload>
#  <viewer_realplayer>false</viewer_realplayer>
#  <viewer_ipod>false</viewer_ipod>
#  <viewer_flash>false</viewer_flash>
#  <use_cwis_metadata>false</use_cwis_metadata>
#  <auto_analyze_slides>false</auto_analyze_slides>
#  <style>default</style>
#  <done_url />
#  </formatting>

    foreach my $field (@$fields_formatting) {
	&debug(6, $field . " -> " . $hash_formatting->{$field});
    }
    
    &debug(5, "exiting retrieve_formatting_metadata");

    return ($hash_formatting);
}

#######################################################
# create_preview
#
#
#######################################################
sub create_preview
{
    my ($lo) = @_;
    my $tags = {};
    my $list_file_video   = [];
    my $list_file_content = [];
    my $list_thumb_video   = [];
    my $list_thumb_content = [];

    &debug(3, "entering create_preview");

    # Run the python script to create animated gifs.
    # Maybe this is all we need, eventually we should stop creating the content....jpg preview files.
    # TODO: Delete the rest of this subroutine, after checking that the content.jpg files are not used any more.
    &syscom(1, "$::create_preview $lo");

    # Create preview subdirectory
    my $dir_preview = $::dir_lo . $::FS . $lo . $::FS . 'preview';
    &syscom(1, "mkdir $dir_preview");
    if (! -d $dir_preview) {
	&pr(0, "Could not create directory $dir_preview\n");
	return;
    }

    # Get image files from images/ directory, make thumbnails, insert into preview dir
    my $dir_images = $::dir_lo . $::FS . $lo . $::FS . 'images';
    if (-d $dir_images) {
	opendir(DIR_IMAGES, "$dir_images")
	    or die "can not open directory $dir_images: $!\n";
	@$list_file_video = sort grep { m/^\d+\./i } readdir DIR_IMAGES;
	closedir(DIR_IMAGES);

	my $ctr_thumb_video = 0;
	foreach my $file_video (@$list_file_video) {
	    my $file_output = sprintf("video%04d\.jpg", ++$ctr_thumb_video);
	    &create_thumb($lo,
			  $dir_images  . $::FS . $file_video,
			  $dir_preview . $::FS . $file_output);
	    push(@$list_thumb_video, $file_output);

	    if ($ctr_thumb_video > 3) { last; }
	}
    }

    # Get content images from slides-capture-analyzed/ directory
    my $dir_analyzed = $::dir_lo . $::FS . $lo . $::FS . 'slides-capture-analyzed';
    if (-d $dir_analyzed) {
	opendir(DIR_CONTENT, $dir_analyzed)
	    or die "can not open directory $dir_analyzed: $!\n";
	@$list_file_content = sort grep { m/^\d+\./i } readdir DIR_CONTENT;
	closedir(DIR_CONTENT);

	my $ctr_thumb_content = 0;
	foreach my $file_content (@$list_file_content) {
	    my $file_output = sprintf("content%04d\.jpg", ++$ctr_thumb_content);
	    &create_thumb($lo,
			  $dir_analyzed . $::FS . $file_content,
			  $dir_preview  . $::FS . $file_output);
	    push(@$list_thumb_content, $file_output);

	    print "$ctr_thumb_content\n";
	    if ($ctr_thumb_content > 6) { last; }
	}
    }


    # Get assorted metadata from LO name, number of slides

    # Build XML file
    my $xml_string = "<preview>\n"
	. "  <date>" . $tags->{'date'} . "</date>\n"
	. "  <time>" . $tags->{'time'} . "</time>\n"
	. "  <duration>" . $tags->{'duration'} . "</duration>\n"
	. "  <device>" . $tags->{'device'} . "</device>\n"
	. "  <filesize>" . $tags->{'filesize'} . "</filesize>\n"
	. "  <numchapters>" . $tags->{'numchapters'} . "</numchapters>\n";

    if (scalar(@$list_thumb_video) > 0) {
	$xml_string .= "  <thumbnails_video>\n";
	
	foreach my $thumbnail_video (@$list_thumb_video) {
	    $xml_string .= "    <video src=\"preview/" . $thumbnail_video . "\" />\n";
	}
	
	$xml_string .= "  </thumbnails_video>\n";
    }
    
    if (scalar(@$list_thumb_content) > 0) {
	$xml_string .= "  <thumbnails_content>\n";
	
	foreach my $thumbnail_content (@$list_thumb_content) {
	    $xml_string .= "    <content src=\"preview/" . $thumbnail_content . "\" />\n";
	}
	
	$xml_string .= "  </thumbnails_content>\n";
    }

    $xml_string .= "</preview>\n";

    my $file_preview_xml = $::dir_lo . $::FS . $lo . $::FS . 'preview.xml';
    open (PREVIEW_XML, ">$file_preview_xml")
	or die "can not open $file_preview_xml for writing: $!\n";
    print PREVIEW_XML $xml_string;
    close (PREVIEW_XML);

    &debug(3, "leaving create_preview");
}

##########################################################
# create_thumb
#
#
##########################################################
sub create_thumb
{
    my ($lo, $path_src, $path_dst) = @_;

    my $image = Image::Magick->new;
    my $result1 = $image->Read($path_src);
    warn "$result1" if "$result1";
	
    # this is where the resizing actually happens
    my $result2 = $image->AdaptiveResize(geometry=>$::PREVIEW_THUMB_WIDTH . 'x' . $::PREVIEW_THUMB_HEIGHT, filter=>'Cubic', blur=>0.9);
    warn "$result2" if "$result2";
    
    # set jpeg quality level
    my $result4 = $image->Set(quality=>$::PREVIEW_THUMB_QUALITY);
    warn "$result4" if "$result4";
    
    my $result3 = $image->Write($path_dst);
    warn "$result3" if "$result2";
    
    undef $image;

}
