# ====================================================================
# Copyright (c) 2010 The Regents of the University of Michigan and
# The European Organization for Nuclear Research (CERN). Licensed under the
# Educational Community License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
# http://www.osedu.org/licenses/ECL-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
# ====================================================================

import subprocess
import os
import sys
import time
import shutil
import glob
import re
import time
import conf
from micala import *
from optparse import OptionParser
import MySQLdb

def main(argv):

    flagFound = False

    # read directory where uploaded lecture.xml files are stored
    Common.pr(2, "Looking for uploaded metadata files...\n")
    for fileMetadata in glob.glob( os.path.join(conf.dirUpload, '*.xml') ):
        flagFound = True

        # first make sure this file is named correctly
        if isValidMetadataFileName(os.path.basename(fileMetadata)):

            # extract LO name from file name
            LOID = extractLOID(os.path.basename(fileMetadata))
            Common.pr(2, "Found metadata for %s\n" % LOID)

            # check to see if the LOID exists
            if os.path.exists(os.path.join(conf.dir_lo, LOID)):

                # check to see if lecture.xml exists in lecture object target directory. if so, make a backup copy
                if os.path.exists(os.path.join(conf.dir_lo, LOID, "lecture.xml")):
                    (timeNow, timeString) = generateTimestamp()
                    shutil.move(os.path.join(conf.dir_lo, LOID, "lecture.xml"), 
                                os.path.join(conf.dir_lo, LOID, "lecture-backup-%s.xml" % timeString))
                # move the lecture.xml file into place
                shutil.move(fileMetadata, os.path.join(conf.dir_lo, LOID, "lecture.xml"))

                Common.pr(1, "  Metadata installed in %s.\n" % LOID)

            else:
                Common.pr(2, "  No lecture object %s. Maybe you need to retrieve it with get_lo.pl first.\n" % LOID)
                Common.pr(2, "  The metadata file will be left intact.\n")

    if not flagFound:
        Common.pr(2, "None found, exiting.\n")

def parse_command_line(arg_list):
    """Parse command line arguments and do basic sanity check."""

    success         = True
    message         = ''
    lecture_object  = ''
    flag_batch_mode = False
    conferenceID    = 0
    contributionID  = 0

    # Create instance of parser
    parser = OptionParser()

    # verbosity level
    parser.add_option("-v", "--verbosity", "--verbose", 
                      dest="verbosity",
                      type="int",
                      action="store", 
                      help="set verbosity level", 
                      metavar="VERBOSITY_LEVEL")

    # debugging level
    parser.add_option("-d", "--debug", 
                      dest="debug",
                      action="store", 
                      help="set debugging level", 
                      metavar="DEBUG_LEVEL")

    # lecture object ID, if user wants to publish a particular lecture
    parser.add_option("--lo", "--lecture", "--lecture_object", 
                      dest="lecture_object",
                      action="store", 
                      help="lecture object ID",
                      metavar="LOID")

    # Parse the command line
    (options, args) = parser.parse_args(arg_list)

    # A lecture object must be specified
    if (options.lecture_object == ''):
        success = False
        flag_batch_mode = True
        message += "No lecture object specified, entering batch mode.\n"
        message += "This feature not yet enabled. Please specify a lecture object on the command line.\n"

    # check flags for validity
    if options.verbosity is not None:
        conf.VERBOSITY_LEVEL = options.verbosity
        if not (int(options.verbosity) >= 0 and int(options.verbosity) <= 10):
            success = False
            message = "verbosity level must be between 0 and 10"

    # check flags for validity
    if options.debug is not None:
        conf.DEBUG_LEVEL     = options.debug
        if not (int(options.debug) >= 0 and int(options.debug) <= 10):
            success = False
            message = "debugging level must be between 0 and 10"

    return (success,
            message,
            flag_batch_mode, 
            options.lecture_object)

def isValidMetadataFileName(fileName):
    """check to make sure given file name fits the format we are expecting."""

    # here's an example:
    # lecture-20100422-cernstream0001-184130
    pattern = re.compile("lecture\-\d\d\d\d\d\d\d\d\-[\w\d\_]+\-\d\d\d\d\d\d\.xml$")

    # match looks for patterns only at the beginning of the string
    match = pattern.match(fileName)

    if match:
        return True
    else:
        return False

def extractLOID(fileName):
    """Given fileName of submitted metadata file, extract the LOID."""

    pattern = re.compile("lecture\-(\d\d\d\d\d\d\d\d\-[\w\d\_]+\-\d\d\d\d\d\d)\.xml$")
    
    # perform match (starting at beginning of string)
    match = pattern.match(fileName)

    if match:
        return match.group(1)
    else:
        return ""

def generateTimestamp():
    """Generate a timestamp string with no spaces corresponding to the current time."""

    timeNow = time.localtime()

    year   = timeNow.tm_year
    month  = timeNow.tm_mon
    day    = timeNow.tm_mday
    hour   = timeNow.tm_hour
    minute = timeNow.tm_min
    second = timeNow.tm_sec

    return (timeNow, "%04d%02d%02d%02d%02d%02d" % \
        (year, month, day, hour, minute, second))

if __name__ == "__main__":
    main(sys.argv[1:])
