# ====================================================================
# Copyright (c) 2010 The Regents of the University of Michigan and
# The European Organization for Nuclear Research (CERN). Licensed under the
# Educational Community License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
# http://www.osedu.org/licenses/ECL-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
# ====================================================================

##########################################################
# lecture_publish_flash_wl.py
#
# Author: Jeremy Herr (herrj@umich.edu)
#
# This script transforms a Lecture Object (version 0.2)
# into a Web Lecture, viewable with the Flash player.
##########################################################
import sys
import subprocess
from optparse import OptionParser
import shutil
import time
import calendar
import re
import glob
import os
import tarfile
import MySQLdb
import Image
import conf
from micala import *

##########################################################
# main
#
#
##########################################################
def main(argv):
    """This is the main subroutine for publishing flash web lecture."""

    path_parent = ''

    currTask    = "publishing flash WL"

    # Parse command line and get settings
    (cl_success, cl_message, flag_batch_mode, \
         lecture_object, tarball, IndicoID) = \
         parse_command_line(argv)

    # Get name of lecture object from tarball name
    # (get rid of the ".tar")
    if tarball is not None and lecture_object is None:
        lecture_object = os.path.basename(tarball)[0:-4]

    Common.pr(1, "Started: " + __file__ + "\n")

    idMachine = MicalaCommunication.getIdMachine(conf.curr_machine)
    idTask    = MicalaCommunication.getIdTask(currTask)

    Common.debug(1, "cl success      = " + str(cl_success))
    Common.debug(1, "cl message      = " + str(cl_message))
    Common.debug(1, "flag_batch_mode = " + str(flag_batch_mode))
    Common.debug(1, "lecture_object  = " + str(lecture_object))
    Common.debug(1, "tarball         = " + str(tarball))

    if not cl_success:
	Common.pr(0, "failure getting command line, exiting.\n")
        Common.pr(0, "%s\n" % cl_message)
       # unix convention is to use exit code 2 for command line errors
        MicalaCommunication.reportStatus('ERROR', "failure getting command line, exiting.\n", idMachine, idTask, None)
	sys.exit(2)

    # If in batch mode, then we need to find an LO to publish
    if flag_batch_mode:
        (flagFoundLO, path_parent, lecture_object) = findLOToBePublished()
    else:
        path_parent = getLectureYear(lecture_object)

    Common.pr(1, " ...will publish lecture object: %s with parent directory %s\n" % \
                  (lecture_object, path_parent))

    # Get internal database id for current lecture object
    # (which might either be in WLAP or IndicoID format)
    idLecture = MicalaCommunication.getIdLecture(lecture_object)
    print "lecture id    = %s" % str(idLecture)

    MicalaCommunication.reportStatus('START', '', idMachine, idTask, idLecture)

    # define full path to the Lecture Object in the Media Archive
    if tarball is not None:
        if os.path.exists(tarball):
            # if we are working with a tarball, first untar into a temp directory
            MicalaCommunication.reportStatus('RUNNING', 'untarring master tarball', idMachine, idTask, idLecture)
            tarIn  = tarfile.open(tarball, 'r:')
            tarIn.extractall(path = conf.dir_temp)
            tarIn.close()
            MicalaCommunication.reportStatus('RUNNING', 'master tarball untarred', idMachine, idTask, idLecture)

            # Set path to the location of the untarred LO (in the temp dir)
            path_master_lo = os.path.join(conf.dir_temp, lecture_object)
        else:
            Common.pr(0, "given tar file not found: %s" % tarball)
            MicalaCommunication.reportStatus('ERROR', 'given tarball not found', idMachine, idTask, idLecture)
            sys.exit(1)
    else:
        # If we are given an already untarred LO, use its location in the media archive
        path_master_lo = \
            os.path.join(conf.dir_media_archive, path_parent, lecture_object)

    # To ensure as much uptime as possible, in case this LO has
    # already been published as a WL, we first build the web lecture
    # in a directory called web_lecture_creation, and once we are
    # done, delete the old one and immediately move the new one into
    # its place.
    web_lecture_creation = lecture_object + "_build"

    MicalaCommunication.reportStatus('RUNNING', "copying files", idMachine, idTask, idLecture)

    # Recursively copy contents of flash template directory to web_lecture_creation
    Common.debug(1, "Copying contents of " + conf.dir_viewer_flash + " to " + 
          os.path.join(conf.dir_public_area, path_parent, web_lecture_creation))

    if os.path.exists(os.path.join(conf.dir_public_area, path_parent, web_lecture_creation)):
        Common.pr(0, "the target directory for publishing this web lecture, %s, already exists. This may mean that the script %s aborted abnormally last time it ran, or that you are trying to run it while an instance is already running. Please either wait until the existing instance stops running before running this script, or delete the directory %s and try again.\n" % \
               (os.path.join(conf.dir_public_area, path_parent, web_lecture_creation), \
               __file__, \
               os.path.join(conf.dir_public_area, path_parent, web_lecture_creation)))
        Common.pr(1, "exiting.\n")
        MicalaCommunication.reportStatus('ERROR', "lecture object build dir already exists", idMachine, idTask, idLecture)
        sys.exit(1)
    else:
        # When we start using Python 2.6 we can use the ignore_patterns feature of copytree to avoid copying .gitignore files.
        shutil.copytree(os.path.join(conf.dir_viewer_flash, "template"),
                        os.path.join(conf.dir_public_area, path_parent, web_lecture_creation))
        # Now replace special tags in flash.html with values
        replace_tags(os.path.join(conf.dir_public_area, path_parent, web_lecture_creation, "flash.html"), 
                     lecture_object, 
                     getLectureYear(lecture_object))

    # Create FLV file inside the Lecture Object.
    Common.pr(1, "Creating FLV file now.\n")
    do_flv_processing(path_master_lo, idMachine, idTask, idLecture)

    # Now that there is an FLV file, copy it from Media Archive
    # into web lecture (slave).
    MicalaCommunication.reportStatus('RUNNING', "copying FLV file", idMachine, idTask, idLecture)
    if os.path.exists(os.path.join(path_master_lo, "media", "master.flv")):
        Common.pr(1, "Copying FLV file into place.\n")
        shutil.copy(os.path.join(path_master_lo, "media", "master.flv"),
                    os.path.join(conf.dir_public_area, path_parent, web_lecture_creation))
    else:
        Common.pr(0, "the file %s does not exist. It should!\n" % \
               os.path.join(path_master_lo, "media", "master.flv"))
        sys.exit(1)

    # Copy lecture.xml from Media Archive into public area
    if os.path.exists(os.path.join(path_master_lo, "lecture.xml")):
        shutil.copy(os.path.join(path_master_lo, "lecture.xml"),
                    os.path.join(conf.dir_public_area, path_parent, web_lecture_creation))
    else:
        Common.pr(0, "missing lecture.xml file!")
        
    # Resize slides and move them into place
    Common.pr(1, "Resizing slides...\n")
    MicalaCommunication.reportStatus('RUNNING', "resizing slides", idMachine, idTask, idLecture)
    resize_images(os.path.join(path_master_lo, 'slides'),
                  os.path.join(conf.dir_public_area, path_parent, 
                               web_lecture_creation, 'real', 'slides'), 
                  conf.SLIDE_WIDTH, 
                  conf.SLIDE_HEIGHT, 
                  conf.SLIDE_QUALITY)

    # Resize thumbnails and move them into place
    Common.pr(1, "Creating thumbnail images...\n")
    MicalaCommunication.reportStatus('RUNNING', "creating thumbnails", idMachine, idTask, idLecture)
    resize_images(os.path.join(path_master_lo, 'slides'),
                  os.path.join(conf.dir_public_area, path_parent, 
                               web_lecture_creation, 'thumbs', 'slides'), 
                  conf.THUMB_WIDTH, 
                  conf.THUMB_HEIGHT, 
                  conf.THUMB_QUALITY)

    # Delete existing web lecture (if any)
    if os.path.exists(os.path.join(conf.dir_public_area, path_parent, lecture_object)):
        Common.pr(1, "Lecture already published. Deleting existing \"slave\".\n")
        try:
            shutil.rmtree(os.path.join(conf.dir_public_area, path_parent, lecture_object))
        except:
            Common.pr(0, "error attempting to delete directory tree %s",
               os.path.join(conf.dir_public_area, path_parent, lecture_object))

    # Move web_lecture_creation into its place
    Common.pr(1, "Installing Flash web lecture to %s.\n" % \
           os.path.join(conf.dir_public_area, path_parent, lecture_object))
    MicalaCommunication.reportStatus('RUNNING', "installing Flash web lecture", idMachine, idTask, idLecture)
    Common.debug(1, "Moving %s to %s.\n" % \
           (os.path.join(conf.dir_public_area, path_parent, web_lecture_creation),
            os.path.join(conf.dir_public_area, path_parent, lecture_object)))
    shutil.move(os.path.join(conf.dir_public_area, path_parent, web_lecture_creation),
                os.path.join(conf.dir_public_area, path_parent, lecture_object))

    # Finally, delete the signal file (if one was used)
    if flag_batch_mode:
        if conf.flag_signal_file:
            Common.debug(1, "Deleting signal file.\n")
            os.remove(os.path.join(conf.dir_awaiting_posting_flash, 
                                   list_signal_files[0]))
        # don't need an else statement because the "COMPLETE" status update 
        # will ensure it doesn't get published again

    # Additional things to do for tarballs
    if tarball is not None:
        # delete LO created from unpacked tarball in temp dir
        # (double check to make sure it's in the temp dir and not the media archive)
        if path_master_lo.rfind(conf.dir_temp) != -1:
            Common.debug(1, "Deleting directory %s\n" % path_master_lo)
            shutil.rmtree(path_master_lo)


    Common.pr(1, "Finished: " + __file__ + "\n")
    MicalaCommunication.reportStatus('COMPLETE', '', idMachine, idTask, idLecture)


##########################################################
# parse_command_line
#
#
##########################################################
def parse_command_line(arg_list):
    """Parse command line arguments and do basic sanity check."""

    success         = True
    message         = ''
    flag_batch_mode = False

    # Create instance of parser
    parser = OptionParser()

    # verbosity level
    parser.add_option("-v", "--verbosity", "--verbose", 
                      dest="verbosity",
                      type="int",
                      action="store", 
                      help="set verbosity level", 
                      metavar="VERBOSITY_LEVEL")

    # debugging level
    parser.add_option("-d", "--debug", 
                      dest="debug",
                      action="store", 
                      help="set debugging level", 
                      metavar="DEBUG_LEVEL")

    # lecture object ID, if user wants to publish a particular lecture
    # (using Michigan naming convention)
    parser.add_option("--lo", "--lecture", "--lecture_object", 
                      dest="lecture_object",
                      action="store", 
                      help="lecture object ID in the format YYYYMMDD-devicename-HHMMSS",
                      metavar="LOID")

    # tarball, if user wants to publish a tarred up lecture object
    parser.add_option("--tarball", "--tar", 
                      dest="tarball",
                      action="store", 
                      help="lecture object tarball",
                      metavar="TARBALL")

    # change the default lecture object directory on local processing PC
    # where lecture objects are stored.
    parser.add_option("--lodir", dest="lodir",
                      action="store", 
                      help="absolute path to default lecture object directory",
                      metavar="DIRNAME")

    # Indico ID
    parser.add_option("--IndicoID", "--indico", 
                      dest="IndicoID",
                      action="store", 
                      type="int",
                      help="Indico ID",
                      metavar="INDICOID")

    # Parse the command line
    (options, args) = parser.parse_args(arg_list)

    # Either a lecture object, IndicoID, or tarball must be specified,
    # If none of these are given, then go into batch mode
    if (options.lecture_object is None \
            and options.IndicoID is None \
            and options.tarball is None):
	flag_batch_mode = True
	Common.pr(1, "No lecture object specified, entering batch mode.\n")

    # check flags for validity
    if options.verbosity is not None:
        conf.VERBOSITY_LEVEL = options.verbosity
        if not (int(options.verbosity) >= 0 and int(options.verbosity) <= 10):
            success = False
            message = "verbosity level must be between 0 and 10"

    # check flags for validity
    if options.debug is not None:
        conf.DEBUG_LEVEL     = options.debug
        if not (int(options.debug) >= 0 and int(options.debug) <= 10):
            success = False
            message = "debugging level must be between 0 and 10"
    
    # should only give lecture object directory OR tarball, not both
    if options.lecture_object is not None and options.tarball is not None:
        success = False
        message = "You have specified both a lecture object directory and a tarball. You should only specify one of these."

    # check for existence of tarball
    if options.tarball is not None and not os.path.exists(options.tarball):
        success = False
        message = "Given tar file does not exist: %s" % options.tarball

    return (success,
            message,
            flag_batch_mode, 
            options.lecture_object,
            options.tarball,
            options.IndicoID)

##########################################################
# findLOToBePublished
#
# Since we are in batch mode, we need to find lecture objects waiting to be published.
# This function returns a flag if one is found, 
#
# If we are in batch mode (no LO given), then we need to
# look in dir_awaiting_posting_flash and use the first
# signal file.
# NOTE: This should be changed to look in the DB for lectures whose
# task "uploading complete LO" is COMPLETE and whose task
# "publishing flash WL" has no status updates at all.
#
##########################################################
def findLOToBePublished():

    # get list of lecture object names in dir_awaiting_posting_flash
    Common.pr(1, "looking for lectures to process in "
       + conf.dir_awaiting_posting_flash + ".\n")

    if not os.path.exists(conf.dir_awaiting_posting_flash):
        Common.pr(0, 'directory ' + conf.dir_awaiting_posting_flash \
               + ' does not exist. Exiting.\n')
        sys.exit(0)

    # pattern to match for the signal file
    # Here we are looking for 
    # CONFID 
    # or CONFIDcCONTRID 
    # or CONFIDcCONTRIDscSCONTRID 
    # or CONFIDsSESSID
    pattern_cern  = re.compile('([sc\d]+)$')
    # Here we are looking for YYYYMMDD-DEVICENAME-HHMMSS
    pattern_umich = re.compile('(\d+\-[\w\d]+\-\d)$')

    # Get list of signal files
    # NOTE: DOES THIS SORT THE FILES IN ALPHABETICAL ORDER??
    list_signal_files = glob.glob(os.path.join(conf.dir_awaiting_posting_flash, '*'))

    # If the signal file is in the Michigan LO format, then we need to
    # get the directory name where it is stored from the XML file. If
    # it is in the Indico ID format, then you
    # have the location right there
    if len(list_signal_files) > 0:
        match_cern  = pattern_cern.search(list_signal_files[0])
        match_umich = pattern_umich.search(list_signal_files[0])
        if match_cern:
            lecture_object = match_cern.group(1)
            path_parent = get_lo_year(lecture_object)
        elif match_umich:
            # I'm pretty sure get_media_archive_location needs to be completely rewritten
            (path_parent, lecture_object) = \
            get_media_archive_location(list_signal_files[0])
        else:
            Common.pr(1, "unable to find signal file, looking for matching LOID, IndicoID in micala database.\n")
            (flagMatchFound, lecture_object_DBID, lecture_object, IndicoID) = \
                MicalaCommuncation.getMatchingLOID(lecture_object)

            if not flagMatchFound:
                Common.pr(0, "signal file " + \
                       list_signal_files[0] + \
                       " is not in a recognized format\n")
    else:
        MicalaCommunication.reportStatus('COMPLETE', '', idMachine, idTask, None)
        Common.pr(1, "There are no lecture objects awaiting posting in Flash. Exiting\n")
        Common.pr(1, "Finished: " + __file__ + "\n")
        sys.exit(0)

    return (flagFoundLO, path_parent, lecture_object)


##########################################################
# replace_tags
#
#
##########################################################
def replace_tags(filehtml, lecture_id, lecture_year):
    """Replace the special tags in the flash.html INSERT_YEAR and INSERT_LECTURE_ID with the appropriate values for the lecture object's year and the lecture_id (IndicoID)."""

    Common.debug(1, "in replace tags\n")
    Common.debug(1, "  replacing INSERT_LECTURE_ID with: %s" % lecture_id)
    Common.debug(1, "  replacing INSERT_YEAR with:       %s" % lecture_year)

    file_handle = open(filehtml, 'r')
    contents_filehtml = file_handle.read(10000)
    file_handle.close()

    # replace up to 10 occurrences of the strings (only need 2, hopefully 10 will cover any future changes)
    contents_replaced_year = contents_filehtml.replace("INSERT_YEAR",            lecture_year, 10)
    contents_replaced_id   = contents_replaced_year.replace("INSERT_LECTURE_ID", lecture_id,   10)

    file_handle_out = open(filehtml, "w")
    file_handle_out.write(contents_replaced_id)
    file_handle_out.close()
    
##########################################################
# getLectureYear
#
#
##########################################################
def getLectureYear(lecture_id):
    """Determine the year in which the lecture object occurred. If it's in the Michigan LOID style, just get it from the LOID itself.
    Otherwise, look it up in the database."""

    # pattern to match for the signal file
    # Here we are looking for 
    # CONFID 
    # or CONFIDcCONTRID 
    # or CONFIDcCONTRIDscSCONTRID 
    # or CONFIDsSESSID
    pattern_cern  = re.compile('([sc\d]+)$')
    # Here we are looking for YYYYMMDD-DEVICENAME-HHMMSS
    pattern_umich = re.compile('(\d\d\d\d)\d\d\d\d\-[\w\d]+\-\d')

    match_cern  = pattern_cern.search(lecture_id)
    match_umich = pattern_cern.search(lecture_id)

    if match_umich:
        year = match_umich.group(1)
    if match_cern:
        resultIDs = MicalaCommunication.getMatchingLOID(lecture_id)
        # If the LOID was successfully found, then get it
        if resultIDs[0]:
            LOID = resultIDs[2]
            match_umich_LOID = pattern_umich.search(LOID)
#            Common.pr(1, "success  = %s\n" % resultIDs[0])
#            Common.pr(1, "idRecord = %s\n" % resultIDs[1])
#            Common.pr(1, "LOID     = %s\n" % resultIDs[2])
#            Common.pr(1, "IndicoID = %s\n" % resultIDs[3])

#            Common.pr(1, "lecture_id = %s\n" % lecture_id)
#            Common.pr(1, "match_obj = %s\n" % match_umich_LOID)

#            sys.exit(0)
            year = match_umich_LOID.group(1)
        # If matching LOID not found, there's probably some corrupt information in the database
        else:
            Common.pr(0, "Unable to find Lecture Object ID that matches %s in the database." % lecture_id)
            sys.exit(1)

    return year

##########################################################
# do_flv_processing
#
# Use ffmpeg to generate the flv file
##########################################################
def do_flv_processing(path_lo, idMachine, idTask, idLecture):

    MicalaCommunication.reportStatus('RUNNING', "creating FLV file", idMachine, idTask, idLecture)

    # We should check for existence of master.avi or master.WHATEVER and process the master file we have.
    # arguments:
    # -y            (overwrite master.flv if it exists. This could happen because we used to store a copy of the flv file in the Lecture Object)
    # -i master.avi (input file)
    # -s '240x160'  (set the output size)
    # -b 150000     (set something I don't recall)
    # -vcodec flv   (set output codec)
    # -ar 44100     (set audio bitrate)
    # master.flv    (output file)
    command = "%s -y -i \"%s\" -s %s -b %d -vcodec %s -ar %d \"%s\"" \
    % (conf.ffmpeg,
       os.path.join(path_lo, "media", "master.avi"),
       '240x160',
       150000,
       'flv',
       44100,
       os.path.join(path_lo, "media", "master.flv"))

    print "running command: " + command + "\n"

    # Spawn a subprocess to run the ffmpeg command and capture output
    proc = subprocess.Popen(command, 
                            shell  = True,
                            stdin  = subprocess.PIPE,
                            stdout = subprocess.PIPE,
                            stderr = subprocess.PIPE,)

    # Remember the starting time
    start_time = calendar.timegm(time.gmtime())

# Here's what output from ffmpeg looks like:
# frame=  185 fps=184 q=2.3 size=     343kB time=6.06 bitrate= 463.4kbits/s    
# And there are extra chars that I guess either delete the line or go back to the beginning of the current line

# For some unknown reason, ffmpeg sends all output to stderr.

    output = ''
    duration = ''
    history = range(10)

    # Define regular expression for getting Duration
    p = re.compile('Duration\:\s*(\d+)\:(\d+)\:(\d+)\.(\d+)', re.MULTILINE)

    # Define regular expression for getting how much time has been processed
    t = re.compile('size\s*\=\s*(\S*)\s*time\s*\=\s*(\S*)\s+', re.MULTILINE)

    i = 0
    keep_going = True
    prev_progress = 0.0
    while keep_going:
        # If you read to many chars at once, the read statement never executes, 
        # I don't understand why
        text_block = proc.stderr.read(1)
        if not text_block: 
            keep_going = False
            # should I EXIT LOOP?

        output = output + text_block
#        print text_block,

        # If duration hasn't already been found, look for it
        if duration == '':
            m = p.search(output)
            if m:
                duration = int(m.group(1)) * 3600 \
                + int(m.group(2)) * 60 \
                + int(m.group(3))
                print "\n*** DURATION: hh:%s mm:%s ss:%s fr:%s = %d seconds***\n" \
                    % (m.group(1), m.group(2), m.group(3), m.group(4), duration)

        # Look for how much time has been processed
        mt = t.search(output[-60:])
        if mt and float(mt.group(2)) != prev_progress:
            # Remember the current time stamps and how far along we are,
            # for the last 10 iterations
#            print "i % 10 = " + str(i % 10)
#            print "mt.group(2) = " + str(mt.group(2))
            history[i % 10] = (calendar.timegm(time.gmtime()), float(mt.group(2)))
            
            TIMESTAMP = 0
            PROGRESS  = 1

            # First build up 10 progress history values, once every time we match the t pattern
            # and report our progress every 100 iterations both to stdout and the monitoring database.
            if i >= 10 and i % 100 == 0:
                # Don't do any of these reports if there's a chance of division by zero.
                # The possible sources of div by zero are if:
                #   duration is zero (maybe this could happen if the regexp fails to match because there are no digits after the decimal)
                #   history[i % 10][PROGRESS] = history[i % 10 - 8][PROGRESS] <-- i.e. no progress is being made by ffmpeg
                #   history[i % 10][TIMESTAMP] - history[i % 10 - 8][TIMESTAMP] <-- can't happen because timestamps always increase
                if duration != 0 and history[i % 10][PROGRESS] - history[i % 10 - 8][PROGRESS] != 0:
                    # Estimate many seconds are left
                    time_left = (float(duration) - history[i % 10][PROGRESS]) \
                        / ( (history[i % 10][PROGRESS] - history[i % 10 - 8][PROGRESS]) \
                                / (float(history[i % 10][TIMESTAMP]) - float(history[i % 10 - 8][TIMESTAMP])) )

                    # Display to STDOUT a summary of current progress
                    print "progress: %d%% (%d/%d)  remaining: %d sec  total time: %d sec" % \
                        (int(round(float(100) * float(mt.group(2)) / float(duration), 0)),
                         int(round(float(mt.group(2)), 0)),
                         duration,
                         time_left,
                         float(time_left) + float(calendar.timegm(time.gmtime())) - float(start_time))

                    # Report to the monitoring database a summary of current progress
                    MicalaCommunication.reportStatus('RUNNING', "FLV conversion: %2d%%" % \
                                      int(round(float(100) * history[i % 10][PROGRESS] / float(duration), 0)), \
                                      idMachine, idTask, idLecture)

            # Increment counter and store current progress value in prev_progress
            prev_progress = history[i % 10][PROGRESS]
            i = i + 1

    end_time = calendar.timegm(time.gmtime())

    print "total time elapsed: " + str(end_time - start_time)
    MicalaCommunication.reportStatus('RUNNING', "FLV file finished", idMachine, idTask, idLecture)

##########################################################
# get_lo_year
#
# TODO: Get rid of this function and use getLectureYear instead
##########################################################
def get_lo_year(lecture_object):
    """This doesn't currently do anything but return 2009.
    But it should find the year that the LO was actually recorded in."""

    year = '2010'

    return year

##########################################################
# get_media_archive_location
#
# Get the conferenceID and contributionID from the filename of the 
# previously saved XML file (generated by get_indico_metadata.pl)
#
##########################################################
def get_media_archive_location(lecture_object):

    conferenceID   = 0
    contributionID = 0

    lodir = os.path.join(conf.dir_to_upload, lecture_object)

    # pattern to match for XML file
    # We are looking for CONFIDcCONTRID.xml etc
    p = re.compile('([sc\d]+)\.xml')

    print 'checking for existence of directory' + lodir

    if not os.path.isdir(lodir):
        print 'ERROR: directory ' + lodir + ' does not exist'
        print 'exiting'
        sys.exit(1)

    print 'looking in ' + lodir

    filename = ''
    for filename in glob.glob(os.path.join(lodir, '*.xml')):
        m = p.search(filename)
        if m:
            conferenceID = m.group(1)
            contributionID = m.group(2)
            break

    if conferenceID == 0:
        Common.pr(0, "ERROR: no XML file found.\n")

    print "conferenceID   = " + str(conferenceID)
    print "contributionID = " + str(contributionID)

    return (conferenceID, contributionID)

##########################################################
# resize_images
#
#
##########################################################
def resize_images(src_dir, dest_dir, target_width, target_height, quality_level):
    """resize an entire given directory of images according to 
    given dimensions, and write the output to another given directory"""

    Common.debug(1, "src_dir       = %s" % src_dir)
    Common.debug(1, "dest_dir      = %s" % dest_dir)
    Common.debug(1, "target_width  = %d" % target_width)
    Common.debug(1, "target_height = %d" % target_height)
    Common.debug(1, "quality_level = %d" % quality_level)

    if not os.path.isdir(src_dir):
        Common.pr(0, "can not open %s. Does not exist (or isn't a directory)\n" %
           src_dir)
        sys.exit(1)

    # read all files in src_dir into a list
    # (this gives a list of absolute paths)
    list_images = glob.glob(os.path.join(src_dir, '*'))

    # regexp to capture just the filename at the end of the path,
    # for any file with the extensions jpg, png, gif or bmp.
    pattern_image = re.compile('^(.*?)\.(jpg|png|gif|bmp)$', 
                               re.IGNORECASE)

    flag_no_slides_found = True

    if len(list_images) > 0:
        Common.pr(1, " ...found %d files in the slides directory.\n" % len(list_images))
        for src_image_path in list_images:
            # returns just the filename, no path
            src_image_base = os.path.basename(src_image_path)
            match_image = pattern_image.search(src_image_base)
            if match_image:
                flag_no_slides_found = False
                Common.pr(2, "resizing %s.\n" % src_image_path)
                im_obj = Image.open(src_image_path)
                # resize image
                im_new = im_obj.resize((target_width, target_height), Image.ANTIALIAS)
                # Save file in destination directory
                Common.debug(6, " saving to %s.jpg" % os.path.join(dest_dir, match_image.group(1)))
                im_new.save(os.path.join(dest_dir, match_image.group(1)) + 
                            ".jpg", "JPEG", quality=quality_level)
    else:
        flag_no_slides_found = True

    if flag_no_slides_found:
        Common.pr(0, "No slides found in %s!\n" % src_dir)
        # instead of exiting here, perhaps I should give a warning 
        # somehow and insert at least a dummy slide?
        sys.exit(1)

##########################################################
# Print a usage statement
#
#
##########################################################
def usage():
    print """Usage message goes here.

Here are more details.
"""

##########################################################
# Run this file as a script
#
#
##########################################################
if __name__ == "__main__":
    main(sys.argv[1:])
