# ====================================================================
# Copyright (c) 2010 The Regents of the University of Michigan and
# The European Organization for Nuclear Research (CERN). Licensed under the
# Educational Community License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
# http://www.osedu.org/licenses/ECL-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
# ====================================================================
#!/usr/bin/perl

use strict;
use XML::DOM;
use Image::Magick;
use File::Path qw(make_path remove_tree);
use Micala::Common ':all';

# Load default variables from global configuration file
do 'c:\\micala\\micala\\etc\\processing.conf';

$Micala::Common::DEBUG_LEVEL = 0;
$Micala::Common::VERBOSITY_LEVEL = 2;

$::USAGE =
    "usage:\n" .
    "$0 --lo=<lecture object ID>\n" .
    "$0 --tarball=<tarred up lecture object>\n" .
    "lecture_object must be the name of a directory containing the file lecture.xml,\n" .
    "as well as the following subdirectories:\n" .
    "   media\n" .
    "   slides\n" .
    "   logs\n" .
    "   resources  - optional\n" .
    "   images     - optional\n" .
    "   chalkboard - optional\n";

&main(@ARGV);

#####################################################################
# sub main
# 
#
#####################################################################
sub main
{
    my (@arglist) = @_;
    &debug(5, "entering main()\n");

    &pr(1, "Starting iPod video script at " . localtime(time()) . "\n");

    # Parse command line input
    &debug(4, "parsing command line...\n");
    my ($lecture_object, $tarball) = &parse_command_line(@arglist);

    # Create a temporary scratch directory in which to do work
    &pr(1, "Creating scratch directories\n");
    my $scratch_dir = &create_scratch_directories();

    if ($tarball ne "") {
	($lecture_object) = &unpack_tarball($tarball, $scratch_dir);
	$::dir_lo = $scratch_dir;
    }

    print "LECTURE OBJECT = $lecture_object\n";

    my $idMachine = &getIdMachine($::curr_machine);
    my $idTask    = &getIdTask("publishing iPod");
    my $idLecture = &getIdLecture($lecture_object);

    my ($get_video_info_success, $duration, $video_width, $video_height) 
	= &get_video_info($::dir_lo . $::FS . $lecture_object . $::FS . "media" . $::FS . "master.avi");

    # If the dimensions of the master video are in the list of known dimensions,
    # then use the prepared intro/title/end screens of that size.
    my $dimensions_screens;
    if ($get_video_info_success && &member_of("${video_width}x${video_height}", $::known_dimensions)) {
	$dimensions_screens = sprintf("%dx%d", $video_width, $video_height);
    }
    # Otherwise use the default dimensions
    # TODO: using the default dimensions is not going to work well 
    # because the logo/credits will be too small for the screen or running off the edge;
    # a better way to do it would be to use smaller logo and end screens and center them on the larger 
    # screen with the appropriate color background.
    else {
	$dimensions_screens = $::podcast_dimensions_default;
    }

    $::IMAGE_INTRO =~ s/DIMENSIONS/$dimensions_screens/;
    $::IMAGE_TITLE =~ s/DIMENSIONS/$dimensions_screens/;
    $::IMAGE_END   =~ s/DIMENSIONS/$dimensions_screens/;

    &reportStatus('START', '', $idMachine, $idTask, $idLecture);

    # Read lecture.xml to get basic metadata and populate list of slide info
    my $slide_list = [];
    my ($lecture_title, $lecture_authors, $lecture_date);
    ($slide_list, $lecture_title, $lecture_authors, $lecture_date)
	= &retrieve_xml_metadata($::dir_lo . $::FS . $lecture_object);

    # Take slides from the Lecture Object and resize them to fit the video
    &reportStatus('RUNNING', 'resizing slides', $idMachine, $idTask, $idLecture);
    &resize_slides($::dir_lo . $::FS . $lecture_object, $slide_list, $scratch_dir, $video_width, $video_height);

    # Separate out the audio track from the master video file
    &reportStatus('RUNNING', 'extracting audio track', $idMachine, $idTask, $idLecture);
    &create_wav_file($::dir_lo . $::FS . $lecture_object, $scratch_dir . $::FS . "audio");

    # Create an AviSynth script
    my $avs_script_file = &create_avs_script($lecture_object, 
					     $slide_list, 
					     $lecture_title, 
					     $lecture_authors, 
					     $lecture_date, 
					     $scratch_dir);

    # Use ffmpeg to create the podcast video file using AviSynth
    &reportStatus('RUNNING', 'creating podcast video', $idMachine, $idTask, $idLecture);
    my $video_file = &create_video_file($lecture_object, $avs_script_file, $scratch_dir);

    # Copy the newly created podcast file to its final destination
    &reportStatus('RUNNING', 'copying file into place', $idMachine, $idTask, $idLecture);
    &syscom(1, "copy $video_file C:\\temp\\");

    # Clean up
    &reportStatus('RUNNING', 'removing scratch directory', $idMachine, $idTask, $idLecture);
#    &remove_scratch_directories($scratch_dir);

    &pr(1, "Finished iPod video script at " . localtime(time()) . "\n");

    &reportStatus('COMPLETE', '', $idMachine, $idTask, $idLecture);

    &debug(5, "exiting main()\n");
}

#####################################################################
# sub parse_command_line
#
#
#####################################################################
sub parse_command_line
{
    my @arglist = @_;
    my $lecture_object = "";
    my $tarball        = "";

    &debug(5, "currently in parse_command_line\n");

    if (scalar(@arglist) > 0) {
	while (scalar(@arglist) > 0) {
	    my $arg = shift @arglist;
	    if ($arg =~ m/^\-\-lo\=(\S+)$/s) {
		$lecture_object = $1;
		if (! &is_lecture_object($::dir_lo . $::FS . $lecture_object)) {
		    &pr(0, $::USAGE);
		    exit(1);
		}
	    }
	    elsif ($arg =~ m/^\-\-tarball\=(\S+)$/s) {
		$tarball = $1;
		if (! -r $tarball) {
		    &pr(0, "Can not find the specified tar file $tarball\n");
		    exit(1);
		}
	    }
	    elsif ($arg =~ m/^\-\-verbose\=(\d+)/s) {
		$Micala::Common::VERBOSITY_LEVEL = $1;
	    }
	    elsif ($arg =~ m/^\-\-verbose$/s) {
		$Micala::Common::VERBOSITY_LEVEL = 2;
	    }
	    elsif ($arg =~ m/^\-v$/s) {
		$Micala::Common::VERBOSITY_LEVEL = 2;
	    }
	    elsif ($arg =~ m/^\-vv$/s) {
		$Micala::Common::VERBOSITY_LEVEL = 3;
	    }
	    elsif ($arg =~ m/^\-vvv$/s) {
		$Micala::Common::VERBOSITY_LEVEL = 4;
	    }
	    elsif ($arg =~ m/^\-\-debug\=(\d+)/s) {
	        $Micala::Common::DEBUG_LEVEL = $1;
	    }
	    elsif ($arg =~ m/^\-\-debug$/s) {
		$Micala::Common::DEBUG_LEVEL = 1;
	    }
	    elsif ($arg =~ m/^\-d$/s) {
		$Micala::Common::DEBUG_LEVEL = 1;
	    }
	}
    }
    else {
	&pr(0, $::USAGE);
	exit(1);
    }
    
#    if ($lecture_object eq '') {
#        $lecture_object = getpwd();
#        $lecture_object =~ s/\n//;
#        &pr(2, "No lecture object specified; using $lecture_object\n");
#        if (! &is_lecture_object($lecture_object)) {
#            &pr(0, $::USAGE);
#            exit(1);
#        }
#     }

    &debug(5, "got lecture_object: $lecture_object\n");
    &debug(5, "exiting parse_command_line\n");

    return ($lecture_object, $tarball);
}

#####################################################################
# sub is_lecture_object
#
#
#####################################################################
sub is_lecture_object
{
    my ($lecture_path) = @_;
    my $success = 1;

    &debug(5, "currently in is_lecture_object\n");

    # is the supplied string an existing directory?
    if (! -d $lecture_path) {
        &pr(0, "$lecture_path is not a directory\n");
        $success = 0;
    }

    # does the file lecture.xml exist in the directory?
    if (! -r $lecture_path . $::FS . 'lecture.xml') {
        &pr(0, "$lecture_path does not contain a lecture.xml file,\n");
        $success = 0;
    }

    # does the directory contain required subdirectories?
    if (! -d $lecture_path . $::FS . 'media') {
        &pr(0, "$lecture_path does not contain required subdirectory 'media'\n");
        $success = 0;
    }
    if (! -d $lecture_path . $::FS . 'slides') {
        &pr(0, "$lecture_path does not contain required subdirectory 'slides'\n");
        $success = 0;
    }

    &debug(5, "exiting is_lecture_object\n");

    return ($success);
}

#####################################################################
# unpack_tarball
#
# Given an absolute path to a tarball, unpack it into the given scratch directory
#####################################################################
sub unpack_tarball
{
    my ($tarball, $scratch_dir) = @_;

    # Set current working directory where we want to put the tar contents
    &pr(1, "changing to scratch directory $scratch_dir\n");
    chdir($scratch_dir);

    my $command_untar_lo = $::tar . " xf " . $tarball;
    
    &pr(1, sprintf("running: [%s]", $command_untar_lo));
    &syscom(1, $command_untar_lo);

    my ($lecture_object) = $tarball =~ m/^.*?[\/\\]+([^\/\\]+)\.tar$/;

    # Get out of the scratch directory so we can delete it later.
    chdir($::dir_lo);

    # return the lecture object ID and the directory where the untarred LO is to be found.
    return ($lecture_object);
}

#####################################################################
# create_scratch_directories
#
#
#####################################################################
sub create_scratch_directories
{
    my $scratch_dirs = ["audio", "slides", "video"];
    
    my $scratch_dir = $::dir_temp . $::FS . sprintf("%s-%d", "podcast", time());

    # If there is already a scratch dir of the same name, remove it first
    if (-d $scratch_dir) {
	&pr(1, "Found a previously existing $scratch_dir, removing...\n");
	remove_tree($scratch_dir);
    }

    foreach my $dir (@$scratch_dirs) {
	&pr(1, "creating dir " . $scratch_dir . $::FS . $dir . "\n");
	make_path($scratch_dir . $::FS . $dir);
    }

    return $scratch_dir;
}

#####################################################################
# remove_scratch_directories
#
#
#####################################################################
sub remove_scratch_directories
{
    my ($scratch_dir) = @_;

    if (-d $scratch_dir) {
	&pr(1, "Removing scratch directories\n");
	remove_tree($scratch_dir);
    }

}

#####################################################################
# sub retrieve_xml_metadata
#
#
#####################################################################
sub retrieve_xml_metadata
{
    my ($lecture_path) = @_;
    my $xmlfile = $lecture_path . $::FS . "lecture.xml";
    my $xmlerr = 0;
    my $metatags = [
                    'TITLE',
                    'CREATOR',
                    'AUTHOR',
                    'SUBJECT',
                    'DESCRIPTION',
                    'PUBLISHER',
                    'CONTRIBUTOR',
                    'DATE',
                    'TYPE',
                    'FORMAT',
                    'SOURCE',
                    'LANGUAGE',
                    'RELATION',
                    'COVERAGE',
                    'RIGHTS',
                    'AFFILIATION',
                    'PHOTO',
                    'TITLESHORT',
                    'PREREQUISITES',
                    'SERIES',
                    'LOCATION',
                    'KEYWORDS',
                    'SEARCHABLE',
                    'TAPE',
                    'NOTES',
                    'NUMSLIDES',
                    'DURATION',
                    'STRUCTURE',
                    'AGGREGATIONLEVEL',
                    'TIME'
                    ];

    &debug(5, "entering retrieve_xml_metadata\n");

    my $parser = new XML::DOM::Parser;
    my $doc = $parser->parsefile($xmlfile)
        or die "unable to parse document $xmlfile\n";

    my $root = $doc->getDocumentElement;

    if ($root->getTagName ne 'LECTURE') {
        $xmlerr++;
        &pr(0, "The top level element of lecture.xml is not <LECTURE>\n");
    }

    my $numslides = 0;
    my $slide_list = [];
    my ($title, $author, $affiliation, $date, $time, $duration, $location, $language);
    my $list_authors = [];

    $title = $root->getAttribute("TITLE");
    &debug(3, "TITLE [" . $root->getAttribute("TITLE") . "]\n");

    for my $node ($root->getChildNodes) {
        if ($node->getNodeType == ELEMENT_NODE) {
            if (&member_of($node->getNodeName, $metatags)) {
#                &debug(3, $node->getNodeName . " [" . $node->getFirstChild->getNodeValue . "]\n");
                if ($node->getNodeName eq 'TITLE') { $title = $node->getFirstChild->getNodeValue; }
                if ($node->getNodeName eq 'AUTHOR') {
		    push(@$list_authors, $node->getFirstChild->getNodeValue);
		}
#                if ($node->getNodeName eq 'AFFILIATION') { $affiliation = $node->getFirstChild->getNodeValue; }
                if ($node->getNodeName eq 'DATE') { $date = $node->getFirstChild->getNodeValue; }
                if ($node->getNodeName eq 'TIME') { $time = $node->getFirstChild->getNodeValue; }
#                if ($node->getNodeName eq 'DURATION') { $duration = $node->getFirstChild->getNodeValue; }
#                if ($node->getNodeName eq 'LOCATION') { $location = $node->getFirstChild->getNodeValue; }
                if ($node->getNodeName eq 'LANGUAGE') { $language = $node->getFirstChild->getNodeValue; }
            }
            elsif ($node->getNodeName == 'PAR') {
                foreach my $child ($node->getChildNodes) {
                    if ($child->getNodeType == ELEMENT_NODE && $child->getNodeName eq "SEQ") {

                        foreach my $slide ($child->getChildNodes) {
                            if ($slide->getNodeType == ELEMENT_NODE && $slide->getNodeName eq "SLIDE") {
                                $numslides++;
				my $attrib_hash = { 
				    'filename' => '',
				    'timestamp' => '',
				    'title' => '',
				    'type' => '',
				    'seconds' => ''  };
                                foreach my $attrib ($slide->getAttributes->getValues) {
                                    if ($attrib->getName eq 'SRC') {
					$attrib_hash->{'filename'} = $attrib->getValue;
					&debug(7, "SRC: [" . $attrib->getValue . "]\n");
                                    }
                                    elsif ($attrib->getName eq 'BEGIN') {
					$attrib_hash->{'timestamp'} = $attrib->getValue;
					$attrib_hash->{'seconds'} 
					= &convert_to_seconds($attrib_hash->{'timestamp'});
					&debug(7, "BEGIN: [" . $attrib->getValue . "]\n");
                                    }
                                    elsif ($attrib->getName eq 'TITLE') {
					$attrib_hash->{'title'} = $attrib->getValue;
					if ($attrib_hash->{'title'} =~ m/\schalkboard$/i) {
					    $attrib_hash->{'type'} = 'chalkboard';
					}
					else {
					    $attrib_hash->{'type'} = 'slide';
					}
					&debug(7, "TITLE: [" . $attrib->getValue . "]\n");
                                    }
                                }
				# For now, totally ignore chalkboard timestamps
				if ($attrib_hash->{'type'} eq 'slide') {
				    push(@$slide_list, $attrib_hash);
				}
                            }
                        }
                    }
                }
            }
        }
    }

    my $formatted_title   = &format_title($title, $::MAX_STRING_LENGTH);
    my $formatted_authors = &format_authors($list_authors, $::MAX_STRING_LENGTH);
    my $formatted_date    = &format_date($date);

    &debug(5, "retrieve_xml_metadata\n");

    return ($slide_list, $formatted_title, $formatted_authors, $formatted_date);
}

#######################################################################
# convert_to_seconds
#
#
#######################################################################
sub convert_to_seconds
{
    my ($timestamp_str) = @_;
    &debug(10, "entering convert_to_seconds()\n");

    my ($hh, $mm, $ss) = $timestamp_str =~ m/^(\d\d)\:(\d\d)\:(\d\d)$/;

    my $seconds = 3600 * $hh + 60 * $mm + $ss;

    &debug(5, "converting $timestamp_str to $seconds seconds\n");
    &debug(10, "exiting convert_to_seconds()\n");

    return ($seconds);
}

#######################################################################
# sub create_wav_file
#
#
#######################################################################
sub create_wav_file
{
    my ($lecture_path, $scratch_dir_audio) = @_;
    my $master_audio_file = $scratch_dir_audio . $::FS . "master-audio.wav";
    my $input_avi_file   = $lecture_path . $::FS . "media" . $::FS . "master.avi";

    &debug(5, "entering create_wav_file\n");

    &pr(1, &syscom(1, "$::ffmpeg -i " . $lecture_path . $::FS . "media" . $::FS . "master.avi" . " $master_audio_file") );

#     # I'm not sure this really belongs in this subroutine, but
#     # since we just created master-audio.wav, we might as well
#     # copy this file to the lecture object for use in making an
#     # audio-only MP3 file.
#     &pr(1, "Copying audio track to media/master-audio.wav\n");
#     &syscom(1, "copy $master_audio_file " . $lecture_path . $::FS . "media" . $::FS);

    &debug(5, "exiting create_wav_file\n");
}

#####################################################################
# resize_slides
#
# resize slides so that they fit the video.
# Currently this just stretches the slides to fit whatever dimensions are given,
# but this should be updated in the future to preserve the slides' aspect ratio,
# scale the slide image so that its widest dimension matches the video dimension,
# center the image, and pad the rest of the area with black.
#####################################################################
sub resize_slides
{
    my ($lecture_path, $slide_list, $scratch_dir, $video_width, $video_height) = @_;
    &debug(5, "entering resize_slides()\n");

    foreach my $slide (@$slide_list) {
	if ($slide->{'type'} eq 'slide') {
	    my $windows_filename = $slide->{'filename'};
	    $windows_filename =~ s!\/!\\!;
	    
	    &pr(1, "resizing " . $windows_filename . "\n");
	    my $curr_image = Image::Magick->new();
	    $curr_image->Read($lecture_path . $::FS . $windows_filename);
	    
	    my ($width) = $curr_image->Get('width');
	    my ($height) = $curr_image->Get('height');
	    
	    $curr_image->Set(quality=>100);

	    $curr_image->Resize(width=>$video_width, height=>$video_height);
#	    $curr_image->Resize(width=>640, height=>360);
	    # I can't tell if this does anything.  All I know is you definitely
	    # don't want to set sigma to 1.0
	    $curr_image->UnsharpMask(radius=>1.0, sigma=>0.0, amount=>40.0, threshold=>0.0);
	    
	    my ($existing_quality) = $curr_image->Get('quality');
	    &debug(5, "quality: $existing_quality\n");
	    
# geometry=>geometry, radius=>double, sigma=>double, amount=>double, threshold=>double
	    
	    $windows_filename =~ s!\.gif!\.jpg!i;
	    
	    $curr_image->Write(filename=>$scratch_dir . $::FS . $windows_filename);
	    
	    undef $curr_image;
	}
    }
    

    &debug(5, "exiting resize_slides()\n");
}

#####################################################################
# create_avs_script
#
#
#####################################################################
sub create_avs_script
{
    my ($lecture_object, $slide_list, $lecture_title, $lecture_authors, $lecture_date, $scratch_dir) = @_;
    # TODO: I need to set the last frame according to the duration of the video.
    # This will mess up when a video with no slides is processed.
    my $last_frame = 100000;

    my $slide_clip_begin;
    my $next_begin;

    &debug(5, "entering create_avs_script()\n");

    print "lecture_object = $lecture_object\n";

    my $str = '';

    $str .= "masterVideo=ChangeFPS(AviSource(\"" 
	. $::dir_lo . $::FS . $lecture_object . $::FS . 'media' . $::FS .  'master.avi'
	. "\", audio=false), " . $::FRAMES_PER_SECOND . ")\n";

    $str .= "wavFile=WAVSource(\"" 
	. $scratch_dir . $::FS . "audio" . $::FS . 'master-audio.wav'
	. "\")\n";

    $str .= "\n";

    $str .= sprintf("clipIntro = Overlay(Trim(masterVideo, %d, %d), ImageSource(\"%s\"))\n", 1, int($::FRAMES_PER_SECOND * $::DURATION_INTRO), $::IMAGE_INTRO);
    $str .= sprintf("clipTitle = Overlay(Trim(masterVideo, %d, %d), ImageSource(\"%s\"))\n", 1, int($::FRAMES_PER_SECOND * $::DURATION_TITLE), $::IMAGE_TITLE);
    $str .= sprintf("clipEnd   = Overlay(Trim(masterVideo, %d, %d), ImageSource(\"%s\"))\n", 1, int($::FRAMES_PER_SECOND * $::DURATION_END),   $::IMAGE_END);

    $str .= "\n";

    my $ctr_clip = 0;

    my $numslides = scalar(@$slide_list);
    # If there are no slides...
    if ($numslides == 0) { 
	# Create one clip for the video
	$str .= &create_clip($ctr_clip++, 'video',
			     int($::FRAMES_PER_SECOND * $::DURATION_INTRO), $last_frame,
			     '', "VIDEO A");

    }
    # If there is at least one slide...
    else {
      for(my $slide_index=0; $slide_index < $numslides; $slide_index++) {
	  # If this is the last slide, then the beginning of the
	  # next slide is the same thing as the end of the video
	  if ($slide_index >= $numslides - 1) {
	      $next_begin = $last_frame;
	  }
	  else {
	      $next_begin = 
		  int(&round($::FRAMES_PER_SECOND * $slide_list->[$slide_index+1]->{'seconds'}));
	      
	      print "next_begin = $next_begin\n";
	  }
	  
	  $slide_clip_begin = &round($::FRAMES_PER_SECOND * $slide_list->[$slide_index]->{'seconds'});
	  
	  
	  # If slide overlaps next slide's beginning (minus some
	  # epsilon), then transition directly to the next slide
	  if (int(&round($::FRAMES_PER_SECOND * $slide_list->[$slide_index]->{'seconds'}) 
		  + $::FRAMES_PER_SECOND * $::DURATION_SLIDE) >= $next_begin - $::TRANSITION_EPSILON) {
	      $slide_list->[$slide_index]->{'filename'} =~ m/([^\/\\\.]+\.\w+)$/;
	      my $path_to_resized_slide = $scratch_dir . $::FS . "slides" . $::FS . $1;
	      $str .= &create_clip($ctr_clip++, 'slide',
				   $slide_clip_begin,
				   &minimum($next_begin - 1, $last_frame),
				   $path_to_resized_slide, "F");
	      
	  }
	  # Otherwise, play video of speaker up until next slide
	  else {
	      &debug(5, "G...\n");
	      &debug(5, "ctr_clip         = $ctr_clip\n");
	      &debug(5, "slide_clip_begin = $slide_clip_begin\n");
	      &debug(5, "                curr_slide_seconds        = " .    $slide_list->[$slide_index]->{'seconds'} . "\n");
	      &debug(5, "    round(FPS * curr_slide_seconds)       = " .     &round($::FRAMES_PER_SECOND * $slide_list->[$slide_index]->{'seconds'}) . "\n");
	      &debug(5, "int(round(FPS * curr_slide_seconds))      = " . int(&round($::FRAMES_PER_SECOND * $slide_list->[$slide_index]->{'seconds'}) - 1) . "\n");
	      &debug(5, "int(round(FPS * curr_slide_seconds) - 1)  = " . int(&round($::FRAMES_PER_SECOND * $slide_list->[$slide_index]->{'seconds'}) - 1) . "\n");
	      $slide_list->[$slide_index]->{'filename'} =~ m/([^\/\\\.]+\.\w+)$/;
	      my $path_to_resized_slide = $scratch_dir . $::FS . "slides" . $::FS . $1;
	      $str .= &create_clip($ctr_clip++, 'slide',
				   $slide_clip_begin,
				   int(&round($::FRAMES_PER_SECOND * $slide_list->[$slide_index]->{'seconds'} + $::FRAMES_PER_SECOND * $::DURATION_SLIDE - 1)),
				   $path_to_resized_slide, "G");
	      $str .= &create_clip($ctr_clip++, 'video',
				   int(&round($::FRAMES_PER_SECOND * $slide_list->[$slide_index]->{'seconds'} + $::FRAMES_PER_SECOND * $::DURATION_SLIDE)),
				   $next_begin - 1,
				   '', "H");
	  }
	  
	  
      } # end loop through slides
	
    } # end else


    # After we're done creating the clips, set total equal to ctr so
    # we can use it again, to concatenate all the clips
    my $total_clips = $ctr_clip;
    $ctr_clip = 0;

    $str .= "\n";
    $str .= sprintf("clip0p%04d", int($ctr_clip + 1)) . " = Dissolve("           # c = 2, 3
	. sprintf("clip%04d", $ctr_clip) . ", "                                # c = 2, 3
	. sprintf("clip%04d", int($ctr_clip+1)) . ", " . $::DISSOLVE_SLIDE . ")\n";   # c = 3, 4
    $ctr_clip++;
    while ($ctr_clip < $total_clips - 1) {
	$str .= sprintf("clip0p%04d", int($ctr_clip + 1)) . " = Dissolve("           # c = 2, 3
	    . sprintf("clip0p%04d", $ctr_clip) . ", "                                # c = 2, 3
	    . sprintf("clip%04d", int($ctr_clip+1)) . ", " . $::DISSOLVE_SLIDE . ")\n";   # c = 3, 4
	$ctr_clip++;
    }

    $str .= "\n";
#	$str .= sprintf("AudioDub(FlipVertical(clip0p%04d), wavFile)\n", $ctr_clip);

# Sometimes necessary to flip the video, I'm not sure under exactly which circumstances this is necessary.
#    $str .= sprintf("Dissolve(Amplify(AudioDub(FlipVertical(clipIntro), wavFile), 0, 0), Amplify(AudioDub(FlipVertical(Subtitle(clipTitle, \"%s\", font=\"%s\", size=%d, text_color=%s, halo_color=%s, align=5, lsp=1)), wavFile), 0, 0), AudioDub(FlipVertical(clip0p%04d), wavFile), Amplify(AudioDub(FlipVertical(clipEnd), wavFile), 0, 0), %d)", "$lecture_title\\n\\n$lecture_authors\\n$lecture_date", "Myriad Pro", 48, "\$000000", "\$ffffff", $ctr_clip, $::DISSOLVE_TITLES);
    $str .= sprintf("Dissolve(Amplify(AudioDub(clipIntro, wavFile), 0, 0), Amplify(AudioDub(Subtitle(clipTitle, \"%s\", font=\"%s\", size=%d, text_color=%s, halo_color=%s, align=5, lsp=1), wavFile), 0, 0), AudioDub(clip0p%04d, wavFile), Amplify(AudioDub(clipEnd, wavFile), 0, 0), %d)", "$lecture_title\\n\\n$lecture_authors\\n$lecture_date", $::podcast_font, $::podcast_font_size, "\$000000", "\$ffffff", $ctr_clip, $::DISSOLVE_TITLES);

    my $path_avisynth_file = $::dir_lo . $::FS 
	. $lecture_object . $::FS . 'logs' . $::FS . $::AVISYNTH_FILE;
    
    print "path_avisynth_file = $path_avisynth_file\n";

    open (AVISYNTH, ">$path_avisynth_file")
	or die "can not open $path_avisynth_file for writing: $!\n";
    print AVISYNTH $str;
    close (AVISYNTH);

    return $path_avisynth_file;

    &debug(5, "exiting create_avs_script()\n");
}

#####################################################################
# create_clip
#
#
#####################################################################
sub create_clip
{
    my ($clip_num, $clip_type, $begin, $end, $filename, $debug_id) = @_;
    my $str = '';
    my $begin_adjusted_for_dissolve;

    if ($clip_num >= 1) {
	$begin_adjusted_for_dissolve = $begin - $::DISSOLVE_SLIDE;
    }
    elsif($clip_num == 0) {
	$begin_adjusted_for_dissolve = $begin;
    }

    if ($clip_type eq 'slide') {
	$str = sprintf("clip%04d = Overlay(Trim(masterVideo, %6s, %6s), ImageSource(\"%20s\")) #          %s\n", 
		       $clip_num, $begin_adjusted_for_dissolve, $end, $filename, $debug_id);
    }
    elsif ($clip_type eq 'video') {
	$str = sprintf("clip%04d =         Trim(masterVideo, %6s, %6s)                                            #          %s\n", 
		       $clip_num, $begin_adjusted_for_dissolve, $end, $debug_id);

    }

    return $str;
}

#####################################################################
# create_video_file
#
# Run the ffmpeg command to generate a video file from the AviSynth script
#####################################################################
sub create_video_file
{
    my ($lecture_object, $avs_file, $scratch_dir) = @_;

    my $file_podcast = $scratch_dir . $::FS . "video" . $::FS . $lecture_object . "-podcast.mp4";

    # Run the ffmpeg command to generate a video file from the given AviSynth script
    my $command = sprintf("%s -i %s -qmax 8 %s",
			  $::ffmpeg,
			  $avs_file,
			  $file_podcast);

    &syscom(1, $command);

    return($file_podcast);
}

#####################################################################
# format_title
#
# truncate string and put ... on it if it's longer than the given max length
#####################################################################
sub format_title
{
    my ($title, $max_length) = @_;
    my $formatted_title = "";

    if (length($title) > $max_length){
	$formatted_title = substr($title, 0, $max_length) . "...";
    }
    else {
	$formatted_title = $title;
    }

    return $formatted_title;
}

#####################################################################
# format_authors
#
# Given a list of author names, create a single string that is 
# no longer than the given maximum length
#####################################################################
sub format_authors
{
    my ($list_authors, $max_length) = @_;
    my $authors = "";
    my $formatted_authors = "";

    # Compile list into string
    if (scalar(@$list_authors) == 1) {
	$authors = $list_authors->[0];
    }
    elsif (scalar(@$list_authors) > 1) {
	$authors = shift @$list_authors;
	foreach my $name (@$list_authors) {
	    $authors .= "; " . $name;
	}
    }
    elsif (scalar(@$list_authors) == 0) {
	$authors = "";
    }

    # Now truncate if necessary
    if (length($authors) > $max_length){
	$formatted_authors = substr($authors, 0, $max_length) . "...";
    }
    else {
	$formatted_authors = $authors
    }


    return $formatted_authors;
}

#####################################################################
# format_date
#
# Given a date of the form YYYY-MM-DD, output a date of day Month, Year
#####################################################################
sub format_date
{
    my ($given_date) = @_;

    my ($year, $month, $day) = $given_date =~ m/(\d+)\-(\d+)\-(\d+)/;

    my $months = ['January',
		  'February',
		  'March',
		  'April',
		  'May',
		  'June',
		  'July',
		  'August',
		  'September',
		  'October',
		  'November',
		  'December',
		  ];

    return sprintf("%d %s %d", int($day), $months->[int($month) - 1], $year);

}

#####################################################################
# get_video_info
#
# Given the full path to a video file, run ffmpeg on it to get the duration and dimensions
#####################################################################
sub get_video_info
{
    my ($filename) = @_;
    my $success = 1;

    if (! -r $filename) {
	$success = 0;
	&pr(0, "File not found: $filename\n");
    }

    # ffmpeg sends output to STDERR so here we capture STDERR along with STDOUT
    my $command = "$::ffmpeg -i $filename 2>&1";

    my $output = qx($command);

    print "command = [$command]\n";
    print "GOT OUTPUT:[$output]\n";

    my ($hh, $mm, $ss, $milli, $width, $height, $duration);

    # find duration string in output and calculate number of seconds
    if ($output =~ m/Duration\:\s*(\d+)\:(\d+)\:(\d+)\.(\d+)/s) {
	$hh     = $1;
	$mm     = $2;
	$ss     = $3;
	$milli  = $4;
	$duration = 3600.0 * $hh + 60.0 * $mm + $ss + $milli / 100.0;
    }
    else {
	$success = 0;
    }

    # find width and height in output
    if ($output =~ m/\,\s*(\d+)x(\d+)\,/s) {
	$width  = $1;
	$height = $2;
    }
    else {
	$success = 0;
    }

    print "hh     = $hh";
    print "mm     = $mm";
    print "ss     = $ss";
    print "milli  = $milli";
    print "width  = $width";
    print "height = $height";
    print "duration = $duration";

    return ($success, $duration, $width, $height);
}

#####################################################################
#
#
#
#####################################################################
sub minimum
{
    my ($a, $b) = @_;

    if ($a <= $b) {
#	&debug(5, "$a <= $b so returning $a\n");
	return ($a);
    }
    else {
#	&debug(5, "$a > $b so returning $b\n");
	return ($b);
    }

}

#####################################################################
# round
#
# for rounding POSITIVE decimals to the nearest integer
#####################################################################
sub round
{
    my ($decimal) = @_;

    return int($decimal + 0.5);
}
