# ====================================================================
# Copyright (c) 2010 The Regents of the University of Michigan and
# The European Organization for Nuclear Research (CERN). Licensed under the
# Educational Community License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
# http://www.osedu.org/licenses/ECL-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
# ====================================================================
##########################################################
# screen_analyze.pl
#
# Author: Jeremy Herr (herrj@umich.edu)
# copyright 2006 by the Regents of the University of Michigan
#
# Usage:
# screen_analyze.pl <-vvv> <--threshold=0.001> INPUT_DIR <OUTPUT_DIR>
#
# Description:
#
# This script reads a directory full of slide images generated
# by VisionRGB, Webcam Surveyor, or other capture software.
# Sorting alphabetically, it compares each image with the one
# before it using ImageMagick.  If there is a change above
# the normalized mean error threshold, it copies this image
# to the output directory.  Thus, given 3600 slide captures
# per hour, it selects only the images that have been changed,
# which might be 50 for a typical lecture, or several hundred
# if there are animations.
#
# last updated 1 May 2006 by herrj@umich.edu
##########################################################

##########################################################
# load the PerlMagick module for image manipulation,
# and the stat module for easy OOP access to file information.
##########################################################
use strict;
use Image::Magick;
use File::stat;
use Micala::Common ':all';

##########################################################
# OS-dependent definitions
##########################################################
$::FILESEP = "\\";
$::OSCOMMAND_COPY  = 'copy';
$::OSCOMMAND_MKDIR = 'mkdir';
$::OSCOMMAND_RM    = 'del';

##########################################################
# set defaults
##########################################################
$::DEFAULT_IN_DIR  = 'C:\\temp\\captured-slides';
$::DEFAULT_OUT_DIR = 'C:\\temp\\captured-slides-analyzed';
$::DEFAULT_NORMALIZED_MEAN_ERROR_THRESHOLD = 0.001;
#$::DEFAULT_NORMALIZED_MEAN_ERROR_THRESHOLD = 0.0025;
#$::DEFAULT_NORMALIZED_MEAN_ERROR_THRESHOLD = 0.007;
$::DEFAULT_NORMALIZED_MEAN_ERROR_THRESHOLD = 0.05;
$::VERBOSITY_LEVEL  = 1;
$Micala::Common::DEBUG_LEVEL      = 5;
$Micala::Common::FLAG_CREATE_BMPS = 0;
$::FLAG_NO_OUTPUT   = 0;

&main(@ARGV);

##########################################################
# main
#
#
##########################################################
sub main
{
    my @args = @_;

    my ($get_command_line_success, $in_dir, $out_dir, $normalized_mean_error_threshold, $mode) 
	= &get_command_line(@args);

    if (!$get_command_line_success) {
	&pr(0, "exiting because of problems with command line arguments\n");
	exit(0);
    }

    if (!opendir (INDIR, "$in_dir")) {
	&pr(0, "input directory", "can not open $in_dir for reading: $!");
	exit(1);
    }

    &debug(3, "reading in and sorting list of image files from $in_dir");

    my @in_file_list_unsorted = grep { m/\.bmp$/i || m/\.jpg$/i || m/\.png$/i } readdir (INDIR);
    closedir (INDIR);

    # now, depending on what the filenames look like, we need
    # to sort them with one of 2 possible methods
    my @in_file_list;
    if($in_file_list_unsorted[0] =~ m/^vrgb/) {
	@in_file_list = sort bynumber @in_file_list_unsorted;
    }
    else {
	@in_file_list = sort @in_file_list_unsorted;
    }

    &debug(3, "done reading in and sorting list of image files: got " . scalar(@in_file_list) . " files");

    if (-r $out_dir . $::FILESEP . "screen-analyze-log.txt") {
	&pr(1, "overwriting existing log file\n");
	&syscom(1, "$::OSCOMMAND_RM " . $out_dir . $::FILESEP . "screen-analyze-log.txt");
    }
    open (LOGFILE, ">$out_dir" . $::FILESEP . "screen-analyze-log.txt")
	or die "can not open $out_dir".$::FILESEP."screen-analyze-log.txt for writing: $!\n";

    my $prev_in_file;
    my $ctr_total = 0;
    my $ctr_picked = 0;
    my $first_pass = 1;
    my $prev_in_path;
    my $curr_in_path;
    my $curr_image_size;
    my $prev_image_size;

    my $prev_image = Image::Magick->new();
    my $curr_image;

    &debug(3, "looping through list of image files.");

    foreach my $in_file (@in_file_list) {
	$curr_image = Image::Magick->new();
	$curr_in_path = $in_dir . $::FILESEP . $in_file;

	$ctr_total++;

	# First get the file size before attempting to open it in
	# ImageMagick, because if it's empty, then this will cause an
	# error in ImageMagick and terminate the script prematurely.
	# If the file IS empty, keep prev_in_path the same and go to
	# the next one.
        # ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks)
        #   = stat($filename);
#	my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = stat($curr_in_path);
	my @stat_array = stat($curr_in_path);
	$curr_image_size = $stat_array[0][7];
#	print "size of $curr_in_path: $curr_image_size bytes\n";

	if ($curr_image_size > 0) {
#	    print "reading curr_image\n";
#	    print "curr_image: " . $curr_image . "\n";
	    $curr_image->Read($curr_in_path);

	}
#	$prev_image->Read($prev_in_path);
	
	if ($first_pass) { 
	    if (! $::FLAG_NO_OUTPUT) {
		if ($in_file =~ m/\.jpg$/) {
		    &syscom(5, "$::OSCOMMAND_COPY \"$curr_in_path\" " 
			    . &generate_out_file_name($out_dir, $curr_in_path, "jpg"));
		}
		else {
		    $curr_image->Set(quality=>100);
		    $curr_image->Write(filename=>&generate_out_file_name($out_dir, $curr_in_path, "jpg"));
		}
		if ($::FLAG_CREATE_BMPS) {
		    &syscom(5, "$::OSCOMMAND_COPY \"$curr_in_path\" " 
			    . &generate_out_file_name($out_dir, $curr_in_path, "bmp"));
		}
	    }
#	system("$::OSCOMMAND_COPY $curr_in_path $out_dir$::FILESEP$in_file");
	    $first_pass = 0; 
	    $prev_in_path = $curr_in_path;
	    $prev_image_size = $curr_image_size;
	    $prev_image = $curr_image;
	    $ctr_picked++;
	    next;
	}
	else {

#    my $newim = Image::Magick->new(size=>"1024x768", pointsize=>24, fill=>'black');
#    $newim->Read("xc:white");
#    $newim->Composite(image);

	    my ($width, $height, $mean_error, $normalized_max_error, $normalized_mean_error);

	    if ($curr_image_size > 0) {
		($width) = $curr_image->Get('width');
		($height) = $curr_image->Get('height');

#		my $curr_filename = $curr_image->Get('filename');
#		my $prev_filename = $prev_image->Get('filename');

		my $w = $curr_image->Compare(image=>$prev_image);
#		if ($w) { print LOGFILE "$w\n"; }
		$mean_error = $w->Get('error');
		$normalized_max_error = $w->Get('maximum-error');
		$normalized_mean_error = $w->Get('mean-error');

		print LOGFILE "$in_file, ". &generate_out_file_name($out_dir, $curr_in_path, "jpg") .", ${width}x${height}, ";
		print LOGFILE sprintf("\%e, %e, %e\n", $mean_error, $normalized_mean_error, $normalized_max_error);
	    }
	    else {
		$width = 0;
		$height = 0;
		$mean_error = 'N/A';
		$normalized_max_error = 'N/A';
		$normalized_mean_error = 'N/A';
		print LOGFILE "$in_file, ". &generate_out_file_name($out_dir, $curr_in_path, "jpg") .", ${width}x${height}, ";
		print LOGFILE "$mean_error, $normalized_mean_error, $normalized_max_error\n";
	    }

#    $curr_image->Write("c:\\temp\\newest.jpg");
	    my $display_in_file;
	    if (length($in_file) > 10) {
		$display_in_file = "..." . substr($in_file, length($in_file)-10, length($in_file));
	    }
	    &pr(2, "[$ctr_total/" . scalar(@in_file_list) . "] comparing " 
		. sprintf("%20s", $display_in_file) . " w/previous");
	    if ($curr_image_size > 0) {
		&pr(3, ", error: " . sprintf("%f", $mean_error));
	    }
	    else {
		&pr(3, ", error: N/A (skip empty file)");
	    }
	    &pr(2, "\n");
	    
#	    if ($curr_image_size > 0 && $normalized_mean_error >= $normalized_mean_error_threshold) {
	    if ($curr_image_size > 0 && $mean_error >= $normalized_mean_error_threshold) {
#		&pr(1, "*** norm mean error ($normalized_mean_error) exceeds threshhold ($normalized_mean_error_threshold)\n");
		&pr(1, "*** norm mean error ($mean_error) exceeds threshhold ($normalized_mean_error_threshold)\n");
		&pr(1, "*** writing to " . &generate_out_file_name($out_dir, $curr_in_path, "jpg") . "\n");

		if (! $::FLAG_NO_OUTPUT) {

		    if ($curr_in_path =~ m/\.jpg$/) {
			&syscom(5, "$::OSCOMMAND_COPY \"$curr_in_path\" " 
				. &generate_out_file_name($out_dir, $curr_in_path, "jpg"));
		    }
		    else {
			$curr_image->Set(quality=>100);
			$curr_image->Write(filename=>&generate_out_file_name($out_dir, $curr_in_path, "jpg"));
		    }
		    if ($::FLAG_CREATE_BMPS) {
			&syscom(5, "$::OSCOMMAND_COPY \"$curr_in_path\" " 
				. &generate_out_file_name($out_dir, $curr_in_path, "bmp"));
		    }
		}
	    
#	system("$::OSCOMMAND_COPY $curr_in_path $out_dir$::FILESEP$in_file");
		$ctr_picked++;
	    }

#	    undef $prev_image;
#	    undef $curr_image;
	    
	    $prev_in_path = $curr_in_path;
	    $prev_image_size = $curr_image_size;
	    $prev_in_file = $in_file;

	    if ($curr_image_size > 0) {
		undef $prev_image;
		$prev_image = $curr_image->Clone();
		undef $curr_image;
	    }

	} # end else block (not first pass)
    } # end foreach loop (looping through files)

    &pr(1, "\n-----------------------------------------------------------\n");
    &pr(1, "$ctr_picked of $ctr_total images selected\n");

    close (LOGFILE)
	or die "can not close the log file: $!\n";

    # after all this is done, go through the output slides and 
    if ($mode eq 'auto') {
	# count how many slides are in output dir

	# 
    }

}

##########################################################
# get_command_line
#
#
##########################################################
sub get_command_line()
{
    my @arg_list = @_;
    my $success = 1;
    my $message = '';
    my $in_dir = '';
    my $out_dir = '';
    my $mode = 'check';
    my $threshold = $::DEFAULT_NORMALIZED_MEAN_ERROR_THRESHOLD;

    &debug(5, "entering get_command_line");

    # parse command line,
    # retrieve flag settings and in/out directories
    while (scalar(@arg_list) > 0) {
	my $arg = shift @arg_list;
	print("arg = [$arg]\n");
	if ($arg =~ m/^\-/s) {
	    if ($arg =~ m/^\-\-verbose\=(\d+)/s) {
		$::VERBOSITY_LEVEL = $1;
	    }
	    elsif ($arg eq "-v") {
		$::VERBOSITY_LEVEL = 2;
	    }
	    elsif ($arg eq "-vv") {
		$::VERBOSITY_LEVEL = 3;
	    }
	    elsif ($arg eq "-vvv") {
		$::VERBOSITY_LEVEL = 5;
	    }
	    elsif ($arg =~ m/^\-\-threshold\=(.+)$/) {
		$threshold = $1;
	    }
	    elsif ($arg eq "--analyze") {
		$::FLAG_NO_OUTPUT = 1;
	    }
	    elsif ($arg =~ m/\-\-mode\=(\S+)/) {
		$mode = $1;
	    }
	    else {
		&pr(1, "command not recorgnized: $arg\n");
	    }
	}
	# if $in_dir not yet defined, then define it
	elsif ($in_dir eq '') {
	    $in_dir = $arg;
	}
	# otherwise define the out_dir
	else {
	    $out_dir = $arg;
	}
    }

    # if no out directory supplied, then simply
    # append -analyzed to the in directory name
    if ($out_dir eq "") {
	$out_dir = $in_dir . "-analyzed";
    }

    # check in/out directories for existence,
    # and flags for validity
    if (!($::VERBOSITY_LEVEL >= 0)) {
	$success = 0;
	&pr(0, "command line", "verbosity level must be between 0 and 5");
    }

    if ($in_dir eq '') {
	&pr(1, "no input/output directories specified, using defaults:\n");
    }

    &pr(1, "Using the following directories:\n");
    &pr(1, "input:  [$::DEFAULT_IN_DIR]\n");
    &pr(1, "output: [$::DEFAULT_OUT_DIR]\n");

    if (! -d $in_dir) {
	$success = 0;
	&pr(0, "command line", "$in_dir is not a valid directory");
    }

    if (! -d $out_dir) {
	&pr(1, "output directory $out_dir does not exist.  Creating it now...\n");
	&syscom(5, "$::OSCOMMAND_MKDIR $out_dir");
	if (! -d $out_dir) {
	    &pr(0, "command line", "could not create output directory $out_dir");
	    $success = 0;
	}
    }

    &debug(5, "exiting get_command_line");

    return ($success, $in_dir, $out_dir, $threshold, $mode);
}

##########################################################
# generate_out_file_name
#
#
##########################################################
sub generate_out_file_name()
{
    my ($out_dir, $file_path, $extension) = @_;

    # The Vision RGB application generates filenames with the following format
    if ($file_path =~ m/vrgb/) {
	my $sb = stat($file_path);
	my $seconds_since_epoch = $sb->mtime;
	my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst)
	    = localtime($seconds_since_epoch);
	my $time_string = $out_dir . $::FILESEP . sprintf("%04d%02d%02d%02d%02d%02d", 
							1900+$year, 1+$mon, $mday, $hour, $min, $sec) . "." . $extension;
	return $time_string;
    }
    # If the slides have already been processed once with this script,
    # Then don't do anything with timestamps, just spit out the same filename
    elsif ($file_path =~ m/^(\d\d\d\d\d\d\d\d\d\d\d\d\d\d\.\w+)$/) {
	my $file_name = $1;
	return $out_dir . $::FILESEP . $file_name;
    }
    # Webcam surveyor generates filenames with the following timestamp format:
    # Vision RGB-PRO Capture_04.13.2006--09.58.58.jpg (or .bmp)
    elsif ($file_path =~ m/Capture\_(\d\d)\.(\d\d)\.(\d\d\d\d)\-\-(\d\d)\.(\d\d)\.(\d\d)\.\w+$/) {
	my ($mon, $mday, $year, $hour, $min, $sec) = ($1, $2, $3, $4, $5, $6);

	my $time_string = $out_dir . $::FILESEP . sprintf("%04d%02d%02d%02d%02d%02d", $year, $mon, $mday, $hour, $min, $sec) . "." . $extension;

	return $time_string;
    }
    # VGA2USB on linux
    elsif ($file_path =~ m/vga\-(\d\d\d\d\d\d\d\d\d\d\d\d\d\d\.\w+)/) {
	my $file_name = $1;

	my $time_string = $out_dir . $::FILESEP . $file_name;

	return $time_string;
    }
    # incremental, once per second (ffmpeg extraction of 1 fps from video file - assuming 8 digits are used)
    elsif ($file_path =~ m/(\d\d\d\d\d\d\d\d\.\w+)/) {
	my $file_name = $1;

	my $time_string = $out_dir . $::FILESEP . $file_name;

	return $time_string;
    }
    # If filename did not fit into any of the above formats,
    # then stat the file to get its creation time and use
    # that information for the timestamp.
    else {
	my $sb = stat($file_path);
	my $seconds_since_epoch = $sb->mtime;
	my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst)
	    = localtime($seconds_since_epoch);
# I'm not sure why I had put this line in ...
#	($hour, $min, $sec) = $file_path =~ m!\-\-(\d+)\.(\d+)\.(\d+)!;
	my $time_string = $out_dir . $::FILESEP . sprintf("%04d%02d%02d%02d%02d%02d", 1900+$year, 1+$mon, $mday, $hour, $min, $sec) . "." . $extension;

	return $time_string;
    }

}


##########################################################
# notes:
#
#    # get the time stamp of when the in_path was last modified
#    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) 
#	= localtime($ctime);

sub bynumber
{
    my ($A) = $a =~ m/vrgb(\d+)/;
    my ($B) = $b =~ m/vrgb(\d+)/;

    return $A <=> $B;
}
