# ====================================================================
# Copyright (c) 2010 The Regents of the University of Michigan and
# The European Organization for Nuclear Research (CERN). Licensed under the
# Educational Community License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
# http://www.osedu.org/licenses/ECL-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
# ====================================================================
use strict;
use Micala::Common ':all';

# Load default variables from global configuration file
do 'c:\\micala\\micala\\etc\\processing.conf';

$Micala::Common::DEBUG_LEVEL = 0;
$Micala::Common::VERBOSITY_LEVEL = 1;

my $hostname          = 'cernppc1'; # CERN processing PC #1

my $excluded_patterns = ['chalkboard-capture', 
			 'media\\master.avi',
			 'slides-capture',
			 'slides-capture-analyzed'];

&main(@ARGV);

##########################################################
# main
#
#
##########################################################
sub main
{
    my @ARGV = @_;

    my ($success, $flag_batch_mode, $lecture_object) = &get_command_line(@ARGV);

    if (! $success) {
	&pr(1, "failure getting command line, exiting.\n");
	exit(0);
    }

    if ($flag_batch_mode) {
	# get list of lecture objects in upload directory
	opendir (DIR_UPLOAD, "$::dir_to_upload")
	    or die "can not open directory $::dir_to_upload: $!\n";
	my @list_los = sort grep { m/^\d\d\d\d/ } readdir (DIR_UPLOAD);
	closedir (DIR_UPLOAD);

	if ( scalar(@list_los) > 0 ) {
	    $lecture_object = $list_los[0];
	}
	else {
	    &pr(0, "There are no lecture objects to be uploaded.  Exiting\n");
	    exit(0);
	}
    }

    my $idMachine = &getIdMachine($::curr_machine);
    my $idTask    = &getIdTask("uploading completed LO");
    my $idLecture = &getIdLecture($lecture_object);
    
    &reportStatus('START', '', $idMachine, $idTask, $idLecture);
    
    # run create_xml_timing.pl if it hasn't already been done
#    &pr(1, "numslides: ". &count_slides_in_lecture_xml($::dir_to_upload, $lecture_object) . "\n");
    if (&count_slides_in_lecture_xml($::dir_to_upload, $lecture_object) == 0) {
	&reportStatus('RUNNING', 'creating timing from slide names', $idMachine, $idTask, $idLecture);
#	print "command: $::create_timing --lodir=$::dir_to_upload $lecture_object\n";
	&syscom(1, "$::create_timing --lodir=$::dir_to_upload $lecture_object");
	&reportStatus('RUNNING', 'finished timing', $idMachine, $idTask, $idLecture);
    }

    # since I just counted the slides, might as well update metadata.xml
    &set_num_slides($::dir_to_upload, $lecture_object); # doesn't do anything yet

    # Get the intermediate path to the lecture object, as well
    # as a target lecture object ID so we know where to upload to.
    my ($path_parent, $lecture_object_dest)
	= &get_media_archive_location($lecture_object);

    print "path_parent         = [$path_parent]\n";
    print "lecture_object_dest = [$lecture_object_dest]\n";

    # If this LO already exists in this location in the Media Archive,
    # rename it (don't delete it, just in case there is some error)
    if (-d $::dir_media_archive . $::FS . $::path_parent . $::FS . $lecture_object_dest) {
	my @td = localtime(time);
	#    0    1    2      3    4    5     6     7     8
	# ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)
	my $lo_backup = "%s\-%s\-" 
	    . sprintf("%04d%02d%02d%02d%02d%02d", 
		      $lecture_object_dest,
		      "backup",
		      $td[5]+1900, $td[4]+1, $td[3], $td[2], $td[1], $td[0]);
	my $command_backup_lo = "move " . $::dir_media_archive 
	    . $::FS . $path_parent
	    . $::FS . $lecture_object_dest
	    . " " . $::dir_media_archive . $::FS . $path_parent . $::FS . $lo_backup;
	&syscom(1, $command_backup_lo);
    }

    # If parent destination directory does not exist, create it
    # (this only happens once per year)
    if (! -d $::dir_media_archive . $::FS . $path_parent) {
	&syscom(1, "mkdir " . $::dir_media_archive . $::FS . $path_parent);
    }

#    # Copy (recursively) the lecture object into the Media Archive on DFS
#    my $command_move_to_archive = "xcopy /I /E " . $::dir_to_upload . $::FS 
#	. $lecture_object . ' ' . $::dir_media_archive . $::FS 
#	 . $path_parent . $::FS . $lecture_object_dest . $::FS;
#    &syscom(1, $command_move_to_archive);

    # Rename lecture object to destination name
    my $command_rename_lo1 = "move " . $::dir_to_upload . $::FS . $lecture_object 
	. ' ' . $::dir_to_upload . $::FS . $lecture_object_dest;
    &syscom(1, $command_rename_lo1);

    # tar up lecture object
    my $tarball     = $lecture_object_dest . ".tar";
    my $tarballpath = $::dir_to_upload . $::FS . $lecture_object_dest . ".tar";
    my $command_tar_up_lo = $::tar . " cf " . $tarball . ' ' . $lecture_object_dest;
    chdir($::dir_to_upload);
    &reportStatus('RUNNING', 'tarring up LO', $idMachine, $idTask, $idLecture);
    &syscom(1, $command_tar_up_lo);
    &reportStatus('RUNNING', 'finished tarring', $idMachine, $idTask, $idLecture);

    # Rename lecture object back to original name
    my $command_rename_lo2 = "move " . $::dir_to_upload . $::FS . $lecture_object_dest
	. ' ' . $::dir_to_upload . $::FS . $lecture_object;
    &syscom(1, $command_rename_lo2);

    # move tarball into the Media Archive on DFS
    my $command_move_tarball = "move $tarballpath $::dir_media_archive$::FS$path_parent$::FS";
    &reportStatus('RUNNING', 'moving LO to Media Archive', $idMachine, $idTask, $idLecture);
    &syscom(1, $command_move_tarball);
    &reportStatus('RUNNING', 'finished move', $idMachine, $idTask, $idLecture);

    # move lecture object from dir_to_upload into "done" folder,
    # overwriting any files that are already there
    if (-d $::dir_uploaded . $::FS . $lecture_object) {
	&syscom(1, "rmdir /S /Q " . $::dir_uploaded . $::FS . $lecture_object);
    }
    &syscom(1, "move $::dir_to_upload$::FS${lecture_object} $::dir_uploaded");

    &reportStatus('COMPLETE', '', $idMachine, $idTask, $idLecture);

    # finally, create a file to signal that this LO needs to be processed
    my $file_signal_lo = $::dir_awaiting_posting_flash 
	 . $::FS . $lecture_object_dest;
    open (DUMMY_FILE, ">$file_signal_lo")
	or die "can not open $file_signal_lo for writing: $!\n";
    print DUMMY_FILE $::dir_media_archive;
    close (DUMMY_FILE);
}

##########################################################
# get_media_archive_location
#
# Given a lecture object ID (either UM style or Indico style)
# return its non-absolute parent path, as well as the destination LO ID
# (which might be different from the original default LO ID).
# All LOs start out their lives with a UM-style ID, but at CERN
# we want to store them as Indico IDs. This is where that new ID is determined.
#
# There are three components to the full absolute path to a lecture object:
# 1) $dir_media_archive
#      the absolute path to the base directory of the media archive.
#      This is defined in the configuration file and never changes
#
# 2) $path_parent
#      The subdirectory or subdirectories holding the lecture object.
#      At CERN, this is just the current year. At Michigan, this is usually
#      a series of subjects, e.g. atlas/meetings/2009
#
# 3) $lo
#      a single subdirectory holding the lecture.xml and other files,
#      e.g. 20091001-umwlcd0010-104126 (UM) or 50273c8 (CERN)
##########################################################
sub get_media_archive_location()
{
    my ($lecture_object) = @_;
    my $path_parent;
    my $destination_lo;

    my $lodir = $::dir_to_upload . $::FS . $lecture_object;

    if ($::naming_convention eq 'cern') {
	# Extract year the LO was recorded in from the LO ID
	$lecture_object =~ m!^(\d\d\d\d)!;
	$path_parent = $1;


	# THIS IS WHERE WE SHOULD QUERY DATABASE SO WE DON'T HAVE TO MESS WITH THE XML FILE
	# .... ->
	# Look for IndicoID style XML file, e.g. 50273c8.xml
	# and extract Indico ID from that
	opendir (LO_DIR, $lodir)
	    or die "can not open directory $lodir: $!\n";
	my @files = grep { m/\d+.*?\.xml/ } readdir LO_DIR;
	closedir (LO_DIR);

	if (scalar(@files) > 0) {
	    $files[0] =~ m/([sc\d]+)\.xml/;
	    $destination_lo = $1;
	}
	else {
	    $destination_lo = &getMatchingIndicoID($lecture_object);
#	    print "ERROR: You are in cern mode, but there is no Indico ID encoded in an";
#	    print "       XML file name, e.g. 50273c8.xml\n";
#	    exit(1);
	}
	
    }
    elsif ($::naming_convention eq 'umich') {
	$destination_lo = $lecture_object;

	# I need to insert the part here that figures out what the path_parent should be.
	# this is normally defined by the <directory> tag in metadata.xml,
	# although I don't remember if that's even defined at this point.
    }
    else {
	print "ERROR: naming_convention is set to an unrecognized value: $::naming_convention.\n";
	print "       You should set it to cern or umich in the configuration file.\n";
	exit(1);
    }

    return ($path_parent, $destination_lo);
}

##########################################################
#
#
#
##########################################################
sub get_lo_year
{
    my ($lecture_object) = @_;

    
    my $year = $1;

    return $year;
}

##########################################################
# get_command_line
#
#
##########################################################
sub get_command_line()
{
    my @arg_list = @_;
    my $success = 1;
    my $message = '';
    my $lecture_object = '';
    my $flag_batch_mode = 0;

    # parse command line,
    # retrieve flag settings
    while (scalar(@arg_list) > 0) {
	my $arg = shift @arg_list;
	if ($arg =~ m/^\-/s) {
	    if ($arg =~ m/^\-\-verbose\=/s) {
		($::VERBOSITY_LEVEL) = $arg =~ m/^\-\-verbose\=(\d+)/s;
	    }
	    elsif ($arg =~ m/^\-\-verbose$/s) {
		$::VERBOSITY_LEVEL = 2;
	    }
	    elsif ($arg =~ m/^\-v$/s) {
		$::VERBOSITY_LEVEL = 2;
	    }
	    elsif ($arg =~ m/^\-vv$/s) {
		$::VERBOSITY_LEVEL = 3;
	    }
	    elsif ($arg =~ m/^\-vvv$/s) {
		$::VERBOSITY_LEVEL = 4;
	    }
	    elsif ($arg =~ m/^\-\-debug\=/s) {
		($::DEBUG_LEVEL) = $arg =~ m/^\-\-debug\=(\d+)/s;
	    }
	    elsif ($arg =~ m/^\-\-debug$/s) {
		$::DEBUG_LEVEL = 1;
	    }
	    elsif ($arg =~ m/^\-d$/s) {
		$::DEBUG_LEVEL = 1;
	    }
	    elsif ($arg =~ m/^\-\-lodir\=\S+$/s) {
		($::dir_lo) = $arg =~ m/^\-\-lodir\=(\S+)$/s;
	    }
	    elsif ($arg =~ m/\-\-lo\=(\S+)/s) {
		$lecture_object = $1;
	    }
	}
	else {
	    $success = 0;
	    &pr(0, "invalid command line argument: $arg\n");
	    exit(0);
	}
    }
    
    # check flags for validity
    if (!($::VERBOSITY_LEVEL >= 0)) {
	$success = 0;
	&pr(0, "verbosity level must be between 0 and 5\n");
    }

    if ($lecture_object eq '') {
	$flag_batch_mode = 1;
	&pr(2, "no lecture object specified, entering batch mode\n");
    }

    return ($success, $flag_batch_mode, $lecture_object);
}

##########################################################
# count_slides_in_lecture_xml
#
#
##########################################################
sub count_slides_in_lecture_xml
{
    my ($path, $lecture_object) = @_;

    my $file = $path . $::FS . $lecture_object . $::FS . 'lecture.xml';

    open (LECTUREXML, "<$file")
	or die "can not open $file for reading: $!\n";
    my $slurp = join('', <LECTUREXML>);
    close (LECTUREXML);

    my $numslides = 0;

    while ($slurp =~ m!\<SLIDE!igs) {
	$numslides++;
    }

    return $numslides;
}

##########################################################
# set_num_slides
#
# update metadata.xml with the given value for the numslides field
##########################################################
sub set_num_slides
{
    my ($path, $lecture_object) = @_;


}

##########################################################
# build_indicoID
#
# Produce an Indico ID string in one of the following formats
#
# 50273c5
# 50273c5sc2
# 50273s10
# 50273
#
##########################################################
sub build_indicoID
{
    my ($conferenceID, $contributionID, $subcontributionID, $sessionID) = @_;
    my $indicoID = '';

    if ($contributionID ne '') {
	if ($subcontributionID ne '') {
	    $indicoID = $conferenceID . 'c' . $contributionID . 'sc' . $subcontributionID;
	}
	else {
	    $indicoID = $conferenceID . 'c' . $contributionID;
	}
    }
    elsif ($sessionID ne '') {
	$indicoID = $conferenceID . 's' . $sessionID;
    }
    elsif ($conferenceID ne '') {
	$indicoID = $conferenceID;
    }
    else {
	print "ERROR: can not create an Indico ID because conference ID is not known.\n";
	exit(1);
    }

    return ($indicoID);
}

##########################################################
# parse_indicoID
# 
# Given an indicoID, get the following variables:
# * conferenceID
# * contributionID
# * subcontributionID
# * sessionID
##########################################################
sub parse_indicoID
{
    my ($indicoID) = @_;
    my ($conferenceID, $contributionID, $subcontributionID, $sessionID);

    # single lecture event
    if ($indicoID =~ m!^(\d+)$!) {
	$conferenceID      = $1;
	$contributionID    = '';
	$subcontributionID = '';
	$sessionID         = '';
    }
    # contrbution
    elsif ($indicoID =~ m!^(\d+)c(\d+)$!) {
	$conferenceID      = $1;
	$contributionID    = $2;
	$subcontributionID = '';
	$sessionID         = '';
    }
    # subcontribution
    elsif ($indicoID =~ m!^(\d+)c(\d+)sc(\d+)$!) {
	$conferenceID      = $1;
	$contributionID    = $2;
	$subcontributionID = $3;
	$sessionID         = '';
    }
    # session
    elsif ($indicoID =~ m!^(\d+)s(\d+)$!) {
	$conferenceID      = $1;
	$contributionID    = '';
	$subcontributionID = '';
	$sessionID         = $2;
    }
    else {
	print "ERROR: invalid Indico ID: [$indicoID]\n";
    }

    return ($conferenceID, $contributionID, $subcontributionID, $sessionID);
}

##########################################################
# report_ip
#
#
##########################################################
sub report_ip
{
    my $ip = &get_this_ip();

    if ($ip ne '') {
	&pr(1, "reporting IP address ($ip) to $::upload_arg_user_host\n");

	# now create a file containing the IP address
	my $ipfile = $::dir_temp . $::FS . $hostname;
	open (FILE, ">$ipfile")
	    or die "can not open $ipfile for writing: $!\n";
	print FILE "$ip\n";
	close (FILE);

	&syscom(1, "$::pscp $::arg_pw $ipfile $::upload_arg_user_host:/home/upload/ip/");
    }
}

##########################################################
# get_this_ip
#
#
##########################################################
sub get_this_ip
{
    my $ip = '';

    my $ipconfig_result = qx("ipconfig");

    # output should look like this:
    # IP Address. . . . . . . . . . . . : 141.211.97.25
    if ($ipconfig_result =~ m/IP\s+Address[\.\s]+\:\s+(\d+\.\d+\.\d+\.\d+)/sg) {
	$ip = $1;
    }
    else {
	# something didn't work, so return an empty string
    }

    return $ip;
}
