# ====================================================================
# Copyright (c) 2010 The Regents of the University of Michigan and
# The European Organization for Nuclear Research (CERN). Licensed under the
# Educational Community License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
# http://www.osedu.org/licenses/ECL-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
# ====================================================================
######################################################
# webcast2lectureobject.py
#
# Author: Jeremy Herr (herrj@umich.edu)
# 
# Description: this script converts 2 f4v video files,
# one video of the speaker and one from the VGA feed,
# and converts it into a raw lecture object, then places
# it in the SSH upload directory. From that point on,
# it is identical to lecture objects created on a
# linux recording laptop/PC
#
# 
######################################################

from optparse import OptionParser
import MySQLdb
import subprocess
import os
import stat
import sys
import time
import shutil
import glob
import re
import conf
from micala import *

def main(argv):
    currTask = "converting video to LO"
    currMachine = conf.curr_machine

    idTask       = MicalaCommunication.getIdTask(currTask)
    idTaskUpload = MicalaCommunication.getIdTask("uploading raw LO")
    idMachine    = MicalaCommunication.getIdMachine(currMachine)

    # This should be specified by the file or path name where the camera/slide files are stored.
    recordingBox = 'cernstream0001'

    # parse command line
    (success,
     message,
     flag_batch_mode,
     pathLecture,
     pathCameraFile,
     pathSlidesFile,
     timeClipBegin,
     timeClipEnd,
     flagFlatten) \
        = parse_command_line(argv)

    # Check for success parsing command line. If not, show error message and quit.
    if not success:
        Common.pr(0, "can not parse command line: %s\n" % message)
        sys.exit(0)

    if flag_batch_mode:
        Common.pr(2, "No command line arguments specified; entering batch mode.\n")
        idTaskRecording  = MicalaCommunication.getIdTask("recording")
        idTaskConverting = MicalaCommunication.getIdTask("converting video to LO")

        listRecordedLectures = MicalaCommunication.getLecturesDoneTaskANotTaskB(
            idTaskRecording,
            idTaskConverting,
            conf.contentTypeWebLecture,
            originPath = "NOT NULL")

        if len(listRecordedLectures) > 0:
            # We want to process the first good lecture found. The for-loop is for skipping bad or non-requested ones.
            flagProcessThis = True
            for lectureRecord in listRecordedLectures:
                idLecture  = lectureRecord["idLecture"]
                LOID       = lectureRecord["LOID"]
                originPath = lectureRecord["OriginPath"]
                timeBegin  = generateTimeBegin(LOID)
                Common.pr(2, "Got LOID from database: %s\n" % LOID)
                Common.debug(3, "original OriginPath: %s" % originPath)
                pathCameraFile = os.path.join(originPath, conf.camera_file)
                pathSlidesFile = os.path.join(originPath, conf.slides_file)

                # If the timeClipBegin and timeClipEnd have not been specified on the command line,
                # Check the database to see if there is anything there.
                if timeClipBegin == '' and timeClipEnd == '':
                    (secondsTimeClipBegin, secondsTimeClipEnd, timeClipBegin, timeClipEnd) = \
                        MicalaCommunication.getCropTimes(idLecture)

                # If either of the OriginPath paths don't exist, then post an error and skip to the next lecture
                if not os.path.exists(pathCameraFile) or not os.path.exists(pathSlidesFile):
                    if not os.path.exists(pathCameraFile):
                        Common.pr(2, "can not find path retrieved from database: %s\n" % pathCameraFile)
                    if not os.path.exists(pathSlidesFile):
                        Common.pr(2, "can not find path retrieved from database: %s\n" % pathSlidesFile)
                    MicalaCommunication.reportStatus("ERROR", "given OriginPath not found", idMachine, idTask, idLecture)
                    flagProcessThis = False
                    continue
                
                # Only process this lecture if it has been requested.
                if not MicalaCommunication.isRequested(idLecture, idTask):
                    flagProcessThis = False

                # Everything looks ok, and we want to only process one lecture at a time, so exit this for loop
                else:
                    flagProcessThis = True
                    break
            # We've come to the last lecture, and it is not to be processed, so exit.
            if not flagProcessThis:
                Common.pr(2, "no valid lectures to be processed.\n")
                sys.exit(0)
        else:
            Common.pr(2, "no lectures awaiting conversion; exiting.\n")
            sys.exit(0)
    else:
        # Set up various variables
        if pathCameraFile is None or pathCameraFile == '':
            pathCameraFile = os.path.join(pathLecture, conf.camera_file)
            pathSlidesFile = os.path.join(pathLecture, conf.slides_file)
            (timeBegin, LOID) = generateLOIDTwoFiles(pathCameraFile, timeClipBegin)
        else:
            (timeBegin, LOID) = generateLOIDTwoFiles(pathCameraFile, timeClipBegin)

        Common.pr(1, "Generated LOID = %s\n" % LOID)

    outPath = os.path.join(conf.dir_temp, LOID)
    outPathSlides = os.path.join(outPath, "slides-capture")
    outPathMedia  = os.path.join(outPath, "media")
    
##########################
# DEBUGGING...

    Common.debug(1, "pathLecture: %s" % pathLecture)
    Common.debug(1, "pathCameraFile: %s" % pathCameraFile)
    if os.path.exists(pathCameraFile):
        Common.debug(1, "%s exists" % pathCameraFile)
    else:
        Common.debug(1, "%s does not exist" % pathCameraFile)
    Common.debug(1, "pathSlidesFile: %s" % pathSlidesFile)
    Common.debug(1, "timeBegin: %s" % timeBegin)
    Common.debug(1, "timeBegin: %s" % time.strftime("%a, %d %b %Y %H:%M:%S +0000", timeBegin))
    Common.debug(1, "LOID: %s" % LOID)
    Common.debug(1, "outPath: %s" % outPath)
    Common.debug(1, "outPathSlides: %s" % outPathSlides)
    Common.debug(1, "outPathMedia: %s" % outPathMedia)

    # If time_begin or time_end have been specified, create a string
    # for insertion into the ffmpeg command to clip the video
    # appropriately (both slide and camera video need to be similarly
    # clipped)
    clipString = ''
    if timeClipBegin != '':
        clipString += " -ss \"%s\"" % timeClipBegin
    if timeClipEnd != '':
        clipString += " -t \"%s\"" % beginEndToDuration(timeClipBegin, timeClipEnd)

    # First make adjustments based on timeClipBegin, if it is specified
    if timeClipBegin is not None and timeClipBegin != '':
        offsetClip = float(timeClipBegin[0:2]) * 3600.0 + float(timeClipBegin[3:5]) * 60.0 + float(timeClipBegin[6:8])
    else:
        offsetClip = 0

    Common.debug(1, "timeClipBegin: %s" % timeClipBegin)
    Common.debug(1, "timeClipEnd:   %s" % timeClipEnd)
    Common.debug(1, "clipString: %s" % clipString)
    Common.debug(1, "offsetClip: %s" % offsetClip)
    Common.debug(1, "timeClipBegin[0:2]: %s" % timeClipBegin[0:2])
    Common.debug(1, "timeClipBegin[3:5]: %s" % timeClipBegin[3:5])
    Common.debug(1, "timeClipBegin[6:8]: %s" % timeClipBegin[6:8])

    Common.debug(1, "idLecture: %s" % idLecture)
    Common.debug(1, "LOID:      %s" % LOID)


##########################

#    sys.exit(0)

    if flag_batch_mode: 
        # Get internal database ID for LOID
        # This shouldn't be necessary
#        idLecture = MicalaCommunication.getIdLecture(LOID)
        pass
    else:
        # Add LOID to database.
        idLecture = MicalaCommunication.createNewMicalaLecture(LOID, conf.contentTypeWebLecture)

    # Report to the database that this task has started
    MicalaCommunication.reportStatus("START", "", idMachine, idTask, idLecture)

    # clean up existing lecture object directory in temp dir
    # (might be left over if this script aborted prematurely)
    if os.path.exists(outPath):
        shutil.rmtree(outPath)

    # create full path for outPath directory
    if not os.path.exists(outPath):
        Common.pr(1, "Creating directory: %s\n" % outPath)
        os.makedirs(outPath)
    
    # copy camera and slide files to lecture object in temp dir
    pathCameraFileCopied = os.path.join(outPathMedia, os.path.basename(pathCameraFile))
    pathSlidesFileCopied = os.path.join(outPathMedia, os.path.basename(pathSlidesFile))
    # Run the f4vpp command on them to "flatten" them
    if flagFlatten:
        commandFlattenCamera = "%s -i %s -o %s" % (conf.f4vpp, pathCameraFile, pathCameraFileCopied)
        commandFlattenSlides = "%s -i %s -o %s" % (conf.f4vpp, pathSlidesFile, pathSlidesFileCopied)
        Common.pr(1, "Flattening camera file\n")
        MicalaCommunication.reportStatus("RUNNING", "flattening camera file", idMachine, idTask, idLecture)
        os.system(commandFlattenCamera)
        Common.pr(1, "Flattening slides file\n")
        MicalaCommunication.reportStatus("RUNNING", "flattening slides file", idMachine, idTask, idLecture)
        os.system(commandFlattenSlides)
    # or just copy them
    else:
        if not os.path.exists(outPathMedia):
            Common.pr(2, "Creating directory: %s\n" % outPathMedia)
            os.makedirs(outPathMedia)
        Common.pr(1, "Copying camera file.\n")
        MicalaCommunication.reportStatus("RUNNING", "copying camera file", idMachine, idTask, idLecture)
        shutil.copy(pathCameraFile, pathCameraFileCopied)
        Common.pr(1, "Copying slides file.\n")
        MicalaCommunication.reportStatus("RUNNING", "copying slides file", idMachine, idTask, idLecture)
        shutil.copy(pathSlidesFile, pathSlidesFileCopied)

    # Get the duration in number of seconds of the camera file.
    # This should really be done earlier, by the recording script, 
    # so we know which lectures to delete before processing them.
    # Make necessary adjustments if timeClipBegin or timeClipEnd are defined.
    if timeClipBegin is not None and timeClipBegin != "" or timeClipEnd is not None and timeClipEnd != "":
        if timeClipEnd is not None and timeClipEnd != "":
            if timeClipBegin is not None and timeClipBegin != "":
                lectureDuration = int(secondsTimeClipEnd) - int(secondsTimeClipBegin)
            else:
                lectureDuration = int(secondsTimeClipEnd)
        else:
            lectureDuration = Common.getVideoDuration(pathCameraFileCopied) - int(secondsTimeClipBegin)
    else:
        lectureDuration = Common.getVideoDuration(pathCameraFileCopied)

    # update lecture record with duration
    # We waited to do this until after the file was flattened because 
    # ffmpeg reports a duration of twice the actual duration on unflattened files.
    # Also, the recording script should get duration based on start/stop time of task.
    MicalaCommunication.updateDuration(idLecture, lectureDuration)

    # create full path for outPathSlides directory
    if not os.path.exists(outPathSlides):
        Common.pr(1, "Creating directory: %s\n" % outPathSlides)
        os.makedirs(outPathSlides)

    # If timeClipBegin or timeClipEnd have been specified, create a string for insertion into the ffmpeg command
    # to clip the video appropriately (both slide and camera video need to be similarly clipped)
    clipString = ''
    if timeClipBegin != '':
        clipString += " -ss \"%s\"" % timeClipBegin
    if timeClipEnd != '':
        clipString += " -t \"%s\"" % beginEndToDuration(timeClipBegin, timeClipEnd)

    # set up slide creation command
    # this will export slide video as JPEGs, one frame per second
    command = """%s %s -i %s -r 1 -an "%s" """ \
        % (conf.ffmpeg,
           clipString,
           pathSlidesFileCopied,
           os.path.join(outPathSlides, "%08d.jpg"))
  
    Common.pr(1, "Generating slide JPEGs (one frame per second)\n")
    Common.pr(1, "running command: " + command + "\n")
    MicalaCommunication.reportStatus("RUNNING", "generating slides", idMachine, idTask, idLecture)
    # Spawn a subprocess to run the ffmpeg command and capture output
    proc = subprocess.Popen(command, 
                            shell  = True,
                            stdin  = subprocess.PIPE,
                            stdout = subprocess.PIPE,
                            stderr = subprocess.PIPE,)

    keep_going = True
    while (keep_going):
        text_block = proc.stderr.read(1)
        if not text_block: keep_going = False
#        print text_block,

    # rename slides appropriately, so that they have time stamps starting with lecture start time
    Common.pr(1, "Renaming slides\n")
    for fileSlide in glob.glob( os.path.join(outPathSlides, '*.jpg') ):
        os.rename(fileSlide,
                  os.path.join(os.path.dirname(fileSlide), generateSlideName(timeBegin, os.path.basename(fileSlide))))

    # Convert flv file to master.avi to fit with existing processing (this should be fixed so we don't transcode it twice)
    command = "%s %s -i \"%s\" -vcodec %s \"%s\"" \
        % (conf.ffmpeg,
           clipString,
           os.path.join(outPathMedia, os.path.basename(pathCameraFileCopied)),
           'msmpeg4v2',
           os.path.join(outPathMedia, "master.avi"))

    Common.pr(1, "Transcoding camera flv file to master.avi.\n")
    Common.pr(1, "running command: " + command + "\n")
    MicalaCommunication.reportStatus("RUNNING", "transcoding flv->avi", idMachine, idTask, idLecture)
    # Spawn a subprocess to run the ffmpeg command and capture output
    proc2 = subprocess.Popen(command, 
                            shell  = True,
                            stdin  = subprocess.PIPE,
                            stdout = subprocess.PIPE,
                            stderr = subprocess.PIPE,)

    keep_going = True
    while (keep_going):
        text_block = proc2.stderr.read(1)
        if not text_block: keep_going = False
#        print text_block,

    # Delete LO from upload directory if it's already been put there
    # (from a previous running of this script that aborted prematurely)
    if os.path.exists(os.path.join(conf.dir_server_uploads_abs, LOID)):
        shutil.rmtree(os.path.join(conf.dir_server_uploads_abs, LOID))

    # Recursively move entire lecture object from temp dir to upload directory.
    Common.pr(1, "Moving lecture object into upload directory.\n")
    MicalaCommunication.reportStatus("RUNNING", "moving to upload dir", idMachine, idTask, idLecture)

    # Also report that uploading the raw LO is starting.  This is a
    # separate task because the original Michigan recording system
    # uses it, without doing the conversion. Reporting it here makes
    # things look consistent on the monitoring page
    MicalaCommunication.reportStatus("START",    "", idMachine, idTaskUpload, idLecture)
    shutil.move(outPath, os.path.join(conf.dir_server_uploads_abs, LOID))

    # Give ssh user lecupload full control permissions over this lecture object
    # so that it can be deleted via ssh later
    # Explanation of ICACLS command:
    #  /T means do it to all subdirs and files recursively
    #  /grant means grant permission
    #  lecupload is the username
    #  F is the designation for Full control permission
    Common.pr(1, "Giving user lecupload necessary permissions\n")
    MicalaCommunication.reportStatus("RUNNING", "fixing permissions", idMachine, idTask, idLecture)
    commandPerms = "ICACLS %s /T /Q /grant %s:F" % (os.path.join(conf.dir_server_uploads_abs, LOID), "lecupload")
    Common.pr(1, "running command: " + commandPerms + "\n")
    Common.syscom(commandPerms)

    # Report that the uploading raw LO task is complete
    MicalaCommunication.reportStatus("COMPLETE", "", idMachine, idTaskUpload, idLecture)

    # Finally, report to the database that the conversion task is complete
    MicalaCommunication.reportStatus("COMPLETE", "", idMachine, idTask, idLecture)

def parse_command_line(arg_list):
    """Parse command line arguments and do basic sanity check."""

    success             = True
    message             = ''
    lecture_path        = ''
    lecture_path_camera = ''
    lecture_path_slides = ''
    time_begin          = ''
    time_end            = ''
    flag_batch_mode     = False

    # Check for non-option args. There are three possibilities: 
    #  0. no non-option args (in this case turn on batch mode, 
    #     to look in the database for a lecture waiting to be converted)
    #  1. one non-option arg (the path of the directory containing camera and slide files)
    #  2. two non-option args (the paths of each camera and slide filename)
    numArgs = 0
    for arg in arg_list:
        if arg[0] != '-':
            numArgs += 1
            if re.search('camera', arg):
                lecture_path_camera = arg
#                lecture_id = os.path.basename(arg)
            elif re.search('slides', arg):
                lecture_path_slides = arg
            else:
                lecture_path = arg

    if numArgs == 0:
        flag_batch_mode = True

    # Create instance of option parser
    parser = OptionParser()

    # verbosity level
    parser.add_option("-v", "--verbosity", "--verbose", 
                      dest="verbosity",
                      type="int",
                      action="store",
                      help="set verbosity level", 
                      metavar="VERBOSITY_LEVEL")

    # debugging level
    parser.add_option("-d", "--debug", 
                      dest="debug",
                      action="store",
                      help="set debugging level", 
                      metavar="DEBUG_LEVEL")

    # begin time
    parser.add_option("-a", "--start", "--begin", 
                      dest="begin",
                      type="string",
                      action="store",
                      help="specify a start time HH:MM:SS", 
                      metavar="BEGIN_TIME")

    # end time
    parser.add_option("-z", "--stop", "--end", 
                      dest="end",
                      type="string",
                      action="store",
                      help="specify an end time HH:MM:SS", 
                      metavar="END_TIME")

    # flag to flatten files using f4vpp
    parser.add_option("-f", "--flatten", "--flat",
                      dest="flagFlatten",
                      action="store_true",
                      default=False,
                      help="flatten the files with f4vpp",
                      metavar="FLAG_FLATTEN")

    # Parse the command line
    (options, args) = parser.parse_args(arg_list)

    # check flags for validity
    if options.verbosity is not None:
        conf.VERBOSITY_LEVEL = options.verbosity
        if not (int(options.verbosity) >= 0 and int(options.verbosity) <= 10):
            success = False
            message = "verbosity level must be between 0 and 10\n"

    # check flags for validity
    if options.debug is not None:
        conf.DEBUG_LEVEL     = options.debug
        if not (int(options.debug) >= 0 and int(options.debug) <= 10):
            success = False
            message += "debugging level must be between 0 and 10\n"

    # check for presence of begin time option
    if options.begin is not None:
        m = re.search('\d\d\:\d\d\:\d\d', options.begin)
        if m:
            time_begin = options.begin

    # check for presence of end time option
    if options.end is not None:
        m = re.search('\d\d\:\d\d\:\d\d', options.end)
        if m:
            time_end = options.end

    # check for existence of given camera file path
    if lecture_path_camera != '':
        if not os.path.exists(lecture_path_camera):
            success = False
            message += "Can not find given camera video file %s\n" % lecture_path_camera

    # check for existence of given slide file path
    if lecture_path_slides != '':
        if not os.path.exists(lecture_path_slides):
            success = False
            message += "Can not find given slides video file %s\n" % lecture_path_slides

    # check for existence of given directory path
    if lecture_path != '':
        if not os.path.exists(lecture_path):
            success = False
            message += "Can not find given directory %s\n" % lecture_path

    return (success,
            message,
            flag_batch_mode,
            lecture_path,
            lecture_path_camera,
            lecture_path_slides,
            time_begin,
            time_end,
            options.flagFlatten)

# don't think I'll be using this anymore, probably should delete it.
def generateLOID(videoFile, machineName):
    """return formatted LOID reflecting time stamp when video file was created."""

    timeCreation = time.localtime(os.stat(videoFile)[stat.ST_CTIME])

    year   = timeCreation.tm_year
    month  = timeCreation.tm_mon
    day    = timeCreation.tm_mday
    hour   = timeCreation.tm_hour
    minute = timeCreation.tm_min
    second = timeCreation.tm_sec

    return (timeCreation, "%04d%02d%02d-%s-%02d%02d%02d" % \
        (year, month, day, machineName, hour, minute, second))

def generateTimeBegin(LOID):
    """Given an LOID, generate a corresponding timestamp (struct_time)."""

    match_datetime = re.search('(20\d\d)(\d\d)(\d\d)[-][\w\d]+[-](\d\d)(\d\d)(\d\d)', LOID)

    if not match_datetime:
        Common.pr(0, "generateTimeBegin failed; The LOID %s is probably badly formed.\n" % LOID)
        sys.exit(1)

    timeBegin = time.localtime(time.mktime((int(match_datetime.group(1)),
                               int(match_datetime.group(2)),
                               int(match_datetime.group(3)),
                               int(match_datetime.group(4)),
                               int(match_datetime.group(5)),
                               int(match_datetime.group(6)),
                               0, 0, -1)))

    return timeBegin

def generateLOIDTwoFiles(videoFile, timeClipBegin):
    """return formatted LOID based on given video file name, adding on any time being cut from the beginning of the video."""

    # Calculate the offset in seconds to postpone the LOID start time,
    # if timeClipBegin is specified
    if timeClipBegin is not None and timeClipBegin != '':
        offsetClip = float(timeClipBegin[0:2]) * 3600.0 + float(timeClipBegin[3:5]) * 60.0 + float(timeClipBegin[6:8])
    else:
        offsetClip = 0.0

    Common.debug(1, "parsing file name %s\n" % videoFile)

    match_date = re.search('((20\d\d)(\d\d)(\d\d))[-]', videoFile)
    match_datetime = re.search('((20\d\d)(\d\d)(\d\d))[-]*.*?[-]*((\d\d)(\d\d)(\d\d))', videoFile)
    match_machine = re.search('(cernpc[^-]*)[-]',    videoFile)

    if not match_datetime or not match_machine:
        Common.pr(0, "generateLOID failed; unable to find either date, time or machine in the filename %s." \
                      % videoFile)
        sys.exit(1)

    timeBegin = time.localtime(time.mktime((int(match_datetime.group(2)), 
                               int(match_datetime.group(3)), 
                               int(match_datetime.group(4)),
                               int(match_datetime.group(6)),
                               int(match_datetime.group(7)),
                               int(match_datetime.group(8)),
                               0, 0, -1)) + offsetClip)

    return (timeBegin, "%s-%s-%s" % \
        (time.strftime("%Y%m%d", timeBegin), 
         match_machine.group(1), 
         time.strftime("%H%M%S", timeBegin)))

def generateSlideName(timeBegin, fileName):
    """Given a starting time stamp (seconds since epoch) of a lecture object and the file name of a slide generated by ffmpeg,
    return a new name for that slide that fits the UM lecture object standard"""

    # Grab the first 8 chars, which is where the sequential numbers are
    fileSubString = fileName[0:8]

    # add the beginning time stamp to the name of the file (minus one, since it is 1-indexed)
    timeStamp = time.localtime(int(time.mktime(timeBegin)) + int(fileSubString) - 1)

    # extract the individual values from the struct_time
    year   = timeStamp.tm_year
    month  = timeStamp.tm_mon
    day    = timeStamp.tm_mday
    hour   = timeStamp.tm_hour
    minute = timeStamp.tm_min
    second = timeStamp.tm_sec

    return ("vga-%04d%02d%02d%02d%02d%02d.jpg" % (year, month, day, hour, minute, second))

def beginEndToDuration(timeClipBegin, timeClipEnd):
    """Given starting time stamp and end time stamp both in the form of a string HH:MM:SS, 
    calculate the duration and return a string HH:MM:SS."""

    # Initialize timeClipBegin if undefined.
    if timeClipBegin == '':
        timeClipBegin = "00:00:00"

    # convert strings to number of seconds
    secondsBegin = float(timeClipBegin[0:2]) * 3600.0 + float(timeClipBegin[3:5]) * 60.0 + float(timeClipBegin[6:8])
    secondsEnd   =   float(timeClipEnd[0:2]) * 3600.0 +   float(timeClipEnd[3:5]) * 60.0 +   float(timeClipEnd[6:8])

    # get duration in seconds
    secondsDuration = secondsEnd - secondsBegin
    durationSS = secondsDuration % 60
    durationMM = ((secondsDuration - durationSS) / 60) % 60
    durationHH = int(secondsDuration / 3600)

    durationString = "%02d:%02d:%02d" % (durationHH, durationMM, durationSS)

    return durationString

if __name__ == "__main__":
    main(sys.argv[1:])
