/* ====================================================================
 * Copyright (c) 2010 The Regents of the University of Michigan and
 * The European Organization for Nuclear Research (CERN). Licensed under the
 * Educational Community License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License. You may
 * obtain a copy of the License at
 *
 * http://www.osedu.org/licenses/ECL-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * permissions and limitations under the License.
  ==================================================================== */

// Written by: Jim Irrer irrer@umich.edu jimirrer@gmail.com

#include "canonVC_C50i.h"

// --------------------------------------------------------

/**
 * Library functions to control the movement of the CANON
 * model VC-C50i web camera.
 */

// --------------------------------------------------------

/**
 * Construct a structure to contain the context of the camera.
 * If there is a problem creating the structure, then an error
 * code will be returned.
 *
 * The VCdelete call should be used to delete the structure
 * (whether or not an error was returned) when it is no
 * longer needed.
 *
 * @param vc Use this space for the struct.
 *
 * @param devName Name of device (for example: /dev/ttyS1 )
 *
 * @return Zero if ok, or error code.
 */

int VCconstruct(VC *vc, char *devName)
{
    struct termios tp;
    int fd = open(devName, O_RDWR | O_NOCTTY | O_NONBLOCK);

    if (fd < 0) {
        return VCsystemError(vc, errno, "Opening camera device");
    }
    memset(vc, 0, sizeof(VC));
    vc->fd = fd;

    // set up the termios struct
    memset(&tp, 0, sizeof(tp));
    tp.c_iflag = IGNPAR;
    tp.c_cflag = CRTSCTS | CS8 | CLOCAL | CREAD | B9600;

    // set the serial port attributes the way the camera needs them
    if (tcsetattr(fd, TCSANOW, &tp) < 0) {
        return VCsystemError(vc, errno, "Could not set serial port attributes");
    }

    vc->tiltMax = 360.0;
    vc->panMax = 360.0;

    vc->panMinSpeed = 0.9;
    vc->panMaxSpeed = 90.0;
    vc->tiltMinSpeed = 0.9;
    vc->tiltMaxSpeed = 69.975;

    if (pipe(vc->threadControl) != 0) {
        VCsystemError(vc, errno, "Unable to create thread control pipes.");
    }

    int threadError = pthread_create(&(vc->serverThread), NULL, &VCthread, vc);
    if (threadError != 0) {
        VCsystemError(vc, threadError, "Unable to start thread.");
    }

    if ((VChostControl  (vc)                                         == 0) &&
        (VCgetPanSpeed  (vc, &vc->panSpeed )                         == 0) &&
        (VCgetTiltSpeed (vc, &vc->tiltSpeed)                         == 0) &&
        (VCgetPanRatio  (vc, &vc->panRatio)                          == 0) &&
        (VCgetTiltRatio (vc, &vc->tiltRatio)                         == 0) &&
        (VCgetFocusRange(vc, &vc->focusMinRange, &vc->focusMaxRange) == 0) &&
        (VCgetZoomRange (vc, &vc->zoomMinRange, &vc->zoomMaxRange)   == 0)
       ) {
       VCsetError(vc, 0, "ok");
    }

    return vc->errorCode;
}

// --------------------------------------------------------

/**
 * Perform whatever logging is desired for debug.
 *
 * @param vc Camera control struct.
 *
 * @msg msg Message to log.
 */
void VClog(VC *vc, char* msg)
{
    // printf("VC %2d %10.6lf command: %s\n", vc->fd, VCgetTime(), msg);
}

// --------------------------------------------------------

/**
 * Report an assertion and exit with a failure status.
 *
 * @param file Source file in which error occurred.
 *
 * @param line Line on which error occurred.
 *
 * @param message Description of failure.
 *
 */
void VCassert(char *file, int line, char *message)
{
    fprintf(stderr, "VC Assertion failure %s:%d : %s\n", file, line, message);
    fprintf(stdout, "VC Assertion failure %s:%d : %s\n", file, line, message);
    exit(1);
}

// --------------------------------------------------------

/**
 * Get the current time in seconds (past 1970) to an accuracy
 * of the nearest microsecond, with an offset starting at the
 * first time the function was called.
 *
 * @return Time in seconds.  On failure, return 0.
 */

static time_t VCbaseSec = 0;
double VCgetTime()
{
    struct timeval tv;
    if (gettimeofday(&tv, null) == 0) {
        if (VCbaseSec == 0) {
            VCbaseSec = tv.tv_sec;
        }
        double t = (tv.tv_sec - VCbaseSec) + (double)(tv.tv_usec / 1000000.0);
        return t;
    }
    return (double)0;
}

// --------------------------------------------------------

/**
 * Get the time elapsed between the VCbaseSec and the time
 * given.  You must call VCgetTime prior to this function. 
 * On failure, return 0.
 *
 * @return Time in seconds.
 */

double VCgetRelativeTime(struct timeval *relTime)
{
    double t = 0;
    if (VCbaseSec == 0) {
        return t;
    }

    t = (relTime->tv_sec - VCbaseSec) + (double)(relTime->tv_usec / 1000000.0);
    return t;
}

// --------------------------------------------------------

/**
 * Convert up to four nibbles of an int16 to ascii.
 *
 * @param h Integer value.
 *
 * @param ascii Put result here.
 *
 * @param size Number of bytes in resulting ascii.
 */

void VChexToAscii(int16 h, unsigned char *ascii, int size)
{
    char *hex = "0123456789ABCDEF";
    int i;
    int offset = 4 - size;

    for (i = 0; i < size; i++) {
        ascii[i] = hex[(h >> (12-(4*(i+offset)))) & 0xf];
    }
}

// --------------------------------------------------------

/**
 * Set the error code and error message.  Make sure that the
 * message is always allocated from the heap.
 *
 * @param vc Control structure.
 *
 * @param code Error code;
 *
 * @param message Error message.
 *
 * @return Error code (as a convenience to the caller).
 */
int VCsetError(VC *vc, int code, char *message)
{
    vc->errorCode = code;
    if (vc->errorMessage != null) {
        free(vc->errorMessage);
    }
    vc->errorMessage = strdup(message);
    if ((code != 0) && (code != VC_E_CAMERA_BUSY)) {
        printf("++++ %d Error : code: %3d  : %s", __LINE__, code, message);
        VCprintCommand(vc);
        printf("\n");
    }
    return code;
}

// --------------------------------------------------------

/**
 * Report an error from the operating system.
 *
 * @param vc Control structure.
 *
 * @param ern Current value of errno.
 *
 * @param message Description of what was going on.
 *
 * @return Error code.
 */
int VCsystemError(VC *vc, int ern, char *message)
{
    char sysMsg[2048];
    if (strerror_r(ern, sysMsg, sizeof(sysMsg)-1) == 0) {
        char *fullMsg = malloc(strlen(sysMsg) + strlen(message) + 100);
        sprintf(fullMsg, "System errno %d - %s : %s", ern, message, sysMsg);
        VCsetError(vc, VC_E_SYSTEM, fullMsg);
        free(fullMsg);
    }
    else {
        char *fullMsg = malloc(strlen(message) + 100);
        sprintf(fullMsg,
            "System errno %d - %s : Unable to obtain system message.",
            ern, message);
        VCsetError(vc, VC_E_SYSTEM, fullMsg);
        free(fullMsg);
    }

    return VC_E_SYSTEM;
}

// --------------------------------------------------------

/**
 * Print the current command.  This is generally used for debugging.
 *
 * @param vc Control structure.
 */
void VCprintCommand(VC *vc)
{
    int i = 0;
    while (true) {
        printf(" %02x", vc->command[i]);
        if ((vc->command[i] == VC_CMD_END) || (i > 30)){
            break;
        }
        i++;
    }
}

// --------------------------------------------------------

/**
 * Convert pan degrees to raw camera native units.
 *
 * @param vc Control structure.
 *
 * @param degrees Degrees to pan (-360 to +360).
 *
 * @param ascii Put 4 ascii characters here.
 *
 * @param size Number of bytes in resulting ascii.
 *
 * @return Hex value of pan speed.
 */

int16 VCpanDegreesToAscii(VC *vc, double degrees, unsigned char *ascii, int size)
{
    int16 p = round(degrees / vc->panRatio);
    VChexToAscii(p+0x8000, ascii, size);
    return p;
}

// --------------------------------------------------------

/**
 * Convert tilt degrees to raw camera native units.
 *
 * @param vc Control structure.
 *
 * @param degrees Degrees to tilt (-360 to +360).
 *
 * @param ascii Put 4 ascii characters here.
 *
 * @param size Number of bytes in resulting ascii.
 *
 * @return Hex value of tilt speed.
 */

int16 VCtiltDegreesToAscii(VC *vc, double degrees, unsigned char *ascii, int size)
{
    int16 t = round(degrees / vc->tiltRatio);
    VChexToAscii(t+0x8000, ascii, size);
    return t;
}


// --------------------------------------------------------

/**
 * Given a response from the camera, decide whether it is a
 * good one or not.  If not, describe what the problem is.
 *
 * @param vc Control structure.
 *
 * @param response Message from camera to computer.
 *
 * @return Zero if good message from camera, error code if
 *         there is a problem.
 */

int VCcameraStatus(VC *vc)
{
    unsigned char good    [] = { 0xfe, 0x30, 0x30, 0x30, 0x30, VC_CMD_END };
    unsigned char busy    [] = { 0xfe, 0x30, 0x30, 0x31, 0x30, VC_CMD_END };
    unsigned char badParam[] = { 0xfe, 0x30, 0x30, 0x35, 0x30, VC_CMD_END };
    unsigned char modeErr [] = { 0xfe, 0x30, 0x30, 0x39, 0x30, VC_CMD_END };

    if (memcmp (vc->response, good, sizeof(good)) == 0) {
        return VCsetError(vc, 0, "ok");
    }

    if (memcmp (vc->response, busy, sizeof(busy)) == 0) {
        return VCsetError(vc, VC_E_CAMERA_BUSY, "Camera is busy.");
    }

    if (memcmp (vc->response, badParam, sizeof(badParam)) == 0) {
        return VCsetError(vc, VC_E_BAD_PARAM, "Bad parameter in command to camera.");
    }

    if (memcmp (vc->response, modeErr, sizeof(modeErr)) == 0) {
        return VCsetError(vc, VC_E_MODE_ERROR, "Camera mode error.");
    }

    // if we get here is means that the command succeded and
    // that the camera is sending back some sort of information,
    // such as the current pan angle.
    return VCsetError(vc, 0, "ok");
}

// --------------------------------------------------------

/**
 * Send a message to the camera.  The command must be
 * terminated with VC_CMD_END, which will be used as the
 * cmomand terminator.  It is assumed that commands are short
 * (less than 100 bytes).
 *
 * After the command is sent, the response will be read from
 * the camera.
 *
 * If a good response is received, then a good status will be returned.
 *
 * If the camera does not respond or responds with an error,
 * then an error code will be returned.
 *
 * @param vc Control structure.
 *
 * @param command Command bytes to send to the camera,
 *        terminated with VC_CMD_END.
 *
 * @param resp If this is not null and the camera gives a
 * good response, then put the resonse here.  Caller must
 * ensure that the buffer is big enough.
 *
 * @return Zero on success, error code on failure.
 */

int VCtoCamera(VC *vc, VC_BUF command)
{
    int writeSize = -1;       // number of bytes written
    int tries;                // loop control flag
    int responseSize = 0;
    int cmdLen = 1;           // length of command

    // build full command in one string
    cmdLen = 1;
    while (command[cmdLen-1] != VC_CMD_END) {
        cmdLen++;
    }

    VC_ASSERT(cmdLen < (sizeof(VC_BUF)-10), "camera command too long");

    memset(vc->command, 0, sizeof(vc->command));
    vc->command[0] = (unsigned char)0xff;
    vc->command[1] = (unsigned char)0x30;
    vc->command[2] = (unsigned char)0x30;
    memcpy(vc->command+3, command, cmdLen);
    writeSize = write(vc->fd, vc->command, 3+cmdLen);
    // TRCHEX(vc->command, 3+cmdLen);

    if (writeSize != (3+cmdLen)) {
        return VCsystemError(vc, errno, "Sending message from computer to camera");
    }

    // read response from camera
    for (tries = 0; tries < 5; tries++) {
        struct timeval timeout;
        fd_set readfds;              // wait on this list of file descriptors
        int sel = -1;                // result of select
        int readSize = -1;           // number of bytes read

        memset(&timeout, 0, sizeof(timeout));
        timeout.tv_sec  = (long)VC_MAX_WAIT_TIME;
        timeout.tv_usec = (long)(round(VC_MAX_WAIT_TIME * 1000 * 1000));

        FD_ZERO(&readfds);
        FD_SET(vc->fd, &readfds);

        sel = select((vc->fd)+1, &readfds, null, null, &timeout);

        switch(sel) {
            case 1:
                // data is ready on input
                readSize = read(vc->fd, vc->response+responseSize, sizeof(VC_BUF)-responseSize);
                if (readSize > 0) {
                    responseSize += readSize;
                    // check for overflow
                    if (responseSize >= sizeof(VC_BUF)) {
                        // mark this message as finished
                        vc->response[sizeof(VC_BUF)-1] = VC_CMD_END;
                        return VCcameraStatus(vc);
                    }
                    if (vc->response[responseSize-1] == VC_CMD_END) {
                        // Got the terminator.  We are all done.
                        return VCcameraStatus(vc);
                    }
                }
                break;
            case 0:
                // Got a timeout.  Must be the camera has failed.
                return VCsetError(vc, VC_E_NO_RESPONSE, "No response from camera.");
                break;
            case -1:
                if (errno == EINTR) {
                    // got interrupted.  Try again.
                    break;
                }
                else {
                    // something is actually wrong.
                    return VCsystemError(vc, errno, "Reading response from camera");
                }
                return VCsystemError(vc, errno, "Waiting for response from camera");
                break;
            default:
                // this should never happen
                return VCsetError(vc, VC_E_UNKNOWN, "Unexpected return from select call");
        }
    }
    // this will only happen if we keep getting interrupted.
    return VCsystemError(vc, errno, "Repeated interrupts while reading from camera");
}

// --------------------------------------------------------

/**
 * Send a command to the camera.  If the camera is busy, keep
 * retrying it until it either succeeds, fails for another
 * reason, or the operation times out
 *
 * @param vc Control structure.
 *
 * @param command Command to send to camera.
 *
 * @return Zero on success, error code on failure.
 */

int VCtoCameraRetryBusy(VC *vc, VC_BUF command)
{
    double timeout = VCgetTime() + VC_MAX_OPERATION_TIME;
    int ok = true;

    VCtoCamera(vc, command);

    while ((vc->errorCode == VC_E_CAMERA_BUSY) && (timeout > VCgetTime())) {
        VCtoCamera(vc, command);
        ok = false;
    }
    if ((!ok) && (vc->errorCode == 0)) {
        ok = true;
    }
    if ((!ok) && (timeout <= VCgetTime())) {
        printf("VCtoCameraRetryBusy failed.  Status: %d\n", vc->errorCode);
        ok = true;
    }
    if (!ok) {
        TRCMSG("busy failed for some other reason");
        ok = true;
    }
    return vc->errorCode;
}

// --------------------------------------------------------

/**
 * Put the camera in host control mode.  This must be done
 * before issuing any other commands.  This mode means that the
 * camera is controlled by the computer and not by the
 * infra-red remote.
 *
 * @param vc Control structure.
 *
 * @return Zero on success, error code on failure.
 */

int VChostControl(VC *vc)
{
    VC_BUF hostControl = { 0x00, 0x90, 0x30, VC_CMD_END };
    return VCtoCamera(vc, hostControl);
}

// --------------------------------------------------------

/**
 * Start or stop camera movement.
 *
 * @param vc Control structure.
 *
 * @param pan Direction to move camera.  Choices are:
 *   VC_PAN_STOP
 *   VC_PAN_RIGHT
 *   VC_PAN_LEFT
 *
 * @param tilt Direction to move camera.  Choices are:
 *   VC_TILT_STOP
 *   VC_TILT_UP
 *   VC_TILT_DOWN
 *
 * @return Zero on success, error code on failure.
 */

int VCmove(VC *vc, int pan, int tilt)
{
    // if nothing to do, then do nothing.
    if ((pan == vc->panDirection) && (tilt == vc->tiltDirection)) {
        return 0;
    }

    switch(pan) {
        case VC_PAN_STOP  :
        case VC_PAN_RIGHT :
        case VC_PAN_LEFT  :
            break;

        default :
        {
            char msg[200];
            sprintf(msg, "Invalid pan direction : %d.", pan);
            return VCsetError(vc, VC_E_PAN_DIRECTION, msg);
        }
    }

    switch(tilt) {
        case VC_TILT_STOP  :
        case VC_TILT_UP    :
        case VC_TILT_DOWN  :
            break;

        default :
        {
            char msg[200];
            sprintf(msg, "Invalid tilt direction : %d.", tilt);
            return VCsetError(vc, VC_E_TILT_DIRECTION, msg);
        }
    }

    unsigned char command[] = { 0x00, 0x60, 0x00, 0x00, VC_CMD_END };
    VChexToAscii((int16)(pan+0x8000), command+2, 1);
    VChexToAscii((int16)(tilt+0x8000), command+3, 1);

    if (VCtoCameraRetryBusy(vc, command) == 0) {
        vc->panDirection = pan;
        vc->tiltDirection = tilt;
    }
    vc->setDirectionTime = VCgetTime();
    return vc->errorCode;
}

// --------------------------------------------------------

/**
 * Get the current pan and tilt angles of the camera.  The
 * results are put in the vc struct.
 *
 * @param vc Control structure.
 *
 * @return Zero on success, error code on failure.
 */
int VCgetPosition(VC *vc)
{
    VC_BUF command = { 0x00, 0x63, VC_CMD_END };

    if (VCtoCameraRetryBusy(vc, command) == 0) {
        // convert pan and tilt to degrees
        vc->panPosition  = (VCrawToReal(vc->response+5, 4)-0x8000) * vc->panRatio;
        vc->tiltPosition = (VCrawToReal(vc->response+9, 4)-0x8000) * vc->tiltRatio;
        vc->getPositionTime = VCgetTime();
    }
    return vc->errorCode;

}

// --------------------------------------------------------

/**
 * Position the camera to the given angles.
 *
 * @param vc Control structure.
 *
 * @param pan Position in degrees at which to point camera.
 *        Positive for right, negative for left.  Zero means straight ahead.
 *
 * @param tilt Position in degrees at which to point camera.
 *        Positive for up, negative for down.  Zero means level.
 *
 * @return Zero on success, error code on failure.
 */
int VCsetPosition(VC *vc,  double pan, double tilt)
{
    VC_BUF command =
        { 0x00, 0x62,
          0x00, 0x00, 0x00, 0x00,
          0x00, 0x00, 0x00, 0x00, VC_CMD_END };

    if ((pan > vc->panMax) || (pan < -vc->panMax)) {
        return VCsetError(vc, VC_E_PAN_POSITION, "Pan degrees out of bounds.");
    }
    if ((tilt > vc->tiltMax) || (tilt < -(vc->tiltMax))) {
        return VCsetError(vc, VC_E_TILT_POSITION, "Tilt degrees out of bounds.");
    }

    VCpanDegreesToAscii (vc, pan , command+2, 4);
    VCtiltDegreesToAscii(vc, tilt, command+6, 4);

    return VCtoCamera(vc, command);
}

// --------------------------------------------------------

/**
 * Determine if the camera is at the current position.  Note
 * that this is rounded to the accuracy to which the camera can
 * be positioned.
 *
 * @param vc Control structure.
 *
 * @param pan Position in degrees at which the camera should be.
 *        Positive for right, negative for left.  Zero means straight ahead.
 *
 * @param tilt Position in degrees at which the camera should be.
 *        Positive for up, negative for down.  Zero means level.
 *
 * @return Zero if in position, -1 if not in that position,
 *        other error code on failure.
 */

int VCisAtPosition(VC *vc, double pan, double tilt)
{
    int  panHex;
    int tiltHex;
    VC_BUF command = { 0x00, 0x63, VC_CMD_END };

    if ((pan > vc->panMax) || (pan < -vc->panMax)) {
        return VCsetError(vc, VC_E_PAN_POSITION, "Pan degrees out of bounds.");
    }
    if ((tilt > vc->tiltMax) || (tilt < -(vc->tiltMax))) {
        return VCsetError(vc, VC_E_TILT_POSITION, "Tilt degrees out of bounds.");
    }

    VCpanDegreesToAscii (vc, pan , (unsigned char*) (&panHex), 4);
    VCtiltDegreesToAscii(vc, tilt, (unsigned char*)(&tiltHex), 4);

    if (VCtoCameraRetryBusy(vc, command) == 0) {
        int p;
        int t;
        memcpy (&p, vc->response+5, 4);
        memcpy (&t, vc->response+9, 4);
        vc->panPosition  = (VCrawToReal(vc->response+5, 4)-0x8000) * vc->panRatio;
        vc->tiltPosition = (VCrawToReal(vc->response+9, 4)-0x8000) * vc->tiltRatio;
        // printf("+ panHex: %8x  p: %8x    tiltHex: %8x  t: %8x\n", panHex, p, tiltHex, t);
        if ((p == panHex) && (t == tiltHex)) {
            return 0;
        }
        else {
            // command was successful, but the camera is not in that position.
            return -1;
        }
    }

    // command error
    return vc->errorCode;
}

// --------------------------------------------------------

/**
 * Position the camera to the given angles.  Do it as quickly
 * as possible, sacrificing some accuracy.
 *
 * @param vc Control structure.
 *
 * @param pan Position in degrees at which to point camera.
 *        Positive for right, negative for left.  Zero means straight ahead.
 *
 * @param tilt Position in degrees at which to point camera.
 *        Positive for up, negative for down.  Zero means level.
 *
 * @return Zero on success, error code on failure.
 */
int VCsetPanPositionFast(VC *vc,  double pan)
{
    if (VCgetPosition(vc) == 0) {

        if (vc->panSpeed != vc->panMaxSpeed) {
            if (VCsetPanSpeed(vc, vc->panMaxSpeed) != 0) {
                return vc->errorCode;
            }
        }

        // determine direction in which to move
        int panDirection = VC_PAN_STOP;
        if (vc->panPosition < pan) {
            panDirection = VC_PAN_RIGHT;
        }
        if (vc->panPosition > pan) {
            panDirection = VC_PAN_LEFT;
        }

        // slope and off set when camera is moving at max pan speed
        double panSlope = 0.011204481792717087;
        double panOffset = 0.078151260504201653;
        // double panSlope = 89.25;
        // double panOffset = -6.975;

        // slope and off set when camera is moving at max tilt speed
        // double tiltSlope = 72;
        // double tiltOffset = -5.85;

        double panDiff  = ABS( pan - vc->panPosition);
        // double tiltDiff = ABS(tilt - vc->tiltPosition);

        double delay = (panSlope * panDiff) + panOffset;

        if (VCmove(vc, panDirection, VC_TILT_STOP) != 0) {
            return vc->errorCode;
        }
        VCsleep(delay);
        VCmove(vc, VC_PAN_STOP, VC_TILT_STOP);
    }
    return vc->errorCode;
}

// --------------------------------------------------------

/**
 * Given an ascii string of hex numbers, convert it to a binary.
 *
 * @param raw Ascii string of hex numbers
 *
 * @param size Number of characters in raw.
 *
 * @return Int value representation of value.
 */
unsigned int VCrawToInt(unsigned char *raw, int size)
{
    char cr[20];
    unsigned int binary;
    memcpy(cr, raw, size);
    cr[size] = '\0';
    sscanf(cr, "%X", &binary);
    return binary;
}

// --------------------------------------------------------

/**
 * Given an ascii string of hex numbers, convert it to a real number.
 *
 * @param raw Ascii string of hex numbers
 *
 * @param size Number of characters in raw.
 *
 * @return floating point representation of value.
 */
double VCrawToReal(unsigned char *raw, int size)
{
    double d = VCrawToInt(raw, size);
    return d;
}

// --------------------------------------------------------

/**
 * Get the pan ratio of the camera.
 *
 * @param vc Control structure.
 *
 * @param ps Pan ratio (returned).
 *
 * @return Zero on success, error code on failure.
 */
int VCgetPanRatio(VC *vc, double *pr)
{
    int errorCode;
    VC_BUF command = { 0x00, 0x5b, 0x30, VC_CMD_END };

    errorCode = VCtoCamera(vc, command);
    if (errorCode == 0) {
        *pr = VCrawToReal(vc->response+5, 4) / 100000.0;
    }
    return errorCode;
}

// --------------------------------------------------------

/**
 * Get the tilt ratio of the camera.
 *
 * @param vc Control structure.
 *
 * @param ps Tilt ratio (returned).
 *
 * @return Zero on success, error code on failure.
 */

int VCgetTiltRatio(VC *vc, double *tr)
{
    int errorCode;
    VC_BUF command = { 0x00, 0x5b, 0x31, VC_CMD_END };

    errorCode = VCtoCamera(vc, command);
    if (errorCode == 0) {
        *tr = VCrawToReal(vc->response+5, 4) / 100000.0;
    }
    return errorCode;
}

// --------------------------------------------------------

/**
 * Set the pan speed of the camera.  If the speed is too fast
 * or slow, then set it to the corresponding high or low speed limit.
 *
 * @param vc Control structure.
 *
 * @param ps Pan speed in degrees/sec.
 *
 * @return Zero on success, error code on failure.
 */
int VCsetPanSpeed(VC *vc, double ps)
{
    unsigned char command[] = { 0x00, 0x50, 0x00, 0x00, 0x00, VC_CMD_END };
    double speed = ps;

    if (ps > vc->panMaxSpeed) {
        speed = vc->panMaxSpeed;
    }
    if (ps < vc->panMinSpeed) {
        speed = vc->panMinSpeed;
    }

    // if the value is zero, make it at least one
    if (VCpanDegreesToAscii(vc, speed, command+2, 3) == 0) {
        command[4] = 0x31;
    }

    if (VCtoCameraRetryBusy(vc, command) == 0) {
        vc->panSpeed = VCrawToReal(vc->command+2, 3) * vc->panRatio;
    }
    return vc->errorCode;
}

// --------------------------------------------------------

/**
 * Set the tilt speed of the camera.  If the speed is too
 * fast or slow, then set it to the corresponding high or low
 * speed limit.
 *
 * @param vc Control structure.
 *
 * @param ts Tilt speed in degrees/sec.
 *
 * @return Zero on success, error code on failure.
 */
int VCsetTiltSpeed(VC *vc, double ts)
{
    unsigned char command[] = { 0x00, 0x51, 0x00, 0x00, 0x00, VC_CMD_END };
    double speed = ts;

    if (ts > vc->tiltMaxSpeed) {
        speed = vc->tiltMaxSpeed;
    }

    if (ts < vc->tiltMinSpeed) {
        speed = vc->tiltMinSpeed;
    }

    if (VCtiltDegreesToAscii(vc, speed, command+2, 3) == 0) {
        command[4] = 0x31;
    }

    if (VCtoCameraRetryBusy(vc, command) == 0) {
        vc->tiltSpeed = ts;
    }
    return vc->errorCode;
}

// --------------------------------------------------------

/**
 * Get the pan speed of the camera in degrees / sec.
 *
 * @param vc Control structure.
 *
 * @param ps Pan speed (returned).
 *
 * @return Zero on success, error code on failure.
 */
int VCgetPanSpeed(VC *vc, double *ps)
{
    VC_BUF command = { 0x00, 0x52, 0x30, VC_CMD_END };
    int errorCode = VCtoCamera(vc, command);
    if (errorCode == 0) {
        *ps = VCrawToReal(vc->response+5, 3) * vc->panRatio;
    }
    return errorCode;
}

// --------------------------------------------------------

/**
 * Get the tilt speed of the camera in degrees / sec.
 *
 * @param vc Control structure.
 *
 * @param ps Tilt speed (returned).
 *
 * @return Zero on success, error code on failure.
 */
int VCgetTiltSpeed(VC *vc, double *ts)
{
    VC_BUF command = { 0x00, 0x52, 0x31, VC_CMD_END };
    int errorCode = VCtoCamera(vc, command);
    if (errorCode == 0) {
        *ts = VCrawToReal(vc->response+5, 3) * vc->tiltRatio;
    }
    return errorCode;
}

// --------------------------------------------------------

/**
 * Convert the pan speed to the actual pan speed that the
 * camera will use.  This accounts for the fact that the speed
 * setting has a granularity.  For example, if the camera is
 * given a pan speed of 3.0000 degrees/sec, then it will
 * actually set it to 3.0375 degress/sec.
 *
 * Setting a value of 0 to stop panning will not work,
 * although the conversion does.
 *
 * @param vc Control structure.
 *
 * @param ps Pan speed.
 *
 * @return Converted value.
 */

double VCconvertPanSpeed(VC *vc, double ps)
{
    unsigned char buf[4];
    VCpanDegreesToAscii(vc, ps, buf, 3);
    double s = VCrawToReal(buf, 3) * vc->panRatio;
    return s;
}

// --------------------------------------------------------

/**
 * Convert the tilt speed to the actual tilt speed that the
 * camera will use.  This accounts for the fact that the speed
 * setting has a granularity.  For example, if the camera is
 * given a tilt speed of 24.0000 degrees/sec, then it will
 * actually set it to 23.96250 degress/sec.  Note that some
 * work out exactly, such as 9, 18, and 27 degrees.
 *
 * Setting a value of 0 to stop tilting will not work,
 * although the conversion does.
 *
 * @param vc Control structure.
 *
 * @param ts Tilt speed.
 *
 * @return Converted value.
 */

double VCconvertTiltSpeed(VC *vc, double ts)
{
    unsigned char buf[4];
    VCtiltDegreesToAscii(vc, ts, buf, 3);
    double s = VCrawToReal(buf, 3) * vc->tiltRatio;
    return s;
}

// --------------------------------------------------------

// --------------------------------------------------------

/**
 * Reset the camera.
 *
 * @param vc Control structure.
 *
 * @return Zero on success, error code on failure.
 */

int VCreset(VC *vc)
{
    VC_BUF command = { 0x00, 0xa1, 0xaa, VC_CMD_END };
    return VCtoCamera(vc, command);
}

// --------------------------------------------------------

/**
 * Set the focus mode.
 *
 * @param vc Control structure.
 *
 * @param mode Focus mode.
 *
 * @return Zero on success, error code on failure.
 */

int VCsetFocusMode(VC *vc, int mode)
{
    VC_BUF command = { 0x00, 0xa1, 0x00, VC_CMD_END };
    switch(mode) {
        case VC_FOCUS_MODE_AUTOMATIC : break;
        case VC_FOCUS_MODE_MANUAL    : break;
        case VC_FOCUS_MODE_NEAR      : break;
        case VC_FOCUS_MODE_FAR       : break;
        default : {
            char msg[200];
            sprintf(msg,
                "Invalid focus mode %X (hex).  Must be one of VC_FOCUS_MODE_* values.",
                mode);
            VCsetError(vc, VC_E_FOCUS_OUT_OF_RANGE, msg);
            return vc->errorCode;
        }
    }

    command[2] = mode;
    return VCtoCamera(vc, command);
}

// --------------------------------------------------------

/**
 * Set the focus position.
 *
 * @param vc Control structure.
 *
 * @param pos Position of focus.
 *
 * @return Zero on success, error code on failure.
 */

int VCsetFocusPosition(VC *vc, int pos)
{
    VC_BUF command = { 0x00, 0xb0, 0x00, 0x00, 0x00, 0x00, VC_CMD_END };
    int16 p = pos;
#if 0
    if ((pos < vc->focusMinRange) || (pos > vc->focusMaxRange)) {
        char msg[200];
        sprintf(msg, "Focus value of %d out of range.  Range: %d through %d",
            pos, vc->focusMinRange, vc->focusMaxRange);
        VCsetError(vc, VC_E_FOCUS_OUT_OF_RANGE, msg);
    }
#endif
    VChexToAscii(p, command+2, 4);
    return VCtoCamera(vc, command);
}

// --------------------------------------------------------

/**
 * Get the focal range.
 *
 * @param vc Control structure.
 *
 * @param minFocus Minimum focus range.
 *
 * @param maxFocus Maximum focus range.
 *
 * @return Zero on success, error code on failure.
 */
int VCgetFocusRange(VC *vc, unsigned int *minFocus, unsigned int *maxFocus)
{
    VC_BUF command = { 0x00, 0xb1, 0x32, VC_CMD_END };

    int errorCode = VCtoCamera(vc, command);

    if (errorCode == 0) {
        *minFocus = VCrawToInt(vc->response+5, 4);
        *maxFocus = VCrawToInt(vc->response+9, 4);
    }
    return errorCode;
}

// --------------------------------------------------------

/**
 * Get the current focus position.
 *
 * @param vc Control structure.
 *
 * @param pos Focus position (returned).
 *
 * @return Zero on success, error code on failure.
 */

int VCgetFocusPosition(VC *vc, unsigned int *pos)
{
    VC_BUF command = { 0x00, 0xb1, 0x30, VC_CMD_END };

    int errorCode = VCtoCamera(vc, command);

    if (errorCode == 0) {
        *pos = VCrawToInt(vc->response+5, 4);
    }
    return errorCode;
}

// --------------------------------------------------------

/**
 * Get the zoom range.
 *
 * @param vc Control structure.
 *
 * @param minFocus Minimum focus range.
 *
 * @param maxFocus Maximum focus range.
 *
 * @return Zero on success, error code on failure.
 */
int VCgetZoomRange(VC *vc, unsigned int *minZoom, unsigned int *maxZoom)
{
    // command to get max zoom.  Min zoom appears to be zero.
    VC_BUF command = { 0x00, 0xb4, 0x33, VC_CMD_END };

    int errorCode = VCtoCamera(vc, command);

    if (errorCode == 0) {
        *minZoom = 0;
        *maxZoom = VCrawToInt(vc->response+5, 4);
    }
    return errorCode;
}

// --------------------------------------------------------

/**
 * Set the zoom.  This uses zoom position 2.
 *
 * @param vc Control structure.
 *
 * @param zoom Amount to zoom.  The larger the value, the
 *        more the image is magnified.
 *
 * @return Zero on success, error code on failure.
 */
int VCsetZoom(VC *vc, int zoom)
{
    VC_BUF command = { 0x00, 0xb3, 0x00, 0x00, 0x00, 0x00, VC_CMD_END };
    int16 z = zoom;
    if ((zoom < vc->zoomMinRange) || (zoom > vc->zoomMaxRange)) {
        char msg[200];
        sprintf(msg, "Zoom value of %d out of range.  Range: %d through %d",
            zoom, vc->zoomMinRange, vc->zoomMaxRange);
        VCsetError(vc, VC_E_ZOOM_OUT_OF_RANGE, msg);
    }
    VChexToAscii(z, command+2, 4);

    VCtoCameraRetryBusy(vc, command);

    if (vc->errorCode == 0) {
        vc->zoomPosition = zoom;
    }

    return vc->errorCode;
}

// --------------------------------------------------------

/**
 * Get the current zoom position.  Use zoom position 2.
 *
 * @param vc Control structure.
 *
 * @param pos Zoom position (returned).
 *
 * @return Zero on success, error code on failure.
 */

int VCgetZoomPosition(VC *vc, unsigned int *pos)
{
    VC_BUF command = { 0x00, 0xb4, 0x30, VC_CMD_END };

    int errorCode = VCtoCamera(vc, command);

    if (errorCode == 0) {
        *pos = VCrawToInt(vc->response+5, 4);
    }
    return errorCode;
}

// --------------------------------------------------------

/**
 * Set the backlight compensation mode.
 *
 * @param vc Control structure.
 *
 * @param mode Backlight compensation mode.
 *
 * @return Zero on success, error code on failure.
 */
int VCsetBacklightCompensationMode(VC *vc, int mode)
{
    unsigned char command[] = { 0x00, 0xa5, 0x00, VC_CMD_END };
    switch (mode) {
        case VC_BACKLIGHT_COMPENSATION_OFF : break;
        case VC_BACKLIGHT_COMPENSATION_ON  : break;
        default : VC_ASSERT(0, "Bad VCsetBacklightCompensationMode parameter");
    }
    command[2] = mode;

    return VCtoCameraRetryBusy(vc, command);
}

// --------------------------------------------------------

/**
 * Set the exposure mode.
 *
 * @param vc Control structure.
 *
 * @param mode Exposure mode.
 *
 * @return Zero on success, error code on failure.
 */
int VCsetExposureMode(VC *vc, int mode)
{
    unsigned char command[] = { 0x00, 0xa5, 0x00, VC_CMD_END };
    switch (mode) {
        case VC_EXPOSURE_AUTOMATIC : break;
        case VC_EXPOSURE_MANUAL    : break;
        default : VC_ASSERT(0, "Bad VCsetExposureMode parameter");
    }
    command[2] = mode;
    return VCtoCameraRetryBusy(vc, command);
}

// --------------------------------------------------------

/**
 * Set the AE lock mode.  Presumably "AE" stands for "auto
 * exposure", but the Canon documentation does not specify.
 *
 * @param vc Control structure.
 *
 * @param mode AE mode.
 *
 * @return Zero on success, error code on failure.
 */
int VCsetAELockMode(VC *vc, int mode)
{
    unsigned char command[] = { 0x00, 0xa5, 0x34, 0x00, VC_CMD_END };
    switch (mode) {
        case VC_AE_LOCK_OFF : break;
        case VC_AE_LOCK_ON  : break;
        default : VC_ASSERT(0, "Bad VCsetAELockMode parameter");
    }
    command[3] = mode;
    return VCtoCameraRetryBusy(vc, command);
}

// --------------------------------------------------------

/**
 * Set the shutter speed mode.
 *
 * @param vc Control structure.
 *
 * @param mode Shutter speed program mode.
 *
 * @return Zero on success, error code on failure.
 */
int VCsetShutterSpeedMode(VC *vc, int mode)
{
    unsigned char command[] = { 0x00, 0xa8, 0x00, VC_CMD_END };
    switch (mode) {
        case VC_SHUTTER_SPEED_PROGRAM : break;
        case VC_SHUTTER_SPEED_1_60    : break;
        case VC_SHUTTER_SPEED_1_100   : break;
        default : VC_ASSERT(0, "Bad VCsetShutterSpeedMode parameter");
    }
    command[5] = mode;
    return VCtoCameraRetryBusy(vc, command);
}

// --------------------------------------------------------

/**
 * Set the iris.
 *
 * @param vc Control structure.
 *
 * @param iris Iris setting.
 *
 * @return Zero on success, error code on failure.
 */
int VCsetIris(VC *vc, char *iris)
{
    unsigned char command[] = { 0x00, 0xa5, 0x39, 0x00, 0x00, VC_CMD_END };
    memcpy(command+3, iris, 2);
    return VCtoCameraRetryBusy(vc, command);
}

// --------------------------------------------------------

/**
 * Point the camera to the straight ahead position.
 *
 * @param vc Control structure.
 */

int VChome(VC *vc)
{
    unsigned char home[] = { 0x00, 0x57, VC_CMD_END } ;
    return VCtoCamera(vc, home);
}

// --------------------------------------------------------

/**
 * Set the infrared cut filter to on or off.
 *
 * Normally, the camera is initialized with the infrared cut
 * filter on.  This filter cuts out infrared light and makes
 * for a nicer visual picture.  It is useful to turn it off so
 * that the camera becomes more sensitive to infrared light and
 * gives a better infrared picture.  This mode is often used
 * for nighttime viewing.
 *
 * @param vc Control structure.
 *
 * @param state VC_IR_CUT_ON or VC_IR_CUT_OFF
 */

int VCirCut(VC *vc, int state)
{
    unsigned char ircut[] = { 0x00, 0xb5, 0x30, VC_CMD_END } ;
    unsigned char irstatus[] = { 0x00, 0xb6, VC_CMD_END } ;
    int status;   // return status
    unsigned char onOff = 0x30;

    switch(state) {
        case VC_IR_CUT_ON  : onOff = 0x31; break;
        case VC_IR_CUT_OFF : onOff = 0x30; break;
        default : VC_ASSERT(0, "Bad VCirCut parameter");
    }

    // get current camera ir cut status
    status = VCtoCamera(vc, irstatus);
    if (status != 0) {
        // error getting status
        return status;
    }

    // if the ir cut is already in the desired state, then do nothing.
    if (vc->response[5] == onOff) {
        return 0;
    }

    // set state to desired state
    ircut[5] = onOff;

    // wait for camera to be not busy
    VCtoCamera(vc, ircut);
    VCsleep(3.0);
    VCtoCameraRetryBusy(vc, ircut);
    // wait for filter to move into position
    return vc->errorCode;
}

// --------------------------------------------------------------------------------

/**
 * Get the number of commands that have yet to be processed by the thread.
 *
 * @param vc Control structure.
 * 
 * @return The number of commands that the thread has not completed.
 */
int VCoutstandingCommands(VC *vc)
{
    return vc->requestedCommands - vc->completedCommands;
}

// --------------------------------------------------------------------------------

/**
 * Wake up the thread that performs camera I/O in the background.
 *
 * @param vc Control structure.
 */

void VCwriteThreadMessage(VC *vc, int command, void *buf, unsigned int size)
{
    unsigned char buffer[100];
    buffer[0] = command;
    VC_ASSERT((size+1) <= sizeof(buffer), "Temporary buffer too small");
    memcpy(buffer+1, buf, size);
    int s = write(vc->threadControl[1], buffer, 1+size);
    vc->requestedCommands++;
    if (s != (1+size)) {
        VCsystemError(vc, errno, "Unable to write to thread control pipe.");
    }
}

// --------------------------------------------------------------------------------

/**
 * Queue the starting or stoping of camera movement.
 *
 * @param vc Control structure.
 *
 * @param pan Direction to move camera.  Choices are:
 *   VC_PAN_STOP
 *   VC_PAN_RIGHT
 *   VC_PAN_LEFT
 *
 * @param tilt Direction to move camera.  Choices are:
 *   VC_TILT_STOP
 *   VC_TILT_UP
 *   VC_TILT_DOWN
 */
void VCqueueMove(VC *vc, int pan, int tilt)
{
    unsigned char buf[2];
    buf[0] = pan;
    buf[1] = tilt;
    vc->queuedPanDirection = pan;
    vc->queuedTiltDirection = tilt;

    VCwriteThreadMessage(vc, VC_Q_MOVE, buf, sizeof(buf));

    char msg[200];
    strcpy(msg, "VCmove ");
    if (pan == VC_PAN_STOP)   strcat(msg, "VC_PAN_STOP");
    if (pan == VC_PAN_RIGHT)  strcat(msg, "VC_PAN_RIGHT");
    if (pan == VC_PAN_LEFT)   strcat(msg, "VC_PAN_LEFT");

    strcat(msg, " : ");

    if (tilt == VC_TILT_STOP) strcat(msg, "VC_TILT_STOP");
    if (tilt == VC_TILT_UP)   strcat(msg, "VC_TILT_UP");
    if (tilt == VC_TILT_DOWN) strcat(msg, "VC_TILT_DOWN");
    VClog(vc, msg);
}

// --------------------------------------------------------------------------------

/*
 * Queue the pan speed to be set.
 *
 * @param vc Control structure.
 *
 * @param ps Pan speed in degrees/sec.
 */
void VCqueueSetPanSpeed(VC *vc, double ps)
{
    vc->queuedPanSpeed = ps;
    VCwriteThreadMessage(vc, VC_Q_PAN_SPEED, &ps, sizeof(ps));

    char msg[200];
    sprintf(msg, "VCqueueSetPanSpeed %8.4lf", ps);
    VClog(vc, msg);
}

// --------------------------------------------------------------------------------

/*
 * Queue the tilt speed to be set.
 *
 * @param vc Control structure.
 *
 * @param ts Tilt speed in degrees/sec.
 */
void VCqueueSetTiltSpeed(VC *vc, double ts)
{
    vc->queuedTiltSpeed = ts;
    VCwriteThreadMessage(vc, VC_Q_TILT_SPEED, &ts, sizeof(ts));

    char msg[200];
    sprintf(msg, "VCqueueSetTiltSpeed %8.4lf", ts);
    VClog(vc, msg);
}

// --------------------------------------------------------------------------------

/*
 * Queue the zoom to be set.
 *
 * @param vc Control structure.
 *
 * @param zoom Zoom units.
 */
void VCqueueSetZoom(VC *vc, int zoom)
{
    VCwriteThreadMessage(vc, VC_Q_ZOOM, &zoom, sizeof(zoom));

    char msg[200];
    sprintf(msg, "VCqueueSetZoom %8d", zoom);
    VClog(vc, msg);
}

// --------------------------------------------------------------------------------

/*
 * Queue the positioning of the camera.
 *
 * @param vc Control structure.
 *
 * @param pan Pan angle.
 *
 * @param tilt Tilt angle.
 */
void VCqueueSetPosition(VC *vc, double pan, double tilt)
{
    double buf[2];
    buf[0] = pan;
    buf[1] = tilt;

    VCwriteThreadMessage(vc, VC_Q_SET_POSITION, buf, sizeof(buf));

    char msg[200];
    sprintf(msg, "VCqueueSetPosition pan: %8.4lf  tilt: %8.4lf", pan, tilt);
    VClog(vc, msg);
}

// --------------------------------------------------------------------------------

/*
 * Queue the positioning of the camera and wait for that
 * position to be attained.
 *
 * @param vc Control structure.
 *
 * @param pan Pan angle.
 *
 * @param tilt Tilt angle.
 */
void VCqueueSetPositionAndWait(VC *vc, double pan, double tilt)
{
    double buf[2];
    buf[0] = pan;
    buf[1] = tilt;

    VCwriteThreadMessage(vc, VC_Q_SET_POSITION_AND_WAIT, buf, sizeof(buf));

    char msg[200];
    sprintf(msg, "VCqueueSetPosition pan: %8.4lf  tilt: %8.4lf", pan, tilt);
    VClog(vc, msg);
}

// --------------------------------------------------------------------------------

/*
 * Queue the command to get the current position of the
 * camera. Values are put in the control structure.
 *
 * @param vc Control structure.
 */
void VCqueueGetPosition(VC *vc)
{
    char *buf = NULL;
    VCwriteThreadMessage(vc, VC_Q_GET_POSITION, buf, 0);
    VClog(vc, "VCqueueGetPosition");
}

// --------------------------------------------------------------------------------

/*
 * Tell the server thread to exit.
 *
 * @param vc Control structure.
 */
void VCqueueExit(VC *vc)
{
    char *buf = NULL;
    VCwriteThreadMessage(vc, VC_Q_EXIT, buf, 0);
    VClog(vc, "VCqueueExit");
}

// --------------------------------------------------------------------------------

/**
 * Read the given number of bytes from the thread pipe.
 *
 * @param vc Control structure.
 *
 * @param buf Read results into here.
 *
 * @param targetSize Number of bytes to read.
 *
 * @return 0 on success, error code on failure.
 */

int VCreadThreadMessage(VC *vc, void *buf, int targetSize)
{
    int s = 0;
    int input = vc->threadControl[0];   // abbreviation

    while (s < targetSize) {
        fd_set readfds;
        FD_ZERO(&readfds);
        FD_SET(input, &readfds);
        int sel = select(input+1, &readfds, null, null, null);
        switch (sel) {
            case 1:
                {
                    int readSize = read(input, ((unsigned char*)buf)+s, targetSize-s);
                    if (readSize > 0) {
                        s += readSize;
                    }
                }
                break;

            case 0:
                // Got a timeout.  This should never happen.
                VC_ASSERT(0, "Unexpected select timeout");
                break;

            case -1:
                if (errno == EINTR) {
                    // got interrupted.  Try again.
                    break;
                }
                else {
                    // something is actually wrong.
                    return VCsystemError(vc, errno, "Reading from pipe");
                }
                return VCsystemError(vc, errno, "Waiting for pipe");
        }
    }
    return VCsetError(vc, 0, "ok");
}

// --------------------------------------------------------------------------------

/**
 * Run in the background and perform I/O with the camera as requested.
 *
 * @param vc Control structure.
 */

void *VCthread(void *v)
{
    // share the same VC with the parent.
    VC *vc = (VC*)v;

    while(true) {
        char buf[40];
        int status = VCreadThreadMessage(vc, buf, 1);
        if (status != 0) {
            pthread_exit(NULL);
        }
        switch(buf[0]) {
            case VC_Q_MOVE:
                {
                    char move[2];
                    status = VCreadThreadMessage(vc, move, sizeof(move));
                    if (status != 0) {
                        VCsystemError(vc, errno, "Unable to read VC_Q_MOVE.");
                        pthread_exit(NULL);
                    }
                    int pan = move[0];
                    int tilt = move[1];
                    VCmove(vc, pan, tilt);
                }
                break;
                
            case VC_Q_PAN_SPEED:
                {
                    double speed;
                    status = VCreadThreadMessage(vc, &speed, sizeof(speed));
                    if (status != 0) {
                        VCsystemError(vc, errno, "Unable to read VC_Q_PAN_SPEED.");
                        pthread_exit(NULL);
                    }
                    VCsetPanSpeed(vc, speed);
                }
                break;

            case VC_Q_TILT_SPEED:
                {
                    double speed;
                    status = VCreadThreadMessage(vc, &speed, sizeof(speed));
                    if (status != 0) {
                        VCsystemError(vc, errno, "Unable to read VC_Q_TILT_SPEED.");
                        pthread_exit(NULL);
                    }
                    VCsetTiltSpeed(vc, speed);
                }
                break;
             
            case VC_Q_ZOOM:
                {
                    int zoom;
                    status = VCreadThreadMessage(vc, &zoom, sizeof(zoom));
                    if (status != 0) {
                        VCsystemError(vc, errno, "Unable to read VC_Q_TILT_SPEED.");
                        pthread_exit(NULL);
                    }
                    VCsetZoom(vc, zoom);
                }
                break;
             
            case VC_Q_SET_POSITION:
                {
                    double setPos[2];
                    status = VCreadThreadMessage(vc, setPos, sizeof(setPos));
                    if (status != 0) {
                        VCsystemError(vc, errno, "Unable to read VC_Q_SET_POSITION.");
                        pthread_exit(NULL);
                    }
                    VCsetPosition(vc, setPos[0], setPos[1]);
                }
                break;

            case VC_Q_SET_POSITION_AND_WAIT:
                {
                    double setPos[2];
                    status = VCreadThreadMessage(vc, setPos, sizeof(setPos));
                    if (status != 0) {
                        VCsystemError(vc, errno, "Unable to read VC_Q_SET_POSITION_AND_WAIT.");
                        pthread_exit(NULL);
                    }
                    VCsetPosition(vc, setPos[0], setPos[1]);
                    while (VCisAtPosition(vc, setPos[0], setPos[1]) < 0) {
                        ;
                    }
                }
                break;

            case VC_Q_GET_POSITION:
                VCgetPosition(vc);
                break;

            case VC_Q_EXIT:
                pthread_exit(NULL);
                break;

            default:
                {
                    char msg[200];
                    sprintf(msg, "Invalid command code: %d", buf[0]);
                    VC_ASSERT(0, msg);
                }
                break;
        }
        vc->completedCommands++;
    }
    // should never get here, but the compiler requires a return value
    return NULL;
}

// --------------------------------------------------------------------------------

/**
 * Delete the control structure.
 *
 * @param vc Control structure.
 *
 * @return 0 on success, error code on failure.  The
 *         structure should be considered deleted whether the return
 *         code was good or bad.
 */
int VCdelete(VC *vc)
{
    int status = close(vc->fd);

    if (status == 0) {
        if (vc->errorMessage != null) {
            free(vc->errorMessage);
            vc->errorMessage = null;
        }
        memset(vc, 0, sizeof(VC));  // not really necessary, but tidy.
        return 0;
    }

    switch (errno) {
        case EBADF : return VC_E_CLOSE_INVALID;
        case EINTR : return VC_E_CLOSE_INTERRUPTED;
        case EIO   : return VC_E_CLOSE_IO_ERR;
        default    : return VC_E_UNKNOWN;
    }
    return VC_E_UNKNOWN;
}

// --------------------------------------------------------

/**
 * Sleep for the given number of seconds.  Time is accurate
 * to sub-second.
 *
 * @param t Time to sleep in seconds.  If less than or equal
 *          to zero then the call will return immediately.
 *
 */

void VCsleep(double t)
{
    if (t <= 0) {
        return;
    }
    struct timespec req;
    struct timespec rem;
    req.tv_sec = t;
    req.tv_nsec = (t - req.tv_sec ) * 1000000000.0;
    nanosleep(&req, &rem);
}

// --------------------------------------------------------

/**
 * Convert the zoom position to the field of view angle in
 * X direction.  Do this by linear interpolation between
 * consectutive entries in the zoomAngles table.
 *
 * @param zoom Zoom unit sent to camera.
 *
 * @return Angle in degrees between center of screen and
 *         right or left edge of screen.k
 */

/** table of camera zoom units and corresponding angle
 * between center and side of screen.
 */
static ZoomAngle zoomAngles[] = {
    {    0, 21.50926 },
    {  100, 19.85521 },
    {  400, 14.50338 },
    {  700, 10.81172 },
    { 1000,  6.34019 },
    { 1300,  3.97250 },
    { 1600,  2.48524 },
    { 1960,  1.04434 }
};

double VCgetZoomAngleX(int zoom)
{
    // number of array entries
    int numZoom = sizeof(zoomAngles) / sizeof(zoomAngles[0]);
    double angle = 0;   // return value
    int z;              // index into zoomAngles table

    for (z = 0; z < (numZoom-1); z++) {
        if ((zoom >= zoomAngles[z].zoom) && (zoom <= zoomAngles[z+1].zoom)) {
            double ratio =
                ((double)(zoom - zoomAngles[z].zoom)) /
                ((double)(zoomAngles[z+1].zoom - zoomAngles[z].zoom));
            angle = ((zoomAngles[z+1].angle - zoomAngles[z].angle) * ratio) +
                zoomAngles[z].angle;
            return angle;
        }
    }

    // if we got here then the the zoom value was bad.
    char msg[200];
    sprintf(msg, "Invalid zoom angle %d.  Must be from %d to %d inclusive.",
        zoom, zoomAngles[0].zoom, zoomAngles[numZoom-1].zoom);
    VC_ASSERT(false, msg);
    return angle;
}

// --------------------------------------------------------

/**
 * Convert the zoom position to the field of view angle in
 * Y direction.
 *
 * @param zoom Zoom unit sent to camera.
 *
 * @return Angle in degrees between center of screen and
 *         right or left edge of screen.k
 */
double VCgetZoomAngleY(int zoom)
{
    double x = VCgetZoomAngleX(zoom);
    double y = RtoD(atan(NTSC_ASPECT_RATIO * tan(DtoR(x))));
    return y;
}

// --------------------------------------------------------

/**
 * Set the shutter speed
 *
 * @param vc Control structure.
 *
 * @param shutter New shutter speed.
 *
 * @return 0 on success, error code on failure.
 */
int VCsetShutterSpeed(VC *vc, int shutter)
{
    unsigned char command[] = { 0x00, 0xa5, 0x35, 0x00, 0x00, VC_CMD_END };

    if (shutter > VC_SHUTTER_SPEED_MAX) {
        return VC_E_SHUTTER_SPEED_TOO_HIGH;
    }

    if (shutter < VC_SHUTTER_SPEED_MIN) {
        return VC_E_SHUTTER_SPEED_TOO_LOW;
    }

    VChexToAscii((int16)shutter, command+3, 2);
    printf("Shutter command: %02x %02x %02x %02x %02x %02x\n",
        command[0], command[1], command[2], command[3], command[4], command[5]);
    return VCtoCameraRetryBusy(vc, command);
}

// --------------------------------------------------------

/**
 * Get the shutter speed
 *
 * @param vc Control structure.
 *
 * @param shutter Current shutter speed (returned)
 *
 * @return 0 on success, error code on failure.
 */
int VCgetShutterSpeed(VC *vc, int *shutter)
{
    unsigned char command[] = { 0x00, 0xa5, 0x36, VC_CMD_END };
    if (VCtoCameraRetryBusy(vc, command) == 0) {
        char buf[4];
        memcpy(buf, vc->response+5, 2);
        buf[2] = '\0';
        sscanf(buf, "%02x", (unsigned int*)shutter);
    }
    return vc->errorCode;
}

// --------------------------------------------------------

/**
 * Convert the given shutter speed value into 1/frame time. 
 * For example, a value of 7 would return 215, which indicates
 * a 1/215 seconds exposure.  The default setup shutter time is
 * 1/60 second.  Setting to greater than that (like 1/30)
 * aggregates frames.
 *
 * @param shutter Value sent to and from camera.
 *
 * @return 1/time in seconds. -1 means that the value of the
 * shutter parameter is out of range.
 */

int VCconvertShutterSpeed(int shutter)
{
    static int speedsFast[][2] = {
        { 0x00,    60 },
        { 0x01,    75 },
        { 0x02,    90 },
        { 0x03,   100 },
        { 0x04,   125 },
        { 0x05,   150 },
        { 0x06,   180 },
        { 0x07,   215 },
        { 0x08,   250 },
        { 0x09,   300 },
        { 0x0A,   350 },
        { 0x0B,   425 },
        { 0x0C,   500 },
        { 0x0D,   600 },
        { 0x0E,   725 },
        { 0x0F,   850 },
        { 0x10,  1000 },
        { 0x11,  1250 },
        { 0x12,  1500 },
        { 0x13,  1750 },
        { 0x14,  2000 },
        { 0x15,  2500 },
        { 0x16,  3000 },
        { 0x17,  4000 },
        { 0x18,  6000 },
        { 0x19, 10000 },
        { 0x1A, 20000 },
        { 0x1B, 80000 }
    };
    #define FAST_SIZE (sizeof(speedsFast) / sizeof(*speedsFast))

    static int speedsSlow[][2] = {
        { 0x80,    30 },
        { 0x81,     8 },
        { 0x82,     2 },
        { 0x83,     1 }
    };
    #define SLOW_SIZE (sizeof(speedsSlow) / sizeof(*speedsSlow))

    if ((shutter >= speedsFast[0][0]) && 
        (shutter <= FAST_SIZE)) {
        return speedsFast[shutter][1];
    }

    if ((shutter >= speedsSlow[0][0]) && 
        (shutter <= SLOW_SIZE)) {
        return speedsSlow[shutter][1];
    }
    return -1;
}

// --------------------------------------------------------

/**
 * Estimate the current position of the camera based on the
 * command history.
 *
 * There is an assumption that after making any changes to
 * the camera's speed or direction that a call is made to get
 * the camera's position.
 *
 * @param vc Control structure.
 *
 * @param pan Position in degrees where the camera is pointed.
 *        Positive for right, negative for left.  Zero means
 *        straight ahead.
 *
 * @param tilt Position in degrees at which the camera is pointed.
 *        Positive for up, negative for down.  Zero means level.
 *
 * @return Zero on success, error code on failure.
 */

int VCestimatePosition(VC *vc, double *pan, double *tilt)
{
    double direction = 0;  // used for sign of direction

    double elapsed = VCgetTime() - vc->getPositionTime;

    // calculate horizontal position
    switch (vc->panDirection) {
        case VC_PAN_RIGHT : direction =  1; break;
        case VC_PAN_LEFT  : direction = -1; break;
        default           : direction =  0; break;
    }

    *pan = vc->panPosition + (vc->panSpeed * direction * elapsed);

    switch (vc->tiltDirection) {
        case VC_TILT_UP   : direction =  1; break;
        case VC_TILT_DOWN : direction = -1; break;
        default           : direction =  0; break;
    }
    *tilt = vc->tiltPosition + (vc->tiltSpeed * direction * elapsed);

    return 0;
}

// --------------------------------------------------------

void ShowError(VC *vc)
{
    printf("VC Error : %d : %s\n", vc->errorCode, vc->errorMessage);
}

#define VC_SHOW_ERROR(V,MSG) printf("%s:%d %s ", __FILE__, __LINE__, MSG), ShowError(V)

// --------------------------------------------------------

#ifndef VC_LIBRARY

#if 0

// main for testing and debugging.
int main(int argc, char **argv)
{
#if 0
{
    int z;
    double angle;
    for (z = 0; z <= 1960; z++) {
        angle = VCgetZoomAngle(z);
        printf("zoom: %4d   angle: %10.5lf\n", z, angle);
    }
    exit(99);
}
#endif
    VC vc;
    double panRatio;
    double tiltRatio;
    int code;
    char *devName = "/dev/ttyS1";

    if (argc > 1) {
        if (strchr(argv[1], '0') != null) {
            devName = "/dev/ttyS0";
        }
        if (strchr(argv[1], '1') != null) {
            devName = "/dev/ttyS1";
        }
    }

    char chmod[200];
    sprintf(chmod, "sudo chmod 666 %s", devName);
    system(chmod);

    int constructCode = VCconstruct(&vc, devName);

#if 1
{
    // speed conversion
    double s;
    double ds;
    unsigned char buf[5];
    int16 is;
    for (s = 0; s < 40; s++) {
        is = VCpanDegreesToAscii(&vc, s, buf, 3);
        ds = VCrawToReal(buf, 3) * vc.panRatio;
        printf("s: %9.5lf    is: %d   ds: %9.5lf\n", s, is, ds);
    }

    printf("\n");
    printf("Pan speed conversions.\n");
    for (s = 0; s < 40; s++) {
        ds = VCconvertPanSpeed(&vc, s);
        printf("s: %9.5lf    ds: %9.5lf\n", s, ds);
    }

    printf("\n");
    printf("Tilt speed conversions.\n");
    for (s = 0; s < 40; s++) {
        ds = VCconvertTiltSpeed(&vc, s);
        printf("s: %9.5lf    ds: %9.5lf\n", s, ds);
    }

    printf("\n");

    printf("Pan  ratio: %20.10lf\n", vc.panRatio);
    printf("Tilt ratio: %20.10lf\n", vc.tiltRatio);
    exit(99);
}
#endif

#if 0
    {
        printf("Camera deacceleration experiment.\n");
        double speed;
        for (speed = 0; speed < 91; speed += 5.0) {
            double maxTime = 2;
            double minTime = 0;

            int tries = 16;
            while ((ABS(maxTime - minTime) > .001) && (tries > 0)) {
                // reset to home
                VCmove(&vc,  VC_PAN_STOP, VC_TILT_STOP);
                VCsetPanSpeed(&vc, 90.0);
                VCsetPosition(&vc, 0.0, 0.0);
                while (VCisAtPosition(&vc, 0.0, 0.0) != 0) {
                }

                VCsetPanSpeed(&vc, speed);
                VCmove(&vc,  VC_PAN_RIGHT, VC_TILT_STOP);
                double avgTime = (maxTime - minTime) / 2.0;
                VCsleep(0.5);   // let camera get up to speed
                VCmove(&vc,  VC_PAN_STOP, VC_TILT_STOP);
                VCsleep(avgTime);   // let camera get up to speed
                VCgetPosition(&vc);
                double oldPan = vc.panPosition;
                VCgetPosition(&vc);
                printf(
    "Speed: %4.1lf  time: %8.6lf  Stopped: %s  minTime: %8.6lf  maxTime: %8.6lf  max-min: %8.6lf\n",
                    speed, avgTime, (oldPan == vc.panPosition) ? "yes" : "no ",
                    minTime, maxTime, maxTime - minTime);
                if (oldPan == vc.panPosition) {
                    maxTime = avgTime;
                }
                else {
                    minTime = avgTime;
                }
                tries--;
            }
        }
        exit(99);
    }
#endif

#if 0
{
    // test the accuracy of moving the camera using "go and stop" commands
    // Tue Dec 20 17:13:41 EST 2005
    double angle;
    double panSlope = 0.011204481792717087;
    double panOffset = 0.078151260504201653;
    double startTime;
    double delayTime;
    double actualAngle;
    double stopTime;
    double oldPosn;
    double finishTime;
    double totalTime;


    VCsetPanSpeed(&vc, vc.panMaxSpeed);
    double startPosn = -45;
    for (angle = 1; angle < 91; angle++) {
        VCsetPosition(&vc, startPosn, 0.0);
        while (VCisAtPosition(&vc, startPosn, 0.0) != 0) {
        }
        VCsleep(0.1);
        delayTime = (panSlope * angle) + panOffset;
        startTime = VCgetTime();
        VCmove(&vc, VC_PAN_RIGHT, VC_TILT_STOP);
        VCsleep(delayTime);
        stopTime = VCgetTime();
        VCmove(&vc, VC_PAN_STOP, VC_TILT_STOP);

        // wait for camera to stop
        VCgetPosition(&vc);
        oldPosn = vc.panPosition + 1;
        while (vc.panPosition != oldPosn) {
            oldPosn = vc.panPosition;
            VCgetPosition(&vc);
        }
        finishTime = VCgetTime();

        VCgetPosition(&vc);
        actualAngle = vc.panPosition - startPosn;
        totalTime = (finishTime - startTime) - 0.033;
        // printf("exp ang: %4.1lf  actual ang: %6.3lf   err: %6.3lf", 
        //     angle, actualAngle, angle - actualAngle);
        printf("act ang: %7.4lf   delay: %9.6lf   total time: %9.6lf   deg/sec: %9.6lf\n",
            actualAngle, delayTime, totalTime, actualAngle / totalTime);
    }
    exit(99);
}
#endif

#if 0

{
    // see how fast we can jerk camera around with positioning
    double panPosn = 45;
    double tiltPosn = 0;
    VCsetPanSpeed(&vc, vc.panMaxSpeed);
    VCsetPosition(&vc, -panPosn, 0.0);
    sleep(3);
    VCgetPosition(&vc);
    double startTime = VCgetTime();
    VCsetPosition(&vc, panPosn, tiltPosn);
    while (VCisAtPosition(&vc, panPosn, tiltPosn) != 0) {
        ;
    }
    double finishTime = VCgetTime();
    double elapsed = finishTime - startTime;
    printf("elapsed time: %9.7lf\n", elapsed);
    exit(99);
}

#endif

#if 0
    {
        double speed;
        for (speed = 0; speed < 20; speed++) {
            VCsetPanSpeed(&vc, speed);
            double sp = VCgetPanSpeed(&vc, &sp);
            printf("speed   requested: %8.5lf   actual: %8.5lf   read:%8.5lf\n",
                speed, vc.panSpeed, sp);
        }
        exit(99);
    }
#endif

#if 0
{
    // create a zoom verses focus range table
    // Answer: it makes no sense at all, and is camera independent.
    // See zf??.log files
    int z;
    VCsetZoom(&vc, 0);
    sleep(5);
    for (z = vc.zoomMinRange; z <= vc.zoomMaxRange; z++) {
        int tries = 100;
        unsigned int zoom = -1;
        VCsetZoom(&vc, z);
        while ((z != zoom) && (tries >= 0)) {
            VCgetZoomPosition(&vc, &zoom);
            tries--;
        }
        VCgetFocusRange(&vc, &(vc.focusMinRange), &(vc.focusMaxRange));
        printf("zoom: %5d    min focus: %5d    max focus: %5d\n",
            z, vc.focusMinRange, vc.focusMaxRange);
    }
    exit(99);

}
#endif

#if 0
{
    // test for simultaneous zoom and pan.  Answer: yes - the
    // camera will simultaneously pan and zoom.
    VCsetPanSpeed(&vc, 40.0);
    VCsetTiltSpeed(&vc, 40.0);
    VChome(&vc);
    VCsleep(2);
    VCsetZoom(&vc, vc.zoomMinRange);
    VCsleep(2);
    VCsetPanSpeed(&vc, 10.0);
    VCsetTiltSpeed(&vc, 10.0);
    VCmove(&vc, VC_PAN_RIGHT, VC_TILT_UP);
    VCsetZoom(&vc, vc.zoomMaxRange);
    int i;
    for (i = 0; i < 100; i++) {
        unsigned int zoom;
        VCgetPosition(&vc);
        VCgetZoomPosition(&vc, &zoom);
        printf("%3d   x : %9.4lf    y : %9.4lf   zoom: %5d\n",
            i, vc.panPosition, vc.tiltPosition, zoom);
    }
    VCmove(&vc, VC_PAN_STOP, VC_TILT_STOP);
    exit(99);
}
#endif

#if 0
    if (constructCode != 0) {
        printf("Return from VCconstruct : %d\n", constructCode);
    }

    printf("min focus: %5d     max focus: %5d\n",
        vc.focusMinRange, vc.focusMaxRange);

    printf("min zoom: %5d     max zoom: %5d\n",
        vc.zoomMinRange, vc.zoomMaxRange);

    printf("pan ratio: %7.4lf     tilt ratio: %7.4lf\n",
        vc.panRatio, vc.tiltRatio);
#endif

#if 0
{
    printf("Camera stabilizing experiment.\n");
    VCsetPanSpeed(&vc, 90.0);
    VChome(&vc);
    VCsleep(2.0);
    // determine time to wait for move
    double angle = 30.0;
    double panSlope = 0.011204481792717087;
    double panOffset = 0.078151260504201653;
    double delay = (panSlope * angle) + panOffset;
    VCmove(&vc, VC_PAN_LEFT, VC_TILT_STOP);
    VCsleep(delay);
    double start = VCgetTime();
    VCmove(&vc, VC_PAN_STOP, VC_TILT_STOP);
    double p = 10000;
    int i;
    VCsleep(0.065);
    for (i = 0; i < 100; i++) {
        VCgetPosition(&vc);
        if (vc.panPosition == p) {
            double finish = VCgetTime();
            printf("i: %3d   elapsed: %9.7lf\n", i, finish-start);
            break;
        }
        p = vc.panPosition;
    }

    exit(99);
}
#endif

#if 0
{
    int i;
    VCsetPanSpeed(&vc, 90.0);
    double incr = 10;
    for (i = 0; i < 10; i++) {
        VChome(&vc);
        VCsleep(2.0);
        double panAngle = incr * i;
        VCsetPanPositionFast(&vc, panAngle);
        sleep(1);
        VCgetPosition(&vc);
        printf("Angle requested: %8.4lf   actual: %8.4lf\n",
            panAngle, vc.panPosition);
    }
    exit(99);
}
#endif

#if 0
{
    int i;
    VCsetPanSpeed(&vc, 90.0);
    double panAngle = 19.8;
    for (i = 0; i < 10; i++) {
        VChome(&vc);
        VCsleep(2.0);
        double startTime = VCgetTime();
        VCsetPosition(&vc, panAngle, 0);
        while (VCisAtPosition(&vc, 19.8, 0) != 0) {
            ;
        }
        double finishTime = VCgetTime();
        VCsleep(2.0);
        printf("panAngle:  %6.3lf    Elapsed time: %8.5lf\n",
            panAngle, finishTime - startTime);
    }
    exit(99);
}
#endif

#if 0
{
    // perform experiment to see how quickly we can pan the
    // camera when telling it to go full speed.
    int i;
    double start = 0;
    double finish = 0;
    double incr = 0.1;
    double moveTime = 0;
    VCsetPanSpeed(&vc, vc.panMaxSpeed);
    for (i = 0; i < 10; i++) {
        VChome(&vc);
        VCsleep(2.0);
        VCgetPosition(&vc);
        start = vc.panPosition;
        VCmove(&vc, VC_PAN_RIGHT, VC_TILT_STOP);
        moveTime = incr * i;
        VCsleep(moveTime);
        VCmove(&vc, VC_PAN_STOP, VC_TILT_STOP);
        VCsleep(2.0);
        VCgetPosition(&vc);
        finish = vc.panPosition;
        printf(" %5.3lf    %8.4lf\n", moveTime, ABS(start - finish));
        TRCINT(i);
    }
    exit(99);
}
#endif

#if 0
{
    // perform experiment to see how quickly we can tilt the
    // camera when telling it to go full speed.
    int i;
    double start = 0;
    double finish = 0;
    double incr = 0.05;
    double moveTime = 0;
    VCsetTiltSpeed(&vc, vc.tiltMaxSpeed / 2.0);
    for (i = 0; i < 10; i++) {
        VChome(&vc);
        VCsleep(2.0);
        VCgetPosition(&vc);
        start = vc.tiltPosition;
        VCmove(&vc, VC_PAN_STOP, VC_TILT_UP);
        moveTime = incr * i;
        VCsleep(moveTime);
        VCmove(&vc, VC_PAN_STOP, VC_TILT_STOP);
        VCsleep(2.0);
        VCgetPosition(&vc);
        finish = vc.tiltPosition;
        printf(" %5.3lf    %8.4lf\n", moveTime, ABS(start - finish));
        TRCINT(i);
    }
    exit(99);
}
#endif

#if 0
    printf("--------------\n");

    printf("speed 90.0\n");
    VCsetPanSpeed(&vc, 10.0);

    printf("move right\n");
    VCmove(&vc, VC_PAN_RIGHT, VC_TILT_STOP);

    printf("sleep 1\n");
    sleep(1);

    printf("move stop\n");
    VCmove(&vc, 0, 0);

    printf("position 90.0\n");
    VCsetPosition(&vc, 90.0, 0);

    printf("sleep 3\n");
    sleep(3);

    printf("move stop\n");
    VCmove(&vc, 0, 0);
    exit(99);
#endif

#if 0
    {
        char pRaw[5];
        double p;
        unsigned char *pp;
        double pd;

        pp = (unsigned char *)pRaw;

        sprintf(pRaw, "%04X", 8);
        pd = VCrawToReal(pp, 4);
        p = pd * vc.panRatio;
        printf(" pRaw: %5d  %5x : pd: %12.5lf   p: %12.5lf\n", pRaw, pRaw, pd, p);

        sprintf(pRaw, "%04X", 800);
        pd = VCrawToReal(pp, 4);
        p = pd * vc.panRatio;
        printf(" pRaw: %5d  %5x : pd: %12.5lf   p: %12.5lf\n", pRaw, pRaw, pd, p);




        sprintf(pRaw, "%04X", 8);
        pd = VCrawToReal(pp, 4);
        p = pd * vc.tiltRatio;
        printf(" pRaw: %5d  %5x : pd: %12.5lf   p: %12.5lf\n", pRaw, pRaw, pd, p);

        sprintf(pRaw, "%04X", 622);
        pd = VCrawToReal(pp, 4);
        p = pd * vc.tiltRatio;
        printf(" pRaw: %5d  %5x : pd: %12.5lf   p: %12.5lf\n", pRaw, pRaw, pd, p);





        exit(99);
    }
#endif

#if 0
    VCstartThread(&vc);
        {
        struct timespec req;
        struct timespec rem;
        req.tv_sec = 0;
        req.tv_nsec = 1000*1000*500;
        // sleep for 1 microsecond.  This actually seems to be 1/50
        // of a second.
        nanosleep(&req, &rem);
        }
    TRC;
    VCschedSetPosition(&vc, 30.0, 30.0);
    sleep(2);
    exit(99);
#endif

#if 0
    VCgetPosition(&vc);
    printf("Position %d pan: %7.3lf  tilt: %7.3lf\n", __LINE__, vc.panPosition, vc.tiltPosition);
    VCsetPanSpeed(&vc, vc.panMaxSpeed);
    VCsetTiltSpeed(&vc, vc.tiltMaxSpeed);
    while (VCsetPosition(&vc, 30.0, 0.0) == VC_E_CAMERA_BUSY) {
    }
    {
        vc.desiredPanPosition = 30.0;
        vc.desiredTiltPosition = 0.0;
        double panDiff = ABS(vc.desiredPanPosition - vc.panPosition);
        double tiltDiff = ABS(vc.desiredTiltPosition - vc.tiltPosition);
        while ((panDiff > 0.2) || (tiltDiff > .2)) {
            VCgetPosition(&vc);
            panDiff = ABS(vc.desiredPanPosition - vc.panPosition);
            tiltDiff = ABS(vc.desiredTiltPosition - vc.tiltPosition);
        }
    }
    VCgetPosition(&vc);
    printf("Position %d pan: %7.3lf  tilt: %7.3lf\n", __LINE__, vc.panPosition, vc.tiltPosition);


    while (VCsetPosition(&vc,  0.0, 10.0) == VC_E_CAMERA_BUSY) {
    }
    VCgetPosition(&vc);
    printf("Position %d pan: %7.3lf  tilt: %7.3lf\n", __LINE__, vc.panPosition, vc.tiltPosition);

    while (VCsetPosition(&vc,  10.0, 10.0) == VC_E_CAMERA_BUSY) {
    }
    VCgetPosition(&vc);
    printf("Position %d pan: %7.3lf  tilt: %7.3lf\n", __LINE__, vc.panPosition, vc.tiltPosition);

    while (VCsetPosition(&vc,  10.0, 0.0) == VC_E_CAMERA_BUSY) {
    }

    VCgetPosition(&vc);
    printf("Position %d pan: %7.3lf  tilt: %7.3lf\n", __LINE__, vc.panPosition, vc.tiltPosition);
    exit(99);
#endif

#if 0
    {
        double start = VCgetTime();
        double stuff;
        int count = 1000;
        int i;
        for (i = 0; i < count; i++) {
            VCgetTiltRatio (&vc, &stuff);
        }
        double elapsed = VCgetTime() - start;
        printf("Time per call: %10.5lf\n", elapsed / count);
        exit(99);
    }
#endif

    code = VCsetIris(&vc, VC_IRIS_F16_0);
#if 1
    int z = vc.zoomMinRange;
    int zStep = ( 1 * (vc.zoomMaxRange - vc.zoomMinRange)) / 3;
    if (VCsetZoom(&vc, zStep) != 0) {
        VC_SHOW_ERROR(&vc, "VCsetZoom");
    }
    exit(99);
    for (z = vc.zoomMinRange; z <= vc.zoomMaxRange; z += zStep) {
        printf("z: %4d\n", z);
        if (VCsetZoom(&vc, z) != 0) {
            VC_SHOW_ERROR(&vc, "VCsetZoom");
        }
        sleep(1);
    }
    if (VCsetZoom(&vc, (vc.zoomMaxRange - vc.zoomMinRange) / 2) != 0) {
        VC_SHOW_ERROR(&vc, "VCsetZoom");
    }
#endif

    double pan = 20.0;
    double tilt = 20.0;

    double start = VCgetTime();

    VChome(&vc);

    VCsetPosition(&vc,  10.0, 10.0);


    if (VCsetExposureMode(&vc, VC_EXPOSURE_MANUAL) != 0) {
        VC_SHOW_ERROR(&vc, "VCsetExposureMode");
    }

    sleep(1);

    if (VCsetAELockMode(&vc, VC_AE_LOCK_OFF) != 0) {
        VC_SHOW_ERROR(&vc, "VCsetAELockMode");
    }


    if (VCsetBacklightCompensationMode(&vc, VC_BACKLIGHT_COMPENSATION_OFF) != 0) {
        VC_SHOW_ERROR(&vc, "VCsetBacklightCompensationMode");
    }

    // code = VCsetIris(&vc, VC_IRIS_F16_0);
    code = VCsetIris(&vc, VC_IRIS_F2_0);
    if (code != 0) {
        printf("Return from VCsetIris : %d\n", code);
    }
    else printf("VCsetIris VC_IRIS_F2_0\n");

    start = VCgetTime();
    int i;
    for (i = 0; i < 100; i++) {
        VCsetPanSpeed (&vc, .01);
    }
    double finish = VCgetTime();
    printf("Elapsed time: %16.12lf\n", finish - start);

    VCsetTiltSpeed(&vc, .01);

    // code = VCirCut(&vc, VC_IR_CUT_ON);
    // printf("Return from VCirCut : %d\n", code);

    // code = VCirCut(&vc, VC_IR_CUT_OFF);
    // printf("Return from VCirCut : %d\n", code);

    for (i = 0; i < 2; i++) {
        while (VCsetPosition(&vc,  0.0, 0.0) == VC_E_CAMERA_BUSY) {
        }

        while (VCsetPosition(&vc,  0.0, 10.0) == VC_E_CAMERA_BUSY) {
        }

        while (VCsetPosition(&vc,  10.0, 10.0) == VC_E_CAMERA_BUSY) {
        }

        while (VCsetPosition(&vc,  10.0, 0.0) == VC_E_CAMERA_BUSY) {
        }
    }
    exit(99);

    // while (VChome(&vc) == VC_E_CAMERA_BUSY) { }

    double ps;
    double ts;
    VCgetPanSpeed(&vc, &ps);
    VCgetTiltSpeed(&vc, &ts);

    int sleepTime = 1;

    VCmove(&vc, VC_PAN_RIGHT, VC_TILT_STOP);
    sleep(sleepTime);

    VCmove(&vc, VC_PAN_STOP, VC_TILT_UP);
    sleep(sleepTime);

    VCmove(&vc, VC_PAN_LEFT, VC_TILT_STOP);
    sleep(sleepTime);

    VCmove(&vc, VC_PAN_STOP, VC_TILT_DOWN);
    sleep(sleepTime);

    VCmove(&vc, VC_PAN_STOP, VC_TILT_STOP);

    while (VChome(&vc) == VC_E_CAMERA_BUSY) {
    }

    double elapsed = VCgetTime() - start;
    printf("Elapsed time: %15.12lf\n", elapsed);
    printf("Return from VCsetPosition : %d\n", code);
    return 0;
}
#endif
#endif
