/* ====================================================================
 * Copyright (c) 2010 The Regents of the University of Michigan and
 * The European Organization for Nuclear Research (CERN). Licensed under the
 * Educational Community License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License. You may
 * obtain a copy of the License at
 *
 * http://www.osedu.org/licenses/ECL-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * permissions and limitations under the License.
  ==================================================================== */

// Written by: Jim Irrer irrer@umich.edu jimirrer@gmail.com

#ifndef CANONVC_C50I_H 

#define CANONVC_C50I_H 

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <fcntl.h>
#include <time.h>
#include <sys/signal.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/time.h>
#include <termios.h>
#include <pthread.h>

// --------------------------------------------------------

#ifndef null
#define null (0)
#endif

#ifndef true
#define true (1==1)
#endif
#ifndef false
#define false (1==0)
#endif

// absolute value function
#define ABS(A) (((A)>=0)?(A):(-(A)))
// min value function
#define MIN(A,B) (((A)<=(B))?(A):(B))
// max value function
#define MAX(A,B) (((A)>=(B))?(A):(B))

// --------------------------------------------------------

/** NTSC aspect ratio.  Pixels are tall and skinny. */
#define NTSC_ASPECT_RATIO (0.9)

// --------------------------------------------------------

// for debugging
#define TRC fprintf(stderr, "%s:%d\n", __FILE__, __LINE__)
#define TRCINT(I)fprintf(stderr, "%s:%d %s : %d\n", __FILE__, __LINE__, #I, (I))
#define TRCDBL(D)fprintf(stderr, "%s:%d %s : %g\n", __FILE__, __LINE__, #D, (D))
#define TRCMSG(T)fprintf(stderr, "%s:%d : %s\n", __FILE__, __LINE__, (T))
#define TRCHEX(M,S)fprintf(stderr, "%s:%d %s : ", __FILE__, __LINE__, #M); { int m; for (m = 0; m < (S); m++) printf(" %02x", 0xff & (M)[m]); } printf("\n")

// --------------------------------------------------------

// seems like math.h should declare this.
extern double round(double);

// --------------------------------------------------------

/** commands to server thread */
#define VC_Q_MOVE                     (1)
#define VC_Q_PAN_SPEED                (2)
#define VC_Q_TILT_SPEED               (3)
#define VC_Q_ZOOM                     (4)
#define VC_Q_SET_POSITION             (5)
#define VC_Q_SET_POSITION_AND_WAIT    (6)
#define VC_Q_GET_POSITION             (7)
#define VC_Q_EXIT                     (8)

// --------------------------------------------------------

/**
 * Maximum time to wait for camera to respond to a message
 * before assuming there was an error
 */
#define VC_MAX_WAIT_TIME (0.3)

/** Time in seconds to wait between tries of reading the camera */
#define VC_RETRY_TIME (0.001)

/**
 * Maximum time in seconds to wait for the camera to finish
 * its previous operation (BUSY status)
 */
#define VC_MAX_OPERATION_TIME (2.0)


/** If the condition fails, then invoke an assertion */
#define VC_ASSERT(COND,MSG) if (!(COND)) VCassert(__FILE__, __LINE__, (MSG))

/** Command terminator for messages sent from the computer to the camera. */
#define VC_CMD_END ((unsigned char)(0xef))

/** Movement controls */
#define VC_PAN_STOP    (0)
#define VC_PAN_RIGHT   (1)
#define VC_PAN_LEFT    (2)
#define VC_PAN_NONE    (-1)

#define VC_TILT_STOP   (0)
#define VC_TILT_UP     (1)
#define VC_TILT_DOWN   (2)
#define VC_TILT_NONE   (-1)

/** Infrared cut controls. */
#define VC_IR_CUT_ON   (0)
#define VC_IR_CUT_OFF  (1)

/** Iris settings */
#define VC_IRIS_CLOSE_MECHANICALLY "00"
#define VC_IRIS_CLOSE_OPTICALLY    "01"
// iris is a very small hole
#define VC_IRIS_F16_0              "02"
#define VC_IRIS_F13_3              "03"
#define VC_IRIS_F11_0              "04"
#define VC_IRIS_F9_4               "05"
#define VC_IRIS_F8_0               "06"
#define VC_IRIS_F6_7               "07"
#define VC_IRIS_F5_6               "08"
#define VC_IRIS_F4_7               "09"
#define VC_IRIS_F4_0               "0A"
#define VC_IRIS_F3_3               "0B"
#define VC_IRIS_F2_8               "0C"
#define VC_IRIS_F2_4               "0D"
#define VC_IRIS_F2_0               "0E"
#define VC_IRIS_F1_9               "0F"
#define VC_IRIS_F1_8               "10"
#define VC_IRIS_F1_6               "11"
// iris is a very large hole
#define VC_IRIS_OPEN               0x12

/** exposure modes */
#define VC_EXPOSURE_AUTOMATIC   (0x32)
#define VC_EXPOSURE_MANUAL      (0x33)

/** AE (auto exposure?) modes */
#define VC_AE_LOCK_OFF (0x30)
#define VC_AE_LOCK_ON  (0x31)

/** backlight compensation mode */
#define VC_BACKLIGHT_COMPENSATION_OFF (0x30)
#define VC_BACKLIGHT_COMPENSATION_ON  (0x31)

/** shutter speed program mode */
#define VC_SHUTTER_SPEED_PROGRAM (0x30)
#define VC_SHUTTER_SPEED_1_60    (0x31)   // 1/60 == (PAL: 1/50) mode.
#define VC_SHUTTER_SPEED_1_100   (0x32)   // 1/100 == (PAL: 1/120) mode.

/** minimum shutter speed */
#define VC_SHUTTER_SPEED_MIN     (0)

/** maximum shutter speed */
#define VC_SHUTTER_SPEED_MAX     (27)

/** Focus modes */
#define VC_FOCUS_MODE_AUTOMATIC  (0x30)
#define VC_FOCUS_MODE_MANUAL     (0x31)
#define VC_FOCUS_MODE_NEAR       (0x32)
#define VC_FOCUS_MODE_FAR        (0x33)

// --------------------------------------------------------

/** invalid position values */
#define VC_INVALID_POSITION (1000.0*1000.0*1000.0)

// --------------------------------------------------------

// error codes

#define VC_E_UNKNOWN                 ( 1)  // Unknown error
#define VC_E_CLOSE_INVALID           ( 2)  // EBADF returned on close
#define VC_E_CLOSE_INTERRUPTED       ( 3)  // EINTR returned on close
#define VC_E_CLOSE_IO_ERR            ( 4)  // EIO returned on close
#define VC_E_PAN_POSITION            ( 5)  // Pan degrees out of bounds
#define VC_E_TILT_POSITION           ( 6)  // Tilt degrees out of bounds
#define VC_E_SYSTEM                  ( 7)  // Error returned from operating system call
#define VC_E_CAMERA_BUSY             ( 8)  // Camera is busy
#define VC_E_BAD_PARAM               ( 9)  // Bad parameter in command to camera
#define VC_E_MODE_ERROR              (10)  // Camera mode error
#define VC_E_UNRECOGNIZED_RESPONSE   (11)  // Unrecognized response from camera
#define VC_E_NO_RESPONSE             (12)  // No response from camera
#define VC_E_PAN_DIRECTION           (13)  // Invalid pan direction
#define VC_E_TILT_DIRECTION          (14)  // Invalid tilt direction
#define VC_E_FOCUS_OUT_OF_RANGE      (15)  // Focus value out of range
#define VC_E_BAD_FOCUS_MODE          (16)  // Invalid mode for focus
#define VC_E_ZOOM_OUT_OF_RANGE       (17)  // Zoom value out of range
#define VC_E_PAN_SPEED_TOO_HIGH      (18)  // Pan speed too high
#define VC_E_PAN_SPEED_TOO_LOW       (19)  // Pan speed too low
#define VC_E_TILT_SPEED_TOO_HIGH     (20)  // Tilt speed too high
#define VC_E_TILT_SPEED_TOO_LOW      (21)  // Tilt speed too low
#define VC_E_SHUTTER_SPEED_TOO_LOW   (22)  // Shutter speed too low
#define VC_E_SHUTTER_SPEED_TOO_HIGH  (23)  // Shutter speed too high

// --------------------------------------------------------

/** un-ambiguous 16 bit type */
typedef short int16;

/** For sending and and receiving messages to and from camera */
typedef unsigned char VC_BUF[200];

// --------------------------------------------------------

// universal constant
#define PI (3.141592653589793238512808959406186204433)

// to convert degrees to radians
#define DtoR(A) (((A)/180.0)*PI)

// to convert radians to degrees
#define RtoD(A) (((A)/PI)*180.0)

// --------------------------------------------------------

typedef struct ZoomAngleStruct
{
    int zoom;      // zoom unit used by camera
    double angle;  // corresponding angle from center to side edge
} ZoomAngle;

// --------------------------------------------------------

typedef struct VCstruct
{
    /** file descriptor used for all I/O */
    int fd;

    /** latest error number */
    int errorCode;

    /** latest error message */
    char *errorMessage;

    /** maximum number of degrees that camera can be panned */
    double panMax;

    /** maximum number of degrees that camera can be tilted */
    double tiltMax;

    /** minimum number of degrees per second that camera can be panned */
    double panMinSpeed;

    /** maximum number of degrees per second that camera can be panned */
    double panMaxSpeed;

    /** minimum number of degrees per second that camera can be tilted */
    double tiltMinSpeed;

    /** maximum number of degrees per second that camera can be tilted */
    double tiltMaxSpeed;

    /** current pan speed in degrees / sec */
    double panSpeed;

    /** queued pan speed in degrees / sec */
    double queuedPanSpeed;

    /** current tilt speed in degrees / sec */
    double tiltSpeed;

    /** queued tilt speed in degrees / sec */
    double queuedTiltSpeed;

    /** pan ratio / 100000 */
    double panRatio;

    /** tilt ratio / 100000 */
    double tiltRatio;

    /** minimum setting for focus */
    unsigned int focusMinRange;

    /** maximum setting for focus */
    unsigned int focusMaxRange;

    /** minimum setting for zoom */
    unsigned int zoomMinRange;

    /** maximum setting for zoom */
    unsigned int zoomMaxRange;

    /** latest command to camera */
    VC_BUF command;

    /** latest response from camera */
    VC_BUF response;

    /** direction that camera is panning */
    int panDirection;

    /** queued direction that camera is to pan */
    int queuedPanDirection;

    /** direction that camera is tilting */
    int tiltDirection;

    /** queued direction that camera is to tilt */
    int queuedTiltDirection;

    /** tima at which direction was changed. */
    double setDirectionTime;

    /** current pan position */
    double panPosition;

    /** current tilt position */
    double tiltPosition;

    /** current zoom position */
    double zoomPosition;

    /** time at which position was fetched */
    double getPositionTime;

    /** pipes to send commands to thread.  Read from 0, write to 1. */
    int threadControl[2];

    /** Thread that sends commands to camera */
    pthread_t serverThread;

    /** Number of commands requested of thread */
    int requestedCommands;

    /** Number of commands completed by thread */
    int completedCommands;

} VC;

// --------------------------------------------------------

extern int VCconstruct(VC *vc, char *devName);
extern void VChexToAscii(int16 h, unsigned char *ascii, int size);
extern void VCassert(char *file, int line, char *message);
extern int VCsetError(VC *vc, int code, char *message);
extern int VCsystemError(VC *vc, int ern, char *message);
extern void VCprintCommand(VC *vc);
extern int16 VCpanDegreesToAscii(VC *vc, double degrees, unsigned char *ascii, int size);
extern int16 VCtiltDegreesToAscii(VC *vc, double degrees, unsigned char *ascii, int size);
extern int VCcameraStatus(VC *vc);
extern int VCtoCamera(VC *vc, VC_BUF command);
extern int VCtoCameraRetryBusy(VC *vc, VC_BUF command);
extern int VChostControl(VC *vc);
extern int VCmove(VC *vc, int pan, int tilt);
extern int VCgetPosition(VC *vc);
extern int VCsetPosition(VC *vc,  double pan, double tilt);
extern int VCisAtPosition(VC *vc, double pan, double tilt);
extern int VCsetPanPositionFast(VC *vc,  double pan);
extern unsigned int VCrawToInt(unsigned char *raw, int size);
extern double VCrawToReal(unsigned char *raw, int size);
extern int VCgetPanRatio(VC *vc, double *pr);
extern int VCgetTiltRatio(VC *vc, double *tr);
extern int VCsetPanSpeed(VC *vc, double ps);
extern int VCsetTiltSpeed(VC *vc, double ts);
extern int VCgetPanSpeed(VC *vc, double *ps);
extern int VCgetTiltSpeed(VC *vc, double *ts);
extern int VCsetFocusMode(VC *vc, int mode);
extern int VCsetFocusPosition(VC *vc, int pos);
extern int VCgetFocusRange(VC *vc, unsigned int *minFocus, unsigned int *maxFocus);
extern int VCgetFocusPosition(VC *vc, unsigned int *pos);
extern int VCgetZoomRange(VC *vc, unsigned int *minZoom, unsigned int *maxZoom);
extern int VCsetZoom(VC *vc, int zoom);
extern int VCgetZoomPosition(VC *vc, unsigned int *pos);
extern int VCsetBacklightCompensationMode(VC *vc, int mode);
extern int VCsetExposureMode(VC *vc, int mode);
extern int VCsetAELockMode(VC *vc, int mode);
extern int VCsetShutterSpeedMode(VC *vc, int mode);
extern int VCsetIris(VC *vc, char *iris);
extern int VChome(VC *vc);
extern int VCirCut(VC *vc, int state);
extern void VCwakeThread(VC *vc);
extern void VCschedMove(VC *vc, int pan, int tilt);
extern void VCschedSetPanSpeed(VC *vc, double ps);
extern void VCschedSetTiltSpeed(VC *vc, double ts);
extern void VCschedSetPosition(VC *vc, double pan, double tilt);
extern int VCdelete(VC *vc);
extern void VCsleep(double t);
extern double VCgetZoomAngleX(int zoom);
extern double VCgetZoomAngleY(int zoom);
extern double VCgetTime();
extern double VCgetRelativeTime(struct timeval *relTime);
extern int main(int argc, char **argv);
extern void *VCthread(void *v);
extern int VCoutstandingCommands(VC *vc);
#ifdef VC_THREADS
extern void VCqueueGetPosition(VC *vc);
#endif

// --------------------------------------------------------

#endif // CANONVC_C50I_H 

