/* ====================================================================
 * Copyright (c) 2010 The Regents of the University of Michigan and
 * The European Organization for Nuclear Research (CERN). Licensed under the
 * Educational Community License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License. You may
 * obtain a copy of the License at
 *
 * http://www.osedu.org/licenses/ECL-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * permissions and limitations under the License.
  ==================================================================== */

// Written by: Jim Irrer irrer@umich.edu jimirrer@gmail.com

#define VC_LIBRARY

#include "canonVC_C50i.h"
#include "canonVC_C50i.c"

// --------------------------------------------------------

/*
 * Give a usage message and exit with bad error status.
 */

void usage(void)
{
    fprintf(stderr, "\nUsage: ptzf -d device [options]\n");
    fprintf(stderr, "    [-p panDegrees]\n");
    fprintf(stderr, "    [ -t tiltDegrees ]\n");
    fprintf(stderr, "    [ -z zoomPosition ]\n");
    fprintf(stderr, "    [ -f focusPosition ]\n");
    fprintf(stderr, "    [ -P panPixels ] Using -P requires the -w specification\n");
    fprintf(stderr, "    [ -T tiltPixels ] Using -T requires the -h specification\n");
    fprintf(stderr, "    [ -w imageWidthInPixels ]\n");
    fprintf(stderr, "    [ -h imageHeightInPixels ]\n");
    fprintf(stderr, "    [ -r relative ] Only applies to -p, -t, and -z options\n");
    exit(1);
}

// --------------------------------------------------------

/**
 * Check the error code.  If it is bad, then print a message
 * and exit.
 *
 * @param msg Text indicating what camera operation was attempted.
 *
 * @param errCode Error code.
 */

void checkErrCode(const char *msg, int errCode)
{
    if (errCode != 0) {
        fprintf(stderr, "Error \"%s\".  Error code: %d\n", msg, errCode);
        usage();
    }
}

// --------------------------------------------------------

/*
 * Get the current settings for the camera and print them to
 * standard output.  The settings are pan, tilt, zoom, and
 * focus, in that order.  Pan and tilt are in degrees, and zoom
 * and focus are in ticks.
 *
 * @param vc Descriptor for camera.
 */

void getSettings(VC *vc)
{
    unsigned int focus;
    unsigned int zoom;

    int errCode;

    errCode = VCsetFocusMode(vc, VC_FOCUS_MODE_AUTOMATIC);
    checkErrCode("set focus mode to automatic", errCode);

    errCode = VCgetPosition(vc);
    checkErrCode("get pan-tilt position", errCode);

    errCode = VCgetZoomPosition(vc, &zoom);
    checkErrCode("get zoom position", errCode);

    errCode = VCgetFocusPosition(vc, &focus);
    checkErrCode("get focus position", errCode);

    printf("%lf %lf %d %d\n", vc->panPosition, vc->tiltPosition, zoom, focus);
}

// --------------------------------------------------------

/**
 * Interpret a text string as a double value.  Exit on error.
 * Bounds check with high and low values.  If high and low
 * are the same, then ignore them.
 *
 * @param text Text to interpret.
 *
 * @param msg Name of parameter to report in case of error.
 *
 * @param low Low limit.
 *
 * @param high High limit.
 *
 * @return double value.
 */

double getDouble(char *text, char *msg, int low, int high)
{
    double d;
    char c;
    if (sscanf(text, "%lf%c", &d, &c) != 1) {
        fprintf(stderr, "%s is an invalid double for %s\n", text, msg);
        usage();
    }

    if (low == high) {
        return d;
    }

    if (d < low) {
        fprintf(stderr, "%s is too low.  Must be %d or higher.\n", msg, low);
        usage();
    }

    if (d > high) {
        fprintf(stderr, "%s is too high.  Must be %d or lower.\n", msg, high);
        usage();
    }
    return d;
}

// --------------------------------------------------------

/**
 * Interpret a text string as an int value.  Exit on error.
 * Bounds check with high and low values.  If high and low
 * are the same, then ignore them.
 *
 * @param text Text to interpret.
 *
 * @param msg Name of parameter to report in case of error.
 *
 * @param low Low limit.
 *
 * @param high High limit.
 *
 * @return int value.
 */

int getInt(char *text, char *msg, int low, int high)
{
    int i;
    char c;
    if (sscanf(text, "%d%c", &i, &c) != 1) {
        fprintf(stderr, "%s is an invalid integer for %s\n", text, msg);
        usage();
    }
    if (low == high) {
        return i;
    }

    if (i < low) {
        fprintf(stderr, "%s is too low.  Must be %d or higher.\n", msg, low);
        usage();
    }
    if (i > high) {
        fprintf(stderr, "%s is too high.  Must be %d or lower.\n", msg, high);
        usage();
    }
    return i;
}

// --------------------------------------------------------

// invalid value for marking parameters as uninitialized
#define UNDEF_VAL (-1000000)

int main(int argc, char **argv)
{
    VC vc;
    double panRatio;
    double tiltRatio;
    int code;
    int errCode;  // error code from camera library
    int deviceInitialized = false;
    int setting = false;
    int positionChanged = false;
    int zoomChanged = false;
    int relative = false;
    double panDegrees  = UNDEF_VAL;
    double tiltDegrees = UNDEF_VAL;
    int zoomPosition   = UNDEF_VAL;
    int focusPosition  = UNDEF_VAL;
    int panPixels      = UNDEF_VAL;
    int tiltPixels     = UNDEF_VAL;
    int width          = UNDEF_VAL;
    int height         = UNDEF_VAL;


    int c;
    while ((c = getopt(argc, argv, "d:p:t:z:f:P:T:w:h:r")) != EOF) {
        switch (c) {
            case 'd':   // device
                if (deviceInitialized) {
                    fprintf(stderr, "Can only specify device once (-d option).");
                    usage();
                }
                errCode = VCconstruct(&vc, optarg);
                checkErrCode("initializing device", errCode);
                deviceInitialized = true;
                break;

            case 'p':   // pan in degrees
                setting = true;
                panDegrees = getDouble(optarg, "pan degrees", -200, 200);
                positionChanged = true;
                break;

            case 't':   // tilt in degrees
                setting = true;
                tiltDegrees = getDouble(optarg, "tilt degrees", -200, 200);
                positionChanged = true;
                break;

            case 'z':   // zoom position
                setting = true;
                zoomPosition = getInt(optarg, "zoom", -1960, 1960);
                zoomChanged = true;
                break;

            case 'f':   // focus position
                setting = true;
                focusPosition = getInt(optarg, "focus", 0, 0);
                break;

            case 'P':   // pan in pixels
                setting = true;
                panPixels = getInt(optarg, "pan pixels", 2, 4096);
                positionChanged = true;
                break;

            case 'T':   // tilt in pixels
                setting = true;
                tiltPixels = getInt(optarg, "tilt pixels", 2, 4096);
                positionChanged = true;
                break;

            case 'w':   // width in pixels
                setting = true;
                width = getInt(optarg, "width", 2, 4096);
                break;

            case 'h':   // width in pixels
                setting = true;
                height = getInt(optarg, "height", 2, 4096);
                break;

            case 'r':   // move relative to current position
                setting = true;
                relative = true;
                break;
        }
    }

    if (!deviceInitialized) {
        fprintf(stderr, "Must specify device with -d option.\n");
        usage();
    }

#if 0
    char chmod[200];
    sprintf(chmod, "sudo chmod 666 %s &", argv[1]);
    system(chmod);
#endif

    // if setting parameters
    if (setting) {
        if (relative || (panDegrees == UNDEF_VAL) || (tiltDegrees == UNDEF_VAL)) {
            errCode = VCgetPosition(&vc);
            checkErrCode("initializing position for pan degrees", errCode);
        }

        if (panDegrees == UNDEF_VAL) {
            panDegrees = vc.panPosition;
        }
        else {
            if (relative) {
                panDegrees += vc.panPosition;
            }
        }

        if (tiltDegrees == UNDEF_VAL) {
            tiltDegrees = vc.tiltPosition;
        }
        else {
            if (relative) {
                tiltDegrees += vc.tiltPosition;
            }
        }

        if (zoomPosition == UNDEF_VAL) {
            unsigned int z;
            errCode = VCgetZoomPosition(&vc, &z);
            checkErrCode("initializing zoom position", errCode);
            zoomPosition = z;
        }
        else {
            if (relative) {
                unsigned int z;
                errCode = VCgetZoomPosition(&vc, &z);
                checkErrCode("initializing zoom position for relative zoom", errCode);
                zoomPosition += z;
            }
        }

        if (panPixels != UNDEF_VAL) {
            if (width == UNDEF_VAL) {
                fprintf(stderr, "Must specify -w <width> in conjuction with -P option.\n");
                usage();
            }
            double zoomAngle = VCgetZoomAngleX(zoomPosition);
            zoomAngle = DtoR(zoomAngle);
            double panDistance = ((width / 2.0) * cos(zoomAngle)) / sin(zoomAngle);
            panDegrees += RtoD(atan(((panPixels - width / 2.0)) / panDistance));
        }

        if (tiltPixels != UNDEF_VAL) {
            if (height == UNDEF_VAL) {
                fprintf(stderr, "Must specify -h <width> in conjuction with -T option.\n");
                usage();
            }
            tiltPixels = height - tiltPixels;
            double zoomAngle = VCgetZoomAngleY(zoomPosition);
            zoomAngle = DtoR(zoomAngle);
            double tiltDistance = ((height / 2.0) * cos(zoomAngle)) / sin(zoomAngle);
            tiltDegrees += RtoD(atan(((tiltPixels - height / 2.0)) / tiltDistance));
        }

        errCode = VCsetPanSpeed(&vc, vc.panMaxSpeed);
        checkErrCode("set pan speed", errCode);

        errCode = VCsetTiltSpeed(&vc, vc.tiltMaxSpeed);
        checkErrCode("set tilt speed", errCode);

        errCode = VCsetPosition(&vc, panDegrees, tiltDegrees);
        checkErrCode("moving camera", errCode);

        errCode = VCsetZoom(&vc, zoomPosition);
        checkErrCode("zooming camera", errCode);

        if (focusPosition == UNDEF_VAL) {
            errCode = VCsetFocusMode(&vc, VC_FOCUS_MODE_AUTOMATIC);
            checkErrCode("set focus mode to automatic", errCode);
        }
        else {
            errCode = VCsetFocusMode(&vc, VC_FOCUS_MODE_MANUAL);
            checkErrCode("set focus mode to manual", errCode);

            errCode = VCsetFocusPosition(&vc, focusPosition);
            checkErrCode("set focus position", errCode);
        }

        // wait for camera to finish movement
        while (positionChanged) {
            errCode = VCisAtPosition(&vc, panDegrees, tiltDegrees);
            if (errCode == 0) {
                break;
            }
            if (errCode != -1) {
                checkErrCode("waiting for position", errCode);
            }
        }

        while (zoomChanged) {
            unsigned int z;
            errCode = VCgetZoomPosition(&vc, &z);
            checkErrCode("waiting for zoom", errCode);
            if (z == zoomPosition) {
                break;
            }
        }
    }

    getSettings(&vc);


    return 0;
}
